{------------------------------------------------------------------------------}
{        This code is copyright (C) 2008-2025 by Michael in der Wiesche;       }
{               stack frame handling is based on the C++ samples               }
{                        provided by Oleg Starodumov at                        }
{             <http://www.debuginfo.com/examples/dbgexamples.html>             }
{               Consts & Types from DbgHlp.h & MiniDumpApiSet.h                }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit DbgHelp;

interface

uses
  Windows, Classes, Rtti, SysUtils, StrUtils, StrTools, FileTools, WindowTools, Processes, DlgProgress, MDResources;

const
  EXCEPTION_CONTINUE_EXECUTION 		= -1;
  EXCEPTION_CONTINUE_SEARCH		=  0;
  EXCEPTION_EXECUTE_HANDLER		=  1;

const
  EXCEPTION_GUARD_PAGE_VIOLATION	= $80000001;
  EXCEPTION_DATATYPE_MISALIGNMENT	= $80000002;
  EXCEPTION_BREAKPOINT			= $80000003;
  EXCEPTION_SINGLE_STEP			= $80000004;
  EXCEPTION_ACCESS_VIOLATION		= $C0000005;
  EXCEPTION_IN_PAGE_ERROR		= $C0000006;
  EXCEPTION_NO_MEMORY			= $C0000017;
  EXCEPTION_ILLEGAL_INSTRUCTION		= $C000001D;
  EXCEPTION_ARRAY_BOUNDS_EXCEEDED	= $C000008C;
  EXCEPTION_FLT_DENORMAL_OPERAND	= $C000008D;
  EXCEPTION_FLT_DIVIDE_BY_ZERO		= $C000008E;
  EXCEPTION_FLT_INEXACT_RESULT		= $C000008F;
  EXCEPTION_FLT_INVALID_OPERATION	= $C0000090;
  EXCEPTION_FLT_OVERFLOW		= $C0000091;
  EXCEPTION_FLT_STACK_CHECK		= $C0000092;
  EXCEPTION_FLT_UNDERFLOW		= $C0000093;
  EXCEPTION_INT_DIVIDE_BY_ZERO		= $C0000094;
  EXCEPTION_INT_OVERFLOW		= $C0000095;
  EXCEPTION_PRIV_INSTRUCTION		= $C0000096;
  EXCEPTION_STACK_OVERFLOW		= $C00000FD;
  EXCEPTION_FLOAT_MULTIPLE_FAULTS	= $C00002B4;
  EXCEPTION_FLOAT_MULTIPLE_TRAPS	= $C00002B5;
  EXCEPTION_STACK_BUFFER_OVERRUN	= $C0000409;
  STATUS_FATAL_USER_CALLBACK_EXCEPTION	= $C000041D;

const // see System unit
  EXCEPTION_CPP_EXCEPTION		= $0EEDFACE;
  EXCEPTION_DELPHI_EXCEPTION		= $0EEDFADE;

type
  PExceptionRecord = ^TExceptionRecord;
  TExceptionRecord = Record
    ExceptionCode: DWord;
    ExceptionFlags: DWord;
    ExceptionRecord: PExceptionRecord;
    ExceptionAddress: Pointer;
    NumberParameters: DWord;
    {$IFDEF WIN64}
    __unusedAlignment: DWord;
    {$ENDIF}
    case {IsOsException:} Boolean of
      True: (ExceptionInformation: Array[0..EXCEPTION_MAXIMUM_PARAMETERS - 1] of ULONG_PTR);
      False: (ExceptAddr: Pointer; ExceptObject: Pointer);
  end;

  PExceptionPointers = ^TExceptionPointers;
  TExceptionPointers = Record
    ExceptionRecord : PExceptionRecord;
    ContextRecord : PContext;
  end;

var
  DebugActiveProcessStop: function(dwProcessId: DWord): Bool; stdcall;
  DebugSetProcessKillOnExit: function(KillOnExit: Bool): Bool; stdcall;

function GetExceptionCode(const ExceptionInfo: TExceptionPointers; var Name: String): Cardinal;
function WriteMiniDump(ParentWindow, ProcessHandle, ProcessId, ThreadId, FileHandle, LangOffset: Cardinal; const ProcessFile: String;
		       const DebugString: String; ExceptionInfo: PExceptionPointers; NoProgressDialog: Longbool): Longbool;
function RemoveBreakpoint(Process, Thread: THandle; const ExceptionInfo: TExceptionPointers; const SymbolFilePath: String): Integer;

implementation

const
  MAX_SYM_NAME = 2000;

const
  SYMOPT_CASE_INSENSITIVE           = $00000001;
  SYMOPT_UNDNAME                    = $00000002;
  SYMOPT_DEFERRED_LOADS             = $00000004;
  SYMOPT_NO_CPP                     = $00000008;
  SYMOPT_LOAD_LINES                 = $00000010;
  SYMOPT_OMAP_FIND_NEAREST          = $00000020;
  SYMOPT_LOAD_ANYTHING              = $00000040;
  SYMOPT_IGNORE_CVREC               = $00000080;
  SYMOPT_NO_UNQUALIFIED_LOADS       = $00000100;
  SYMOPT_FAIL_CRITICAL_ERRORS       = $00000200;
  SYMOPT_EXACT_SYMBOLS              = $00000400;
  SYMOPT_ALLOW_ABSOLUTE_SYMBOLS     = $00000800;
  SYMOPT_IGNORE_NT_SYMPATH          = $00001000;
  SYMOPT_INCLUDE_32BIT_MODULES      = $00002000;
  SYMOPT_PUBLICS_ONLY               = $00004000;
  SYMOPT_NO_PUBLICS                 = $00008000;
  SYMOPT_AUTO_PUBLICS               = $00010000;
  SYMOPT_NO_IMAGE_SEARCH            = $00020000;
  SYMOPT_SECURE                     = $00040000;
  SYMOPT_NO_PROMPTS                 = $00080000;
  SYMOPT_OVERWRITE                  = $00100000;
  SYMOPT_IGNORE_IMAGEDIR            = $00200000;
  SYMOPT_FLAT_DIRECTORY             = $00400000;
  SYMOPT_FAVOR_COMPRESSED           = $00800000;
  SYMOPT_ALLOW_ZERO_ADDRESS         = $01000000;
  SYMOPT_DISABLE_SYMSRV_AUTODETECT  = $02000000;
  SYMOPT_READONLY_CACHE             = $04000000;
  SYMOPT_SYMPATH_LAST               = $08000000;
  SYMOPT_DISABLE_FAST_SYMBOLS       = $10000000;
  SYMOPT_DISABLE_SYMSRV_TIMEOUT     = $20000000;
  SYMOPT_DISABLE_SRVSTAR_ON_STARTUP = $40000000;
  SYMOPT_DEBUG                      = $80000000;

type
  DWord64 = UInt64;
  ULong64 = UInt64;
  ULong32 = UInt32;

type
  TAddressMode = (
    AddrMode1616,
    AddrMode1632,
    AddrModeReal,
    AddrModeFlat
  );

type
  PAddress64 = ^TAddress64;
  TAddress64 = Record
    Offset: DWord64;
    Segment: Word;
    Mode: TAddressMode;
  end;

type
  PKdHelp64 = ^TKdHelp64;
  TKdHelp64 = Record
    Thread: DWord64;
    ThCallbackStack: DWord;
    ThCallbackBStore: DWord;
    NextCallback: DWord;
    FramePointer: DWord;
    KiCallUserMode: DWord64;
    KeUserCallbackDispatcher: DWord64;
    SystemRangeStart: DWord64;
    KiUserExceptionDispatcher: DWord64;
    StackBase: DWord64;
    StackLimit: DWord64;
    Reserved: Array[0..4] of DWord64;
  end;

type
  PStackFrame64 = ^TStackFrame64;
  TStackFrame64 = Record
    AddrPC: TAddress64;
    AddrReturn: TAddress64;
    AddrFrame: TAddress64;
    AddrStack: TAddress64;
    AddrBStore: TAddress64;
    FuncTableEntry: Pointer;
    Params: Array[0..3] of DWord64;
    IsFar: Bool;
    IsVirtual: Bool;
    Reserved: Array[0..2] of DWord64;
    KdHelp: TKdHelp64;
  end;

type
  TSymbolInfo = Record
    SizeOfStruct: ULong;
    TypeIndex: ULong;
    Reserved: Array[0..1] of ULong64;
    Index: ULong;
    Size: ULong;
    ModBase: ULong64;
    Flags: ULong;
    Value: ULong64;
    Address: ULong64;
    Register: ULong;
    Scope: ULong;
    Tag: ULong;
    NameLen: ULong;
    MaxNameLen: ULong;
    Name: Array[0..MAX_SYM_NAME] of Char;
  end;

type
  TMiniDumpType = (
    MiniDumpNormal				= $00000000,
    MiniDumpWithDataSegs			= $00000001,
    MiniDumpWithFullMemory			= $00000002,
    MiniDumpWithHandleData			= $00000004,
    MiniDumpFilterMemory			= $00000008,
    MiniDumpScanMemory				= $00000010,
    MiniDumpWithUnloadedModules			= $00000020,
    MiniDumpWithIndirectlyReferencedMemory	= $00000040,
    MiniDumpFilterModulePaths			= $00000080,
    MiniDumpWithProcessThreadData		= $00000100,
    MiniDumpWithPrivateReadWriteMemory		= $00000200,
    MiniDumpWithoutOptionalData			= $00000400,
    MiniDumpWithFullMemoryInfo			= $00000800,
    MiniDumpWithThreadInfo			= $00001000,
    MiniDumpWithCodeSegs			= $00002000,
    MiniDumpWithoutAuxiliaryState		= $00004000,
    MiniDumpWithFullAuxiliaryState		= $00008000,
    MiniDumpWithPrivateWriteCopyMemory		= $00010000,
    MiniDumpIgnoreInaccessibleMemory		= $00020000,
    MiniDumpWithTokenInformation		= $00040000,
    MiniDumpWithModuleHeaders			= $00080000,
    MiniDumpFilterTriage			= $00100000,
    MiniDumpWithAvxXStateContext		= $00200000,
    MiniDumpWithIptTrace			= $00400000,
    MiniDumpValidTypeFlags			= $0007FFFF
  );

  PMinidumpExceptionInformation = ^TMinidumpExceptionInformation;
  TMinidumpExceptionInformation = Record
    ThreadId: DWord;
    ExceptionPointers: PExceptionPointers;
    ClientPointers: Bool;
  end;

type
  TMinidumpStreamType = (

    UnusedStream		=  0,
    ReservedStream0		=  1,
    ReservedStream1		=  2,

    ThreadListStream		=  3,
    ModuleListStream		=  4,
    MemoryListStream		=  5,
    ExceptionStream		=  6,
    SystemInfoStream		=  7,
    ThreadExListStream		=  8,
    Memory64ListStream		=  9,

    CommentStreamA		= 10,
    CommentStreamW		= 11,

    HandleDataStream		= 12,
    FunctionTableStream		= 13,
    UnloadedModuleListStream	= 14,
    MiscInfoStream		= 15,
    MemoryInfoListStream	= 16,
    ThreadInfoListStream	= 17,
    HandleOperationListStream	= 18,
    TokenStream			= 19,
    JavaScriptDataStream	= 20,
    SystemMemoryInfoStream      = 21,
    ProcessVmCountersStream     = 22,
    IptTraceStream              = 23,
    ThreadNamesStream           = 24,

    ceStreamNull		= $8000,
    ceStreamSystemInfo		= $8001,
    ceStreamException		= $8002,
    ceStreamModuleList		= $8003,
    ceStreamProcessList		= $8004,
    ceStreamThreadList		= $8005,
    ceStreamThreadContextList	= $8006,
    ceStreamThreadCallStackList	= $8007,
    ceStreamMemoryVirtualList	= $8008,
    ceStreamMemoryPhysicalList	= $8009,
    ceStreamBucketParameters	= $800A,
    ceStreamProcessModuleMap	= $800B,
    ceStreamDiagnosisList	= $800C,

    LastReservedStream		= $FFFF
  );

  PMinidumpUserStream = ^TMinidumpUserStream;
  TMinidumpUserStream = Record
    StreamType: TMinidumpStreamType;
    BufferSize: ULong;
    Buffer: Pointer
  end;

  PMinidumpUserStreamInformation = ^TMinidumpUserStreamInformation;
  TMinidumpUserStreamInformation = Record
    UserStreamCount: ULong;
    UserStreamArray: PMinidumpUserStream;
  end;

type
  TMinidumpCallbackType = (
    ModuleCallback,
    ThreadCallback,
    ThreadExCallback,
    IncludeThreadCallback,
    IncludeModuleCallback,
    MemoryCallback,
    CancelCallback,
    WriteKernelMinidumpCallback,
    KernelMinidumpStatusCallback,
    RemoveMemoryCallback,
    IncludeVmRegionCallback,
    IoStartCallback,
    IoWriteAllCallback,
    IoFinishCallback,
    ReadMemoryFailureCallback,
    SecondaryFlagsCallback,
    IsProcessSnapshotCallback,
    VmStartCallback,
    VmQueryCallback,
    VmPreReadCallback,
    VmPostReadCallback
  );

  PMiniDumpThreadCallback = ^TMiniDumpThreadCallback;
  TMiniDumpThreadCallback = Packed Record
    ThreadId: ULong;
    ThreadHandle: THandle;
    Context: TContext;
    SizeOfContext: ULong;
    StackBase: ULong64;
    StackEnd: ULong64;
  end;

  PMiniDumpThreadExCallback = ^TMiniDumpThreadExCallback;
  TMiniDumpThreadExCallback = Packed Record
    ThreadId: ULong;
    ThreadHandle: THandle;
    Context: TContext;
    SizeOfContext: ULong;
    StackBase: ULong64;
    StackEnd: ULong64;
    BackingStoreBase: ULong64;
    BackingStoreEnd: ULong64;
  end;

  PMinidumpModuleCallback = ^TMinidumpModuleCallback;
  TMinidumpModuleCallback = Packed Record
    FullPath: PWChar;
    BaseOfImage: ULong64;
    SizeOfImage: ULong;
    CheckSum: ULong;
    TimeDateStamp: ULong;
    VersionInfo: TVSFixedFileInfo;
    CvRecord: Pointer;
    SizeOfCvRecord: ULong;
    MiscRecord: Pointer;
    SizeOfMiscRecord: ULong;
  end;

  PMinidumpIncludeThreadCallback = ^TMinidumpIncludeThreadCallback;
  TMinidumpIncludeThreadCallback = Packed Record
    ThreadId: ULong;
  end;

  PMinidumpIncludeModuleCallback = ^TMinidumpIncludeModuleCallback;
  TMinidumpIncludeModuleCallback = Packed Record
    BaseOfImage: ULong64;
  end;

  PMinidumpIoCallback = ^TMinidumpIoCallback;
  TMinidumpIoCallback = Packed Record
    Handle: THandle;
    Offset: ULong64;
    Buffer: Pointer;
    BufferBytes: ULong;
  end;

  PMinidumpReadMemoryFailureCallback = ^TMinidumpReadMemoryFailureCallback;
  TMinidumpReadMemoryFailureCallback = Packed Record
    Offset: ULong64;
    Bytes: ULong;
    FailureStatus: HResult;
  end;

  PMinidumpCallbackInput = ^TMinidumpCallbackInput;
  TMinidumpCallbackInput = Packed Record
    ProcessId: ULong;
    ProcessHandle: THandle;
    CallbackType: ULong;
    CallbackParam: Packed Record case Cardinal of
      0: (Status: HResult);
      1: (Thread: TMinidumpThreadCallback);
      2: (ThreadEx: TMinidumpThreadExCallback);
      3: (Module: TMinidumpModuleCallback);
      4: (IncludeThread: TMinidumpIncludeThreadCallback);
      5: (IncludeModule: TMinidumpIncludeModuleCallback);
      6: (Io: TMinidumpIoCallback);
      7: (ReadMemoryFailure: TMinidumpReadMemoryFailureCallback);
      8: (SecondaryFlags: ULong);
    end;
  end;

type
  PMinidumpMemoryInfo = ^TMinidumpMemoryInfo;
  TMinidumpMemoryInfo = Packed Record
    BaseAddress: ULong64;
    AllocationBase: ULong64;
    AllocationProtect: ULong32;
    __alignment1: ULong32;
    RegionSize: ULong64;
    State: ULong32;
    Protect: ULong32;
    PageType: ULong32;
    __alignment2: ULong32;
  end;

  PMinidumpCallbackOutput = ^TMinidumpCallbackOutput;
  TMinidumpCallbackOutput = Packed Record case Cardinal of
    0: (ModuleWriteFlags: ULong);
    1: (ThreadWriteFlags: ULong);
    2: (SecondaryFlags: ULong);
    3: (MemoryParams: Packed Record
      MemoryBase: ULong64;
      MemorySize: ULong;
    end);
    4: (CancelParams: Packed Record
      CheckCancel: Bool;
      Cancel: Bool;
    end);
    5: (Handle: THandle);
    6: (VmParams: Packed Record
      VmRegion: TMinidumpMemoryInfo;
      Continue: Bool;
    end);
    7: (Status: HResult);
  end;

type
  PMiniDumpCallback = ^TMiniDumpCallback;
  TMiniDumpCallback = function(CallbackParam: Pointer; const CallbackInput: TMinidumpCallbackInput;
			       var CallbackOutput: TMinidumpCallbackOutput): Bool; stdcall;

  PMinidumpCallbackInformation = ^TMinidumpCallbackInformation;
  TMinidumpCallbackInformation = Packed Record
    CallbackRoutine: TMiniDumpCallback;
    CallbackParam: Pointer;
  end;

var
  MiniDumpWriteDump: function(ProcessHandle: THandle; ProcessId: DWord; FileHandle: THandle; MiniDumpType: TMiniDumpType;
			      ExceptionParam: PMinidumpExceptionInformation; UserStreamParam: PMinidumpUserStreamInformation;
			      const CallbackParam: TMinidumpCallbackInformation): Bool; stdcall;
  SymGetOptions: function: DWord; stdcall;
  SymSetOptions: function(SymOptions: DWord): DWord; stdcall;
  SymInitialize: function(hProcess: THandle; UserSearchPath: PChar; fInvadeProcess: Bool): Bool; stdcall;
  SymCleanup: function(hProcess: THandle): Bool; stdcall;
  SymFromAddr: function(hProcess: THandle; dwAddr: DWord64; var dwDisplacement: DWord64;
			var Symbol: TSymbolInfo): Bool; stdcall;
  SymFunctionTableAccess64: function(hProcess: THandle; AddrBase: DWord64): Pointer; stdcall;
  SymGetModuleBase64: function(hProcess: THandle; dwAddr: DWord64): DWord; stdcall;
  StackWalk64: function(MachineType: DWord; hProcess: THandle; hThread: THandle;
			var StackFrame: TStackFrame64; var ContextRecord: TContext;
			ReadMemoryRoutine64: TFarProc;
			FunctionTableAccessRoutine64: TFarProc;
			GetModuleBaseRoutine64: TFarProc;
			TranslateAddress64: TFarProc): Bool; stdcall;


// GetExceptionCode section ------------------------------------------------------------------------

function GetExceptionCode(const ExceptionInfo: TExceptionPointers; var Name: String): Cardinal;
begin
  Result := ExceptionInfo.ExceptionRecord.ExceptionCode;
  case Result of
    EXCEPTION_GUARD_PAGE_VIOLATION:	Name := '(guard page violation)';
    EXCEPTION_DATATYPE_MISALIGNMENT:	Name := '(datatype misalignment)';
    EXCEPTION_BREAKPOINT:		Name := '(breakpoint)';
    EXCEPTION_SINGLE_STEP:		Name := '(single step)';
    EXCEPTION_ACCESS_VIOLATION:		Name := '(access violation)';
    EXCEPTION_IN_PAGE_ERROR:		Name := '(in page error)';
    EXCEPTION_NO_MEMORY:		Name := '(no memory)';
    EXCEPTION_ILLEGAL_INSTRUCTION:	Name := '(illegal instruction)';
    EXCEPTION_ARRAY_BOUNDS_EXCEEDED:	Name := '(array bounds exceeded)';
    EXCEPTION_FLT_DENORMAL_OPERAND:	Name := '(float denormal operand)';
    EXCEPTION_FLT_DIVIDE_BY_ZERO:	Name := '(float divide by zero)';
    EXCEPTION_FLT_INEXACT_RESULT:	Name := '(float inexact result)';
    EXCEPTION_FLT_INVALID_OPERATION:	Name := '(float invalid operation)';
    EXCEPTION_FLT_OVERFLOW:		Name := '(float overflow)';
    EXCEPTION_FLT_STACK_CHECK:		Name := '(float stack check)';
    EXCEPTION_FLT_UNDERFLOW:		Name := '(float underflow)';
    EXCEPTION_INT_DIVIDE_BY_ZERO:	Name := '(integer divide by zero)';
    EXCEPTION_INT_OVERFLOW:		Name := '(integer overflow)';
    EXCEPTION_PRIV_INSTRUCTION:		Name := '(private instruction)';
    EXCEPTION_STACK_OVERFLOW:		Name := '(stack overflow)';
    EXCEPTION_FLOAT_MULTIPLE_FAULTS:	Name := '(float multiple faults)';
    EXCEPTION_FLOAT_MULTIPLE_TRAPS:	Name := '(float multiple traps)';
    EXCEPTION_STACK_BUFFER_OVERRUN:	Name := '(stack buffer overrun)';
  else
    Name := LB + '0x' + IntToHex(Result, 8) + RB;
  end;
end;


// WriteMiniDump section ---------------------------------------------------------------------------

type
  PDialogParams = ^TDialogParams;
  TDialogParams = Record
    ParentHandle: THandle;
    LangOffset: Integer;
  end;

function MiniDumpCallback(CallbackParam: Pointer; const CallbackInput: TMinidumpCallbackInput;
			  var CallbackOutput: TMinidumpCallbackOutput): Bool; stdcall;
var
  dwWritten: DWord;
begin
  Result := TRUE;
  case TMinidumpCallbackType(CallbackInput.CallbackType) of
    IoStartCallback: with PDialogParams(CallbackParam)^ do begin
      ShowProgressDialog(ParentHandle, LangOffset);
      CallbackOutput.Status := S_FALSE;
    end;
    IoWriteAllCallback: with CallbackInput.CallbackParam.Io do begin
      SetFilePointer(Handle, Int64Rec(Offset).Lo, NIL, FILE_BEGIN);
      WriteFile(Handle, Buffer^, BufferBytes, dwWritten, NIL);
      CallbackOutput.Status := S_OK;
    end;
    IoFinishCallback: begin
      CallbackOutput.Status := S_OK;
      HideProgressDialog;
    end;
  end;
  ProcessMessages;
end;

function WriteMiniDump(ParentWindow, ProcessHandle, ProcessId, ThreadId, FileHandle, LangOffset: Cardinal; const ProcessFile: String;
		       const DebugString: String; ExceptionInfo: PExceptionPointers; NoProgressDialog: Longbool): Longbool;
var
  dp: TDialogParams;
  mci: TMinidumpCallbackInformation;
  musi: TMinidumpUserStreamInformation;
  mus: TMinidumpUserStream;
  mei: TMinidumpExceptionInformation;
begin
  if ParentWindow = 0 then ParentWindow := FindProcessWindow(ProcessId, ProcessFile, ThreadId);
  dp.ParentHandle := ParentWindow;
  dp.LangOffset := LangOffset;
  if not NoProgressDialog then begin
    mci.CallbackRoutine := MiniDumpCallback;
    mci.CallbackParam := @dp;
  end
  else begin
    mci.CallbackRoutine := NIL;
    mci.CallbackParam := NIL;
  end;
  if DebugString <> E then begin
    musi.UserStreamCount := 1;
    musi.UserStreamArray := @mus;
    mus.StreamType := CommentStreamW;
    mus.BufferSize := succ(Length(DebugString)) * SizeOf(Char);
    mus.Buffer := Pointer(DebugString);
  end
  else FillChar(musi, SizeOf(TMinidumpUserStreamInformation), 0);
  if ExceptionInfo <> NIL then begin
    mei.ThreadID := ThreadId;
    mei.ExceptionPointers := ExceptionInfo;
    mei.ClientPointers := FALSE;
    Result := MiniDumpWriteDump(ProcessHandle, ProcessId, FileHandle, MiniDumpWithIndirectlyReferencedMemory, @mei, @musi, mci);
  end
  else Result := MiniDumpWriteDump(ProcessHandle, ProcessId, FileHandle, MiniDumpWithIndirectlyReferencedMemory, NIL, @musi, mci);
end;


// RemoveBreakpoint section ------------------------------------------------------------------------

type
  PFrame64 = ^TFrame64;
  TFrame64 = Record
    EIP: DWord64;
    ESP: DWord64;
    EBP: DWord64;
    ReturnAddress: DWord64;
  end;
  TStackData = Class(TList)
  private
    function GetItem(Index: Integer): PFrame64;
  public
    function Add(const StackFrame: TStackFrame64): Integer; reintroduce; overload;
    procedure Insert(Index: Integer; const Frame: TFrame64); reintroduce; overload;
    procedure Clear; override;
    function FindFrame(Address: Pointer): Integer;
    property Items[Index: Integer]: PFrame64 read GetItem; default;
  end;

function TStackData.GetItem(Index: Integer): PFrame64;
begin
  Result := PFrame64(inherited Items[Index]);
end;

function TStackData.Add(const StackFrame: TStackFrame64): Integer;
var
  pf: PFrame64;
begin
  New(pf);
  with pf^, StackFrame do begin
    EIP := AddrPC.Offset;
    ESP := AddrStack.Offset;
    EBP := AddrFrame.Offset;
    ReturnAddress := AddrReturn.Offset;
  end;
  Result := inherited Add(pf);
end;

procedure TStackData.Insert(Index: Integer; const Frame: TFrame64);
var
  pf: PFrame64;
begin
  New(pf);
  pf^ := Frame;
  inherited Insert(Index, pf);
end;

procedure TStackData.Clear;
var
  iIndex: Integer;
begin
  for iIndex := 0 to pred(Count) do if Items[iIndex] <> NIL then Dispose(Items[iIndex]);
  inherited;
end;

function TStackData.FindFrame(Address: Pointer): Integer;
var
  iIndex: Integer;
begin
  Result := -1;
  for iIndex := 0 to pred(Count) do begin
    if UInt64(Address) = Items[iIndex]^.EIP then begin
      Result := iIndex;
      Break;
    end;
  end;
end;

function RemoveDebugDialog(Process, Thread: THandle; const ExceptionInfo: TExceptionPointers; const SymbolFilePath: String): Longbool;

  function GetFrameSymbol(Process: THandle; Frames: TStackData; Index: Integer): String;
  var
    si: TSymbolInfo;
    dwDisplacement: DWord64;
  begin
    with si do begin
      SizeOfStruct := SizeOf(TSymbolInfo);
      MaxNameLen := MAX_SYM_NAME;
    end;
    if SymFromAddr(Process, Frames[Index]^.EIP, dwDisplacement, si) then
      Result := si.name
    else Result := E;
  end;

type
  TFarCall = Array[0..4] of Byte;

const
  DBG_BREAK_POINT = 'DbgBreakPoint';
  FAIL_WITH_MSG = 'failwithmessage';
  RTC = '_RTC_';
  NOP = $90;
  FC = $E8;
var
  sd: TStackData;
  ctxt: TContext;
  sf: TStackFrame64;
  {$IF CompilerVersion > 21.0}
  dwSize: NativeUInt;
  {$ELSE]}
  dwSize: DWord;
  {$IFEND}
  iIndex: Integer;
  dwAddr: DWord64;
  aCode: TFarCall;
begin
  Result := FALSE;
  sd := TStackData.Create;
  if sd <> NIL then begin
    try
      SymSetOptions(SymGetOptions or SYMOPT_DEFERRED_LOADS);
      if SymInitialize(Process, PChar(SymbolFilePath), TRUE) then begin
	try
	  ctxt := ExceptionInfo.ContextRecord^;
	  FillChar(sf, SizeOf(TStackFrame64), 0);
	  with sf do begin
	    with AddrPC do begin
	      Offset := ctxt.Eip;
	      Mode := AddrModeFlat;
	    end;
	    with AddrFrame do begin
	      Offset := ctxt.Ebp;
	      Mode := AddrModeFlat;
	    end;
	    with AddrStack do begin
	      Offset := ctxt.Esp;
	      Mode := AddrModeFlat;
	    end;
	  end;
	  with sf do begin
	    while StackWalk64(IMAGE_FILE_MACHINE_I386, Process, Thread, sf, ctxt, NIL,
			      @SymFunctionTableAccess64, @SymGetModuleBase64, NIL) do begin
	      if (AddrPC.Offset <> 0) and (AddrPC.Offset <> AddrReturn.Offset) then sd.Add(sf);
	    end;
	  end;
	  if sd.Count > 0 then begin
	    if (sd[0]^.ESP <> 0) and (GetFrameSymbol(Process, sd, 0) = DBG_BREAK_POINT) then begin
	      sd.Insert(1, sd[0]^);
	      if ReadProcessMemory(Process, Pointer(sd[0]^.ESP), @sd[1]^.EIP, SizeOf(Pointer), dwSize)
	      and (GetFrameSymbol(Process, sd, 1) = FAIL_WITH_MSG) then begin
		iIndex := 2;
		while (iIndex < sd.Count) and StrTools.StartsStr(RTC, GetFrameSymbol(Process, sd, iIndex)) do inc(iIndex);
		if (iIndex > 2) and (iIndex < sd.Count) then begin
		  dwAddr := sd[iIndex]^.EIP - Length(aCode);
		  if ReadProcessMemory(Process, Pointer(dwAddr), @aCode, SizeOf(TFarCall), dwSize)
		  and (aCode[0] = FC) then begin
		    FillChar(aCode, SizeOf(TFarCall), NOP);
		    if WriteProcessMemory(Process, Pointer(dwAddr), @aCode, SizeOf(TFarCall), dwSize)
		    and FlushInstructionCache(Process, Pointer(dwAddr), dwSize) then begin
		      Result := TRUE;
		    end;
		  end;
		end;
	      end;
	    end;
	  end;
	finally
	  SymCleanup(Process);
	end;
      end;
    finally
      sd.Free;
    end;
  end;
end;

function RemoveBreakpoint(Process, Thread: THandle; const ExceptionInfo: TExceptionPointers; const SymbolFilePath: String): Integer;
const
  NOP: Byte = $90;
var
  {$IF CompilerVersion > 21.0}
  dwSize: NativeUInt;
  {$ELSE]}
  dwSize: DWord;
  {$IFEND}
begin
  Result := EXCEPTION_EXECUTE_HANDLER;
  with ExceptionInfo.ExceptionRecord^ do begin
    if (ExceptionCode = EXCEPTION_BREAKPOINT) and (ExceptionAddress <> NIL) then begin
      RemoveDebugDialog(Process, Thread, ExceptionInfo, SymbolFilePath);
      if WriteProcessMemory(Process, ExceptionAddress, @NOP, SizeOf(NOP), dwSize)
      and FlushInstructionCache(Process, ExceptionAddress, dwSize) then begin
	Result := EXCEPTION_CONTINUE_EXECUTION;
      end;
    end;
  end;
end;


// Initialization section --------------------------------------------------------------------------

function InitDbgHelp: THandle;
const
  DBGHELP = 'DbgHelp.dll';
var
  hKernel: THandle;
begin
  hKernel := GetOrLoadLibrary(KERNEL32);
  Result := LoadLibrary(PChar(GetMiniDumpInstallPath + DBGHELP));
  if Result = 0 then Result := LoadLibrary(PChar(GetSysDir + DBGHELP));
  DebugActiveProcessStop := GetProcAddress(hKernel, 'DebugActiveProcessStop');
  DebugSetProcessKillOnExit := GetProcAddress(hKernel, 'DebugSetProcessKillOnExit');
  MiniDumpWriteDump := GetProcAddress(Result, 'MiniDumpWriteDump');
  SymGetOptions := GetProcAddress(Result, 'SymGetOptions');
  SymSetOptions := GetProcAddress(Result, 'SymSetOptions');
  SymInitialize := GetProcAddress(Result, 'SymInitializeW');
  SymCleanup := GetProcAddress(Result, 'SymCleanup');
  SymFromAddr := GetProcAddress(Result, 'SymFromAddrW');
  SymFunctionTableAccess64 := GetProcAddress(Result, 'SymFunctionTableAccess64');
  SymGetModuleBase64 := GetProcAddress(Result, 'SymGetModuleBase64');
  StackWalk64 := GetProcAddress(Result, 'StackWalk64');
end;


begin
  InitDbgHelp;
end.

