{------------------------------------------------------------------------------}
{        This code is copyright (C) 2008-2025 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit DlgCrashReport;

interface

uses
  Windows, Messages, StrTools, WindowTools, DlgMsgBoxes, MDResources;

type
  TDialogData = Record
    ParentHandle: THandle;
    LangOffset: Integer;
    HelpText: String;
    HelpCaption: String;
    ExceptionText: String;
    PMailAddressText: PString;
    PDescriptionText: PString;
  end;

function ShowReportDialog(var DialogData: TDialogData): Integer;

implementation

const
  MsgFilterHook: THandle = 0;
  DlgData: TDialogData = (
    ParentHandle : 0;
    LangOffset: 0;
    HelpText: E;
    HelpCaption: E;
    ExceptionText: E;
    PMailAddressText: NIL;
    PDescriptionText: NIL;
  );

function MsgFilterProc(nCode: Integer; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
begin
  if nCode = MSGF_DIALOGBOX then with PMsg(lParam)^ do begin
    case Message of
      WM_KEYDOWN: begin
	case wParam of
	  VK_ESCAPE: PostMessage(GetParent(hWnd), WM_SYSCOMMAND, SC_CLOSE, 0);
	  Ord('A'): if (GetKeyState(VK_CONTROL) < 0) and (ClassString(hWnd) = CLASS_EDIT) then begin
	    SendMessage(hWnd, EM_SETSEL, 0, -1);
	    Result := -1;
	    Exit;
	  end;
	end;
      end;
    end;
  end;
  Result := CallNextHookEx(MsgFilterHook, nCode, wParam, lParam);
end;

function ReportDialogProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
begin
  Result := 0;
  case uMsg of
    WM_DESTROY: with DlgData do begin
      UnhookWindowsHookEx(MsgFilterHook);
      PMailAddressText^ := WindowStringW(GetDlgItem(hWnd, EDIT_ADDRESS));
      PDescriptionText^ := WindowStringW(GetDlgItem(hWnd, EDIT_DESCRIPTION));
    end;
    WM_SHOWWINDOW: if wParam <> 0 then SetDialogForeground(hWnd, DlgData.ParentHandle);
    WM_HELP, WM_CONTEXTMENU: with DlgData do ShowHelpDialog(hWnd, HelpText, HelpCaption);
    WM_INITDIALOG: begin
      if lParam > 0 then begin
	SetControlText(hWnd, lParam, LBL_ADDRESS, E);
	SetControlText(hWnd, lParam, EDIT_ADDRESS, DlgData.PMailAddressText^);
	SetControlText(hWnd, lParam, LBL_DESCRIPTION, E);
	SetControlText(hWnd, lParam, EDIT_DESCRIPTION, E);
	SetControlText(hWnd, lParam, BTN_SAVE_REPORT, E);
	SetControlText(hWnd, lParam, BTN_SEND_REPORT, E);
	SetControlText(hWnd, lParam, BTN_CANCEL_REPORT, E);
      end;
      MessageBeep(MB_ICONERROR);
      SetFocus(GetDlgItem(hWnd, EDIT_DESCRIPTION));
      SendMessage(GetDlgItem(hWnd, EDIT_DESCRIPTION), EM_SETSEL, 0, -1);
      SetWindowText(hWnd, PChar(WindowStringW(hWnd) + SP + DlgData.ExceptionText));
      MsgFilterHook := SetWindowsHookEx(WH_MSGFILTER, @MsgFilterProc, 0, GetCurrentThreadID);
    end;
    WM_COMMAND: if (lParam <> 0) and (HiWord(wParam) = BN_CLICKED) then begin
      case LoWord(wParam) of
	BTN_SAVE_REPORT: Result := Ord(EndDialog(hWnd, BTN_SAVE_REPORT));
	BTN_SEND_REPORT: Result := Ord(EndDialog(hWnd, BTN_SEND_REPORT));
	BTN_CANCEL_REPORT: Result := Ord(EndDialog(hWnd, BTN_CANCEL_REPORT));
      end;
    end;
    WM_SYSCOMMAND: begin
      case wParam of
	SC_CLOSE: Result := Ord(EndDialog(hWnd, BTN_CANCEL_REPORT));
	SC_CONTEXTHELP: with DlgData do ShowHelpDialog(hWnd, HelpText, HelpCaption);
      end;
    end;
  end;
end;

function ShowReportDialog(var DialogData: TDialogData): Integer;
begin
  try
    DlgData := DialogData;
    Result := DialogBoxParam(HInstance, MakeIntResource(DLG_DUMP_REPORT), DlgData.ParentHandle, @ReportDialogProc, DlgData.LangOffset);
  except
    Result := 0;
  end;
end;

end.

