{------------------------------------------------------------------------------}
{        This code is copyright (C) 2008-2025 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}
{$J+,Z4}
program MDSwitch;

{$R 'MDSwitch.res' 'MDSwitch.rc'}

uses
  Windows, SysUtils, StrUtils, StrTools, FileTools, PMMessages, PMOptions, PMTools, PMInterface, MDResources, DlgMsgBoxes;

type
  TSuccess = (sNone, sEnabled, sDisabled, sUndefined);

function RenameFormFactFiles(const Directory: String): TSuccess;
const
  MD_FILTER  = 'MiniDump.*';
var
  hSearch: THandle;
  fd: TWin32FindData;
  pExt, pNew: PChar;
  sFileName: String;
begin
  Result := sNone;
  hSearch := FindFirstFile(PChar(Directory + MD_FILTER), fd);
  if hSearch <> INVALID_HANDLE_VALUE then begin
    try
      repeat
	if (fd.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) = 0 then begin
	  pExt := StrTools.StrRScan(fd.cFileName, DOT);
	  if (pExt <> NIL) and (StrTools.StrLen(pExt) > 3) then begin
	    if (pExt[1] = TLD) and (Upcase(pExt[2]) = 'F') then begin
	      sFileName := fd.cFileName;
	      inc(pExt);
	      pNew := pExt;
	      inc(pExt);
	      StrTools.StrMove(pNew, pExt, succ(StrTools.StrLen(pExt)));
	      if MoveFile(Directory + sFileName, Directory + fd.cFileName, FALSE) then Result := sEnabled;
	    end
	    else if (Upcase(pExt[1]) = 'F') then begin
	      sFileName := fd.cFileName;
	      pNew := pExt;
	      inc(pNew);
	      StrTools.StrMove(pNew, pExt, succ(StrTools.StrLen(pExt)));
	      pNew[0] := TLD;
	      if MoveFile(Directory + sFileName, Directory + fd.cFileName, FALSE) then Result := sDisabled;
	    end;
	  end;
	end;
      until not FindNextFile(hSearch, fd);
    finally
      Windows.FindClose(hSearch);
    end;
  end;
end;

function ShowSuccessMessage(Success: TSuccess): TSuccess;
const
  WPM32 = 'WPM32';
var
  iLen, iOffset: Integer;
  sTLD, sHint: String;
begin
  Result := Success;
  if Success <> sNone then begin
    iLen := GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_SISO639LANGNAME, NIL, 0);
    SetString(sTLD, NIL, pred(iLen));
    GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_SISO639LANGNAME, PChar(sTLD), iLen);
    GetLangIndex(WPM32 + sTLD, PMFormApi);
    iOffset := GetLangOffset;
    sHint := Str(iOffset + STR_RESTART_HINT);
    if Success = sEnabled then
      ShowInfoDialog(0, Str(iOffset + STR_ENABLED_SUCCESS) + CRLF + sHint, ExtractBaseName(ParamStr(0)))
    else ShowInfoDialog(0, Str(iOffset + STR_DISABLED_SUCCESS) + CRLF + sHint, ExtractBaseName(ParamStr(0)));
  end;
end;

var
  BaseSuccess, UserSuccess, Result: TSuccess;
begin
  Result := sNone;
  BaseSuccess := RenameFormFactFiles(PMFormApi.BaseDir);
  UserSuccess := RenameFormFactFiles(PMFormApi.MailBoxDir);
  if BaseSuccess <> UserSuccess then begin
    if BaseSuccess = sNone then Result := UserSuccess;
    if UserSuccess = sNone then Result := BaseSuccess;
  end
  else Result := BaseSuccess;
  if (Result <> sNone) and (ShowSuccessMessage(Result) <> sNone) then Exit;
  ShowErrorDialog(0, Str(GetLangOffset + STR_FAILURE), ExtractBaseName(ParamStr(0)));
end.

