{------------------------------------------------------------------------------}
{        This code is copyright (C) 2008-2025 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit MakeCab;

interface

uses
  Windows, SysUtils, Classes, StrTools, FileTools, Processes, MDResources;

const
  CAB = '.cab';
  GZIP = '.gzip';

function CompressFile(const SourceFile, DestFile, WorkDir: String; ForegroundWindow: THandle): String;
function MakeCabinet(const SourceFiles, DestFile, DestDir: String; ForegroundWindow: THandle): String;

implementation

uses
  ZLibExGZ;

const
  MakeCabPath: String = 'MakeCab.exe';

function CompressZip(const SourceFile, DestFile, WorkDir: String): String;
var
  sCurDir, sZipFile: String;
  dwLastError: DWord;
begin
  sCurDir := GetCurDir;
  try
    SetCurDir(WorkDir);
    sZipFile := FileTools.ChangeFileExt(DestFile, GZIP);
    if FileTools.ExtractFilePath(sZipFile) = E then
      Result := WorkDir + sZipFile
    else Result := sZipFile;
    GZCompressFile(SourceFile, sZipFile);
  finally
    dwLastError := GetLastError;
    SetCurDir(sCurDir);
    if Result <> E then
      DeleteFile(WorkDir + SourceFile)
    else Result := WorkDir + SourceFile;
    SetLastError(dwLastError);
  end;
end;

function CompressFile(const SourceFile, DestFile, WorkDir: String; ForegroundWindow: THandle): String;
const
  QQ = QU + SP + QU;
var
  dwLastError: DWord;
  sCabFile: String;
begin
  SetLastError(0);
  Result := CompressZip(SourceFile, DestFile, WorkDir);
  if (Result = E) or not FileExists(Result) then begin
    try
      if DestFile <> E then
	sCabFile := DestFile
      else sCabFile := FileTools.ChangeFileExt(DestFile, CAB);
      if CreateUserProcess(QU + MakeCabPath + QQ + SourceFile + QQ + sCabFile + QU, WorkDir,
			   ForegroundWindow, SW_HIDE, cuprExitCode) <> 0 then begin
	Result := E;
	Exit;
      end;
      if FileTools.ExtractFilePath(sCabFile) = E then
	Result := WorkDir + sCabFile
      else Result := sCabFile;
    finally
      dwLastError := GetLastError;
      if Result <> E then
	DeleteFile(WorkDir + SourceFile)
      else Result := WorkDir + SourceFile;
      SetLastError(dwLastError);
    end;
  end;
end;

function MakeCabinet(const SourceFiles, DestFile, DestDir: String; ForegroundWindow: THandle): String;

  function DeleteSourceFiles(const SourceFiles: String): Integer;
  var
    sFiles: String;
    pFiles: PChar;
    iBeg, iEnd, iLen: Integer;
  begin
    Result := 0;
    if SourceFiles = E then Exit;
    sFiles := SourceFiles;
    UniqueString(sFiles);
    pFiles := Pointer(sFiles);
    iBeg := 1;
    repeat
      iEnd := iBeg;
      while not CharInSet(sFiles[iEnd], [NUL, CR]) do inc(iEnd);
      iLen := iEnd - iBeg;
      pFiles[iLen] := NUL;
      if DeleteFile(pFiles) then inc(Result);
      inc(pFiles, iLen + 2);
      iBeg := iEnd + 2;
    until sFiles[iBeg] = NUL;
  end;

const
  MC = ' /f ';
  CMD = '.cmds';
  INF = '.info';
  RPT = '.rept';
  CMDS_MAKECAB = '.set CabinetFileCountThreshold=1' + CRLF +
		 '.set FolderFileCountThreshold=0' + CRLF +
		 '.set FolderSizeThreshold=0' + CRLF +
		 '.set MaxCabinetSize=0' + CRLF +
		 '.set DiskDirectory1=%s' + CRLF +
		 '.set CabinetName1=%s' + CRLF +
		 '.set InfFileName=%s' + CRLF +
		 '.set RptFileName=%s' + CRLF;
var
  dwLastError: DWord;
  sWorkDir, sCabFile, sCmdFile, sInfFile, sRptFile, sCurDir: String;
begin
  SetLastError(0);
  if DestDir = E then begin
    sWorkDir := FileTools.ExtractFilePath(DestFile);
    sCabFile := FileTools.ExtractFileName(DestFile);
  end
  else begin
    sWorkDir := DestDir;
    sCabFile := DestFile;
  end;
  if GetLineCount(StrTools.Trim(SourceFiles)) = 1 then Result := CompressZip(StrTools.Trim(SourceFiles), DestFile, sWorkDir);
  if (Result = E) or not FileExists(Result) then begin
    dwLastError := GetLastError;
    sCmdFile := sWorkDir + FileTools.ChangeFileExt(sCabFile, CMD);
    sInfFile := FileTools.ChangeFileExt(sCabFile, INF);
    sRptFile := FileTools.ChangeFileExt(sCabFile, RPT);
    try
      if not WriteTextFile(sCmdFile, AnsiString(StrTools.Format(CMDS_MAKECAB, [sWorkDir, sCabFile, sInfFile, sRptFile]) + SourceFiles))
      or (CreateUserProcess(QU + MakeCabPath + QU + MC + sCmdFile, sWorkDir, ForegroundWindow, SW_HIDE, cuprExitCode) <> 0) then begin
	dwLastError := GetLastError;
	Result := E;
	Exit;
      end;
      Result := sWorkDir + sCabFile;
    finally
      sCurDir := GetCurDir;
      SetCurDir(sWorkDir);
      try
	DeleteFile(sCmdFile);
	DeleteFile(sInfFile);
	DeleteFile(sRptFile);
	DeleteSourceFiles(SourceFiles);
      finally
	SetCurDir(sCurDir);
      end;
      SetLastError(dwLastError);
    end;
  end;
end;

procedure InitMakeCab;
var
  sPath: String;
begin
  sPath := GetMiniDumpInstallPath;
  if FileExists(sPath + MakeCabPath) then Insert(sPath, MakeCabPath, 1);
end;

begin
  InitMakeCab;
end.

