{------------------------------------------------------------------------------}
{        This code is copyright (C) 2008-2025 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}
{$J+,Z4}
library MiniDump;

{$R 'MiniDump.res' 'MiniDump.rc'}

uses
  Windows, Messages, ActiveX, SysUtils, StrTools, FileTools, WindowTools, Processes, DbgHelp, MakeCab,
  PMTools, PMOptions, PMMessages, PMInterface, MDResources, DlgCrashReport, DlgMsgBoxes, DlgSaveDump;

const
  CallWndProcHook: THandle = 0;
  ExceptionFilter: TFarProc = NIL;
  ExtensionState: (esIdle, esFiltering, esRetrieving) = esIdle;

function CallWndProc(nCode: Integer; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
var
  hControl: THandle;
  sInfo, sFile: String;
begin
  if nCode = HC_ACTION then with PCWPStruct(lParam)^ do begin
    case Message of
      WM_SHOWWINDOW: if ExtensionState > esIdle then begin
	if (wParam <> 0) and (IsDialogWindow(hWnd)) then begin
	  if ExtensionState = esFiltering then
	    CenterDialog(hWnd, GetParent(hWnd))
	  else begin
	    hControl := GetDlgItem(hWnd, CTL_BUTTON_ABOUT_INFO);
	    if hControl <> 0 then begin
	      PostMessage(hControl, BM_CLICK, 0, 0);
	      PostMessage(GetDlgItem(hWnd, ID_OK), BM_CLICK, 0, 0);
	    end
	    else begin
	      hControl := GetDlgItem(hWnd, CTL_EDIT_ABOUT_INFO);
	      if hControl <> 0 then begin
		ExtensionState := esFiltering;
		sInfo := WindowStringW(hControl);
		if sInfo <> E then begin
		  Delete(sInfo, 1, FirstStrPos(PARA, sInfo) + pred(Length(PARA)));
		  sFile := PWideChar(GetProp(FindProcessWindow(GetCurrentProcessID, PM_CLASS_FRAME), PWideChar(GetVersionInformation)));
		  if sFile <> E then WriteTextFile(sFile, AnsiString(sInfo));
		end;
		PostMessage(GetDlgItem(hWnd, ID_OK), BM_CLICK, 0, 0);
	      end;
	    end;
	  end;
	end;
      end;
    end;
  end;
  Result := CallNextHookEx(CallWndProcHook, nCode, wParam, lParam);
end;

function MiniDumpExceptionFilter(const ExceptionInfo: TExceptionPointers): Longint; stdcall;

  function GetDebugString(ExceptionAddress: Cardinal; ExceptionObject: Exception): String;

    function GetExceptionMessage(ExceptionObject: Exception): String;
    begin
      Result := TrimCustom(GetAnyString(GetStringAddress(ExceptionObject, 'FMessage')), TRIMCHARS + [DOT]);
    end;

    function GetSafeCallError(ExceptionObject: Exception): String;
    var
      ErrorInfo: IErrorInfo;
      wsErrorMsg: WideString;
    begin
      if (ExceptionObject is ESafeCallException) and (GetErrorInfo(0, ErrorInfo) = S_OK)
      and (ErrorInfo.GetDescription(wsErrorMsg) = S_OK) and (wsErrorMsg <> E) then begin
	Result := TrimCustom(wsErrorMsg, TRIMCHARS + [DOT]);
	SetErrorInfo(0, ErrorInfo);
	Exit;
      end;
      Result := E;
    end;

  begin
    if Exception <> NIL then begin
      Result := AppendHexAddress(GetSafeCallError(ExceptionObject), Pointer(ExceptionAddress));
      if Result = E then Result := AppendHexAddress(GetExceptionMessage(ExceptionObject), Pointer(ExceptionAddress));
      Exit;
    end;
    Result := E;
  end;

  function CreateMiniDump(const EP: TExceptionPointers; const DumpFile: String; ParentWindow: THandle): Integer;
  var
    hDumpFile: THandle;
    sDebugString: String;
  begin
    Result := ERROR_SUCCESS;
    hDumpFile := CreateWriteFile(DumpFile);
    if hDumpFile <> INVALID_HANDLE_VALUE then begin
      try
	with ExceptionInfo.ExceptionRecord^ do begin
	  if (ExceptionCode = EXCEPTION_CPP_EXCEPTION) or (ExceptionCode = EXCEPTION_DELPHI_EXCEPTION) then begin
	    sDebugString := GetDebugString(ExceptionInformation[0], Exception(ExceptionInformation[1]));
	  end;
	end;
	if not WriteMiniDump(ParentWindow, GetCurrentProcess, GetCurrentProcessID, GetCurrentThreadID,
			     hDumpFile, GetLangOffset, E, sDebugString, @EP, FALSE) then begin
	  Result := GetLastError;
	end;
      except
      end;
      CloseFile(hDumpFile);
    end
    else Result := GetLastError;
  end;

  procedure Terminate(ExitCode: Cardinal);
  const
    LOCK_FILE = 'MAILBOXP.LCK';
  begin
    try
      CloseProcessFileHandles(GetCurrentProcessID, GetCurrentProcess);
      DeleteFile(PMFormApi.NewMailDir + LOCK_FILE);
    finally
      FatalExit(ExitCode);
    end;
  end;

var
  sBase, sDmpFile, sArchive, sTempPath, sSaveFile, sInfoFile, sMailAddress, sDescription, sVersionInfo: String;
  iSysError, iDlgResult: Integer;
  hParentWindow: THandle;
  dd: TDialogData;
begin
  Result := EXCEPTION_EXECUTE_HANDLER;
  ExtensionState := esFiltering;
  sBase := GetBaseDTFileName;
  sDmpFile := sBase + DMP;
  sArchive := sBase + CAB;
  sTempPath := GetTmpDirW;
  hParentWindow := FindProcessWindow(GetCurrentProcessID, PM_CLASS_FRAME);
  iSysError := CreateMiniDump(ExceptionInfo, sTempPath + sDmpFile, hParentWindow);
  if iSysError = 0 then begin
    try
      CallWndProcHook := SetWindowsHookEx(WH_CALLWNDPROC, @CallWndProc, 0, GetCurrentThreadID);
      try
	with dd do begin
	  ParentHandle := hParentWindow;
	  LangOffset := GetLangOffset;
	  GetHelpStrings(HelpText, HelpCaption);
	  GetExceptionCode(ExceptionInfo, ExceptionText);
	  PMailAddressText := @sMailAddress;
	  PDescriptionText := @sDescription;
	end;
	iDlgResult := ShowReportDialog(dd);
	if iDlgResult in [BTN_SAVE_REPORT, BTN_SEND_REPORT] then begin
	  if iSysError = 0 then begin
	    if sMailAddress = E then iDlgResult := BTN_SAVE_REPORT;
	    case iDlgResult of
	      BTN_SAVE_REPORT: begin
		try
		  sSaveFile := ShowSaveFileDialog(hParentWindow, Str(dd.LangOffset + STR_SAVE_REPORT), sDmpFile);
		  sDmpFile := sTempPath + sDmpFile;
		  if (sSaveFile = E) or not MoveFile(sDmpFile, sSaveFile, TRUE) then sSaveFile := sDmpFile;
		finally
		  ShowInfoDialog(hParentWindow, sSaveFile, Str(dd.LangOffset + STR_REPORT_SAVED));
		  Result := RemoveBreakpoint(GetCurrentProcess, GetCurrentThread, ExceptionInfo, PMFormApi.BaseDir);
		  if Result <> EXCEPTION_CONTINUE_EXECUTION then Terminate(ExceptionInfo.ExceptionRecord.ExceptionCode);
		end;
	      end;
	      BTN_SEND_REPORT: begin
		try
		  ExtensionState := esRetrieving;
		  sSaveFile := CompressFile(sDmpFile, sArchive, sTempPath, PMFormApi.FrameHandle);
		  sInfoFile := FileTools.ChangeFileExt(sSaveFile, CFG);
		  sVersionInfo := GetVersionInformation;
		  if SetProp(hParentWindow, PWideChar(sVersionInfo), THandle(sInfoFile)) then begin
		    try
		      SendMessage(hParentWindow, WM_COMMAND, IDM_ABOUT_PEGASUS_MAIL, 0);
		      ProcessMessages;
		    finally
		      RemoveProp(hParentWindow, PWideChar(sVersionInfo));
		    end;
		  end
		  else sInfoFile := E;
		  if (ExceptionInfo.ExceptionRecord.ExceptionCode = EXCEPTION_BREAKPOINT)
		  and (CreateMessage(sMailAddress, sDescription, sSaveFile, sInfoFile, CallWndProcHook <> 0) = 0) then
		    Result := RemoveBreakpoint(GetCurrentProcess, GetCurrentThread, ExceptionInfo, PMFormApi.BaseDir)
		  else iSysError := CreateReport(sMailAddress, sDescription, sSaveFile, sInfoFile, E, CallWndProcHook <> 0);
		finally
		  if Result <> EXCEPTION_CONTINUE_EXECUTION then begin
		    if iSysError = 0 then begin
		      ShowWarnDialog(hParentWindow, Str(dd.LangOffset + STR_SEND_MESSAGE_HINT),
				     Str(dd.LangOffset + STR_REPORT_QUEUED));
		    end;
		    Terminate(ExceptionInfo.ExceptionRecord.ExceptionCode);
		  end;
		end;
	      end;
	    end;
	  end;
	end
	else begin
	  Result := RemoveBreakpoint(GetCurrentProcess, GetCurrentThread, ExceptionInfo, PMFormApi.BaseDir);
	  if Result <> EXCEPTION_CONTINUE_EXECUTION then Result := EXCEPTION_CONTINUE_SEARCH;
	  DeleteFile(sTempPath + sDmpFile);
	  Exit;
	end;
      finally
	UnhookWindowsHookEx(CallWndProcHook);
      end;
    except
      CreateReport(sMailAddress, sDescription, CompressFile(sDmpFile, sArchive, sTempPath, PMFormApi.FrameHandle), E, E, FALSE);
      Terminate(ExceptionInfo.ExceptionRecord.ExceptionCode);
    end;
  end;
end;

procedure ExceptionFilterHandler(Reason: Integer);
begin
  case Reason of
    DLL_PROCESS_ATTACH: begin
      DllProc := @ExceptionFilterHandler;
      ExceptionFilter := SetUnhandledExceptionFilter(@MiniDumpExceptionFilter);
      if Assigned(RegisterApplicationRestart) then RegisterApplicationRestart(Pointer(PMFormApi.Commandline), RESTART_NO_REBOOT);
    end;
    DLL_PROCESS_DETACH: if ExceptionFilter <> NIL then SetUnhandledExceptionFilter(ExceptionFilter);
  end;
end;

exports
  FORMINIT,
  FORMPROC;

begin
  ExceptionFilterHandler(DLL_PROCESS_ATTACH);
end.

