{------------------------------------------------------------------------------}
{        This code is copyright (C) 2008-2025 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}
{$J+,Z4}
program PMDebug;

{$R 'PMDebug.res' 'PMDebug.rc'}

uses
  WOW64, Windows, Messages, SysUtils, StrUtils, Classes,
  DbgHelp, MakeCab, StrTools, FileTools, WindowTools, Processes,
  PMTools, PMOptions, PMMessages, PMInterface, MDResources, DlgMsgBoxes;


// TMiniDumpMode -----------------------------------------------------------------------------------

type
  TMiniDumpMode = (
    mdmDefault,
    mdmSnapshot,
    mdmCatchAll
  );


// TProcessInfos -----------------------------------------------------------------------------------

type
  PProcessInfos = ^TProcessInfos;
  TProcessInfos = Record
    FilePath: String;
    Attached: TFileTime;
    DebugInfo: TCreateProcessDebugInfo;
  end;


// TThreadInfos ------------------------------------------------------------------------------------

type
  PThreadInfo = ^TThreadInfo;
  TThreadInfo = Record
    ThreadID: Cardinal;
    DebugInfo: TCreateThreadDebugInfo;
  end;
  TThreadInfos = Class(TList)
  private
    function GetThreadInfo(ThreadID: Cardinal): TCreateThreadDebugInfo;
  public
    function Add(ThreadID: Cardinal; const ProcessInfo: TCreateProcessDebugInfo): Integer; reintroduce; overload;
    function Add(ThreadID: Cardinal; const ThreadInfo: TCreateThreadDebugInfo): Integer; reintroduce; overload;
    procedure Clear; override;
    procedure Delete(ThreadID: Cardinal); reintroduce;
    function FindThread(ThreadID: Cardinal): Integer;
    property DebugInfos[ThreadID: Cardinal]: TCreateThreadDebugInfo read GetThreadInfo;
  end;

function TThreadInfos.GetThreadInfo(ThreadID: Cardinal): TCreateThreadDebugInfo;
var
  iIndex: Integer;
begin
  iIndex := FindThread(ThreadID);
  if iIndex > -1 then
    Result := PThreadInfo(Items[iIndex])^.DebugInfo
  else FillChar(Result, SizeOf(TCreateThreadDebugInfo), 0);
end;

function TThreadInfos.Add(ThreadID: Cardinal; const ProcessInfo: TCreateProcessDebugInfo): Integer;
var
  pti: PThreadInfo;
begin
  New(pti);
  pti^.ThreadID := ThreadID;
  with pti^.DebugInfo do begin
    hThread := ProcessInfo.hThread;
    lpThreadLocalBase := ProcessInfo.lpThreadLocalBase;
    lpStartAddress  := ProcessInfo.lpStartAddress;
  end;
  Result := inherited Add(pti);
end;

function TThreadInfos.Add(ThreadID: Cardinal; const ThreadInfo: TCreateThreadDebugInfo): Integer;
var
  pti: PThreadInfo;
begin
  New(pti);
  pti^.ThreadID := ThreadID;
  pti^.DebugInfo := ThreadInfo;
  Result := inherited Add(pti);
end;

procedure TThreadInfos.Clear;
var
  iIndex: Integer;
begin
  for iIndex := 0 to pred(Count) do if Items[iIndex] <> NIL then begin
    Dispose(PThreadInfo(Items[iIndex]));
    Items[iIndex] := NIL;
  end;
  inherited;
end;

procedure TThreadInfos.Delete(ThreadID: Cardinal);
var
  iIndex: Integer;
begin
  iIndex := FindThread(ThreadID);
  if (iIndex > -1) and (Items[iIndex] <> NIL) then begin
    Dispose(PThreadInfo(Items[iIndex]));
    Items[iIndex] := NIL;
  end
  else Exit;
  inherited Delete(iIndex);
end;

function TThreadInfos.FindThread(ThreadID: Cardinal): Integer;
var
  iIndex: Integer;
begin
  Result := -1;
  for iIndex := 0 to pred(Count) do if Items[iIndex] <> NIL then begin
    if ThreadID = PThreadInfo(Items[iIndex])^.ThreadID then begin
      Result := iIndex;
      Break;
    end;
  end;
end;


// TDllInfos ---------------------------------------------------------------------------------------

type
  TDllInfos = Class(TStringList)
  private
    function GetDebugInfo(BaseAddress: Pointer): TLoadDllDebugInfo;
    function GetDllHandle(Index: Integer): THandle;
  public
    destructor Destroy; override;
    function Add(const FileName: String; const DllInfo: TLoadDllDebugInfo): Integer; reintroduce; overload;
    procedure Delete(BaseAddress: Pointer); reintroduce;
    function FindInstance(BaseAddress: Pointer): Integer;
    function FindModule(const FileName: String): Integer;
    property DebugInfos[BaseAddress: Pointer]: TLoadDllDebugInfo read GetDebugInfo;
    property DllHandles[Index: Integer]: THandle read GetDllHandle;
  end;

function TDllInfos.GetDebugInfo(BaseAddress: Pointer): TLoadDllDebugInfo;
var
  iIndex: Integer;
begin
  iIndex := FindInstance(BaseAddress);
  if iIndex > -1 then
    Result := PLoadDllDebugInfo(Objects[iIndex])^
  else FillChar(Result, SizeOf(TLoadDllDebugInfo), 0);
end;

function TDllInfos.GetDllHandle(Index: Integer): THandle;
begin
  if Index > -1 then
    Result := THandle(PLoadDllDebugInfo(Objects[Index])^.lpBaseOfDll)
  else Result := 0;
end;

destructor TDllInfos.Destroy;
var
  iIndex: Integer;
begin
  for iIndex := 0 to pred(Count) do Dispose(PLoadDllDebugInfo(Objects[iIndex]));
  inherited Destroy;
end;

function TDllInfos.Add(const FileName: String; const DllInfo: TLoadDllDebugInfo): Integer;
var
  pdi: PLoadDllDebugInfo;
begin
  New(pdi);
  pdi^ := DllInfo;
  Result := inherited AddObject(FileName, TObject(pdi));
end;

procedure TDllInfos.Delete(BaseAddress: Pointer);
var
  iIndex: Integer;
begin
  iIndex := FindInstance(BaseAddress);
  if (iIndex > -1) and (Objects[iIndex] <> NIL) then begin
    Dispose(PLoadDllDebugInfo(Objects[iIndex]));
    Objects[iIndex] := NIL;
  end
  else Exit;
  inherited Delete(iIndex);
end;

function TDllInfos.FindInstance(BaseAddress: Pointer): Integer;
var
  iIndex: Integer;
begin
  Result := -1;
  for iIndex := 0 to pred(Count) do if Objects[iIndex] <> NIL then begin
    if BaseAddress = PLoadDllDebugInfo(Objects[iIndex])^.lpBaseOfDll then begin
      Result := iIndex;
      Break;
    end;
  end;
end;

function TDllInfos.FindModule(const FileName: String): Integer;
var
  iIndex: Integer;
begin
  Result := -1;
  for iIndex := 0 to pred(Count) do if Objects[iIndex] <> NIL then begin
    if StrTools.AnsiCompareText(FileName, FileTools.ExtractFileName(Strings[iIndex])) = 0 then begin
      Result := iIndex;
      Break;
    end;
  end;
end;


var
  MiniDumpMode: TMiniDumpMode;
  ProcessInfos: TProcessInfos;
  ThreadInfos: TThreadInfos;
  DllInfos: TDllInfos;


// Implementation ----------------------------------------------------------------------------------

procedure ShowSystemError(Code: Cardinal);
begin
  ShowErrorDialog(0, SysErrorMessage(Code), Unquote(GetVersionInformation, LB, RB));
end;

function ReadDebugString(TargetProcess: THandle; const DebugStringInfo: TOutputDebugStringInfo): String;
var
  pDebugString: Pointer;
  {$IF CompilerVersion > 21.0}
  dwRead: NativeUInt;
  {$ELSE]}
  dwRead: DWord;
  {$IFEND}
begin
  with DebugStringInfo do begin
    if fUnicode <> 0 then begin
      pDebugString := WideStrAlloc(nDebugStringLength);
      try
	ReadProcessMemory(TargetProcess, lpDebugStringData, pDebugString, StrTools.StrBufSize(PWideChar(pDebugString)), dwRead);
	Result := String(PWideChar(pDebugString));
	Exit;
      finally
	StrTools.StrDispose(PWideChar(pDebugString));
      end;
    end
    else begin
      pDebugString := AnsiStrAlloc(nDebugStringLength);
      try
	ReadProcessMemory(TargetProcess, lpDebugStringData, pDebugString, StrTools.StrBufSize(PAnsiChar(pDebugString)), dwRead);
	Result := String(PAnsiChar(pDebugString));
	Exit;
      finally
	StrTools.StrDispose(PAnsiChar(pDebugString));
      end;
    end;
    Result := E;
  end;
end;

function ReadExceptionMessage(TargetProcess: THandle; const ExceptionDebugInfo: TExceptionDebugInfo): String;

  function ReadMessage(TargetProcess: THandle; StringAddress: Pointer): String;
  var
    {$IF CompilerVersion > 21.0}
    dwDone, dwSize, dwWSize, dwASize: NativeUInt;
    {$ELSE]}
    dwDone, dwSize, dwWSize, dwASize: DWord;
    {$IFEND}
    pc: Pointer;
  begin
    if StringAddress <> NIL then begin
      // decrease by 32 bits for including string length
      StringAddress := Pointer(UInt64(StringAddress) - SizeOf(Cardinal));
      if ReadProcessMemory(TargetProcess, StringAddress, @dwSize, SizeOf(dwSize), dwDone) and (dwSize <> 0) then begin
	// allocate for length, characters and terminating null
	dwWSize := succ(dwSize) * SizeOf(WideChar) + SizeOf(Cardinal);
	dwASize := succ(dwSize) * SizeOf(AnsiChar) + SizeOf(Cardinal);
	pc := AllocMem(dwWSize);
	try
	  // try reading ansi string if reading wide string fails
	  if ReadProcessMemory(TargetProcess, Pointer(StringAddress), pc, dwWSize, dwDone)
	  or ReadProcessMemory(TargetProcess, Pointer(StringAddress), pc, dwASize, dwDone) then begin
	    // pass on a pointer mimicking the default Pointer(String)
	    Result := GetAnyString(Pointer(UInt64(pc) + SizeOf(Cardinal)));
	    Exit;
	  end;
	finally
	  FreeMem(pc);
	end;
      end;
    end;
    Result := E;
  end;

var
  {$IF CompilerVersion > 21.0}
  dwSize, dwDone: NativeUInt;
  {$ELSE}
  dwSize, dwDone: DWord;
  {$IFEND}
  pe: Pointer;
begin
  with ExceptionDebugInfo.ExceptionRecord do begin
    if (ExceptionCode = EXCEPTION_CPP_EXCEPTION) or (ExceptionCode = EXCEPTION_DELPHI_EXCEPTION) then begin
      dwSize := Exception.InstanceSize;
      pe := AllocMem(dwSize);
      try
	if ReadProcessMemory(TargetProcess, Pointer(ExceptionInformation[1]), pe, dwSize, dwDone) then begin
	  Result := AppendHexAddress(ReadMessage(TargetProcess, GetStringAddress(TObject(pe), 'FMessage', Exception.ClassInfo)),
				     Pointer(ExceptionInformation[0]));
	  Exit;
	end;
      finally
	FreeMem(pe);
      end;
    end;
  end;
  Result := E;
end;

type
  TFileType = (ftCAB, ftDMP);

function GetFileName(FileType: TFileType): String;
const
  FileTypes: Array[TFileType] of String = ('cab', 'dmp');
begin
  if FileType = ftCAB then
    Result := GetBaseDTFileName + DOT + FileTypes[FileType]
  else Result := GetBaseDTFileName + FormatDateTime('-zzz', Now) + DOT + FileTypes[FileType];
end;

type
  TUserData = Record
    CompareTime: TFileTime;
    SearchPath: String;
    Text: String;
  end;
  TFileFoundData = Record
    PFindData: PWin32FindData;
    UserData: TUserData;
  end;
  TOnFileFound = function(var FileFoundData: TFileFoundData): Longbool;

function FileSearch(const FilePattern: String; OnFileFound: TOnFileFound; var UserData: TUserData): Integer;
var
  ffd: TFileFoundData;
  fd: TWin32FindData;
  hSearch: THandle;
begin
  Result := 0;
  if Assigned(OnFileFound) then begin
    ffd.PFindData := @fd;
    ffd.UserData := UserData;
    hSearch := FindFirstFile(PWideChar(UserData.SearchPath + FilePattern), fd);
    if hSearch <> INVALID_HANDLE_VALUE then begin
      try
	repeat
	  if (fd.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY = 0) and OnFileFound(ffd) then inc(Result);
	until not FindNextFile(hSearch, fd);
	UserData.Text := ffd.UserData.Text;
      finally
	Windows.FindClose(hSearch);
      end;
    end;
  end;
end;

function MiniDumpFileCheck(const VersionInfo: String): String;

  function AddFileName(var FileFoundData: TFileFoundData): Longbool;
  begin
    Result := FALSE;
    with FileFoundData do begin
      if CompareFileTime(UserData.CompareTime, PFindData^.ftCreationTime) <= 0 then begin
	UserData.Text := UserData.Text + PFindData^.cFileName + CRLF;
	Result := TRUE;
      end;
    end;
  end;

const
  FILTER_DMP = AST + HT + AST + DMP;
var
  ud: TUserData;
  iCount: Integer;
begin
  Result := E;
  ud.CompareTime:= ProcessInfos.Attached;
  ud.SearchPath := GetTmpDirW;
  ud.Text := E;
  iCount := FileSearch(FILTER_DMP, @AddFileName, ud);
  if iCount > 0 then begin
    Result := MakeCabinet(ud.Text, GetFileName(ftCAB), ud.SearchPath, PMFormApi.FrameHandle);
    if Result <> E then CreateReport(E, E, Result, E, VersionInfo, FALSE);
  end;
end;

function CreateMiniDump(const ProcessID, ThreadID: Cardinal; DebugString: String;
			PEDI: PExceptionDebugInfo; NoProgressDialog: Longbool = FALSE): Cardinal;
const
  NOP: Byte = $90;
var
  hDumpFile, hProcess, hThread: THandle;
  iLangOffset: Integer;
  ctxt: TWOW64Context;
  ep: TExceptionPointers;
begin
  Result := DBG_EXCEPTION_NOT_HANDLED;
  hDumpFile := CreateWriteFile(GetTmpDirW + GetFileName(ftDMP));
  if hDumpFile <> INVALID_HANDLE_VALUE then begin
    try
      iLangOffset := GetLangOffset;
      hProcess := ProcessInfos.DebugInfo.hProcess;
      if hProcess <> 0 then begin
	hThread := ThreadInfos.DebugInfos[ThreadID].hThread;
	if hThread <> 0 then begin
	  if WOW64SuspendThread(hThread) <> $FFFFFFFF then begin
	    try
	      FillChar(ctxt, SizeOf(TContext), 0);
	      ctxt.W32Context.ContextFlags := WOW64_CONTEXT_ALL;
	      if (PEDI <> NIL) and WOW64GetThreadContext(hThread, ctxt) then begin
		if DebugString = E then DebugString := ReadExceptionMessage(hProcess, PEDI^);
		ep.ExceptionRecord := @PEDI^.ExceptionRecord;
		ep.ContextRecord := @ctxt;
		if WriteMiniDump(0, hProcess, ProcessID, ThreadID, hDumpFile, iLangOffset,
				 ProcessInfos.FilePath, DebugString, @ep, NoProgressDialog) then begin
		  if PEDI^.ExceptionRecord.ExceptionCode = EXCEPTION_BREAKPOINT then begin
		    RemoveBreakpoint(hProcess, hThread, ep, FileTools.ExtractFilePath(ProcessInfos.FilePath));
		    Result := DBG_CONTINUE;
		  end;
		  Exit;
		end;
	      end;
	      if WriteMiniDump(0, hProcess, ProcessID, ThreadID, hDumpFile, iLangOffset,
			       ProcessInfos.FilePath, DebugString, NIL, NoProgressDialog) then Exit;
	    finally
	      ResumeThread(hThread);
	    end;
	  end;
	end;
	if WriteMiniDump(0, hProcess, ProcessID, ThreadID, hDumpFile, iLangOffset,
			 ProcessInfos.FilePath, DebugString, NIL, NoProgressDialog) then Exit;
      end;
    finally
      CloseFile(hDumpFile);
    end;
  end;
end;

function GetPMProcessID(var MiniDumpMode: TMiniDumpMode; var NewPMInstance: Longbool; var Parameter: String): Cardinal;

  function GetAppCommands(var ExePath: String): String;
  var
    iPos: Integer;
  begin
    Result := GetCommandLine;
    iPos := succ(NextCharPos(SP, Result, Length(ParamStr(0))));
    if StrTools.EndsText(PM_FILE_NAME, ParamStr(1)) then begin
      iPos := succ(NextCharPos(SP, Result, iPos + Length(ParamStr(1))));
      ExePath := ParamStr(1);
    end;
    Delete(Result, 1, pred(iPos));
  end;

const
  SNAP_SHOT = 'Snapshot';
  CATCH_ALL = 'CatchAll';
var
  sExePath, sCommands: String;
  iPos: Cardinal;
  pi: TProcessInfo;
begin
  Result := 0;
  sExePath := E;
  Parameter := E;
  NewPMInstance := FALSE;
  MiniDumpMode := mdmDefault;
  sCommands := GetAppCommands(sExePath);
  iPos := FirstTextPos(DA + SNAP_SHOT, sCommands);
  if iPos = 0 then iPos := FirstTextPos(SL + SNAP_SHOT, sCommands);
  if iPos > 0 then begin
    Delete(sCommands, iPos, succ(Length(SNAP_SHOT)));
    MiniDumpMode := mdmSnapshot;
  end;
  iPos := FirstTextPos(DA + CATCH_ALL, sCommands);
  if iPos = 0 then iPos := FirstTextPos(SL + CATCH_ALL, sCommands);
  if iPos > 0 then begin
    Delete(sCommands, iPos, succ(Length(CATCH_ALL)));
    MiniDumpMode := mdmCatchAll;
  end;
  TrimSelf(sCommands);
  if sCommands <> E then begin
    if MiniDumpMode > mdmDefault then begin
      if sCommands[1] = LS then begin
	iPos := NextCharPos(RS, sCommands, 1);
	if iPos > 0 then begin
	  Parameter := String(FromBase64(AnsiString(Copy(sCommands, 2, iPos - 2))));
	  Delete(sCommands, 1, iPos);
	  TrimSelf(sCommands);
	end;
      end;
    end;
  end;
  if PMFormApi <> NIL then with PMFormApi do begin
    FillChar(pi, SizeOf(pi), 0);
    try
      if not (StrTools.EndsText(PM_FILE_NAME, sExePath) and FileExists(sExePath)) then sExePath := BaseDir + PM_FILE_NAME;
      if (sCommands = E) or StrTools.SameText(sCommands, Commandline) then begin
	pi := FindProcessInfo(sExePath, TRUE);
	Result := pi.ProcessInfo.dwProcessId;
      end;
      if Result = 0 then with pi.ProcessInfo do begin
	if MainWinVersion > 6 then begin
	  hProcess := CreateUserProcess(QU + sExePath + QU + SP + sCommands, FileTools.ExtractFilePath(sExePath),
					FrameHandle, SW_SHOWNORMAL, cuprProcessHandle);
	end
	else begin
	  hProcess := CreateProcessHandle(QU + sExePath + QU + SP + sCommands,
					  FileTools.ExtractFilePath(sExePath),
					  FrameHandle, SW_SHOWNORMAL);
	end;
	if hProcess <> 0 then begin
	  Result := GetProcessId(hProcess);
	  pi.ProcessInfo.dwProcessId := Result;
	  pi := FindProcessInfo(sExePath, TRUE, @pi.ProcessInfo);
	end;
	NewPMInstance := TRUE;
      end;
    finally
      CloseHandle(pi.ProcessInfo.hProcess);
      Initialize(0);
    end;
  end;
end;

function TerminateRunningInstance: Longbool;
var
  cDebugger: Cardinal;
  hDebugger: THandle;
begin
  Result := FALSE;
  cDebugger := FindProcessInfo(GetProcessNameOfModule(0), FALSE).ProcessInfo.dwProcessId;
  if cDebugger <> 0 then begin
    hDebugger := OpenProcess(PROCESS_TERMINATE or SYNCHRONIZE, FALSE, cDebugger);
    if hDebugger <> 0 then begin
      try
	if TerminateProcess(hDebugger, 0) then begin
	  WaitForSingleObject(hDebugger, INFINITE);
	  Result := TRUE;
	end;
      finally
	CloseHandle(hDebugger);
      end;
    end;
  end;
end;

function Debug: Longbool;
var
  sParameter, sDebugInfo, sVersionInfo: String;
  hAccessToken, hSynchronizeProcess: THandle;
  fNewInstance, fInitialBreak: Longbool;
  pBreakpoint: Pointer;
  cContinue: Cardinal;
  de: TDebugEvent;
begin
  Result := FALSE;
  hAccessToken := 0;
  ThreadInfos := TThreadInfos.Create;
  if ThreadInfos <> NIL then begin
    try
      DllInfos := TDllInfos.Create;
      if DllInfos <> NIL then begin
	TerminateRunningInstance;
	hAccessToken := ElevateDebugPrivileges(GetCurrentProcess, TRUE);
	if DebugActiveProcess(GetPMProcessID(MiniDumpMode, fNewInstance, sParameter)) then begin
	  hSynchronizeProcess := 0;
	  fInitialBreak := FALSE;
	  pBreakpoint := NIL;
	  while WaitForDebugEvent(de, INFINITE) do with de do begin
	    cContinue := DBG_CONTINUE;
	    case dwDebugEventCode of
	      EXCEPTION_DEBUG_EVENT: with Exception.ExceptionRecord do begin
		if not fInitialBreak and (ExceptionCode = EXCEPTION_BREAKPOINT) then begin
		  fInitialBreak := TRUE;
		  if MiniDumpMode = mdmSnapshot then begin
		    if (CreateMiniDump(dwProcessID, dwThreadID, sParameter, NIL) = DBG_EXCEPTION_NOT_HANDLED)
		    and (MiniDumpFileCheck(sVersionInfo) <> E) then begin
		      ShowInfoDialog(0, Str(GetLangOffset + STR_SNAPSHOT), Unquote(GetVersionInformation, LB, RB));
		    end;
		    Exit;
		  end
		  else if not fNewInstance then ShowInfoDialog(0, Str(GetLangOffset + STR_ATTACHED), Unquote(GetVersionInformation, LB, RB));
		  SendNotifyMessage(FindProcessWindow(dwProcessID, PM_CLASS_FRAME), DebugMessage, 0, 0);
		  ContinueDebugEvent(dwProcessID, dwThreadID, cContinue);
		  Continue;
		end;
		if (Exception.dwFirstChance <> 0) and (ExceptionAddress <> pBreakpoint) then begin
		  if ExceptionCode = EXCEPTION_BREAKPOINT then begin
		    if MiniDumpMode = mdmDefault then begin
		      if (CreateMiniDump(dwProcessID, dwThreadID, E, @Exception) = DBG_CONTINUE)
		      and (MiniDumpFileCheck(sVersionInfo) <> E) then begin
			ShowInfoDialog(0, Str(GetLangOffset + STR_SNAPSHOT), Unquote(GetVersionInformation, LB, RB));
		      end;
		      MiniDumpFileCheck(sVersionInfo);
		    end;
		    pBreakpoint := ExceptionAddress;
		  end
		  else begin
		    if MiniDumpMode = mdmCatchAll then CreateMiniDump(dwProcessID, dwThreadID, sDebugInfo, @Exception, TRUE);
		    pBreakpoint := NIL;
		  end;
		  cContinue := DBG_EXCEPTION_NOT_HANDLED;
		end
		else begin
		  cContinue := CreateMiniDump(dwProcessID, dwThreadID, E, @Exception);
		  if Exception.dwFirstChance = 0 then
		    Result := TRUE
		  else TerminateProcess(ProcessInfos.DebugInfo.hProcess, ExceptionCode);
		end;
		sDebugInfo := E;
	      end;
	      CREATE_THREAD_DEBUG_EVENT: begin
		ThreadInfos.Add(dwThreadID, CreateThread);
	      end;
	      CREATE_PROCESS_DEBUG_EVENT: begin
		with ProcessInfos do begin
		  with PMFormApi do SetLangIndex(StrTools.UpperCase(LanguageCode), LanguageModule);
		  FilePath := GetModulePathByFileHandle(CreateProcessInfo.hFile);
		  GetSystemTimeAsFileTime(Attached);
		  DebugInfo := CreateProcessInfo;
		end;
		CloseHandle(ProcessInfos.DebugInfo.hFile);
		ThreadInfos.Add(dwThreadID, CreateProcessInfo);
		if Assigned(DebugSetProcessKillOnExit) then DebugSetProcessKillOnExit(FALSE);
	      end;
	      EXIT_THREAD_DEBUG_EVENT: begin
		ThreadInfos.Delete(dwThreadID);
	      end;
	      EXIT_PROCESS_DEBUG_EVENT: begin
		if Result then CloseProcessFileHandles(dwProcessID, ProcessInfos.DebugInfo.hProcess);
		hSynchronizeProcess := OpenProcess(SYNCHRONIZE, FALSE, dwProcessID);
	      end;
	      LOAD_DLL_DEBUG_EVENT: begin
		with DllInfos do begin
		  Add(GetModulePathByFileHandle(LoadDll.hFile), LoadDll);
		  if Count > 0 then GetLangIndex(Strings[pred(Count)], PMFormApi);
		  CloseHandle(LoadDll.hFile);
		end;
	      end;
	      UNLOAD_DLL_DEBUG_EVENT: begin
		DllInfos.Delete(UnloadDll.lpBaseOfDll);
	      end;
	      OUTPUT_DEBUG_STRING_EVENT: begin
		sDebugInfo := ReadDebugString(ProcessInfos.DebugInfo.hProcess, DebugString);
		if StrTools.ContainsText(sDebugInfo, 'Terminate') then begin
		  sDebugInfo := E;
		  if Assigned(DebugSetProcessKillOnExit) then DebugSetProcessKillOnExit(FALSE);
		  if Assigned(DebugActiveProcessStop) then DebugActiveProcessStop(dwProcessID);
		  hSynchronizeProcess := OpenProcess(SYNCHRONIZE, FALSE, dwProcessID);
		end
		else if StrTools.ContainsText(sDebugInfo, PEGASUS_MAIL) then begin
		  sVersionInfo := sDebugInfo;
		  sDebugInfo := E;
		end;
	      end;
	      RIP_EVENT: begin
		ShowSystemError(RIPInfo.dwError);
	      end;
	    end;
	    ContinueDebugEvent(dwProcessID, dwThreadID, cContinue);
	    if hSynchronizeProcess <> 0 then begin
	      WaitForSingleObject(hSynchronizeProcess, INFINITE);
	      CloseHandle(hSynchronizeProcess);
	      MiniDumpFileCheck(sVersionInfo);
	      Exit;
	    end;
	  end;
	end
	else ShowSystemError(GetLastError);
      end;
    finally
      DllInfos.Free;
      ThreadInfos.Free;
      CloseHandle(hAccessToken);
    end;
  end;
end;

begin
  Debug;
end.

