{------------------------------------------------------------------------------}
{        This code is copyright (C) 2008-2025 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit PMInterface;

interface

uses
  Windows, ShellApi, Messages, Classes, SysUtils, StrTools, FileTools, RegistryTools,
  WindowTools, PMDDEClient, PMMessages, PMOptions, PMTools, MDResources, DlgVersionInfo;

const
  PMFormApi: TPMFormApi = NIL;
  DebugMessage: Cardinal = 0;

function FORMINIT(wVersion: Word; iVariant: Integer; hParent: THandle;
		  pData: PAnsiChar; var hDialog: THandle; pCallback: PAnsiChar): Word; stdcall;
function FORMPROC(hWindow: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
function CreateMessage(const Address, Description, DumpFile, InfoFile: String; GetSysInfo: Longbool): Integer;
function CreateReport(const Address, Description, DumpFile, InfoFile, VersionInfo: String; GetSysInfo: Longbool): Integer;

implementation

uses
  MakeCab;

const
  CallWndHook: THandle = 0;
  GetMessageHook: THandle = 0;

function GetVersionInfoString: String; forward;

function CallWndProc(nCode: Integer; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
var
  pcwps: PCWPStruct;
begin
  if (PMFormApi <> NIL) and (nCode = HC_ACTION) then begin
    pcwps := PCWPStruct(lParam);
    case pcwps^.Message of
      WM_DESTROY: with pcwps^ do if hWnd = PMFormApi.FrameHandle then ShowWindow(hWnd, SW_HIDE);
    else
      if (pcwps^.Message = DebugMessage) and (pcwps^.hWnd = PMFormApi.FrameHandle) then begin
	OutputDebugString(PChar(GetVersionInfoString));
      end;
    end;
  end;
  Result := CallNextHookEx(CallWndHook, nCode, wParam, lParam);
end;

function GetMessageProc(nCode: Integer; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
var
  pMessage: PMsg;
begin
  Result := CallNextHookEx(GetMessageHook, nCode, wParam, lParam);
  if (PMFormApi <> NIL) and (nCode = HC_ACTION) then begin
    pMessage := PMsg(lParam);
    case pMessage^.Message of
      WM_COMMAND: with pMessage^ do begin
	if (lParam = 0) and (MenuString(PMFormApi.PMMenus[pmmExtensions], wParam, MF_BYCOMMAND) = MD_BASE_NAME) then begin
	  if ShowVersionDialog(PMFormApi.FrameHandle, GetLangOffset) = BTN_VER_DOWNLOAD then begin
	    ShellExecute(0, NIL, MD_DOWNLOAD_URL, NIL, NIL, SW_SHOWNORMAL);
	  end;
	end;
      end;
    end;
  end;
end;

function FORMINIT(wVersion: Word; iVariant: Integer; hParent: THandle;
		  pData: PAnsiChar; var hDialog: THandle; pCallback: PAnsiChar): Word; stdcall;
const
  FORMPROC = 'FORMPROC';
begin
  if (wVersion and $FF00) <= $0100 then begin
    Result := 1;
    hDialog := 0;
    with PMFormApi do begin
      Initialize(hParent);
      SetLangIndex(StrTools.UpperCase(LanguageCode), LanguageModule);
    end;
    StrTools.StrLCopy(pCallback, FORMPROC, Length(FORMPROC));
  end
  else Result := 0;
end;

function FORMPROC(hWindow: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
begin
  Result := 0;
  case uMsg of
    WM_FM_INIT: begin
      if IsDebuggerPresent then OutputDebugString(PChar(GetVersionInfoString));
      CallWndHook := SetWindowsHookEx(WH_CALLWNDPROC, @CallWndProc, 0, GetCurrentThreadID);
      GetMessageHook := SetWindowsHookEx(WH_GETMESSAGE, @GetMessageProc, 0, GetCurrentThreadID);
    end;
    WM_FM_RESTOREFOCUS: PMFormApi.MdiChildren.ActiveChild := GetWindow(PMFormApi.FormHandle, GW_HWNDNEXT);
    WM_FM_LOGEVENT: case wParam of
      WLOG_SHUTDOWN: if lParam = 0 then begin
        ProcessMessages;
	if CallWndHook <> 0 then begin
	  UnhookWindowsHookEx(CallWndHook);
	  CallWndHook := 0;
	end;
	if GetMessageHook <> 0 then begin
	  UnhookWindowsHookEx(GetMessageHook);
	  GetMessageHook := 0;
	end;
      end;
    end;
  end;
end;

function GetRendererVersion: String;
const
  RPM = 'Renderer.pm';
  IER = 'IERenderer.ini';
  BEAR = 'BearHtml.dll';
  USE_BEAR = 'UseBearHtml';
  IDLE_CHECKS = 'DoIdleChecks';
  LOAD_WINSOCK = 'LoadWinSock';
  AUTO_GRAPHICS = 'AutoLoadGraphics';
  NEW_MAIL_OPEN = 'NewMailFolderOpen';
var
  fvi: TFileVersionInfo;
  sOption: String;
begin
  Result := E;
  with fvi do begin
    QueryParameters := [fviFixed, fviFileVersion, fviOriginalFilename];
    QueryFile := StrTools.Trim(String(ReadTextFile(PMFormApi.BaseDir + RPM)));
    if Length(QueryFile) > 1 then begin
      if (QueryFile[1] <> BS) and (QueryFile[2] <> CL) then QueryFile := PMFormApi.BaseDir + QueryFile;
      if FileVersionInfo(fvi) then Result := ExtractBaseName(OriginalFilename) + SP + FileVersion + SP + LB + '%s' + RB;
    end;
    QueryFile := PMFormApi.BaseDir + BEAR;
    if FileVersionInfo(fvi) then begin
      if OriginalFilename = E then with Fixed do begin
	FileVersion := IntToStr(HiWord(dwFileVersionMS)) + DOT + IntToStr(LoWord(dwFileVersionMS)) + DOT +
		       IntToStr(HiWord(dwFileVersionLS)) + DOT + IntToStr(LoWord(dwFileVersionLS));
	if Result = E then
	  Result := ExtractBaseName(QueryFile) + SP + FileVersion
	else Result := StrTools.Format(Result, [ExtractBaseName(QueryFile) + SP + FileVersion]);
      end
      else Result := ExtractBaseName(OriginalFilename) + SP + FileVersion;
    end;
  end;
  if Result <> E then with PMFormApi do begin
    Result := Result + BrowserNames;
    sOption := BoolToStr((GetIniDWord(NewMailDir + PMAIL_INI, SEC_DEFAULT, ID_READER_OPTIONS) and $4000) <> 0, TRUE);
    if sOption <> E then Result := Result + CRLF + AUTO_GRAPHICS + EQ + StrTools.LowerCase(sOption);
    sOption := BoolToStr((GetIniDWord(NewMailDir + PMail_INI, SEC_DEFAULT, ID_POP3_OPTIONS_WORD) and $400) <> 0, TRUE);
    if sOption <> E then Result := Result + CRLF + IDLE_CHECKS + EQ +  StrTools.LowerCase(sOption);
    sOption := GetIniString(MailBoxDir + IER, E, LOAD_WINSOCK, FALSE);
    if sOption = E then sOption := GetIniString(NewMailDir + IER, E, LOAD_WINSOCK, FALSE);
    if sOption <> E then Result := Result + CRLF + LOAD_WINSOCK + EQ + StrTools.LowerCase(sOption);
    sOption := GetIniString(MailBoxDir + IER, E, USE_BEAR, FALSE);
    if sOption = E then sOption := GetIniString(NewMailDir + IER, E, USE_BEAR, FALSE);
    if sOption <> E then Result := Result + CRLF + USE_BEAR + EQ + StrTools.LowerCase(sOption);
    if MdiChildren.NewMailFolderName <> E then with MdiChildren do begin
      sOption := BoolToStr(StrTools.EndsText(NewMailFolderName, WindowStringW(MdiChildren.FindMdiFolMan(0))) or
      			  (MdiChildren.FindMdiFolder(NewMailFolderName) <> 0), TRUE);
      if sOption <> E then Result := Result + CRLF + NEW_MAIL_OPEN + EQ + StrTools.LowerCase(sOption);
    end;
  end;
end;

function GetDefaultPrinter: String;
var
  aPrinter: TPathArray;
begin
  if GetProfileString('Windows', 'Device', NIL, aPrinter, SizeOf(TPathArray)) > 0 then begin
    SetString(Result, aPrinter, StrTools.StrLen(aPrinter));
    Delete(Result, Pos(CM, Result), MAXINT);
  end
  else Result := '(???)';
end;

function GetVersionInfoString: String;
const
  ASP = ' & ';
begin
  Result := GetEnvironmentInfo + ASP + GetMSHTMLVersion + CRLF +
	    PMFormApi.VerStr + ASP + GetDefaultPrinter + CRLF +
	    GetRendererVersion + PARA;
end;

function CreateMsgTextFile(const Description, MsgTextFile, InfoFile, VersionInfo: String; LangOffset: Integer; DDE: Longbool): TStringArray;
var
  sVersionInfo, sSilentCrash, sDescription: String;
begin
  Result.New;
  sVersionInfo := VersionInfo;
  if sVersionInfo = E then sVersionInfo := GetVersionInfoString;
  sSilentCrash := '=> ' + Str(LangOffset + STR_SILENT_CRASH_HINT_4);
  sDescription := StrTools.Trim(Description);
  if InfoFile = E then begin
    Result.AddText(Str(LangOffset + STR_SILENT_CRASH_HINT_1) + SP +
		   Str(LangOffset + STR_SILENT_CRASH_HINT_2) + SP +
		   Str(LangOffset + STR_SILENT_CRASH_HINT_3));
    if sDescription = E then
      Result.AddLines([String(sSilentCrash)])
    else Result.AddLines([String(sDescription)]);
  end
  else begin
    if (sDescription = E) or ((LangOffset > 0) and (sDescription = StrTools.Trim(Str(LangOffset + EDIT_DESCRIPTION)))
    or (sDescription = StrTools.Trim(DialogControlString(HInstance, MakeIntResource(DLG_DUMP_REPORT), EDIT_DESCRIPTION)))) then
      Result.AddText(sSilentCrash)
    else Result.AddText(sDescription);
  end;
  sVersionInfo := Str(LangOffset + STR_SILENT_CRASH_HINT_5) + PARA + sVersionInfo;
  if not DDE then begin
    try
      WriteTextFile(MsgTextFile, AnsiString(sVersionInfo + ReplaceStrings(Result.GetText, CRLF, PARA) + PMFormApi.GetDefaultSignature(TRUE) + PARA));
    finally
      Result.Clear;
    end;
  end
  else Result.InsertLine(sVersionInfo, 0);
end;

function DuplicateBackslashes(const FileName: String): String;
var
  iLen, iEnd, iBeg: Integer;
  pResult: PChar;
begin
  if FileName <> E then begin
    iLen := Length(FileName);
    SetString(Result, NIL, iLen shl 1);
    pResult := Pointer(Result);
    iEnd := 1;
    repeat
      iBeg := iEnd;
      while not CharInSet(FileName[iEnd], [NUL, BS]) do inc(iEnd);
      if iEnd >= iBeg then begin
	iLen := succ(iEnd - iBeg);
	StrTools.StrLCopy(pResult, @FileName[iBeg], iLen);
	inc(pResult, iLen);
	if FileName[iEnd] = BS then begin
	  pResult^ := BS;
	  inc(pResult);
	  inc(iEnd);
	end
	else Break;
      end;
    until iEnd = 0;
    Delete(Result, pResult - Pointer(Result), MAXINT);
    Exit;
  end;
  Result := E;
end;

function CreateMiniDumpAttachment(const DumpFile: String; DDE: Longbool): String;
const
  CAB = 'CAB-archive,0';
  GZIP = 'GZIP-archive,0';
  A_ARCHIVE = 'AT:"%s",';
  DDE_ARCHIVE = '"%s",-,';
var
  sA_ARCH, sD_ARCH: String;
begin
  if FileTools.ExtractFileExt(DumpFile) = MakeCab.GZIP then begin
    sA_ARCH := A_ARCHIVE + GZIP;
    sD_ARCH := DDE_ARCHIVE + GZIP;
  end
  else begin
    sA_ARCH := A_ARCHIVE + CAB;
    sD_ARCH := DDE_ARCHIVE + CAB;
  end;
  if not DDE then
    Result := StrTools.Format(sA_ARCH, [DuplicateBackslashes(DumpFile)]) + PARA
  else Result := StrTools.Format(sD_ARCH, [DuplicateBackslashes(DumpFile)]) + PARA;
end;

function CreateSysInfoAttachment(const InfoFile: String; DDE: Longbool): String;
const
  TXT = 'Text,0';
  A_TEXT = 'AT:"%s",' + TXT;
  DDE_TEXT = '"%s",-,' + TXT;
begin
  if (InfoFile <> E) and FileExists(InfoFile) then begin
    if not DDE then
      Result := StrTools.Format(A_TEXT, [DuplicateBackslashes(InfoFile)]) + CRLF
    else Result := StrTools.Format(DDE_TEXT, [DuplicateBackslashes(InfoFile)]) + CRLF;
  end
  else Result := E;
end;

procedure GetPMFileNames(const DumpFile: String; var DraftFile, MsgTextFile: String);
begin
  repeat
    DraftFile := PMFormApi.MailBoxDir + IntToHex(GetTickCount, 8) + EXT_MSG_DRAFT;
  until not FileExists(DraftFile);
  MsgTextFile := FileTools.ChangeFileExt(DraftFile, EXT_MSG_TEXT);
end;

function GetMDFormData: String;
var
  sFilePath: String;
  iBeg, iEnd: Integer;
begin
  sFilePath := PMFormApi.MailBoxDir + MD_FORMFACTS;
  if not FileExists(sFilePath) then sFilePath := PMFormApi.BaseDir + MD_FORMFACTS;
  Result := String(ReadTextFile(sFilePath));
  iBeg := FirstTextPos(FFF_DATA, Result);
  if iBeg > 0 then begin
    iBeg := NextCharPos(EQ, Result, iBeg + Length(FFF_DATA));
    if iBeg > 0 then begin
      inc(iBeg);
      iEnd := NextCharPos(CR, Result, iBeg);
      Result := Unquote(Copy(Result, iBeg, iEnd - iBeg));
    end;
  end;
end;

function GetReportName(const DumpFile: String): String;
const
  INFO = 'Crash dump report';
begin
  Result := INFO + SP + ExtractBaseName(DumpFile);
end;

function CreateMessage(const Address, Description, DumpFile, InfoFile: String; GetSysInfo: Longbool): Integer;
var
  ddeccpmm: TDDECCPMMessage;
begin
  Result := ERROR_SUCCESS;
  try
    ddeccpmm := TDDECCPMMessage.Create(pmmiWindow);
    if ddeccpmm <> NIL then with ddeccpmm do begin
      try
	Mime := TRUE;
	AddrTo := Address;
	Subject := GetReportName(DumpFile) + SP + GetVersionInformation;
	AddAttachments([CreateSysInfoAttachment(InfoFile, TRUE), CreateMiniDumpAttachment(DumpFile, TRUE)]);
	AddBodyText(CreateMsgTextFile(Description, E, InfoFile, E, GetLangOffset, TRUE));
	Restore;
      finally
	Result := DDEError;
	Free;
      end;
    end;
  except
  end;
end;

function CreateReport(const Address, Description, DumpFile, InfoFile, VersionInfo: String; GetSysInfo: Longbool): Integer;
const
  AUTO = 'ASV:';
  MIME = 'MI:1';
  ADDR = 'TO:';
  SUBJ = 'SU:';
  BODY = 'RX:';
var
  sReport, sAddress, sDraftFile, sMsgTextFile, sAttachments: String;
begin
  Result := ERROR_SUCCESS;
  try
    sReport := GetReportName(DumpFile);
    if sAddress = E then sAddress := GetMDFormData;
    GetPMFileNames(DumpFile, sDraftFile, sMsgTextFile);
    if GetSysInfo then sAttachments := CreateSysInfoAttachment(InfoFile, FALSE);
    sAttachments := sAttachments + CreateMiniDumpAttachment(DumpFile, FALSE);
    CreateMsgTextFile(Description, sMsgTextFile, InfoFile, E, GetLangOffset, FALSE);
    WriteTextFile(sDraftFile, AnsiString(AUTO + sReport + CRLF + MIME + CRLF + ADDR + sAddress + CRLF +
					 SUBJ + sReport + SP + GetVersionInformation + CRLF +
					 BODY + FileTools.ExtractFileName(sMsgTextFile) + CRLF + sAttachments));
  except
    Result := ERROR_WRITE_FAULT;
  end;
end;

initialization
  PMFormApi := TPMFormApi.Create;
  DebugMessage := RegisterWindowMessage(PChar(HashedString(MD_BASE_NAME, 32)));

finalization
  if PMFormApi <> NIL then FreeAndNil(PMFormApi);

end.

