{------------------------------------------------------------------------------}
{        This code is copyright (C) 2008-2025 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit Processes;

interface

uses
  Windows, TlHelp32, ShlObj, ShellApi, SysUtils, Variants, StrTools, FileTools, WindowTools;


// KERNEL32
const
  PROCESS_DEP_ENABLE				= $00000001;
  PROCESS_DEP_DISABLE_ATL_THUNK_EMULATION	= $00000002;

const
  RESTART_NO_CRASH				= $00000001;
  RESTART_NO_HANG				= $00000002;
  RESTART_NO_PATCH				= $00000004;
  RESTART_NO_REBOOT				= $00000008;

type
  TDEPSystemPolicyType = (
    DEPPolicyAlwaysOff,
    DEPPolicyAlwaysOn,
    DEPPolicyOptIn,
    DEPPolicyOptOut
  );

var
  GetProcessId: function(Process: THandle): DWord; stdcall;
  GetProcessIdOfThread: function(Thread: THandle): DWord; stdcall;
  GetProcessHandleFromHwnd: function(Window: THandle): THandle; stdcall;

  GetProcessDEPPolicy: function(hProcess: THandle; var lpFlags: DWord; var lpPermanent: Bool): Bool; stdcall;
  SetProcessDEPPolicy: function(dwFlags: DWord): Bool; stdcall;

  GetSystemDEPPolicy: function: TDEPSystemPolicyType; stdcall;

  RegisterApplicationRestart: function(pwzCommandline: PWideChar; dwFlags: DWord): HResult; stdcall;
  UnregisterApplicationRestart: procedure; stdcall;


// USER32
const
  WM_POWERBROADCAST	= $0218;

  PBT_APMSUSPEND        = $0004;

  PBT_APMRESUMECRITICAL = $0006;
  PBT_APMRESUMESUSPEND  = $0007;

var
  GetShellWindow: function: THandle; stdcall;

  RegisterSuspendResumeNotification: function(hRecipient: THandle; Flags: DWord): THandle; stdcall;
  UnregisterSuspendResumeNotification: function(Handle: THandle): Bool; stdcall;


// My stuff
type
  PProcessInfo = ^TProcessInfo;
  TProcessInfo = Record
    piProcessPath: String;
    piCommandLine: String;
    piVisible: Longbool;
    piWindow: THandle;
    ProcessInfo: TProcessInformation;
  end;

type
  TCreateUserProcessReturn = (cuprExitCode, cuprProcessHandle, cuprProcessID);
  TCreateProcessReturn = (cprConsoleExitCode, cprWaitExitCode, cprProcessHandle, cprProcessID);
  TShellExecReturn = (serHInstErrorCode, serWaitExitCode, serProcessID);

function IsPMDebugged: Longbool;
procedure EnableDataExecutionPrevention;
function CloseProcessFileHandles(SourceProcessId, SourceProcessHandle: Cardinal): Cardinal;
function CreateProcessId(const CommandLine, PathName: String; ForegroundWindow: THandle; Show: Integer = SW_SHOW;
			 CreateProcessReturn: TCreateProcessReturn = cprProcessID): Cardinal;
function CreateProcessHandle(const CommandLine, PathName: String; ForegroundWindow: THandle; Show: Integer = SW_SHOW;
			     CreateProcessReturn: TCreateProcessReturn = cprProcessHandle): Cardinal;
function CreateProcess(const CommandLine, PathName: String; ForegroundWindow: THandle; Show: Integer; var ConsoleOutput: String;
		       WaitMilliSeconds: Cardinal = 100; CreateProcessReturn: TCreateProcessReturn = cprWaitExitCode): Cardinal;
function CreateUserProcess(const CommandLine, PathName: String; ForegroundWindow: THandle; Show: Integer;
			   CreateUserProcessReturn: TCreateUserProcessReturn = cuprProcessID): Cardinal;
function ShellExecuteAndWait(ForegroundWindow: THandle; const Operation, FileName: String;
			     const Parameters: String = E; const Directory: String = E;
			     ShowCmd: Integer = SW_SHOWNORMAL; WaitMilliSeconds: Cardinal = 100;
			     ShellExecReturn: TShellExecReturn = serHInstErrorCode): Cardinal;
function ElevateDebugPrivileges(ProcessOrToken: THandle; Enable: Boolean; HandleIsToken: Boolean = FALSE): THandle;
function HasElevatedDebugPrivilege(Process: THandle): Longbool;
function IsDebugged(const ExcludeDebugger: String): Longbool;
function GetProcessNameByID(ProcessId: DWord): String;
function GetProcessPathByID(ProcessId: DWord): String;
function GetProcessNameOfModule(ModuleHandle: THandle): String;
function GetProcessByID(const PI: TProcessInfo; ProcessId: DWord): Longbool;
function FindChildProcess(const ProcessName: String; ProcessId: DWord): Cardinal;
function FindParentProcess(const ProcessId: DWord; ReturnId: Longbool = FALSE): Variant;
function FindProcessId(const ProcessName: String; CheckId: DWord = 0): Cardinal;
function FindProcessModule(ProcessId: Cardinal; const Filter: String; ReturnName: Longbool): Variant;
function FindProcessWindow(ProcessId: Cardinal; const ProcessPath: String; Thread: THandle = 0): THandle;
function FindProcessInfo(const ProcessPath: String; Visible: Longbool = TRUE; PProcessInfo: PProcessInformation = NIL): TProcessInfo;
function GetProcessImageFullPath(Process: THandle): String;
function GetMappedFileName(Process: THandle; FileView: Pointer): String; overload;
function GetModuleBaseName(Process, Module: THandle): String; overload;
function GetModulePathByFileHandle(FileHandle: THandle): String;
function GetShellProcessId: Cardinal;
function GetShellAccessToken: THandle;
function GetCommmandLineParams: String;
function ProcessRunning(const Application: String; ExtractFromCommandLine: Longbool = FALSE): Longbool;
function RunningAsAdmin(ProcessId: Cardinal): Longbool;
function UserProfileLoad(Token: THandle; const User: String): THandle;
function UserProfileUnload(Token, Profile: THandle): Longbool;
function UserImpersonate: Longbool;
function UserRevert: Longbool;
function NewInstance: Longbool;


implementation

uses
  NtStatus, UserEnv;

// ADVAPI32
const
  SE_GROUP_MANDATORY				= $00000001;
  SE_GROUP_ENABLED_BY_DEFAULT			= $00000002;
  SE_GROUP_ENABLED				= $00000004;
  SE_GROUP_OWNER				= $00000008;
  SE_GROUP_USE_FOR_DENY_ONLY			= $00000010;
  SE_GROUP_INTEGRITY				= $00000020;
  SE_GROUP_INTEGRITY_ENABLED			= $00000040;
  SE_GROUP_LOGON_ID				= $C0000000;
  SE_GROUP_RESOURCE				= $20000000;

  SE_GROUP_VALID_ATTRIBUTES			= SE_GROUP_MANDATORY		or
						  SE_GROUP_ENABLED_BY_DEFAULT	or
						  SE_GROUP_ENABLED		or
						  SE_GROUP_OWNER		or
						  SE_GROUP_USE_FOR_DENY_ONLY	or
						  SE_GROUP_LOGON_ID		or
						  SE_GROUP_RESOURCE		or
						  SE_GROUP_INTEGRITY		or
						  SE_GROUP_INTEGRITY_ENABLED;

const
  SE_CREATE_TOKEN_NAME				= 'SeCreateTokenPrivilege';
  SE_ASSIGNPRIMARYTOKEN_NAME			= 'SeAssignPrimaryTokenPrivilege';
  SE_LOCK_MEMORY_NAME				= 'SeLockMemoryPrivilege';
  SE_INCREASE_QUOTA_NAME			= 'SeIncreaseQuotaPrivilege';
  SE_UNSOLICITED_INPUT_NAME			= 'SeUnsolicitedInputPrivilege';
  SE_MACHINE_ACCOUNT_NAME			= 'SeMachineAccountPrivilege';
  SE_TCB_NAME					= 'SeTcbPrivilege';
  SE_SECURITY_NAME				= 'SeSecurityPrivilege';
  SE_TAKE_OWNERSHIP_NAME			= 'SeTakeOwnershipPrivilege';
  SE_LOAD_DRIVER_NAME				= 'SeLoadDriverPrivilege';
  SE_SYSTEM_PROFILE_NAME			= 'SeSystemProfilePrivilege';
  SE_SYSTEMTIME_NAME				= 'SeSystemtimePrivilege';
  SE_PROF_SINGLE_PROCESS_NAME			= 'SeProfileSingleProcessPrivilege';
  SE_INC_BASE_PRIORITY_NAME			= 'SeIncreaseBasePriorityPrivilege';
  SE_CREATE_PAGEFILE_NAME			= 'SeCreatePagefilePrivilege';
  SE_CREATE_PERMANENT_NAME			= 'SeCreatePermanentPrivilege';
  SE_BACKUP_NAME				= 'SeBackupPrivilege';
  SE_RESTORE_NAME				= 'SeRestorePrivilege';
  SE_SHUTDOWN_NAME				= 'SeShutdownPrivilege';
  SE_DEBUG_NAME					= 'SeDebugPrivilege';
  SE_AUDIT_NAME					= 'SeAuditPrivilege';
  SE_SYSTEM_ENVIRONMENT_NAME			= 'SeSystemEnvironmentPrivilege';
  SE_CHANGE_NOTIFY_NAME				= 'SeChangeNotifyPrivilege';
  SE_REMOTE_SHUTDOWN_NAME			= 'SeRemoteShutdownPrivilege';
  SE_UNDOCK_NAME				= 'SeUndockPrivilege';
  SE_SYNC_AGENT_NAME				= 'SeSyncAgentPrivilege';
  SE_ENABLE_DELEGATION_NAME			= 'SeEnableDelegationPrivilege';
  SE_MANAGE_VOLUME_NAME				= 'SeManageVolumePrivilege';
  SE_IMPERSONATE_NAME				= 'SeImpersonatePrivilege';
  SE_CREATE_GLOBAL_NAME				= 'SeCreateGlobalPrivilege';
  SE_TRUSTED_CREDMAN_ACCESS_NAME		= 'SeTrustedCredManAccessPrivilege';
  SE_RELABEL_NAME				= 'SeRelabelPrivilege';
  SE_INC_WORKING_SET_NAME			= 'SeIncreaseWorkingSetPrivilege';
  SE_TIME_ZONE_NAME				= 'SeTimeZonePrivilege';
  SE_CREATE_SYMBOLIC_LINK_NAME			= 'SeCreateSymbolicLinkPrivilege';

const
  SECURITY_MANDATORY_UNTRUSTED_RID		= 'S-1-16-0';		// $00000000
  SECURITY_MANDATORY_LOW_RID			= 'S-1-16-4096';	// $00001000
  SECURITY_MANDATORY_MEDIUM_RID			= 'S-1-16-8192';	// $00002000
  SECURITY_MANDATORY_MEDIUM_PLUS_RID		= 'S-1-16-8448';	// $00002100
  SECURITY_MANDATORY_HIGH_RID			= 'S-1-16-12288';	// $00003000
  SECURITY_MANDATORY_SYSTEM_RID			= 'S-1-16-16384';	// $00004000
  SECURITY_MANDATORY_PROTECTED_PROCESS_RID	= 'S-1-16-20480';	// $00005000

const
  ANYSIZE_ARRAY 				= 01;
  SID_REVISION					= 01;	// Current revision level
  SID_MAX_SUB_AUTHORITIES			= 15;
  SID_RECOMMENDED_SUB_AUTHORITIES		= 01;	// Will change to around 6 in a future release.

type
  PSidIdentifierAuthority = ^TSidIdentifierAuthority;
  TSidIdentifierAuthority = Record
    Value: Array[0..5] of Byte;
  end;

type
  PSid = ^TSid;
  TSid = Record
    Revision: Byte;
    Subauthoritycount: Byte;
    Identifierauthority: TSidIdentifierAuthority;
    Subauthority: Array[0..ANYSIZE_ARRAY - 1] of DWord;
  end;

const
  SECURITY_MAX_SID_SIZE	= (SizeOf(TSid) - SizeOf(DWORD) + (SID_MAX_SUB_AUTHORITIES * Sizeof(DWORD)));

type
  PTokenMandatoryLabel = ^TTokenMandatoryLabel;
  TTokenMandatoryLabel = Record
    TMLabel: TSIDAndAttributes;
  end;

var
  ConvertStringSidToSid: function(StringSid: LPCTSTR; var Sid: PSid): Bool; stdcall;
  CreateRestrictedToken: function(ExistingTokenHandle: THandle; Flags: DWord;
    DisableSidCount: DWord; SidsToDisable: PSIDAndAttributes;
    DeletePrivilegeCount: DWord; PrivilegesToDelete: PLUIDAndAttributes;
    RestrictedSidCount: DWord; SidsToRestrict: PSIDAndAttributes;
    var NewTokenHandle: THandle): Bool; stdcall;

// NTDLL
type
  TProcessInfoClass = (
    ProcessBasicInformation = 0,
    ProcessDebugPort = 7,
    ProcessWow64Information = 26,
    ProcessImageFileName = 27,
    ProcessBreakOnTermination = 29,
    ProcessProtectionInformation = 61
  );

  TSystemInformationClass = (
    SystemBasicInformation = 0,
    SystemPerformanceInformation = 2,
    SystemTimeOfDayInformation = 3,
    SystemProcessInformation = 5,
    SystemProcessorPerformanceInformation = 8,
    SystemHandleInformation = 16,
    SystemInterruptInformation = 23,
    SystemExceptionInformation = 33,
    SystemRegistryQuotaInformation = 37,
    SystemLookasideInformation = 45
  );

  TAccessMask = Cardinal;

  PSystemHandle = ^TSystemHandle;
  TSystemHandle = Record
    ProcessId: Cardinal;
    ObjectTypeNumber: Byte;
    Flags: Byte;
    Handle: Word;
    PObject: Pointer;
    GrantedAccess: TAccessMask;
  end;

  PSystemHandleInformation = ^TSystemHandleInformation;
  TSystemHandleInformation = Record
    HandleCount: Cardinal;
    Handles: Array[0..0] of TSystemHandle;
  end;

  TObjectInformationClass = (
    ObjectBasicInformation = 0,
    ObjectTypeInformation = 2
  );

  PPublicObjectBasicInformation = ^TPublicObjectBasicInformation;
  TPublicObjectBasicInformation = Record
    Attributes: Cardinal;
    GrantedAccess: TAccessMask;
    HandleCount: Cardinal;
    PointerCount: Cardinal;
    Reserved: Array[0..9] of Cardinal;
  end;

  PUnicodeString = ^TUnicodeString;
  TUnicodeString = Record
    Length: Word;
    MaximumLength: Word;
    Buffer: PChar;
  end;

  PPublicObjectTypeInformation = ^TPublicObjectTypeInformation;
  TPublicObjectTypeInformation = Record
    TypeName: TUnicodeString;
    Reserved: Array[0..21] of Cardinal;
  end;

  PPebLdrData = ^TPebLdrData;
  TPebLdrData = Record
    Reserved1: Array[0..7] of Byte;
    Reserved2: Array[0..2] of Pointer;
    InMemoryOrderModuleList: TListentry;
  end;

  PRtlUserProcessParameters = ^TRtlUserProcessParameters;
  TRtlUserProcessParameters = Record
    Reserved1: Array[0..15] of Byte;
    Reserved2: Array[0..9] of Pointer;
    ImagePathName: TUnicodeString;
    CommandLine: TUnicodeString;
  end;

  PPpsPostProcessInitRoutine = Procedure; stdcall;

  PPeb = ^TPeb;
  TPeb = Record
    Reserved1: Array[0..1] of Byte;
    BeingDebugged: Byte;
    Reserved2: Array[0..0] of Byte;
    Reserved3: Array[0..1] of Pointer;
    Ldr: PPebLdrData;
    ProcessParameters: PRtlUserProcessParameters;
    Reserved4: Array[0..2] of Pointer;
    AtlThunkSListPtr: Pointer;
    Reserved5: Pointer;
    Reserved6: Cardinal;
    Reserved7: Pointer;
    Reserved8: Cardinal;
    AtlThunkSListPtr32: Cardinal;
    Reserved9: Array[0..44] of Pointer;
    Reserved10: Array[0..95] of Byte;
    PostProcessInitRoutine: PPpsPostProcessInitRoutine;
    Reserved11: Array[0..127] of Byte;
    Reserved12: Array[0..0] of Pointer;
    SessionId: Cardinal;
  end;

  PProcessBasicInformation = ^TProcessBasicInformation;
  TProcessBasicInformation = Record
    Reserved1: Pointer;
    PebBaseAddress: PPEB;
    Reserved2: Array[0..1] of Pointer;
    UniqueProcessId: Cardinal;
    Reserved3: Pointer;
  end;

var
  NtClose: function(Handle: THandle): TNtStatus; stdcall;
  NtDuplicateObject: function(SourceProcessHandle, SourceHandle: THandle;
			      TargetProcessHandle: THandle; var TargetHandle: THandle;
			      DesiredAccess: TAccessMask; Attributes, Options: Cardinal): TNtStatus; stdcall;
  NtQueryObject: function(Handle: THandle; ObjectInformationClass: TObjectInformationClass;
			  ObjectInformation:  Pointer; ObjectInformationLength: Cardinal;
			  var ReturnLength: Cardinal): TNtStatus; stdcall;
  NtQueryInformationProcess: function(ProcessHandle: THandle; ProcessInformationClass: TProcessInfoClass;
				      ProcessInformation: Pointer; ProcessInformationLength: Cardinal;
				      var ReturnLength: Cardinal): TNtStatus; stdcall;
  NtQuerySystemInformation: function(SystemInformationClass: TSystemInformationClass; SystemInformation: Pointer;
				     SystemInformationLength: Cardinal; var ReturnLength: Cardinal): TNtStatus; stdcall;

// PSAPI
var
  EnumProcessModules: function (hProcess: THandle; lphModule: LPDWORD; cb: DWORD; var lpcbNeeded: DWORD): BOOL stdcall;
  GetProcessImageFileNameW: function (hProcess: THandle; lpImageFileName: PWideChar; nSize: DWORD): DWORD; stdcall;
  GetMappedFileNameW: function(hProcess: THandle; lpv: Pointer; lpFilename: PWideChar; nSize: DWORD): DWORD; stdcall;
  GetModuleBaseNameW: function (hProcess: THandle; hModule: HMODULE; lpBaseName: PWideChar; nSize: DWORD): DWORD stdcall;


// My stuff --------------------------------------------------------------------

const
  SWP_NOMOVE_NOSIZE = SWP_NOMOVE or SWP_NOSIZE;
  SWP_NA_NOMOVE_NOSIZE = SWP_NOACTIVATE or SWP_NOMOVE_NOSIZE;

function IsPMDebugged: Longbool;
begin
  Result := IsDebugged('PMDebug.exe');
end;

procedure EnableDataExecutionPrevention;
var
  dwFlags: DWord;
  fDummy: Longbool;
begin
  if Assigned(GetSystemDEPPolicy) and GetProcessDEPPolicy(GetCurrentProcess, dwFlags, fDummy) then begin
    if (GetSystemDEPPolicy = DEPPolicyOptOut) and (dwFlags and PROCESS_DEP_ENABLE = 0) then Exit;
    SetProcessDEPPolicy(dwFlags or PROCESS_DEP_ENABLE);
  end;
end;

function CloseProcessFileHandles(SourceProcessId, SourceProcessHandle: Cardinal): Cardinal;
var
  iStatus: Integer;
  cSize, cResult, cIndex: Cardinal;
  pshi: PSystemHandleInformation;
  ppoti: PPublicObjectTypeInformation;
  hTarget, hFile: THandle;
begin
  Result := 0;
  cSize := SizeOf(pshi);
  if Assigned(NtQuerySystemInformation) and Assigned(NtQueryObject)
  and Assigned(NtDuplicateObject) and Assigned(NtClose) then begin
    pshi := AllocMem(cSize);
    if pshi = NIL then Exit;
    try
      repeat
	iStatus := NtQuerySystemInformation(SystemHandleInformation, pshi, cSize, cResult);
	if iStatus = STATUS_INFO_LENGTH_MISMATCH then begin
	  cSize := cSize shl 1;
	  ReallocMem(pshi, cSize);
	end
	else begin
	  if iStatus = STATUS_SUCCESS then Break;
	  Exit;
	end;
      until pshi = NIL;
      if pshi = NIL then Exit;
      for cIndex := pred(pshi^.HandleCount) downto 0 do with pshi^.Handles[cIndex] do begin
	if (ProcessId <> SourceProcessId) or (GrantedAccess = $0012019F) then Continue;
	iStatus := NtDuplicateObject(SourceProcessHandle, Handle, GetCurrentProcess, hTarget, 0, 0, 0);
	if iStatus <> STATUS_SUCCESS then Exit;
	cSize := SizeOf(ppoti);
	ppoti := NIL;
	try
	  ppoti := AllocMem(cSize);
	  if ppoti = NIL then Exit;
	  repeat
	    iStatus := NtQueryObject(hTarget, ObjectTypeInformation, ppoti, cSize, cResult);
	    if iStatus = STATUS_INFO_LENGTH_MISMATCH then begin
	      cSize := cSize shl 1;
	      ReallocMem(ppoti, cSize);
	    end
	    else begin
	      if iStatus = STATUS_SUCCESS then Break;
	      Exit;
	    end;
	  until ppoti = NIL;
	  if ppoti = NIL then Exit;
	  if not StrTools.SameText(ppoti^.TypeName.Buffer, 'FILE')
	  or ((GrantedAccess and (GENERIC_WRITE or GENERIC_EXECUTE)) <> GENERIC_WRITE) then Continue;
	  if DuplicateHandle(SourceProcessHandle, Handle, GetCurrentProcess, @hFile, 0, FALSE,
			     DUPLICATE_CLOSE_SOURCE or DUPLICATE_SAME_ACCESS) then begin
	    try
	      FlushFileBuffers(hFile);
	    finally
	      CloseHandle(hFile);
	    end;
	    inc(Result);
	  end;
	finally
	  FreeMem(ppoti);
	  NtClose(hTarget);
	end;
      end;
    finally
      FreeMem(pshi);
      pshi := NIL;
    end;
  end;
end;

function GetLogicalDriveStrings: String;
var
  dwSize: DWord;
begin
  dwSize := Windows.GetLogicalDriveStrings(0, NIL);
  if dwSize > 0 then begin
    SetString(Result, NIL, dwSize);
    Delete(Result, succ(Windows.GetLogicalDriveStrings(dwSize, PChar(Result))), MAXINT);
    Exit;
  end;
  Result := E;
end;

function CreateProcessId(const CommandLine, PathName: String; ForegroundWindow: THandle; Show: Integer = SW_SHOW;
			 CreateProcessReturn: TCreateProcessReturn = cprProcessID): Cardinal;
var
  sDummy: String;
begin
  Result := CreateProcess(CommandLine, PathName, ForegroundWindow, Show, sDummy, 100, CreateProcessReturn);
end;

function CreateProcessHandle(const CommandLine, PathName: String; ForegroundWindow: THandle; Show: Integer = SW_SHOW;
			     CreateProcessReturn: TCreateProcessReturn = cprProcessHandle): Cardinal;
var
  sDummy: String;
begin
  Result := CreateProcess(CommandLine, PathName, ForegroundWindow, Show, sDummy, 100, CreateProcessReturn);
end;

function CreateProcess(const CommandLine, PathName: String; ForegroundWindow: THandle; Show: Integer; var ConsoleOutput: String;
		       WaitMilliSeconds: Cardinal = 100; CreateProcessReturn: TCreateProcessReturn = cprWaitExitCode): Cardinal;
type
  TConsoleOutput = Array[Word] of AnsiChar;
var
  si: TStartupInfo;
  sa: TSecurityAttributes;
  hInRead, hInWrite: THandle;
  hOutRead, hOutWrite: THandle;
  pi: TProcessInformation;
  pwCommandLine: TCommandLine;
  pwPathName: PWideChar;
  sAppPath: String;
  hProcessWindow: THandle;
  coPipe: TConsoleOutput;
  dwOutput, dwCreateFlags, dwLoopCount: DWord;
begin
  Result := 0;
  ConsoleOutput := E;
  if (CommandLine <> E) and (Length(CommandLine) < SizeOf(TCommandLine) div SizeOf(Char)) then begin
    FillChar(si, SizeOf(si), 0);
    with si do begin
      cb := SizeOf(si);
      dwFlags := STARTF_FORCEOFFFEEDBACK or STARTF_USESHOWWINDOW;
      wShowWindow := Show;
      if CreateProcessReturn = cprConsoleExitCode then begin
	with sa do begin
	  nLength := SizeOf(sa);
	  bInheritHandle := TRUE;
	  lpSecurityDescriptor := NIL;
	end;
	if CreatePipe(hInRead, hInWrite, @sa, 0) then SetHandleInformation(hInWrite, HANDLE_FLAG_INHERIT, 0);
	if CreatePipe(hOutRead, hOutWrite, @sa, 0) then SetHandleInformation(hOutRead, HANDLE_FLAG_INHERIT, 0);
	hStdInput := hInRead;
	hStdOutput := hOutWrite;
	hStdError := hOutWrite;
	dwFlags := dwFlags or STARTF_USESTDHANDLES;
      end;
    end;
    FillChar(pi, SizeOf(pi), 0);
    StrTools.StrLCopy(pwCommandLine, Pointer(CommandLine), Length(CommandLine));
    dwCreateFlags := Byte((CreateProcessReturn = cprProcessID) and (Show = SW_SHOWNA)) * CREATE_SUSPENDED;
    if PathName <> E then
      pwPathName := PChar(PathName)
    else pwPathName := PChar(GetSystemFolderPath(CSIDL_MYDOCUMENTS ));
    sAppPath := GetAppPathFromCommandLine(CommandLine);
    if Show = SW_SHOWNA then begin
      SetForegroundWindow(ForegroundWindow);
      LockSetForegroundWindow(LSFW_LOCK);
    end
    else LockSetForegroundWindow(LSFW_UNLOCK);
    if Windows.CreateProcess(NIL, pwCommandLine, NIL, NIL, CreateProcessReturn <> cprProcessID, dwCreateFlags, NIL, pwPathName, si, pi) then begin
      with pi do begin
	try
	  if CreateProcessReturn > cprWaitExitCode then begin
	    if Show = SW_SHOWNA then begin
	      SetForegroundWindow(ForegroundWindow);
	      LockSetForegroundWindow(LSFW_LOCK);
	      ResumeThread(hThread);
	      dwLoopCount := 0;
	      repeat
		inc(dwLoopCount);
		hProcessWindow := FindProcessWindow(dwProcessId, sAppPath, hThread);
		if hProcessWindow <> 0 then begin
		  SetWindowPos(hProcessWindow, ForegroundWindow, 0, 0, 0, 0, SWP_NA_NOMOVE_NOSIZE);
		  if GetForegroundWindow = ForegroundWindow then begin
		    LockSetForegroundWindow(LSFW_LOCK);
		    SetActiveWindow(ForegroundWindow);
		  end;
		end;
	      until ((hProcessWindow <> 0) and (WaitForInputIdle(hProcess, 0) <> WAIT_TIMEOUT)) or (dwLoopCount = 30);
	    end
	    else if Show <> SW_HIDE then begin
	      dwLoopCount := 0;
	      repeat
		inc(dwLoopCount);
		hProcessWindow := FindProcessWindow(dwProcessId, sAppPath, hThread);
	      until ((hProcessWindow <> 0) and (WaitForInputIdle(hProcess, 0) <> WAIT_TIMEOUT)) or (dwLoopCount = 10);
	      if (hProcessWindow <> 0) and (GetForegroundWindow <> hProcessWindow) then begin
		SetForegroundWindow(hProcessWindow);
		SetActiveWindow(hProcessWindow);
	      end;
	    end;
	  end
	  else if WaitForSingleObject(hProcess, INFINITE) = WAIT_FAILED then begin
	    Result := GetLastError;
	    Exit;
	  end
	  else if CreateProcessReturn = cprConsoleExitCode then begin
	    dwOutput := 0;
	    if PeekNamedPipe(hOutRead, NIL, 0, NIL, @dwOutput, NIL) and (dwOutput > 0)
	    and ReadFile(hOutRead, coPipe, SizeOf(coPipe), dwOutput, NIL) then begin
	      ConsoleOutput := String(PAnsiChar(@coPipe));
	    end;
	  end;
	finally
	  if CreateProcessReturn > cprWaitExitCode then begin
	    SetLastError(Result);
	    CloseHandle(hThread);
	    if CreateProcessReturn > cprProcessHandle then begin
	      CloseHandle(hProcess);
	      Result := dwProcessId;
	    end
	    else Result := hProcess;
	  end
	  else GetExitCodeProcess(hProcess, Result);
	end;
      end;
    end;
  end;
end;

function GetAccountSid(const SystemName, AccountName: String; var pSid: PSid): Longbool;
var
  dwSid, dwLen, dwUse: DWord;
  sDomain: String;
begin
  pSid := NIL;
  Result := FALSE;
  if AccountName <> E then begin
    dwSid := 0;
    dwLen := 0;
    dwUse := 0;
    if not LookupAccountName(NIL, Pointer(AccountName), NIL, dwSid, NIL, dwLen, dwUse) and (dwSid > 0) then begin
      GetMem(pSid, dwSid);
      SetString(sDomain, NIL, pred(dwLen));
      Result := LookupAccountName(NIL, Pointer(AccountName), pSid, dwSid, Pointer(sDomain), dwLen, dwUse);
    end;
  end;
end;

function SetTokenIntegrityLevel(hToken: THandle; const SecurityLevel: String): Longbool;
var
  pIntegritySid: PSid;
  tml: TTokenMandatoryLabel;
begin
  Result := ConvertStringSidToSid(PChar(SecurityLevel), pIntegritySid);
  if Result then begin
    try
      with tml.TMLabel do begin
	Sid := pIntegritySid;
	Attributes := SE_GROUP_INTEGRITY;
      end;
      Result := SetTokenInformation(hToken, TokenIntegrityLevel, @tml, SizeOf(tml) + GetLengthSid(pIntegritySid));
    finally
      if pIntegritySid <> NIL then LocalFree(THandle(pIntegritySid));
    end;
  end;
end;

function CreateUserProcess(const CommandLine, PathName: String; ForegroundWindow: THandle; Show: Integer;
			   CreateUserProcessReturn: TCreateUserProcessReturn = cuprProcessID): Cardinal;
var
  hToken, hImpersonate, hProfile, hProcessWindow: THandle;
  si: TStartupInfo;
  pi: TProcessInformation;
  pwCommandLine: TCommandLine;
  pwPathName: PWideChar;
  sAppPath: String;
  dwCreateFlags, dwLoopCount: DWord;
 begin
  Result := 0;
  if (CommandLine <> E) and (Length(CommandLine) < SizeOf(pwCommandLine) div SizeOf(Char)) then begin
    if OpenProcessToken(GetCurrentProcess, TOKEN_QUERY or TOKEN_DUPLICATE or TOKEN_ADJUST_PRIVILEGES, hToken)
    and DuplicateTokenEx(hToken, MAXIMUM_ALLOWED, NIL, SecurityImpersonation, TokenPrimary, hImpersonate) then begin
      try
	if SetTokenIntegrityLevel(hImpersonate, SECURITY_MANDATORY_MEDIUM_RID) then begin
	  hProfile := UserProfileLoad(ElevateDebugPrivileges(hImpersonate, FALSE, TRUE), UserName);
	  try
	    FillChar(si, SizeOf(si), 0);
	    with si do begin
	      cb := SizeOf(si);
	      dwFlags := STARTF_FORCEOFFFEEDBACK or STARTF_USESHOWWINDOW;
	      wShowWindow := Show;
	    end;
	    FillChar(pi, SizeOf(pi), 0);
	    StrTools.StrLCopy(pwCommandLine, Pointer(CommandLine), Length(CommandLine));
	    dwCreateFlags := Byte((CreateUserProcessReturn > cuprExitCode) and (Show = SW_SHOWNA)) * CREATE_SUSPENDED;
	    if PathName <> E then
	      pwPathName := PChar(PathName)
	    else pwPathName := PChar(GetSystemFolderPath(CSIDL_MYDOCUMENTS));
	    sAppPath := GetAppPathFromCommandLine(CommandLine);
	    if Show = SW_SHOWNA then begin
	      SetForegroundWindow(ForegroundWindow);
	      LockSetForegroundWindow(LSFW_LOCK);
	    end
	    else LockSetForegroundWindow(LSFW_UNLOCK);
	    if CreateProcessAsUser(hImpersonate, NIL, pwCommandLine, NIL, NIL, FALSE, dwCreateFlags, NIL, pwPathName, si, pi) then begin
	      with pi do begin
		try
		  SetPriorityClass(hProcess, (GetPriorityClass(GetCurrentProcess)));
		  if CreateUserProcessReturn > cuprExitCode then begin
		    if Show = SW_SHOWNA then begin
		      SetForegroundWindow(ForegroundWindow);
		      LockSetForegroundWindow(LSFW_LOCK);
		      ResumeThread(hThread);
		      dwLoopCount := 0;
		      repeat
			inc(dwLoopCount);
			hProcessWindow := FindProcessWindow(dwProcessId, sAppPath, hThread);
			if hProcessWindow <> 0 then begin
			  SetWindowPos(hProcessWindow, ForegroundWindow, 0, 0, 0, 0, SWP_NA_NOMOVE_NOSIZE);
			  if GetForegroundWindow = ForegroundWindow then begin
			    LockSetForegroundWindow(LSFW_LOCK);
			    SetActiveWindow (ForegroundWindow);
			  end;
			end;
		      until ((hProcessWindow <> 0) and (WaitForInputIdle(hProcess, 0) <> WAIT_TIMEOUT)) or (dwLoopCount = 30);
		    end
		    else if Show <> SW_HIDE then begin
		      dwLoopCount := 0;
		      repeat
			inc(dwLoopCount);
			hProcessWindow := FindProcessWindow(dwProcessId, sAppPath, hThread);
		      until ((hProcessWindow <> 0) and (WaitForInputIdle(hProcess, 0) <> WAIT_TIMEOUT)) or (dwLoopCount = 10);
		      if (hProcessWindow <> 0) and (GetForegroundWindow <> hProcessWindow) then begin
			SetForegroundWindow(hProcessWindow);
			SetActiveWindow(hProcessWindow);
		      end;
		    end;
		  end
		  else if WaitForSingleObject(hProcess, INFINITE) = WAIT_FAILED then begin
		    Result := GetLastError;
		    Exit;
		  end;
		finally
		  if CreateUserProcessReturn > cuprExitCode then begin
		    SetLastError(Result);
		    CloseHandle(hThread);
		    if CreateUserProcessReturn > cuprProcessHandle then begin
		      CloseHandle(hProcess);
		      Result := dwProcessId;
		    end
		    else Result := hProcess;
		  end
		  else begin
		    GetExitCodeProcess(hProcess, Result);
		    CloseHandle(hProcess);
		    CloseHandle(hThread);
		  end;
		end;
	      end;
	    end;
	  finally
	    if hProfile <> 0 then UserProfileUnload(hImpersonate, hProfile);
	  end;
	end;
      finally
	if hImpersonate <> 0 then CloseHandle(hImpersonate);
	if hToken <> 0 then CloseHandle(hToken);
      end;
    end;
  end;
end;

function FindProcessByWindow(Process: THandle; ProcessID: DWord; const ProcessNamePath: String; var ProcessWindow: THandle): DWord;
var
  sProcName: String;
begin
  Result := 0;
  ProcessWindow := 0;
  sProcName := Unquote(ProcessNamePath);
  repeat
    ProcessWindow := FindProcessWindow(ProcessID, sProcName);
    if ProcessWindow <> 0 then begin
      GetWindowThreadProcessId(ProcessWindow, ProcessID);
      if (FindParentProcess(ProcessID, TRUE) = GetCurrentProcessId)
      or SameFileNames(sProcName, GetProcessPathByID(ProcessID)) then
	Result := ProcessID
      else Result := GetProcessId(Process);
    end;
  until ((ProcessWindow <> 0) and (WaitForInputIdle(Process, 0) <> WAIT_TIMEOUT)) or (Result <> 0);
end;

function ShellExecuteAndWait(ForegroundWindow: THandle; const Operation, FileName: String;
			     const Parameters: String = E; const Directory: String = E;
			     ShowCmd: Integer = SW_SHOWNORMAL; WaitMilliSeconds: Cardinal = 100;
			     ShellExecReturn: TShellExecReturn = serHInstErrorCode): Cardinal;
var
  sei: TShellExecuteInfo;
  hProcessWindow: THandle;
begin
  if ForegroundWindow = 0 then ForegroundWindow := GetForegroundWindow;
  FillChar(sei, SizeOf(sei), 0);
  with sei do begin
    cbSize := SizeOf(sei);
    fMask := SEE_MASK_NOASYNC or SEE_MASK_NOCLOSEPROCESS or SEE_MASK_FLAG_NO_UI;
    Wnd := ForegroundWindow;
    lpVerb := PChar(Operation);
    lpFile := PChar(FileName);
    lpParameters := PChar(Parameters);
    lpDirectory := PChar(Directory);
    nShow := ShowCmd;
  end;
  if ShellExecuteEx(@sei) then with sei do begin
    try
      if hProcess <> 0 then begin
	case ShellExecReturn of
	  serHInstErrorCode: Result := hInstApp;
	  serWaitExitCode: begin
	    Result := WaitForSingleObject(hProcess, WaitMilliSeconds);
	    case Result of
	      WAIT_OBJECT_0: GetExitCodeProcess(hProcess, Result);
	      WAIT_FAILED: Result := GetLastError;
	    end;
	    Exit;
	  end;
	  serProcessID: Result := FindProcessByWindow(hProcess, GetProcessId(hProcess), FileName, hProcessWindow);
	end;
      end;
    finally
      if hProcess <> 0 then CloseHandle(hProcess);
    end;
  end;
end;

function ElevateDebugPrivileges(ProcessOrToken: THandle; Enable: Boolean; HandleIsToken: Boolean = FALSE): THandle;
var
  tp: TTokenPrivileges;
  cLength: Cardinal;
begin
  if HandleIsToken then
    Result := ProcessOrToken
  else OpenProcessToken(ProcessOrToken, TOKEN_ADJUST_PRIVILEGES or TOKEN_QUERY, Result);
  if (Result <> 0) and LookupPrivilegeValue(NIL, SE_DEBUG_NAME, tp.Privileges[0].Luid) then begin
    tp.PrivilegeCount := 1;
    tp.Privileges[0].Attributes := Ord(Enable) * SE_PRIVILEGE_ENABLED;
    AdjustTokenPrivileges(Result, FALSE, tp, 0, NIL, cLength);
  end;
end;

function HasElevatedDebugPrivilege(Process: THandle): Longbool;
var
  hToken: THandle;
  ps: TPrivilegeSet;
begin
  if OpenProcessToken(Process, TOKEN_QUERY, hToken) then begin
    try
      with ps do begin
	if LookupPrivilegeValue(NIL, SE_DEBUG_NAME, Privilege[0].Luid) then begin
	  Privilege[0].Attributes := SE_PRIVILEGE_ENABLED;
	  Control := PRIVILEGE_SET_ALL_NECESSARY;
	  PrivilegeCount := 1;
	  Result := PrivilegeCheck(hToken, ps, Result);
	  Exit;
	end;
      end;
    finally
      CloseHandle(hToken);
    end;
  end;
  Result := FALSE;
end;

function IsDebugged(const ExcludeDebugger: String): Longbool;
begin
  Result := IsDebuggerPresent and not StrTools.SameText(FindParentProcess(GetCurrentProcessId), ExcludeDebugger);
end;

function GetProcessNameByID(ProcessId: DWord): String;
begin
  Result := FileTools.ExtractFileName(GetProcessPathByID(ProcessId));
end;

function GetProcessPathByID(ProcessId: DWord): String;
var
  hProcess: THandle;
begin
  hProcess := OpenProcess(PROCESS_ALL_ACCESS, FALSE, ProcessId);
  try
    Result := GetProcessImageFullPath(hProcess);
  finally
    CloseHandle(hProcess);
  end;
end;

function GetProcessNameOfModule(ModuleHandle: THandle): String;
begin
  Result := GetModuleName(ModuleHandle);
end;

function GetProcessByID(const PI: TProcessInfo; ProcessId: DWord): Longbool;
var
  dwThreadProcessId: DWord;
begin
  Result := StrTools.SameText(PI.piProcessPath, GetProcessPathByID(ProcessId));
  if not Result and (PI.ProcessInfo.hThread <> 0) and (MainWinVersion >= 6) then begin
    dwThreadProcessId := GetProcessIdOfThread(PI.ProcessInfo.hThread);
    if (dwThreadProcessId <> 0) and (dwThreadProcessId <> ProcessId) then begin
      Result := StrTools.SameText(PI.piProcessPath, GetProcessPathByID(ProcessId));
    end;
  end;
end;

function FindChildProcess(const ProcessName: String; ProcessId: DWord): Cardinal;
var
  hSnapshot: THandle;
  pe: TProcessEntry32;
begin
  hSnapshot := CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  if hSnapshot <> INVALID_HANDLE_VALUE then begin
    try
      pe.dwSize := SizeOf(pe);
      if Process32First(hSnapshot, pe) then begin
	repeat
	  if (pe.th32ParentProcessID <> 0)
	  and SameFileNames(ProcessName, GetProcessPathByID(pe.th32ParentProcessID)) then begin
	    if ProcessId = pe.th32ProcessID then begin
	      Result := ProcessId;
	      Exit;
	    end;
	  end;
	until not Process32Next(hSnapshot, pe);
      end;
    finally
      CloseHandle(hSnapshot);
    end;
  end;
  Result := 0;
end;

function FindParentProcess(const ProcessId: DWord; ReturnId: Longbool = FALSE): Variant;
var
  hSnapshot: THandle;
  pe: TProcessEntry32;
begin
  hSnapshot := CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  if hSnapshot <> INVALID_HANDLE_VALUE then begin
    try
      pe.dwSize := SizeOf(pe);
      if Process32First(hSnapshot, pe) then begin
	repeat
	  if ProcessId = pe.th32ProcessID then begin
	    if ReturnId then
	      Result := pe.th32ParentProcessID
	    else Result := GetProcessNameByID(pe.th32ParentProcessID);
	    Exit;
	  end;
	until not Process32Next(hSnapshot, pe);
      end;
    finally
      CloseHandle(hSnapshot);
    end;
  end;
  Result := E;
end;

function FindProcessId(const ProcessName: String; CheckId: DWord = 0): Cardinal;
var
  hSnapshot: THandle;
  pe: TProcessEntry32;
  sProcessFileName: String;
  cProcessIdBuffer: Cardinal;
begin
  Result := 0;
  cProcessIdBuffer := 0;
  hSnapshot := CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
  if hSnapshot <> INVALID_HANDLE_VALUE then begin
    try
      pe.dwSize := SizeOf(pe);
      if Process32First(hSnapshot, pe) then begin
	sProcessFileName := FileTools.ExtractFileName(ProcessName);
	if StrTools.SameText(sProcessFileName, ProcessName) then sProcessFileName := E;
	repeat
	  if StrTools.SameText(ProcessName, GetProcessPathByID(pe.th32ProcessId))
	  or ((CheckId = pe.th32ProcessId) and SameFileNames(sProcessFileName, pe.szExeFile)) then begin
	    Result := pe.th32ProcessId;
	    Exit;
	  end
	  else if SameFileNames(sProcessFileName, pe.szExeFile) then cProcessIdBuffer := pe.th32ProcessId;
	until not Process32Next(hSnapshot, pe);
	if Result = 0 then Result := cProcessIdBuffer;
      end;
    finally
      CloseHandle(hSnapshot);
    end;
  end;
end;

function FindProcessModule(ProcessId: Cardinal; const Filter: String; ReturnName: Longbool): Variant;
var
  aModules: Array[ShortInt] of THandle;
  hProcess: THandle;
  dwNeeded: DWord;
  iIndex: Integer;
  sName: String;
begin
  FillChar(aModules, SizeOf(aModules), 0);
  hProcess := OpenProcess(PROCESS_ALL_ACCESS, FALSE, ProcessId);
  try
    dwNeeded := 0;
    if EnumProcessModules(hProcess, @aModules, SizeOf(aModules), dwNeeded) then begin
      if dwNeeded > MAXWORD then dwNeeded := MAXWORD;
      for iIndex := 0 to (dwNeeded div SizeOf(THandle)) do if (aModules[iIndex] <> 0) then begin
	sName := GetModuleBaseName(hProcess, aModules[iIndex]);
	if (sName <> E) and StrTools.StartsText(Filter, sName) then begin
	  if ReturnName then
	    Result := sName
	  else Result := aModules[iIndex];
	  Exit;
	end;
      end;
    end;
  finally
    if hProcess <> 0 then CloseHandle(hProcess);
  end;
  if ReturnName then
    Result := E
  else Result := 0;
end;

type
  PWindowData = ^TWindowData;
  TWindowData = Record
    ProcessId: Cardinal;
    ProcessPath: String;
    Thread: THandle;
    ClassName: String;
    FoundWindow: THandle;
  end;

function EnumFindProcessWindow(Window: THandle; PWD: PWindowData): Bool; stdcall;
var
  dwProcessId, dwParentProcessId: DWord;
  hParent: THandle;
begin
  GetWindowThreadProcessId(Window, @dwProcessId);
  hParent := GetParent(Window);
  if (hParent <> 0) then GetWindowThreadProcessId(hParent, @dwParentProcessId);
  with PWD^ do if IsWindowVisible(Window) and ((hParent = 0) or (dwProcessId = dwParentProcessId)) then begin
    if (ProcessId = dwProcessId) or ((MainWinVersion >= 6) and (dwProcessId = GetProcessIdOfThread(Thread))) then begin
      FoundWindow := Window;
      Result := FALSE;
      Exit;
    end;
    if (ProcessPath <> E) and (FindProcessId(ProcessPath, dwProcessId) = dwProcessId) then begin
      FoundWindow := Window;
      Result := FALSE;
      Exit;
    end;
  end;
  PWD^.FoundWindow := 0;
  Result := TRUE;
end;

function FindProcessWindow(ProcessId: Cardinal; const ProcessPath: String; Thread: THandle = 0): THandle;
var
  wd: TWindowData;
begin
  wd.ProcessId := ProcessId;
  wd.ProcessPath := ProcessPath;
  wd.Thread := Thread;
  EnumWindows(@EnumFindProcessWindow, LPARAM(@wd));
  Result := wd.FoundWindow;
end;

function EnumFindProcessInfo(FoundWindow: THandle; PPI: PProcessInfo): Bool; stdcall;
var
  dwProcessId: Cardinal;
begin
  Result := TRUE;
  with PPI^ do begin
    if not piVisible or IsWindowVisible(FoundWindow) then begin
      ProcessInfo.dwThreadID := GetWindowThreadProcessId(FoundWindow, @dwProcessId);
      if (dwProcessId <> GetCurrentProcessId)
      and (GetProcessByID(PPI^, dwProcessId) or (FindProcessModule(dwProcessId, PPI^.piProcessPath, TRUE) <> E)) then begin
	if StrTools.SameText(PPI^.piProcessPath, GetProcessPathByID(dwProcessId)) then begin
	  ProcessInfo.dwProcessId := dwProcessId;
	  piWindow := FoundWindow;
	  Result := FALSE;
	  Exit;
	end;
      end
      else ProcessInfo.dwThreadID := 0;
    end;
  end;
end;

function FindProcessInfo(const ProcessPath: String; Visible: Longbool = TRUE; PProcessInfo: PProcessInformation = NIL): TProcessInfo;
begin
  with Result do begin
    if PProcessInfo = NIL then
      piProcessPath := ProcessPath
    else piProcessPath := GetProcessImageFullPath(PProcessInfo^.hProcess);
    piVisible := Visible;
    piWindow := 0;
    if PProcessInfo <> NIL then
      ProcessInfo := PProcessInfo^
    else FillChar(ProcessInfo, SizeOf(ProcessInfo), 0);
  end;
  EnumWindows(@EnumFindProcessInfo, LPARAM(@Result));
end;

function GetDrivePathFromDevicePath(const DevicePath: String): String;
var
  dwSizeLo, dwSizeHi: DWord;
  sDriveList: String;
  aDrive: TPathArray;
  pDriveList: PChar;
begin
  Result := DevicePath;
  sDriveList := GetLogicalDriveStrings;
  if sDriveList <> E then begin
    pDriveList := PChar(sDriveList);
    dwSizeHi := Length(sDriveList);
    dwSizeLo := 0;
    repeat
      if succ(dwSizeLo) < dwSizeHi then pDriveList[dwSizeLo + 2] := NUL;
      if (QueryDosDevice(@pDriveList[dwSizeLo], aDrive, MAX_PATH) > 0) then begin
	if StrTools.StartsText(aDrive, Result) then begin
	  inc(pDriveList, dwSizeLo);
	  Result := pDriveList + Copy(Result, succ(StrTools.StrLen(aDrive)), MAXINT);
	  Break;
	end;
      end;
      while pDriveList[dwSizeLo] <> NUL do inc(dwSizeLo);
      while (dwSizeLo < dwSizeHi) and (pDriveList[dwSizeLo] = NUL) do inc(dwSizeLo);
      if dwSizeLo < dwSizeHi then
	Continue
      else Break;
    until dwSizeLo = dwSizeHi;
  end;
end;

function GetProcessImageFullPath(Process: THandle): String;
var
  aModuleName: TUNCPath;
begin
  SetString(Result, aModuleName, GetProcessImageFileNameW(Process, @aModuleName, SizeOf(aModuleName)));
  if Result <> E then Result := GetDrivePathFromDevicePath(Result);
end;

function GetMappedFileName(Process: THandle; FileView: Pointer): String;
var
  aModuleName: TUNCPath;
begin
  SetString(Result, aModuleName, GetMappedFileNameW(Process, FileView, @aModuleName, SizeOf(aModuleName)));
end;

function GetModuleBaseName(Process, Module: THandle): String;
var
  aModuleName: TUNCPath;
begin
  SetString(Result, aModuleName, GetModuleBaseNameW(Process, Module, @aModuleName, SizeOf(aModuleName)));
end;

function GetModulePathByFileHandle(FileHandle: THandle): String;
var
  dwSizeLo, dwSizeHi: DWord;
  hFileMap: THandle;
  pFileView: Pointer;
begin
  if FileHandle <> 0 then begin
    dwSizeLo := GetFileSize(FileHandle, @dwSizeHi);
    if ((dwSizeLo <> INVALID_FILE_SIZE) or (GetLastError = NO_ERROR))
    and ((dwSizeLo <> 0) or (dwSizeHi <> 0)) then begin
      hFileMap := CreateFileMapping(FileHandle, NIL, PAGE_READONLY, 0, 1, NIL);
      if hFileMap <> 0 then begin
	pFileView := NIL;
	try
	  pFileView := MapViewOfFile(hFileMap, FILE_MAP_READ, 0, 0, 1);
	  if pFileView <> NIL then begin
	    Result := GetDrivePathFromDevicePath(GetMappedFileName(GetCurrentProcess, pFileView));
	    Exit;
	  end;
	finally
	  UnmapViewOfFile(pFileView);
	  CloseHandle(hFileMap);
	end;
      end;
    end;
  end;
  Result := E;
end;

function GetShellProcessId: Cardinal;
begin
  if Assigned(GetShellWindow) then
    GetWindowThreadProcessId(GetShellWindow, @Result)
  else Result := 0;
end;

function GetShellAccessToken: THandle;
var
  dwProcessId: DWord;
  hProcess: THandle;
begin
  dwProcessId := GetShellProcessId;
  if dwProcessId <> 0 then begin
    hProcess := OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, dwProcessId);
    if hProcess <> 0 then begin
      try
	if OpenProcessToken(hProcess, TOKEN_QUERY or TOKEN_IMPERSONATE or TOKEN_DUPLICATE, Result) then Exit;
      finally
	CloseHandle(hProcess);
      end;
    end;
  end;
  Result := 0;
end;

function GetCommmandLineParams: String;
var
  ppCmdLine: PPWideChar;
  iNum: Integer;
  cLen: Cardinal;
  pParams: PWideChar;
begin
  ppCmdLine := CommandLineToArgvW(GetCommandLineW, iNum);
  if ppCmdLine <> NIL then begin
    try
      if iNum > 1 then begin
	pParams := ppCmdLine^ + succ(StrTools.StrLen(ppCmdLine^));
	cLen := 0;
	for iNum := iNum downto 2 do begin
	  inc(cLen, StrTools.StrLen(pParams + cLen));
	  pParams[cLen] := SP;
	  inc(cLen);
	end;
	SetString(Result, pParams, pred(cLen));
      end;
    finally
      LocalFree(THandle(ppCmdLine));
    end;
  end;
end;

function ProcessRunning(const Application: String; ExtractFromCommandLine: Longbool = FALSE): Longbool;
var
  sProcessName: String;
begin
  Result := FALSE;
  sProcessName := E;
  if Application = E then Exit;
  if ExtractFromCommandLine then begin
    sProcessName := GetAppPathFromCommandLine(Application);
    Result := StrTools.SameText(sProcessName, GetProcessPathByID(FindProcessId(FileTools.ExtractFileName(sProcessName))));
  end
  else Result := not TestWriteFile(Application);
end;

function RunningAsAdmin(ProcessId: Cardinal): Longbool;
var
  hProcess, hToken: THandle;
  te: TTokenElevation;
  dwSize: DWord;
begin
  if ProcessId <> 0 then begin
    hProcess := OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, ProcessId);
    if hProcess <> 0 then begin
      hToken := 0;
      try
	if OpenProcessToken(hProcess, TOKEN_QUERY, hToken)
	and GetTokenInformation(hToken, TokenElevationType, @te, SizeOf(te), dwSize) then begin
	  Result := (TTokenElevationType(te.TokenIsElevated) = TokenElevationTypeFull);
	  Exit;
	end;
      finally
	CloseHandle(hToken);
	CloseHandle(hProcess);
      end;
    end;
  end;
  Result := FALSE;
end;

function UserProfileLoad(Token: THandle; const User: String): THandle;
var
  pi: TProfileInfo;
begin
  if MainWinVersion >= 5 then begin
    FillChar(pi, SizeOf(pi), 0);
    with pi do begin
      dwSize := SizeOf(pi);
      dwFlags := PI_NOUI;
      lpUserName := PChar(User);
    end;
    if LoadUserProfile(Token, pi) then begin
      Result := pi.hProfile;
      Exit;
    end;
  end;
  Result := HKEY_CURRENT_USER;
end;

function UserProfileUnload(Token, Profile: THandle): Longbool;
begin
  Result := (MainWinVersion >= 5) and (Profile <> HKEY_CURRENT_USER) and UnloadUserProfile(Token, Profile);
end;

function UserImpersonate: Longbool;
var
  hToken: THandle;
begin
  Result := FALSE;
  hToken := GetShellAccessToken;
  try
    Result := ImpersonateLoggedOnUser(hToken);
  finally
    if Result then CloseHandle(hToken);
  end;
end;

function UserRevert: Longbool;
begin
  Result := RevertToSelf;
end;

function NewInstance: Longbool;
var
  Mutex: THandle;
begin
  Result := FALSE;
  Mutex := CreateMutex(NIL, TRUE, PChar(ParamStr(0)));
  if GetLastError = ERROR_ALREADY_EXISTS then begin
    CloseHandle(Mutex);
    Result := TRUE;
  end;
end;

function LoadADVAPI: THandle;
begin
  Result := GetOrLoadLibrary(ADVAPI32);
  ConvertStringSidToSid := GetProcAddress(Result, 'ConvertStringSidToSidW');
  CreateRestrictedToken := GetProcAddress(Result, 'CreateRestrictedToken');
end;

function LoadKERNEL: THandle;
begin
  Result := GetOrLoadLibrary(KERNEL32);
  GetProcessDEPPolicy := GetProcAddress(Result, 'GetProcessDEPPolicy');
  SetProcessDEPPolicy := GetProcAddress(Result, 'SetProcessDEPPolicy');
  GetProcessId := GetProcAddress(Result, 'GetProcessId');
  GetProcessIdOfThread := GetProcAddress(Result, 'GetProcessIdOfThread');
  GetSystemDEPPolicy := GetProcAddress(Result, 'GetSystemDEPPolicy');
  RegisterApplicationRestart := GetProcAddress(Result, 'RegisterApplicationRestart');
  UnregisterApplicationRestart := GetProcAddress(Result, 'UnregisterApplicationRestart');
  GetProcessHandleFromHwnd := GetProcAddress(GetOrLoadLibrary('OleAcc.dll'), 'GetProcessHandleFromHwnd');
end;

function LoadNTDLL: THandle;
begin
  Result := GetOrLoadLibrary('NTDLL.DLL');
  NtClose := GetProcAddress(Result, 'NtClose');

  NtDuplicateObject := GetProcAddress(Result, 'NtDuplicateObject');
  NtQueryObject := GetProcAddress(Result, 'NtQueryObject');

  NtQueryInformationProcess := GetProcAddress(Result, 'NtQueryInformationProcess');
  NtQuerySystemInformation := GetProcAddress(Result, 'NtQuerySystemInformation');
end;

function LoadPSAPI: THandle;
var
  sPrefix: String;
begin
  if MainWinVersion >= 7 then begin
    Result := GetOrLoadLibrary(KERNEL32);
    sPrefix := 'K32';
  end
  else begin
    Result := GetOrLoadLibrary('PSAPI.DLL');
    sPrefix := E;
  end;
  EnumProcessModules := GetProcAddress(Result, PChar(sPrefix + 'EnumProcessModules'));
  GetProcessImageFileNameW := GetProcAddress(Result, PChar(sPrefix + 'GetProcessImageFileNameW'));
  GetMappedFileNameW := GetProcAddress(Result, PChar(sPrefix + 'GetMappedFileNameW'));
  GetModuleBaseNameW := GetProcAddress(Result, PChar(sPrefix + 'GetModuleBaseNameW'));
end;

function LoadUSER32: THandle;
begin
  Result := GetOrLoadLibrary(USER32);
  GetShellWindow := GetProcAddress(Result, 'GetShellWindow');
  if MainWinVersion >= 8 then begin
    RegisterSuspendResumeNotification := GetProcAddress(Result, 'RegisterSuspendResumeNotification');
    UnregisterSuspendResumeNotification := GetProcAddress(Result, 'UnregisterSuspendResumeNotification');
  end;
end;

begin
  LoadADVAPI;
  LoadKERNEL;
  LoadNTDLL;
  LoadPSAPI;
  LoadUSER32;
end.

