{------------------------------------------------------------------------------}
{        This code is copyright (C) 2010-2025 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit WindowTools;

interface

uses
  Windows, ShlObj, ShLwApi, Messages, Types, SysUtils, Math, Classes, Graphics, UxThemes;

const
  CLASS_DIALOG = '#32770';
  CLASS_BUTTON = VSCLASS_BUTTON;
  CLASS_EDIT = VSCLASS_EDIT;
  CLASS_STATIC = VSCLASS_STATIC;

const
  ZERO_SIZE: TSize = (cx: 0; cy:0);
  NO_SIZE: TSize = (cx: -1; cy: -1);
  ZERO_POINT: TPoint = (X: 0; Y: 0);
  OFFSCREEN_POS: TPoint = (X: -1; Y: -1);

const
  EMPTY_RECT: TRect = (Left: 0; Top: 0; Right: 0; Bottom: 0);

type
  TFontParams = (
    fpSize,
    fpWidth,
    fpEscapement,
    fpOrientation,
    fpBold,
    fpItalic,
    fpUnderline,
    fpStrikeOut,
    fpCharSet,
    fpOutPrecision,
    fpClipPrecision,
    fpQuality,
    fpPitchAndFamily,
    fpPitch,
    fpFamily,
    fpName,
    fpColor
  );

type
  PClassNames = ^TClassNames;
  TClassNames = Array of String;

function GetCursorPos: TPoint;
function SysCursorSize: TSize;
function SetWaitCursor: THandle;
function MakePoints(Pos: Cardinal): TPoint;
function MakeSize(Val: Cardinal): TSize; overload;
function MakeSize(cx, cy: Integer): TSize; overload;
function EqualSize(const Sz1, Sz2: TSize): Longbool;
function NullSize(const Size: TSize): Longbool;
function NullPoint(const Point: TPoint): Longbool;
function PointToPos(const Point: TPoint): Cardinal;
function CenterPoint(const Rect: TRect): TPoint;
function SizeToPoint(const Size: TSize): TPoint;
function PointToSize(const Point: TPoint): TSize;
function IsOffScreenPos(const Point: TPoint): Longbool;
function GetPointOffset(const Pt1, Pt2: TPoint): TPoint;
function OffsetPoint(const Point: TPoint; DX, DY: Integer): TPoint;
function OffsetSize(const Size: TSize; DX, DY: Integer): TSize;
function SamePoints(const Point1, Point2: TPoint): Longbool;
function SameSizes(const Size1, Size2: TSize): Longbool;
function Get_X_lParam(lParam: Integer): SmallInt;
function Get_Y_lParam(lParam: Integer): SmallInt;
function Make_lParam(Low, High: Word): Longint; overload;
function Make_lParam(const Point: TPoint): Longint; overload;
function GetAsyncMouseBtnUnswitched(VKCode: Integer): SmallInt;
function PointsFromPixels(Pixels: Integer): Integer;
function PixelsFromPoints(Points: Integer): Integer;
function PixelsToTwips(Pixels: Integer): Integer;
function PixelsFromTwips(Twips: Integer): Integer;
function MilliMetersFromTwips(Twips: Integer): Integer;
function TwipsFromMilliMeters(MilliMeters: Integer): Integer;
function MMtoPixWidth(DC: THandle; MM: Integer): Cardinal;
function MMtoPixHeight(DC: THandle; MM: Integer): Cardinal;
function GetPixDimsFromTwips(DC: THandle; const TwipsDims: TSize): TSize;
function GetTwipsDimsFromPixels(DC: THandle; const Pixels: TSize): TSize;
function GetShellDlgFontName(Theme: THandle; FontID: Integer): String;
function GetValueFromLogFont(const LogFont: TLogFont; FontParams: TFontParams; AsString: Longbool = FALSE): Variant;
procedure ProcessMessages(NoUserInput: Longbool = FALSE);
function GetSysScreenRect: TRect;
function GetSysScreenDimensions: TSize;
function GetFocusWindowPos: TPoint;
function GetWindowAtCursor: THandle;
function GetClientPos(Window: THandle; const Pos: TPoint): TPoint;
function GetWindowPos(Window: THandle): TPoint;
function GetClientSizes(Window: THandle): TSize;
function GetControlRect(Control: THandle): TRect;
function GetChildWindowRect(Child: THandle; Parent: THandle = 0): TRect;
function ClientToScreen(Window: THandle; const Point: TPoint): TPoint; overload;
function ScreenToClient(Window: THandle; const Point: TPoint): TPoint; overload;
function ClientToScreen(Window: THandle; var Rect: TRect): Longbool; overload;
function ScreenToClient(Window: THandle; var Rect: TRect): Longbool; overload;
function ClientToScreenRect(Window: THandle; const Rect: TRect): TRect;
function ScreenToClientRect(Window: THandle; const Rect: TRect): TRect;
function AddBorder(const Rect, Border: TRect): TRect; overload;
function AddBorder(const Rect: TRect; L, T, R, B: Integer): TRect; overload;
function InflateRect(const Rect: TRect; DX, DY: Integer): TRect; overload;
function OffsetRect(const Rect: TRect; DX, DY: Integer): TRect; overload;
function IntersectRect(const Rect1, Rect2: TRect): TRect; overload;
function GetBoundsRect(const Rect: TRect): TRect; overload;
function GetBoundsRect(Window: THandle): TRect; overload;
function GetClientRect(Window: THandle): TRect; overload;
function GetWindowRect(Window: THandle): TRect; overload;
function GetClientCenter(Window: THandle): TPoint;
function GetClientWidth(Window: THandle): Integer;
function GetClientHeight(Window: THandle): Integer;
function IsAllRectEmpty(const Rect: TRect): Longbool;
function IsZeroSize(const Size: TSize): Longbool;
function IsNoSize(const Size: TSize): Longbool;
function GetWindowModuleHandle(Window: THandle): THandle;
function GetWindowModuleName(Window: THandle): String;
function CenterDialog(Dialog, Owner: THandle): Longbool;
function CenterInClientRect(const ChildRect: TRect; Child, Parent: THandle; Draw: Longbool): Longbool;
function FrameRect3D(DC: THandle; const Rect: TRect; TopLeftColor, BottomRightColor: TColorRef): Longbool;
function FindChildWindowAtPos(Parent: THandle; const ScreenPos: TPoint; const Executable: String;
			      const ClassOrWndText: String = ''; CheckWindowText: Longbool = FALSE): THandle;
function FindControlByID(Parent: THandle; const ClassOrWndText: String; ControlId: Integer = 0; CheckWindowText: Longbool = FALSE): THandle;
function FindChildWindow(Parent: THandle; const ClassOrWndText: String; ControlId: Integer = 0;
			 CheckRegEx: Longbool = FALSE; CheckWindowText: Longbool = FALSE;
			 IsDialogControl: Longbool = FALSE; Instance: Cardinal = 1): THandle;
function FindChildWindowInClass(Parent: THandle; const ClassNames: TClassNames; LastInstance: Longbool = FALSE): THandle;
function FindThreadWindow(ThreadId: Cardinal; ExtStyles: Cardinal; const ClassOrWndText: String = '';
			  CheckWindowText: Longbool = FALSE): THandle;
function GetTopWindow(ParentWindow: THandle): THandle; overload;
function GetTopWindow(ParentWindow: THandle; SkipScrollBar: Longbool): THandle; overload;
function GetWindowState(Window: THandle): Cardinal; overload;
function SetWindowPlacement(WindowHandle: THandle; ShowParam: Cardinal): Longbool; overload;
function IsClassName(Window: THandle; const Name: String; CheckRegEx: Longbool = FALSE): Longbool;
function IsDialogWindow(Window: THandle; Child: Longbool = FALSE): Longbool;
function CreateDialogWindow(const Name: String): THandle;
procedure SetDialogForeground(Dialog, Parent: THandle);
function TrackMouseEvents(Window: THandle; Events: DWord = TME_HOVER or TME_LEAVE; HoverTime: DWord = HOVER_DEFAULT): Longbool;


const
  LSFW_LOCK = 1;
  LSFW_UNLOCK = 2;

function LockSetForegroundWindow(uLockCode: UINT): BOOL; stdcall;
	 external 'user32.dll' name 'LockSetForegroundWindow';

function GetMenuPosFromID(hmenu: HMENU; id: UINT): Integer; stdcall;
	 external 'shlwapi.dll' name 'GetMenuPosFromID';

const // for TWindowPlacement
  WPF_SETMINPOSITION = $0001;
  WPF_RESTORETOMAXIMIZED = $0002;
  WPF_ASYNCWINDOWPLACEMENT = $0004;


implementation

uses
  StrTools, FileTools;

function GetCursorPos: TPoint; inline;
begin
  if not Windows.GetCursorPos(Result) then Result := OFFSCREEN_POS;
end;

function SysCursorSize: TSize; inline;
begin
  Result.cx := GetSystemMetrics(SM_CXCURSOR);
  Result.cy := GetSystemMetrics(SM_CYCURSOR);
end;

function SetWaitCursor: THandle;
begin
  Result := SetCursor(LoadCursor(0, IDC_WAIT));
end;

function MakePoints(Pos: Cardinal): TPoint; assembler;
asm
  {$IFDEF CPUX64}
  // LoWord(ECX) = X, HiWord(ECX) = Y, EAX replaces Result
  MOVZX	RAX,CX
  XOR	CX,CX
  SHL	RCX,10h
  OR	RAX,RCX
  {$ELSE}
  // LoWord(EAX) = X, HiWord(EAX) = Y, EDX = @Result
  MOVZX	ECX,AX
  MOV	[EDX],ECX
  SHR	EAX,10h
  MOV	[EDX + 04h],EAX
  MOV	EAX,EDX
  {$ENDIF}
end;

function MakeSize(Val: Cardinal): TSize; inline;
begin
  with Result do begin
    cx := Val;
    cy := Val;
  end;
end;

function MakeSize(cx, cy: Integer): TSize; inline;
begin
  Result.cx := cx;
  Result.cy := cy;
end;

function EqualSize(const Sz1, Sz2: TSize): Longbool; inline;
begin
  Result := (Sz1.cx = Sz2.cx) and (Sz1.cy = Sz2.cy);
end;

function NullSize(const Size: TSize): Longbool; inline;
begin
  with Size do Result := (cx = 0) and (cy = 0);
end;

function NullPoint(const Point: TPoint): Longbool;
begin
  with Point do Result := (X = 0) and (Y = 0);
end;

function PointToPos(const Point: TPoint): Cardinal; inline;
begin
  with Point do Result := (Y shl 16) or (X and $FFFF);
end;

function CenterPoint(const Rect: TRect): TPoint; inline;
begin
  Result := Types.CenterPoint(Rect);
end;

function SizeToPoint(const Size: TSize): TPoint; inline;
begin
  Result := TPoint(Size);
end;

function PointToSize(const Point: TPoint): TSize; inline;
begin
  Result := TSize(Point);
end;

function IsOffScreenPos(const Point: TPoint): Longbool; inline;
begin
  with Point do Result := (X = -1) or (Y = -1);
end;

function GetPointOffset(const Pt1, Pt2: TPoint): TPoint; inline;
begin
  with Result do begin
    X := Pt1.X - Pt2.X;
    Y := Pt1.Y - Pt2.Y;
  end;
end;

function OffsetPoint(const Point: TPoint; DX, DY: Integer): TPoint;
begin
  Result := Point;
  with Result do begin
    inc(X, DX);
    inc(Y, DY);
  end;
end;

function OffsetSize(const Size: TSize; DX, DY: Integer): TSize;
begin
  Result := Size;
  with Result do begin
    inc(cx, DX);
    inc(cy, DY);
  end;
end;

function SamePoints(const Point1, Point2: TPoint): Longbool; inline;
begin
  Result := PointsEqual(Point1, Point2);
end;

function SameSizes(const Size1, Size2: TSize): Longbool;
begin
  Result := (Size1.cx = Size2.cx) and (Size1.cy = Size2.cy);
end;

function Get_X_lParam(lParam: Integer): SmallInt; inline;
begin
  Result := LoWord(lParam);
end;

function Get_Y_lParam(lParam: Integer): SmallInt; inline;
begin
  Result := HiWord(lParam);
end;

function Make_lParam(Low, High: Word): Longint; inline;
begin
  Result := High shl 16 or Low;
end;

function Make_lParam(const Point: TPoint): Longint; inline;
begin
  with Point do Result := Word(X) shl 16 or Word(Y);
end;

function GetAsyncMouseBtnUnswitched(VKCode: Integer): SmallInt;
begin
  if GetSystemMetrics(SM_SWAPBUTTON) <> 0 then begin
    case VKCode of
      VK_LBUTTON: VKCode := VK_RBUTTON;
      VK_RBUTTON: VKCode := VK_LBUTTON;
    end;
  end;
  Result := GetAsyncKeyState(VKCode);
end;

function PointsFromPixels(Pixels: Integer): Integer;
var
  hDC: THandle;
begin
  hDC := GetDC(0);
  try
    Result := Abs(MulDiv(Pixels, 72, GetDeviceCaps(hDC, LOGPIXELSY)));
  finally
    ReleaseDC(0, hDC);
  end;
end;

function PixelsFromPoints(Points: Integer): Integer;
var
  hDC: THandle;
begin
  hDC := GetDC(0);
  try
    Result := Abs(MulDiv(Points, GetDeviceCaps(hDC, LOGPIXELSY), 72));
  finally
    ReleaseDC(0, hDC);
  end;
end;

function PixelsToTwips(Pixels: Integer): Integer;
var
  hDC: THandle;
begin
  hDC := GetDC(0);
  try
    Result := MulDiv(Pixels, 1440, GetDeviceCaps(hDC, LOGPIXELSY));
  finally
    ReleaseDC(0, hDC);
  end;
end;

function PixelsFromTwips(Twips: Integer): Integer;
var
  hDC: THandle;
begin
  hDC := GetDC(0);
  try
    Result := MulDiv(Twips, GetDeviceCaps(hDC, LOGPIXELSY), 1440);
  finally
    ReleaseDC(0, hDC);
  end;
end;

function MilliMetersFromTwips(Twips: Integer): Integer;
var
  rm: TFPURoundingMode;
begin
  rm := SetRoundMode(rmNearest);
  try
    Result := Round(Twips / 25.4);
  finally
    SetRoundMode(rm);
  end;
end;

function TwipsFromMilliMeters(MilliMeters: Integer): Integer;
var
  rm: TFPURoundingMode;
begin
  rm := SetRoundMode(rmNearest);
  try
    Result := Round(MilliMeters * 25.4);
  finally
    SetRoundMode(rm);
  end;
end;

function MMtoPixWidth(DC: THandle; MM: Integer): Cardinal;
begin
  if DC = 0 then DC := GetDC(0);
  try
    Result := MulDiv(MM, GetDeviceCaps(DC, HORZRES), GetDeviceCaps(DC, HORZSIZE));
  finally
    ReleaseDC(0, DC);
  end;
end;

function MMtoPixHeight(DC: THandle; MM: Integer): Cardinal;
begin
  if DC = 0 then DC := GetDC(0);
  try
    Result := MulDiv(MM, GetDeviceCaps(DC, VERTRES), GetDeviceCaps(DC, VERTSIZE));
  finally
    ReleaseDC(0, DC);
  end;
end;

function GetPixDimsFromTwips(DC: THandle; const TwipsDims: TSize): TSize;
begin
  if DC = 0 then DC := GetDC(0);
  try
    Result.cx := MulDiv(TwipsDims.cx, GetDeviceCaps(DC, LOGPIXELSX), 1440);
    Result.cy := MulDiv(TwipsDims.cy, GetDeviceCaps(DC, LOGPIXELSY), 1440);
  finally
    ReleaseDC(0, DC);
  end;
end;

function GetTwipsDimsFromPixels(DC: THandle; const Pixels: TSize): TSize;
begin
  if DC = 0 then DC := GetDC(0);
  try
    Result.cx := MulDiv(Pixels.cx, 1440, GetDeviceCaps(DC, LOGPIXELSX));
    Result.cy := MulDiv(Pixels.cy, 1440, GetDeviceCaps(DC, LOGPIXELSY));
  finally
    ReleaseDC(0, DC);
  end;
end;

function GetShellDlgFontName(Theme: THandle; FontID: Integer): String;
const
  MsShellDlg = 'MS Shell Dlg';
  MsShellDlg2 = MsShellDlg + SP + '2';
var
  lfSystem: TLogFont;
begin
  if Theme <> 0 then begin
    FillChar(lfSystem, SizeOf(TLogFont), 0);
    if GetThemeSysFont(Theme, FontID, lfSystem) = S_OK then begin
      Result := GetValueFromLogFont(lfSystem, fpName);
      if Result <> E then Exit;
    end;
  end;
  Result := GetRegString(HKEY_LOCAL_MACHINE, 'Software\Microsoft\Windows NT\CurrentVersion\FontSubstitutes', MsShellDlg2);
  if Result <> E then Exit;
  Result := MsShellDlg;
end;

type
  TFontFamily = (
    FF_DONTCARE   = (0 shl 4),
    FF_ROMAN      = (1 shl 4),
    FF_SWISS      = (2 shl 4),
    FF_MODERN     = (3 shl 4),
    FF_SCRIPT     = (4 shl 4),
    FF_DECORATIVE = (5 shl 4),
    FF_TECH       = (6 shl 4),
    FF_BIDI       = (7 shl 4)
  );

function GetFontFamilyName(FontFamily: TFontFamily): String;
begin
  case FontFamily of
    FF_DONTCARE   : Result := 'nil';
    FF_ROMAN      : Result := 'roman';
    FF_SWISS      : Result := 'swiss';
    FF_MODERN     : Result := 'modern';
    FF_SCRIPT     : Result := 'script';
    FF_DECORATIVE : Result := 'decorative';
    FF_TECH       : Result := 'tech';
    FF_BIDI       : Result := 'bidi';
  end;
end;

function GetValueFromLogFont(const LogFont: TLogFont; FontParams: TFontParams; AsString: Longbool = FALSE): Variant;
begin
  with LogFont do begin
    case FontParams of
      fpSize: Result := PointsFromPixels(lfHeight);
      fpWidth: Result := lfWidth;
      fpEscapement: Result := lfEscapement;
      fpOrientation: Result := lfOrientation;
      fpBold: Result := (lfWeight >= FW_BOLD);
      fpItalic: Result := lfItalic;
      fpUnderline: Result := lfUnderline;
      fpStrikeOut: Result := lfStrikeOut;
      fpCharSet: Result := lfCharSet;
      fpOutPrecision: Result := lfOutPrecision;
      fpClipPrecision: Result := lfClipPrecision;
      fpQuality: Result := lfQuality;
      fpPitchAndFamily: Result := lfPitchAndFamily;
      fpPitch: Result := lfPitchAndFamily and $03;
      fpFamily: begin
	if not AsString then
	  Result := lfPitchAndFamily and $F0
	else Result := GetFontFamilyName(TFontFamily(lfPitchAndFamily and $F0));
	Exit;
      end;
      fpName: begin
	Result := String(lfFaceName);
	Exit;
      end;
      fpColor: Result := clWindowText;
    end;
  end;
  if AsString then Result := IntToStr(Result);
end;

procedure ProcessMessages(NoUserInput: Longbool = FALSE);
var
  msg: TMsg;
begin
  while PeekMessage(msg, 0, 0, 0, PM_REMOVE) do begin
    if NoUserInput then begin
      case msg.Message of
	WM_KEYFIRST..WM_KEYLAST: Sleep(0);
	WM_MOUSEFIRST..WM_MOUSELAST: Sleep(0);
      else
	TranslateMessage(msg);
	DispatchMessage(msg);
      end;
    end
    else begin
      TranslateMessage(msg);
      DispatchMessage(msg);
    end;
  end;
end;

function GetSysScreenRect: TRect;
begin
  SystemParametersInfo(SPI_GETWORKAREA, 0, @Result, 0);
end;

function GetSysScreenDimensions: TSize;
begin
  with GetSysScreenRect do Result := MakeSize(Right - Left, Bottom - Top);
end;

function GetFocusWindowPos: TPoint;
begin
  Result := GetWindowPos(GetFocus);
end;

function GetWindowAtCursor: THandle;
begin
  Result := WindowFromPoint(GetCursorPos);
end;

function GetClientPos(Window: THandle; const Pos: TPoint): TPoint;
begin
  Result := Pos;
  if not Windows.ScreenToClient(Window, Result) then Result := OFFSCREEN_POS;
end;

function GetWindowPos(Window: THandle): TPoint;
var
  rect: TRect;
begin
  if Windows.GetWindowRect(Window, rect) then
    Result := rect.TopLeft
  else Result := OFFSCREEN_POS;
end;

function GetClientSizes(Window: THandle): TSize;
begin
  Result := TSize(GetClientRect(Window).BottomRight);
end;

function GetControlRect(Control: THandle): TRect;
var
  hParent: THandle;
begin
  hParent := GetParent(Control);
  Result := GetClientRect(Control);
  if Windows.ClientToScreen(Control, Result.TopLeft) and Windows.ScreenToClient(hParent, Result.TopLeft) then with Result do begin
    inc(Right, Left);
    inc(Bottom, Top);
    Exit;
  end;
  Result := EMPTY_RECT;
end;

function GetChildWindowRect(Child: THandle; Parent: THandle = 0): TRect;
begin
  if Parent = 0 then Parent := GetParent(Child);
  Result := WindowTools.IntersectRect(GetWindowRect(Parent), GetWindowRect(Child));
end;

function ClientToScreen(Window: THandle; const Point: TPoint): TPoint;
begin
  Result := Point;
  if not Windows.ClientToScreen(Window, Result) then Result := OFFSCREEN_POS;
end;

function ScreenToClient(Window: THandle; const Point: TPoint): TPoint;
begin
  Result := Point;
  if not Windows.ScreenToClient(Window, Result) then Result := OFFSCREEN_POS;
end;

function ClientToScreen(Window: THandle; var Rect: TRect): Longbool;
begin
  with Rect do Result := Windows.ClientToScreen(Window, TopLeft) and Windows.ClientToScreen(Window, BottomRight);
end;

function ScreenToClient(Window: THandle; var Rect: TRect): Longbool;
begin
  with Rect do Result := Windows.ScreenToClient(Window, TopLeft) and Windows.ScreenToClient(Window, BottomRight);
end;

function ClientToScreenRect(Window: THandle; const Rect: TRect): TRect;
begin
  Result := Rect;
  if not ClientToScreen(Window, Result) then Result := EMPTY_RECT;
end;

function ScreenToClientRect(Window: THandle; const Rect: TRect): TRect;
begin
  Result := Rect;
  if not ScreenToClient(Window, Result) then Result := EMPTY_RECT;
end;

function AddBorder(const Rect, Border: TRect): TRect;
begin
  Result := Rect;
  with Result do begin
    dec(Left, Border.Left);
    dec(Top, Border.Top);
    inc(Right, Border.Right);
    inc(Bottom, Border.Bottom);
  end;
end;

function AddBorder(const Rect: TRect; L, T, R, B: Integer): TRect;
begin
  Result := Rect;
  with Result do begin
    dec(Left, L);
    dec(Top, T);
    inc(Right, R);
    inc(Bottom, B);
  end;
end;

function InflateRect(const Rect: TRect; DX, DY: Integer): TRect;
begin
  Result := Rect;
  if not Windows.InflateRect(Result, DX, DY) then Result := Rect;
end;

function OffsetRect(const Rect: TRect; DX, DY: Integer): TRect;
begin
  Result := Rect;
  if not Windows.OffsetRect(Result, DX, DY) then Result := Rect;
end;

function IntersectRect(const Rect1, Rect2: TRect): TRect;
begin
  if not Types.IntersectRect(Result, Rect1, Rect2) then Result := EMPTY_RECT;
end;

function GetBoundsRect(const Rect: TRect): TRect;
begin
  Result := Rect;
  with Result do begin
    dec(Right, Left);
    dec(Bottom, Top);
  end;
end;

function GetBoundsRect(Window: THandle): TRect;
begin
  Result := GetBoundsRect(GetWindowRect(Window));
end;

function GetClientRect(Window: THandle): TRect;
begin
  if not Windows.GetClientRect(Window, Result) then Result := EMPTY_RECT;
end;

function GetWindowRect(Window: THandle): TRect;
begin
  if not Windows.GetWindowRect(Window, Result) then Result := EMPTY_RECT;
end;

function GetClientCenter(Window: THandle): TPoint;
begin
  with Result, GetClientRect(Window) do begin
    X := Right div 2;
    Y := Bottom div 2;
  end;
end;

function GetClientWidth(Window: THandle): Integer;
begin
  Result := GetClientRect(Window).Right;
  if Result > 0 then dec(Result, GetSystemMetrics(SM_CXVSCROLL));
end;

function GetClientHeight(Window: THandle): Integer;
begin
  Result := GetClientRect(Window).Bottom;
  if Result > 0 then dec(Result, GetSystemMetrics(SM_CYHSCROLL));
end;

function IsAllRectEmpty(const Rect: TRect): Longbool;
begin
  Result := EqualRect(Rect, EMPTY_RECT);
end;

function IsZeroSize(const Size: TSize): Longbool;
begin
  with Size do Result := (cx = 0) and (cy = 0);
end;

function IsNoSize(const Size: TSize): Longbool;
begin
  with Size do Result := (Abs(cx) = 1) and (Abs(cy) = 1);
end;

function GetWindowModuleHandle(Window: THandle): THandle;
begin
  Result := GetWindowLongPtr(Window, GWLP_HINSTANCE);
end;

function GetWindowModuleName(Window: THandle): String;
var
  aModuleName: TPathArray;
begin
  SetString(Result, aModuleName, GetWindowModuleFileName(Window, @aModuleName, SizeOf(aModuleName)));
end;

function CenterDialog(Dialog, Owner: THandle): Longbool;
var
  hFocus: THandle;
  rDialog, rOwner: TRect;
begin
  Result := FALSE;
  if Owner = 0 then Owner := GetDesktopWindow;
  if (Dialog <> 0) and (Owner <> 0) then begin
    rOwner := GetWindowRect(Owner);
    if not IsAllRectEmpty(rOwner) then begin
      with rOwner do begin
	dec(Right, pred(Left));
	dec(Bottom, pred(Top));
      end;
      rDialog := GetWindowRect(Dialog);
      hFocus := GetFocus;
      with rDialog do begin
	dec(Right, pred(Left));
	dec(Bottom, pred(Top));
	Left := rOwner.Left + (rOwner.Right - Right) div 2;
	Top := rOwner.Top + (rOwner.Bottom - Bottom) div 2;
	Result := SetWindowPos(Dialog, HWND_TOP, Left, Top, 0, 0, SWP_NOACTIVATE or SWP_NOSIZE);
	if not IsWindowEnabled(Owner) and (hFocus <> Dialog) and not IsChild(Dialog, hFocus) then Windows.SetFocus(Dialog);
      end;
    end;
  end;
end;

function CenterInClientRect(const ChildRect: TRect; Child, Parent: THandle; Draw: Longbool): Longbool;
var
  rParent, rChild: TRect;
begin
  Result := FALSE;
  if Parent = 0 then Parent := GetDesktopWindow;
  if not IsAllRectEmpty(ChildRect) and (Parent <> 0) then begin
    rParent := GetClientRect(Parent);
    if not IsAllRectEmpty(rParent) then begin
      rChild := ChildRect;
      with rChild do begin
	Left := rParent.Left + (rParent.Right - Right) div 2;
	Top := rParent.Top + (rParent.Bottom - Bottom) div 2;
	Result := MoveWindow(Child, Left, Top, Right, Bottom, Draw);
      end;
    end;
  end;
end;

function FindControlByID(Parent: THandle; const ClassOrWndText: String; ControlId: Integer = 0; CheckWindowText: Longbool = FALSE): THandle;
begin
  Result := FindChildWindow(Parent, ClassOrWndText, ControlId, FALSE, CheckWindowText, TRUE);
end;

type
  TChildAtPosParams = Record
    Executable: String;
    ClassOrWndText: String;
    TextLength: Integer;
    CheckWndTxt: Longbool;
    Position: TPoint;
    ChildHandle: THandle;
  end;

function EnumFindChildWndAtPos(Window: THandle; var SP: TChildAtPosParams): Bool; stdcall;
var
  sClassOrWndText: String;
begin
  Result := TRUE;
  if not (IsWindowEnabled(Window) and IsWindowVisible(Window)) then Exit;
  with SP do begin
    if not StrTools.SameText(Executable, GetWindowModuleName(Window)) then Exit;
    if ClassOrWndText <> E then begin
      if not CheckWndTxt then
	sClassOrWndText := ClassString(Window)
      else sClassOrWndText := DelAmp(WindowStringW(Window));
      if (sClassOrWndText = E) or (Length(sClassOrWndText) <> TextLength) then Exit;
      if not StrTools.SameText(sClassOrWndText, ClassOrWndText) then Exit;
    end;
    if PtInRect(GetWindowRect(Window), Position) then begin
      ChildHandle := Window;
      Result := FALSE;
    end;
  end;
end;

function FrameRect3D(DC: THandle; const Rect: TRect; TopLeftColor, BottomRightColor: TColorRef): Longbool;
type
  PPoints = ^TPoints;
  TPoints = Array[0..0] of TPoint;

  function Polyline(DC: THandle; const Points: Array of TPoint): Longbool;
  begin
    Result := Windows.Polyline(DC, PPoints(@Points)^, succ(High(Points)));
  end;

var
  hCurPen, hTLPen, hBRPen: THandle;
begin
  if DC <> 0 then begin
    hCurPen := GetCurrentObject(DC, OBJ_PEN);
    hTLPen := 0;
    hBRPen := 0;
    try
      hTLPen := CreatePen(PS_SOLID, 1, TopLeftColor);
      SelectObject(DC, hTLPen);
      with Rect do Result := Polyline(DC, [Point(Left, Bottom), Point(Left, Top), Point(Left, Top), Point(Right, Top)]);
      if not Result then Exit;
      hBRPen := CreatePen(PS_SOLID, 1, BottomRightColor);
      SelectObject(DC, hBRPen);
      with Rect do Result := Polyline(DC, [Point(Right, Top - 1), Point(Right, Bottom), Point(Right, Bottom), Point(Left - 2, Bottom)]);
      if not Result then Exit;
    finally
      SelectObject(DC, hCurPen);
      if hBRPen <> 0 then DeleteObject(hBRPen);
      if hTLPen <> 0 then DeleteObject(hTLPen);
    end;
    Exit;
  end;
  Result := FALSE;
end;

function FindChildWindowAtPos(Parent: THandle; const ScreenPos: TPoint; const Executable: String;
			      const ClassOrWndText: String = E; CheckWindowText: Longbool = FALSE): THandle;
var
  sp: TChildAtPosParams;
begin
  sp.Executable := Executable;
  sp.ClassOrWndText := ClassOrWndText;
  sp.TextLength := Length(ClassOrWndText);
  sp.CheckWndTxt := CheckWindowText;
  sp.Position := ScreenPos;
  sp.ChildHandle := 0;
  EnumChildWindows(Parent, @EnumFindChildWndAtPos, LPARAM(@sp));
  Result := sp.ChildHandle;
end;

type
  TChildWndParams = Record
    ClassOrWndText: String;
    TextLength: Integer;
    ControlId: Integer;
    CheckRegEx: Longbool;
    CheckWndTxt: Longbool;
    IsDlgControl: Longbool;
    Instance: Cardinal;
    ChildHandle: THandle;
  end;

function EnumFindChildWindow(Window: THandle; var SP: TChildWndParams): Bool; stdcall;
var
  sClassOrWndText: String;
begin
  Result := TRUE;
  with SP do begin
    if IsDlgControl and (ClassString(Window) = CLASS_DIALOG) then begin
      if (ControlId <> 0) and (GetDlgCtrlID(Window) <> ControlId) then Exit;
      sClassOrWndText := DelAmp(DialogString(Window, ControlId));
      if (sClassOrWndText = E) or ((not CheckRegEx) and (Length(sClassOrWndText) <> TextLength)) then Exit;
      if not (CheckRegEx and CheckAST_QMRegEx(sClassOrWndText, ClassOrWndText))
      and not StrTools.SameText(sClassOrWndText, ClassOrWndText) then Exit;
      Window := GetDlgItem(Window, ControlId);
    end
    else begin
      if not CheckWndTxt then
	sClassOrWndText := ClassString(Window)
      else begin
	sClassOrWndText := DelAmp(WindowStringW(Window));
	if IsStringUpperCase(ClassOrWndText) then sClassOrWndText := StrTools.AnsiUpperCase(sClassOrWndText);
      end;
      if (sClassOrWndText = E) or ((not CheckRegEx) and (Length(sClassOrWndText) <> TextLength)) then Exit;
      if not (CheckRegEx and CheckAST_QMRegEx(sClassOrWndText, ClassOrWndText))
      and not StrTools.SameText(sClassOrWndText, ClassOrWndText) then Exit;
      if (ControlId <> 0) and (GetDlgCtrlID(Window) <> ControlId) then Exit;
    end;
    dec(Instance);
    if Instance = 0 then begin
      ChildHandle := Window;
      Result := FALSE;
    end;
  end;
end;

function FindChildWindow(Parent: THandle; const ClassOrWndText: String; ControlId: Integer = 0;
			 CheckRegEx: Longbool = FALSE; CheckWindowText: Longbool = FALSE;
			 IsDialogControl: Longbool = FALSE; Instance: Cardinal = 1): THandle;
var
  sp: TChildWndParams;
begin
  sp.ClassOrWndText := ClassOrWndText;
  sp.TextLength := Length(ClassOrWndText);
  sp.ControlId := ControlId;
  sp.CheckRegEx := CheckRegEx;
  sp.CheckWndTxt := CheckWindowText;
  sp.IsDlgControl := IsDialogControl;
  sp.Instance := Instance;
  sp.ChildHandle := 0;
  EnumChildWindows(Parent, @EnumFindChildWindow, LPARAM(@sp));
  Result := sp.ChildHandle;
end;

type
  TChildWndInClassParams = Record
    ClassNameList: TClassNames;
    LastInstance: Longbool;
    ChildHandle: THandle;
    ChildClass: String;
  end;

function EnumFindChildWindowInClass(Window: THandle; var SP: TChildWndInClassParams): Bool; stdcall;
var
  sClassName: String;
begin
  Result := TRUE;
  with SP do begin
    sClassName := ClassString(Window);
    if MatchText(sClassName, ClassNameList, Length(sClassName)) then begin
      if Window <> 0 then begin
	Result := LastInstance;
	ChildHandle := Window;
      end;
    end;
  end;
end;

function FindChildWindowInClass(Parent: THandle; const ClassNames: TClassNames; LastInstance: Longbool = FALSE): THandle;
var
  sp: TChildWndInClassParams;
begin
  sp.ClassNameList := ClassNames;
  sp.LastInstance := LastInstance;
  sp.ChildHandle := 0;
  sp.ChildClass := E;
  EnumChildWindows(Parent, @EnumFindChildWindowInClass, LPARAM(@sp));
  Result := sp.ChildHandle;
end;

type
  TThreadWndParams = Record
    ExtendedStyles: Cardinal;
    ClassOrWndText: String;
    TextLength: Integer;
    CheckWndTxt: Longbool;
    FoundHandle: THandle;
  end;

function EnumFindThreadWindow(Window: THandle; var SP: TThreadWndParams): Bool; stdcall;
var
  sClassOrWndText: String;
begin
  Result := TRUE;
  with SP do begin
    if (GetWindowLongPtr(Window, GWL_EXSTYLE) and ExtendedStyles) <> ExtendedStyles then Exit;
    if not CheckWndTxt then
      sClassOrWndText := ClassString(Window)
    else begin
      sClassOrWndText := DelAmp(WindowStringW(Window));
      if IsStringUpperCase(ClassOrWndText) then sClassOrWndText := StrTools.AnsiUpperCase(sClassOrWndText);
    end;
    if (sClassOrWndText = E) or (Length(sClassOrWndText) <> TextLength) then Exit;
    if not StrTools.SameText(sClassOrWndText, ClassOrWndText) then Exit;
    FoundHandle := Window;
    Result := FALSE;
  end;
end;

function FindThreadWindow(ThreadId: Cardinal; ExtStyles: Cardinal; const ClassOrWndText: String = '';
			  CheckWindowText: Longbool = FALSE): THandle;
var
  sp: TThreadWndParams;
begin
  sp.ExtendedStyles := ExtStyles;
  sp.ClassOrWndText := ClassOrWndText;
  sp.TextLength := Length(ClassOrWndText);
  sp.CheckWndTxt := CheckWindowText;
  sp.FoundHandle := 0;
  EnumThreadWindows(ThreadId, @EnumFindThreadWindow, LPARAM(@sp));
  Result := sp.FoundHandle;
end;

function GetTopWindow(ParentWindow: THandle): THandle;
begin
  Result := Windows.GetTopWindow(ParentWindow);
end;

function GetTopWindow(ParentWindow: THandle; SkipScrollBar: Longbool): THandle;
begin
  Result := Windows.GetTopWindow(ParentWindow);
  if ClassString(Result) = 'SCROLLBAR' then Result := GetParent(Result);
end;

function GetWindowState(Window: THandle): Cardinal;
var
  wp: TWindowPlacement;
begin
  FillChar(wp, SizeOf(wp), 0);
  wp.length := SizeOf(wp);
  if GetWindowPlacement(Window, wp) then begin
    Result := wp.showCmd;
    Exit;
  end;
  Result := MAXDWORD;
end;

function SetWindowPlacement(WindowHandle: THandle; ShowParam: Cardinal): Longbool;
var
  wp: TWindowPlacement;
begin
  FillChar(wp, SizeOf(wp), 0);
  wp.length := SizeOf(wp);
  if GetWindowPlacement(WindowHandle, wp) then begin
    wp.showCmd := ShowParam;
    wp.rcNormalPosition := GetSysScreenRect;
    Result := Windows.SetWindowPlacement(WindowHandle, wp);
    Exit;
  end;
  Result := FALSE;
end;

function IsClassName(Window: THandle; const Name: String; CheckRegEx: Longbool = FALSE): Longbool;
var
  sClassName: String;
begin
  sClassName := ClassString(Window);
  if not CheckRegEx then
    Result := StrTools.SameText(sClassName, Name)
  else Result := CheckAST_QMRegEx(sClassName, Name);
end;

function IsDialogWindow(Window: THandle; Child: Longbool = FALSE): Longbool;
begin
  Result := (Child or ((GetWindowLongPtr(Window, GWL_STYLE) and WS_CHILD) <> WS_CHILD)) and (ClassString(Window) = CLASS_DIALOG);
end;

function CreateDialogWindow(const Name: String): THandle;
begin
  Result := CreateWindow(CLASS_DIALOG, PChar(Name), 0, 0, 0, 0, 0, 0, 0, HInstance, NIL);
end;

procedure SetDialogForeground(Dialog, Parent: THandle);
var
  wp: TWindowPlacement;
begin
  if Parent <> 0 then begin
    FillChar(wp, SizeOf(wp), 0);
    wp.length := SizeOf(wp);
    if GetWindowPlacement(Parent, wp) and (wp.showCmd = SW_SHOWMINIMIZED) then with wp do begin
      flags := WPF_ASYNCWINDOWPLACEMENT;
      showCmd := SW_SHOWNOACTIVATE;
      SetWindowPlacement(Parent, wp);
    end;
  end;
  CenterDialog(Dialog, Parent);
end;

function TrackMouseEvents(Window: THandle; Events: DWord = TME_HOVER or TME_LEAVE; HoverTime: DWord = HOVER_DEFAULT): Longbool;
var
  tme: Windows.TTrackMouseEvent;
begin
  with tme do begin
    cbSize := SizeOf(tme);
    dwFlags := Events;
    hwndTrack := Window;
    dwHoverTime := HoverTime;
  end;
  Result := Windows.TrackMouseEvent(tme);
end;

end.

