{------------------------------------------------------------------------------}
{        This code is copyright (C) 2010-2025 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit StrTools;

interface

uses
  Generics.Defaults, Windows, WinError, NtStatus, ShLwApi, WinHTTP, WinInet, WinCrypt, Variants, ActiveX, Rtti,
  Math, Graphics, GraphUtil, Classes, SysUtils, StrUtils, AnsiStrings, WideStrUtils, Character;

type
  RAW = RawByteString;
  PAChar = PAnsiChar;
  PWChar = PWideChar;

const // string length offset
  SLO32 = SizeOf(Integer);

const // web document doctypes
  BASE_64 = 'base64';
  DOC_HTML = 'html';
  DOC_SVG = 'svg';

const // common char constants
  E = '';
  SP = ' ';
  DDOT = '..';
  DBS = SP + SP;
  QBS = DBS + DBS;

  AT = '@';
  AM = '&';
  PC = '%';
  HT = '#';
  US = '$';

  CL = ':';
  CM = ',';
  SC = ';';

  EM = '!';
  QM = '?';

  BS = '\';
  SL = '/';

  DS = SL + SL;
  TS = DS + SL;

  LT = '<';
  GT = '>';

  LB = '(';
  RB = ')';

  LS = '[';
  RS = ']';

  LC = '{';
  RC = '}';

  DA = '-';
  UD = '_';

  DD = DA + DA;

  EQ = '=';
  QU = '"';
  SQ = '''';

  DQ = QU + QU;
  DSQ = SQ + SQ;

  LQ = '&#xBB;';
  RQ = '&#xAB;';

  AST = '*';
  DOT = '.';
  TLD = '~';

  PLS = '+';
  MNS = '-';

  NUL = #0;
  TAB = #9;

  LF = #10;
  CR = #13;

  NL = #13#10;

  SDS = ' - ';

  CRLF = NL;
  ZERO = '0';
  NBSP = #160;
  PARA = NL + NL;

  QUOTES = [QU, SQ];
  SLASHES = [BS, SL];

  DIGITS = ['0'..'9'];
  LO_ASCII = ['a'..'z'];
  HI_ASCII = ['A'..'Z'];
  TRIMCHARS = [#01..#32];
  WEB_TRIM = TRIMCHARS + [NBSP];

  LO_CHARS = [$61..$7A, $E0..$F6];
  HI_CHARS = [$41..$5A, $C0..$D6];

  WSP = [LF, CR, SP];
  WSPT = WSP + [TAB];
  NBLANKS = [NUL..SP];
  BLANKS = [TAB, LF, CR, SP];
  BRACKETS = [LB, RB, LT, GT, LS, RS, LC, RC];

  NON_ASCII = [NUL, #$80..#$FF];
  ALL = [TAB, QU, AM, SQ, LT, GT] + NON_ASCII;

  HEXCHARS = DIGITS + ['A'..'F'];
  FLOATCHARS = [' ', '+', '-', '.'] + DIGITS;

  WORDDELIMITERS = [#0..#255] - ['a'..'z', 'A'..'Z', '0'..'9', #$C0..#$D6, #$E0..#$F6];

  HEX_MARKER = '0x';
  RTF_MARKER = '{\rtf';

  CHARSET = 'charset=';

  FONT_ARIAL = 'Arial';
  FONT_SEGOE = 'Segoe UI';
  FONT_TIMES = 'Times New Roman';

  FONT_CONSOLE = 'Consolas';
  FONT_COURIER = 'Courier New';

  FONT_SYMBOL = 'Segoe UI Symbol';
  FONT_EMOJI = 'Segoe UI Emoji';

const // URL protocols
  PROT_ABOUT	= 'about:';
  PROT_CID	= 'cid:';
  PROT_DATA	= 'data:';
  PROT_FILE	= 'file:';
  PROT_FTP	= 'ftp:';
  PROT_FTPS	= 'ftps:';
  PROT_GOPHER	= 'gopher:';
  PROT_HTTP	= 'http:';
  PROT_HTTPS	= 'https:';
  PROT_SOCKS	= 'socks:';
  PROT_JSCRIPT	= 'javascript:';
  PROT_VBSCRIPT	= 'vbscript:';
  PROT_MAIL	= 'mailto:';
  PROT_NEWS	= 'news:';
  PROT_RES	= 'res:';

const // ... with slashes
  PROT_A_DS  = PROT_ABOUT + DS;
  PROT_C_DS  = PROT_CID + DS;
  PROT_D_DS  = PROT_DATA + DS;
  PROT_M_DS  = PROT_MAIL + DS;
  PROT_F_TS  = PROT_FILE + TS;
  PROT_H_DS  = PROT_HTTP + DS;
  PROT_HS_DS = PROT_HTTPS + DS;
  PROT_R_DS  = PROT_RES + DS;

const
  SDOM_WWW = 'www.';
  ABBL = 'about:blank';

type
  TTrimOption = (toLeft, toRight);
  TTrimOptions = Set of TTrimOption;

const
  TO_BOTH = [toLeft, toRight];

type
  TReplaceStrArrayA = Array of AnsiString;
  TReplaceStringsA = Record
    FromStr: TReplaceStrArrayA;
    ToStr: TReplaceStrArrayA;
  end;

  TReplaceStrArrayW = Array of UnicodeString;
  TReplaceStringsW = Record
    FromStr: TReplaceStrArrayW;
    ToStr: TReplaceStrArrayW;
  end;

type
  TStringSeachOption = (soDown, soMatchCase, soWholeWord);
  TStringSearchOptions = Set of TStringSeachOption;

type
  TWideCharArray = Array of WideChar;
  TAnsiCharArray = Array of AnsiChar;

type
  TQuoteChars = Set of AnsiChar;

type
  DAChar = Array[0..1] of AnsiChar;
  PDAChar = ^DAChar;
  QAChar = Array[0..3] of AnsiChar;
  PQAChar = ^QAChar;

  DWChar = Array[0..1] of WideChar;
  PDWChar = ^DWChar;
  QWChar = Array[0..3] of WideChar;
  PQWChar = ^QWChar;

type
  PDigitArrayA = ^TDigitArrayA;
  TDigitArrayA = Array[0..2] of AnsiChar;
  PDigitArrayW = ^TDigitArrayA;
  TDigitArrayW = Array[0..2] of WideChar;

type
  PPathArrayA = ^TPathArrayA;
  TPathArrayA = Array[0..MAX_PATH] of AnsiChar;
  PPathArrayW = ^TPathArrayW;
  TPathArrayW = Array[0..MAX_PATH] of WideChar;

type
  PPathArray = ^TPathArray;
  TPathArray = TPathArrayW;

type
  TUrlPathA = Array[0..MAXSHORT] of AnsiChar;
  TUrlPathW = Array[0..MAXSHORT] of WideChar;

type
  TCommandLine = Array[0..MAXSHORT] of WideChar;
  TUNCPath = TCommandLine;

type
  TNCLFontType = (ftCaption, ftSmCaption, ftMenu, ftStatus, ftMessage);
  TNCLFontNameSize = Record
    fnsName: String;
    fnsSize: Integer;
  end;

type // redefined here because of type mismatch of default RichEdit unit definition !
  PCharFormat = ^TCharFormat;
  TCharFormat = Record
    cbSize: UInt;
    dwMask: DWord;
    dwEffects: DWord;
    yHeight: Longint;
    yOffset: Longint;
    crTextColor: TColorRef;
    bCharSet: Byte;
    bPitchAndFamily: Byte;
    szFaceName: Array[0..LF_FACESIZE - 1] of WideChar;
  end;

 type // redefined here because of wrong declaration as packed in Windows unit !
  POutlineTextmetric = ^TOutlineTextmetric;
  TOutlineTextmetric = Record
    otmSize: UINT;
    otmTextMetrics: TTextMetric;
    otmFiller: Byte;
    otmPanoseNumber: TPanose;
    otmfsSelection: UINT;
    otmfsType: UINT;
    otmsCharSlopeRise: Integer;
    otmsCharSlopeRun: Integer;
    otmItalicAngle: Integer;
    otmEMSquare: UINT;
    otmAscent: Integer;
    otmDescent: Integer;
    otmLineGap: UINT;
    otmsCapEmHeight: UINT;
    otmsXHeight: UINT;
    otmrcFontBox: TRect;
    otmMacAscent: Integer;
    otmMacDescent: Integer;
    otmMacLineGap: UINT;
    otmusMinimumPPEM: UINT;
    otmptSubscriptSize: TPoint;
    otmptSubscriptOffset: TPoint;
    otmptSuperscriptSize: TPoint;
    otmptSuperscriptOffset: TPoint;
    otmsStrikeoutSize: UINT;
    otmsStrikeoutPosition: Integer;
    otmsUnderscoreSize: Integer;
    otmsUnderscorePosition: Integer;
    otmpFamilyName: PWideChar;
    otmpFaceName: PWideChar;
    otmpStyleName: PWideChar;
    otmpFullName: PWideChar;
  end;


// TStringRecs -----------------------------------------------------------------
type
  TStringRecW = Record
    PWChar: PWideChar;
    PWLen: Integer;
  end;
  function StrLECatW(PResult: PWideChar; const Strings: Array of TStringRecW): PWideChar; overload;

type
  TStringRecA = Record
    PAChar: PAnsiChar;
    PALen: Integer;
  end;
  function StrLECatA(PResult: PAnsiChar; const Strings: Array of TStringRecA): PAnsiChar; overload;



// TStringArray ----------------------------------------------------------------
type
  PAStrings = ^TAStrings;
  TAStrings = Array[0..0] of String;
  TStringAdjust = function(const Str: String): String;
  TStringArray = Object
  private
    FCount: Integer;
    FStrings: PAStrings;
    FStringAdjust: TStringAdjust;
    function Grow(Items: Integer): Integer;
  protected
    function AdjustString(const Str: String): String;
  public
    procedure New;
    procedure Clear;
    function  Count: Integer;
    function  GetString(Index: Integer): String;
    procedure SetString(Index: Integer; const Value: String);
    procedure InsertLine(const Value: String; Index: Integer);
    procedure AddLines(const Values: Array of String);
    procedure AddText(const Value: String);
    function  GetText: String;
    property  StringAdjust: TStringAdjust read FStringAdjust write FStringAdjust;
  end;


// TRawByteStringStream --------------------------------------------------------
type
  TRawByteStringStream = class(TStream)
  private
    FDataString: RawByteString;
    FPosition: Integer;
  protected
    procedure SetSize(NewSize: Longint); override;
  public
    constructor Create(const AString: RawByteString);
    function Read(var Buffer; Count: Longint): Longint; override;
    function ReadString(Count: Longint): RawByteString;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    procedure WriteString(const AString: RawByteString);
    property DataString: RawByteString read FDataString;
  end;


procedure Beep(BeepType: Cardinal = MB_ICONHAND);

// URL protocol stuff ----------------------------------------------------------

function GetProtocolFromUrl(const Url: RawByteString; Start: Integer = 1): RawByteString; overload;
function GetProtocolFromUrl(const Url: UnicodeString; Start: Integer = 1): UnicodeString; overload;
function IsProtocolUrl(const Url: RawByteString): Boolean; overload;
function IsProtocolUrl(const Url: UnicodeString): Boolean; overload;
function IsAboutUrl(const Url: RawByteString): Boolean; overload;
function IsAboutUrl(const Url: UnicodeString): Boolean; overload;
function IsBrowserUrl(const Url: RawByteString; CheckDelims: Longbool = FALSE): Boolean; overload;
function IsBrowserUrl(const Url: UnicodeString; CheckDelims: Longbool = FALSE): Boolean; overload;
function IsCidUrl(const Url: RawByteString): Boolean; overload;
function IsCidUrl(const Url: UnicodeString): Boolean; overload;
function IsDataUrl(const Url: RawByteString): Boolean; overload;
function IsDataUrl(const Url: UnicodeString): Boolean; overload;
function IsHttpUrl(const Url: RawByteString): Boolean; overload;
function IsHttpUrl(const Url: UnicodeString): Boolean; overload;
function IsMailToUrl(const Url: RawByteString): Boolean; overload;
function IsMailToUrl(const Url: UnicodeString): Boolean; overload;
function IsResUrl(const Url:  RawByteString): Boolean; overload;
function IsResUrl(const Url: UnicodeString): Boolean; overload;
function IsWWWUrl(const Url:  RawByteString; CheckDelims: Longbool = FALSE): Boolean; overload;
function IsWWWUrl(const Url: UnicodeString; CheckDelims: Longbool = FALSE): Boolean; overload;


// All the rest of it ----------------------------------------------------------

procedure SilentSafeCallException(ErrorCode: Integer; ErrorAddr: Pointer);

function IsContextMenuShortCut: Longbool;

function GetContextMenuShortCut(KeyData: Longint): TShortCut;

function SendKeys(const Keys: Array of Cardinal): Longbool;

type
  TShiftStateEx = Set of (sseShift, sseAlt, sseCtrl, sseLeft, sseRight, sseMiddle, sseDouble, sseTouch, ssePen,
			  sseLShift, sseRShift, sseLCtrl, sseRCtrl, sseLAlt, sseRAlt);

function GetKeyDataToShiftState(KeyData: Longint): TShiftState;
function GetShiftStateFromShiftStateEx(SSE: TShiftStateEx): TShiftState;

function KeyboardStateToShiftStateEx: TShiftStateEx; overload;
function KeyboardStateToShiftStateEx(const KeyboardState: TKeyboardState): TShiftStateEx; overload;

type
  TRegNameFiltered = function(const Name: String): Longbool;

function SCodeError(SCode: Longint): Cardinal;
function HResultError(HRes: HResult): Cardinal;
function MakeSCode(Severity1Bit, Facility11Bits, Code8Bits: ULong): HResult;
function MakeHResult(Severity1Bit, Facility11Bits, Code8Bits: ULong): HResult;
function GetRegKeyList(Key: HKEY; const Path: String; List: TStrings; const ReturnValue: String = E;
		       const RegNameFiltered: TRegNameFiltered = NIL): Cardinal;

function WinNT: Longbool;
function XPSP3: Longbool;
function IsLower(C: WideChar): Longbool;
function IsUpper(C: WideChar): Longbool;
function Asc(C: AnsiChar): Byte; inline; overload;
function Asc(C: WideChar): Byte; inline; overload;
function CharInSet(C: AnsiChar; const S: TSysCharSet): Longbool; overload;
function CharInSet(C: WideChar; const S: TWideCharArray): Longbool; overload;
function CharMove(Dest, Source: PAnsiChar): PAnsiChar; inline; overload;
function CharMove(Dest, Source: PWideChar): PWideChar; inline; overload;
function LoChar(C: AnsiChar): AnsiChar; inline; overload;
function LoChar(C: WideChar): WideChar; inline; overload;
function HiChar(C: AnsiChar): AnsiChar; inline; overload;
function HiChar(C: WideChar): WideChar; inline; overload;
function SamePDAChars(PDAC1, PDAC2: PDAChar): Longbool;
function SamePDWChars(PDWC1, PDWC2: PDWChar): Longbool;
function HexToDigitA(Hex: AnsiChar): Byte;
function HexToDigitW(Hex: WideChar): Byte;
function DigitToHexA(Digit: Byte): AnsiChar;
function DigitToHexW(Digit: Byte): WideChar;
function HexToCharA(Hex: PAnsiChar): AnsiChar;
function HexToCharW(Hex: PWideChar): WideChar;
function HexByteToDigit(HexByte: PAnsiChar): Byte; overload;
function HexByteToDigit(HexByte: PWideChar): Byte; overload;
function ByteToDigitChars(Value: Byte; PResult: PAnsiChar): Byte; overload;
function ByteToDigitChars(Value: Byte; PResult: PWideChar): Byte; overload;
function DecCharsValue(DecChars: PAnsiChar; Count: Cardinal): Cardinal; overload;
function DecCharsValue(DecChars: PWideChar; Count: Cardinal): Cardinal; overload;
function HexCharsValue(HexChars: PAnsiChar; Count: Cardinal): Cardinal; overload;
function HexCharsValue(HexChars: PWideChar; Count: Cardinal): Cardinal; overload;
function HexToBytes(const Hex: RawByteString): RawByteString; overload;
function HexToBytes(const Hex: UnicodeString): UnicodeString; overload;
function BytesToHex(const Bytes: RawByteString): RawByteString; overload;
function BytesToHex(const Bytes: UnicodeString): UnicodeString; overload;
function BytesToHexA(const Bytes: PByte; Count: Cardinal; UpperCase: Longbool = TRUE): RawByteString; overload;
function BytesToHexW(const Bytes: PByte; Count: Cardinal; UpperCase: Longbool = TRUE): UnicodeString; overload;
function BytesToHexChars(const Bytes: PByte; Count: Cardinal; HexChars: PAnsiChar; UpperCase: Longbool = TRUE): Integer; overload;
function BytesToHexChars(const Bytes: PByte; Count: Cardinal; HexChars: PWideChar; UpperCase: Longbool = TRUE): Integer; overload;
function HexCharsToDigitCharsA(Digits, HexChars: PAnsiChar; HexLen: Cardinal): PAnsiChar; overload;
function HexCharsToDigitCharsW(Digits, HexChars: PWideChar; HexLen: Cardinal): PWideChar; overload;
function HexCharsToDigitsA(HexChars: PAnsiChar; HexLen: Cardinal): RawByteString; overload;
function HexCharsToDigitsW(HexChars: PWideChar; HexLen: Cardinal): UnicodeString; overload;
function HexCharsGroup(const Hex: RawByteString; GroupChars: Integer): RawByteString; overload;
function HexCharsGroup(const Hex: UnicodeString; GroupChars: Integer): UnicodeString; overload;
function CardToHexA(Value: UInt64; Digits: Cardinal = 8; const Prefix: AnsiString = '0x'): AnsiString;
function CardToHexW(Value: UInt64; Digits: Cardinal = 8; const Prefix: UnicodeString = '0x'): UnicodeString;
function CardToStrA(Value: UInt64): AnsiString;
function CardToStrW(Value: UInt64): UnicodeString;
function StrToCardEx(const Str: RawByteString): UInt64; overload;
function StrToCardEx(const Str: UnicodeString): UInt64; overload;
function StrToCardDef(const Str: RawByteString; Def: UInt64): UInt64; overload;
function StrToCardDef(const Str: UnicodeString; Def: UInt64): UInt64; overload;
function StrToCard(const Str: RawByteString; AddHexPrefix: Longbool = FALSE): UInt64; overload;
function StrToCard(const Str: UnicodeString; AddHexPrefix: Longbool = FALSE): UInt64; overload;
function StrToInt64Def(const Str: RawByteString; Default: Int64 = 0): UInt64; overload;
function StrToInt64Def(const Str: UnicodeString; Default: Int64 = 0): UInt64; overload;
function StrLenToCard(const Str: RawByteString; Pos, Len: Integer; AddHexPrefix: Longbool = FALSE): UInt64; overload;
function StrLenToCard(const Str: UnicodeString; Pos, Len: Integer; AddHexPrefix: Longbool = FALSE): UInt64; overload;
function StrToFloatEx(const Value: RawByteString; Default: Extended = -1): Extended; overload;
function StrToFloatEx(const Value: UnicodeString; Default: Extended = -1): Extended; overload;
function FixUnicodeLineBreaks(const Text: String; var Fixed: Boolean): String;
function DelAmp(const Str: RawByteString): RawByteString; overload;
function DelAmp(const Str: UnicodeString): UnicodeString; overload;
function SetPasString(Text: PAnsiChar; Size: Integer; Trim: Longbool = FALSE): RawByteString; overload;
function SetPasString(Text: PWideChar; Size: Integer; Trim: Longbool = FALSE): UnicodeString; overload;
function Crop(const Text: RawByteString; Count: Integer): RawByteString; overload;
function Crop(const Text: RawByteString; Index, Count: Integer): RawByteString; overload;
function Crop(const Text: UnicodeString; Count: Integer): UnicodeString; overload;
function Crop(const Text: UnicodeString; Index, Count: Integer): UnicodeString; overload;
function InsStr(const Insert, Text: RawByteString; Index: Integer): RawByteString; overload;
function InsStr(const Insert, Text: UnicodeString; Index: Integer): UnicodeString; overload;
function NoText(const Str: RawByteString): Longbool; overload;
function NoText(const Str: UnicodeString): Longbool; overload;
function SetStr(Text: PAnsiChar; Size: Integer): RawByteString; overload;
function SetStr(Text: PWideChar; Size: Integer): UnicodeString; overload;
procedure Capitalize(var Text: RawByteString; AllWords: Longbool = false); overload;
procedure Capitalize(var Text: UnicodeString; AllWords: Longbool = false); overload;
procedure Hyphenize(var Text: RawByteString; AllWords: Longbool = false); overload;
procedure Hyphenize(var Text: UnicodeString; AllWords: Longbool = false); overload;
function SpaceCompress(const Text: RawByteString): RawByteString; overload;
function SpaceCompress(const Text: UnicodeString): UnicodeString; overload;
function Trim(const Str: RawByteString): RawByteString; overload;
function Trim(const Str: UnicodeString): UnicodeString; overload;
function TrimLeft(const Str: RawByteString): RawByteString; overload;
function TrimLeft(const Str: UnicodeString): UnicodeString; overload;
function TrimRight(const Str: RawByteString): RawByteString; overload;
function TrimRight(const Str: UnicodeString): UnicodeString; overload;
procedure TrimSelf(var Str: RawByteString); overload;
procedure TrimSelf(var Str: UnicodeString); overload;
function TrimCopy(const Str: RawByteString; Index, Count: Integer): RawByteString; overload;
function TrimCopy(const Str: UnicodeString; Index, Count: Integer): UnicodeString; overload;
function TrimCustom(const Str: RawByteString; TrimCharSet: TSysCharSet;
		    TrimOptions: TTrimOptions = TO_BOTH): RawByteString; overload;
function TrimCustom(const Str: UnicodeString; TrimCharSet: TSysCharSet;
		    TrimOptions: TTrimOptions = TO_BOTH): UnicodeString; overload;
function DoWrap(const Source: RawByteString; LineLen: Cardinal = 80): RawByteString; overload;
function DoWrap(const Source: UnicodeString; LineLen: Cardinal = 80): UnicodeString; overload;
function Unwrap(Source: PAnsiChar; SrcLen: Integer; TrimCharSet: TSysCharSet = []; ReplaceChar: AnsiChar = NUL): RawByteString; overload;
function Unwrap(Source: PWideChar; SrcLen: Integer; TrimCharSet: TSysCharSet = []; ReplaceChar: WideChar = NUL): UnicodeString; overload;
function Unwrap(const Str: RawByteString; TrimCharSet: TSysCharSet = []; ReplaceChar: AnsiChar = NUL): RawByteString; overload;
function Unwrap(const Str: UnicodeString; TrimCharSet: TSysCharSet = []; ReplaceChar: WideChar = NUL): UnicodeString; overload;
function Unquote(const Str: RawByteString; CBeg: AnsiChar = QU; CEnd: AnsiChar = QU): RawByteString; overload;
function Unquote(const Str: UnicodeString; CBeg: WideChar = QU; CEnd: WideChar = QU): UnicodeString; overload;
function Unquote(const Str: RawByteString; QuoteChar: AnsiChar; var WasQuoted: Longbool): RawByteString; overload;
function Unquote(const Str: UnicodeString; QuoteChar: WideChar; var WasQuoted: Longbool): UnicodeString; overload;
function SetUnquote(const Str: RawByteString; SBeg: TQuoteChars = QUOTES; SEnd: TQuoteChars = QUOTES): RawByteString; overload;
function SetUnquote(const Str: UnicodeString; SBeg: TQuoteChars = QUOTES; SEnd: TQuoteChars = QUOTES): UnicodeString; overload;
function Unescape(const Str: RawByteString; SkipSpaces: Longbool = FALSE; EscapeChar: AnsiChar = PC;
		  MakeIXBreaks: Longbool = FALSE): Utf8String; overload;
function Unescape(const Str: UnicodeString; SkipSpaces: Longbool = FALSE; EscapeChar: WideChar = PC;
		  MakeIXBreaks: Longbool = FALSE): UnicodeString; overload;
function UrlEscapeText(const Text: RawByteString; MakeIXBreaks: Longbool = TRUE): RawByteString; overload;
function UrlEscapeText(const Text: UnicodeString; MakeIXBreaks: Longbool = TRUE): UnicodeString; overload;
function CheckAST_QMRegEx(const Target, RegEx: String): Longbool;
function GetAnyString(StringPointer: Pointer): String;
function GetStringAddress(AnyObject: TObject; const FieldName: String; ClassInfo: Pointer = NIL): Pointer;
function StringFromHandleA(Hdl: THandle; Size: Integer = -1; Free: Longbool = FALSE; Local: Longbool = FALSE): RawByteString; overload;
function StringFromHandleW(Hdl: THandle; Size: Integer = -1; Free: Longbool = FALSE; Local: Longbool = FALSE): UnicodeString; overload;
function GetPCharHandle(Text: PAnsiChar; Size: Integer = -1; ReallocHandle: THandle = 0; Local: Longbool = FALSE): THandle; overload;
function GetPCharHandle(Text: PWideChar; Size: Integer = -1; ReallocHandle: THandle = 0; Local: Longbool = FALSE): THandle; overload;
function GetStringHandle(const Str: RawByteString; ReallocHandle: THandle = 0; Local: Longbool = FALSE): THandle; overload;
function GetStringHandle(const Str: UnicodeString; ReallocHandle: THandle = 0; Local: Longbool = FALSE): THandle; overload;
function GetTextAsHandle(const Strings: Array of RawByteString): THandle; overload;
function GetTextAsHandle(const Strings: Array of UnicodeString): THandle; overload;
function Copy(const Str: RawByteString; Pos, Len: Integer): RawByteString; overload;
function Copy(const Str: UnicodeString; Pos, Len: Integer): UnicodeString; overload;
function CopyLine(const Str: RawByteString; Num: Integer): RawByteString; overload;
function CopyLine(const Str: UnicodeString; Num: Integer): UnicodeString; overload;
function GetFirstLine(const Str: RawByteString): RawByteString; overload;
function GetFirstLine(const Str: UnicodeString): UnicodeString; overload;
function GetLastLine(const Str: RawByteString): RawByteString; overload;
function GetLastLine(const Str: UnicodeString): UnicodeString; overload;
function LowerCase(const Str: RawByteString): RawByteString; overload;
function LowerCase(const Str: UnicodeString): UnicodeString; overload;
function UpperCase(const Str: RawByteString): RawByteString; overload;
function UpperCase(const Str: UnicodeString): UnicodeString; overload;
procedure StrLower(var Str: RawByteString); overload;
procedure StrLower(var Str: UnicodeString); overload;
procedure StrUpper(var Str: RawByteString); overload;
procedure StrUpper(var Str: UnicodeString); overload;
function ToLower(Chr: AnsiChar): AnsiChar; overload;
function ToLower(Chr: WideChar): WideChar; overload;
function ToUpper(Chr: AnsiChar): AnsiChar; overload;
function ToUpper(Chr: WideChar): WideChar; overload;
function ToLower(var Str: RawByteString; Pos: Integer = 1; Len: Integer = 1): Integer; overload;
function ToLower(var Str: UnicodeString; Pos: Integer = 1; Len: Integer = 1): Integer; overload;
function ToUpper(var Str: RawByteString; Pos: Integer = 1; Len: Integer = 1): Integer; overload;
function ToUpper(var Str: UnicodeString; Pos: Integer = 1; Len: Integer = 1): Integer; overload;
function StrScan(const Str: PAnsiChar; Char: AnsiChar): PAnsiChar; overload;
function StrScan(const Str: PWideChar; Char: WideChar): PWideChar; overload;
function StrRScan(const Str: PAnsiChar; Char: AnsiChar): PAnsiChar; overload;
function StrRScan(const Str: PWideChar; Char: WideChar): PWideChar; overload;
function ScanBack(const Str: PAnsiChar; Char: AnsiChar): PAnsiChar; overload;
function ScanBack(const Str: PWideChar; Char: WideChar): PWideChar; overload;
function ScanFore(const Str: PAnsiChar; Char: AnsiChar): PAnsiChar; overload;
function ScanFore(const Str: PWideChar; Char: WideChar): PWideChar; overload;
function StrLen(const Str: PAnsiChar): Cardinal; overload;
function StrLen(const Str: PWideChar): Cardinal; overload;
function StrEnd(const Str: PAnsiChar): PAnsiChar; overload;
function StrEnd(const Str: PWideChar): PWideChar; overload;
function StrEnd(const Str: RawByteString): PAnsiChar; overload;
function StrEnd(const Str: UnicodeString): PWideChar; overload;
function StrPos(Str, SubStr: PAnsiChar): PAnsiChar; overload;
function StrPos(Str, SubStr: PWideChar): PWideChar; overload;
function StrPos(const Str, SubStr: RawByteString): Integer; overload;
function StrPos(const Str, SubStr: UnicodeString): Integer; overload;
function TextPos(Str, SubStr: PAnsiChar): PAnsiChar; overload;
function TextPos(Str, SubStr: PWideChar): PWideChar; overload;
function TextPos(const Str, SubStr: RawByteString): Integer; overload;
function TextPos(const Str, SubStr: UnicodeString): Integer; overload;
function StrCharLength(const Str: PAnsiChar): Integer; overload;
function StrCharLength(const Str: PWideChar): Integer; overload;
function StrComp(S1, S2: PAnsiChar): Integer; overload;
function StrComp(S1, S2: PWideChar): Integer; overload;
function StrIComp(S1, S2: PAnsiChar): Integer; overload;
function StrIComp(S1, S2: PWideChar): Integer; overload;
function StrLComp(S1, S2: PAnsiChar; MaxLen: Cardinal): Integer; overload;
function StrLComp(S1, S2: PWideChar; MaxLen: Cardinal): Integer; overload;
function StrAtPos(S1, S2: PAnsiChar; MaxLen: Cardinal): Longbool; overload;
function StrAtPos(S1, S2: PWideChar; MaxLen: Cardinal): Longbool; overload;
function StrLIComp(S1, S2: PAnsiChar; MaxLen: Cardinal): Integer; overload;
function StrLIComp(S1, S2: PWideChar; MaxLen: Cardinal): Integer; overload;
function StrIAtPos(S1, S2: PAnsiChar; MaxLen: Cardinal): Longbool; overload;
function StrIAtPos(S1, S2: PWideChar; MaxLen: Cardinal): Longbool; overload;
function StrCat(Dest: PAnsiChar; const Source: PAnsiChar): PAnsiChar; overload;
function StrCat(Dest: PWideChar; const Source: PWideChar): PWideChar; overload;
function StrLCat(Dest: PAnsiChar; const Source: PAnsiChar; MaxLen: Cardinal): PAnsiChar; overload;
function StrLCat(Dest: PWideChar; const Source: PWideChar; MaxLen: Cardinal): PWideChar; overload;
function StrCopy(Dest: PAnsiChar; const Source: PAnsiChar): PAnsiChar; overload;
function StrCopy(Dest: PWideChar; const Source: PWideChar): PWideChar; overload;
function StrECopy(Dest: PAnsiChar; const Source: PAnsiChar): PAnsiChar; overload;
function StrECopy(Dest: PWideChar; const Source: PWideChar): PWideChar; overload;
function StrLCopy(Dest: PAnsiChar; const Source: PAnsiChar; MaxLen: Cardinal): PAnsiChar; overload;
function StrLCopy(Dest: PWideChar; const Source: PWideChar; MaxLen: Cardinal): PWideChar; overload;
function StrLECopy(Dest: PAnsiChar; const Source: PAnsiChar; Length: Cardinal): PAnsiChar; overload;
function StrLECopy(Dest: PWideChar; const Source: PWideChar; Length: Cardinal): PWideChar; overload;
function StrPLCopy(Dest: PAnsiChar; const Source: RawByteString; MaxLen: Cardinal): PAnsiChar; overload;
function StrPLCopy(Dest: PWideChar; const Source: UnicodeString; MaxLen: Cardinal): PWideChar; overload;
function StrMove(Dest: PAnsiChar; const Source: PAnsiChar; Count: Cardinal): PAnsiChar; overload;
function StrMove(Dest: PWideChar; const Source: PWideChar; Count: Cardinal): PWideChar; overload;
function StrLPasCopy(Source: PAnsiChar; Length: Integer): RawByteString; overload;
function StrLPasCopy(Source: PWideChar; Length: Integer): UnicodeString; overload;
function AnsiStrAlloc(Size: Cardinal): PAnsiChar; overload;
function StrBufSize(const Str: PAnsiChar): Cardinal; overload;
function StrBufSize(const Str: PWideChar): Cardinal; overload;
function StrNew(const Str: PAnsiChar): PAnsiChar; overload;
function StrNew(const Str: PWideChar): PWideChar; overload;
procedure StrDispose(Str: PAnsiChar); overload;
procedure StrDispose(Str: PWideChar); overload;
function AnsiStrEnd(Str: PAnsiChar): PAnsiChar; overload;
function AnsiStrEnd(Str: PWideChar): PWideChar; overload;
function AnsiStrPos(Str, SubStr: PAnsiChar): PAnsiChar; overload;
function AnsiStrPos(Str, SubStr: PWideChar): PWideChar; overload;
function AnsiStrRPos(Str, SubStr: PAnsiChar): PAnsiChar; overload;
function AnsiStrRPos(Str, SubStr: PWideChar): PWideChar; overload;
function AnsiTextPos(Str, SubStr: PAnsiChar): PAnsiChar; overload;
function AnsiTextPos(Str, SubStr: PWideChar): PWideChar; overload;
function AnsiTextRPos(Str, SubStr: PAnsiChar): PAnsiChar; overload;
function AnsiTextRPos(Str, SubStr: PWideChar): PWideChar; overload;
function AnsiStrComp(S1, S2: PAnsiChar): Integer; overload;
function AnsiStrComp(S1, S2: PWideChar): Integer; overload;
function AnsiStrIComp(S1, S2: PAnsiChar): Integer; overload;
function AnsiStrIComp(S1, S2: PWideChar): Integer; overload;
function AnsiStrLComp(S1, S2: PAnsiChar; MaxLen: Cardinal): Integer; overload;
function AnsiStrLComp(S1, S2: PWideChar; MaxLen: Cardinal): Integer; overload;
function AnsiStrAtPos(S1, S2: PAnsiChar; MaxLen: Cardinal): Longbool; overload;
function AnsiStrAtPos(S1, S2: PWideChar; MaxLen: Cardinal): Longbool; overload;
function AnsiStrLIComp(S1, S2: PAnsiChar; MaxLen: Cardinal): Integer; overload;
function AnsiStrLIComp(S1, S2: PWideChar; MaxLen: Cardinal): Integer; overload;
function AnsiStrIAtPos(S1, S2: PAnsiChar; MaxLen: Cardinal): Longbool; overload;
function AnsiStrIAtPos(S1, S2: PWideChar; MaxLen: Cardinal): Longbool; overload;
function AnsiExtractQuotedStr(Src: PAnsiChar; Quote: AnsiChar = QU): AnsiString; overload;
function AnsiExtractQuotedStr(Src: PWideChar; Quote: WideChar = QU): UnicodeString; overload;
function IsStringUpperCase(const Str: RawByteString): Longbool; overload;
function IsStringUpperCase(const Str: UnicodeString): Longbool; overload;
function IsUTF8String(const Str: RawByteString): Longbool;
function SearchBuf(Buf: PAnsiChar; BufLen: Integer; SelStart, SelLength: Integer;
		   SearchString: RawByteString; Options: TStringSearchOptions;
		   const WordDelimiters: TSysCharSet = WORDDELIMITERS): PAnsiChar; overload;
function SearchBuf(Buf: PWideChar; BufLen: Integer; SelStart, SelLength: Integer;
		   SearchString: UnicodeString; Options: TStringSearchOptions): PWideChar; overload;
function StringPos(const SearchStr, Str: RawByteString; BegPos: Integer; Options: TStringSearchOptions = [soDown]): Integer; overload;
function StringPos(const SearchStr, Str: UnicodeString; BegPos: Integer; Options: TStringSearchOptions = [soDown]): Integer; overload;
function TextPos(const SearchText, Text: RawByteString; BegPos: Integer; Options: TStringSearchOptions = [soDown]): Integer; overload;
function TextPos(const SearchText, Text: UnicodeString; BegPos: Integer; Options: TStringSearchOptions = [soDown]): Integer; overload;
function CollapseWhiteSpace(const Str, MaxLineBreak: RawByteString): RawByteString; overload;
function CollapseWhiteSpace(const Str, MaxLineBreak: UnicodeString): UnicodeString; overload;
function Contains(Str, SubStr: AnsiString; BegPos: Integer = 1; EndPos: Integer = MAXINT): Longbool; overload;
function Contains(Str, SubStr: UnicodeString; BegPos: Integer = 1; EndPos: Integer = MAXINT): Longbool; overload;
function StrContains(PBegPos, PEndPos, PSubStr: PAnsiChar; SubStrLen: Cardinal): PAnsiChar; overload;
function StrContains(PBegPos, PEndPos, PSubStr: PWideChar; SubStrLen: Cardinal): PWideChar; overload;
function GetPadding(const Text: RawByteString; var Left, Right: RawByteString): TTrimOptions; overload;
function GetPadding(const Text: UnicodeString; var Left, Right: UnicodeString): TTrimOptions; overload;
function GetLineCount(const Text: RawByteString; TestTrimmed: Longbool = FALSE): Integer; overload;
function GetLineCount(const Text: UnicodeString; TestTrimmed: Longbool = FALSE): Integer; overload;
function SameChar(Chr1, Chr2: AnsiChar): Longbool; overload;
function SameChar(Chr1, Chr2: WideChar): Longbool; overload;
function CharDiff(OpenChar, CloseChar: AnsiChar; BegChars, EndChars: PAnsiChar): Integer; overload;
function CharDiff(OpenChar, CloseChar: WideChar; BegChars, EndChars: PWideChar): Integer; overload;
function GetCharsCount(const Chars: TSysCharSet; const Text: RawByteString): Integer; overload;
function GetCharsCount(const Chars: TSysCharSet; const Text: UnicodeString): Integer; overload;
function GetCharCount(Chr: AnsiChar; const Text: RawByteString): Integer; overload;
function GetCharCount(Chr: WideChar; const Text: UnicodeString): Integer; overload;
function FirstCharInSet(const Text: RawByteString; Chars: TSysCharSet): Integer; overload;
function FirstCharInSet(const Text: UnicodeString; Chars: TSysCharSet): Integer; overload;
function FirstCharPos(Chr: AnsiChar; const Text: RawByteString): Integer; overload;
function FirstCharPos(Chr: WideChar; const Text: UnicodeString): Integer; overload;
function LastCharPos(Chr: AnsiChar; const Text: RawByteString): Integer; overload;
function LastCharPos(Chr: WideChar; const Text: UnicodeString): Integer; overload;
function NextCharPos(Chr: AnsiChar; const Text: RawByteString; BegPos: Integer): Integer; overload;
function NextCharPos(Chr: WideChar; const Text: UnicodeString; BegPos: Integer): Integer; overload;
function PrevCharPos(Chr: AnsiChar; const Text: RawByteString; EndPos: Integer): Integer; overload;
function PrevCharPos(Chr: WideChar; const Text: UnicodeString; EndPos: Integer): Integer; overload;
function RangeCharPos(Chr: AnsiChar; const Text: RawByteString; BegPos, EndPos: Integer): Integer; overload;
function RangeCharPos(Chr: WideChar; const Text: UnicodeString; BegPos, EndPos: Integer): Integer; overload;
function AnsiPos(const SubStr, Str: RawByteString): Integer; overload;
function AnsiPos(const SubStr, Str: UnicodeString): Integer; overload;
function AnsiQuotedStr(const Str: RawByteString; Quote: AnsiChar = QU): RawByteString; overload;
function AnsiQuotedStr(const Str: UnicodeString; Quote: WideChar = QU): UnicodeString; overload;
function AnsiSameStr(const S1, S2: RawByteString): Longbool; overload;
function AnsiSameStr(const S1, S2: UnicodeString): Longbool; overload;
function AnsiSameText(const S1, S2: RawByteString): Longbool; overload;
function AnsiSameText(const S1, S2: UnicodeString): Longbool; overload;
function AnsiCompareStr(const S1, S2: RawByteString): Integer; overload;
function AnsiCompareStr(const S1, S2: UnicodeString): Integer; overload;
function AnsiCompareText(const S1, S2: RawByteString): Integer; overload;
function AnsiCompareText(const S1, S2: UnicodeString): Integer; overload;
function AnsiLowerCase(const Str: RawByteString): RawByteString; overload;
function AnsiLowerCase(const Str: UnicodeString): UnicodeString; overload;
function AnsiUpperCase(const Str: RawByteString): RawByteString; overload;
function AnsiUpperCase(const Str: UnicodeString): UnicodeString; overload;
function GetStrCount(const SearchStr, Str: RawByteString): Integer; overload;
function GetStrCount(const SearchStr, Str: UnicodeString): Integer; overload;
function FirstStrPos(const SearchStr, Str: RawByteString): Integer; overload;
function FirstStrPos(const SearchStr, Str: UnicodeString): Integer; overload;
function LastStrPos(const SearchStr, Str: RawByteString): Integer; overload;
function LastStrPos(const SearchStr, Str: UnicodeString): Integer; overload;
function NextStrPos(const SearchStr, Str: RawByteString; BegPos: Integer): Integer; overload;
function NextStrPos(const SearchStr, Str: UnicodeString; BegPos: Integer): Integer; overload;
function PrevStrPos(const SearchStr, Str: RawByteString; EndPos: Integer): Integer; overload;
function PrevStrPos(const SearchStr, Str: UnicodeString; EndPos: Integer): Integer; overload;
function RangeStrPos(const SearchStr, Str: RawByteString; BegPos, EndPos: Integer): Integer; overload;
function RangeStrPos(const SearchStr, Str: UnicodeString; BegPos, EndPos: Integer): Integer; overload;
function GetTextCount(const SearchText, Text: RawByteString): Integer; overload;
function GetTextCount(const SearchText, Text: UnicodeString): Integer; overload;
function FirstTextPos(const SearchText, Text: RawByteString): Integer; overload;
function FirstTextPos(const SearchText, Text: UnicodeString): Integer; overload;
function LastTextPos(const SearchText, Text: RawByteString): Integer; overload;
function LastTextPos(const SearchText, Text: UnicodeString): Integer; overload;
function NextTextPos(const SearchText, Text: RawByteString; BegPos: Integer): Integer; overload;
function NextTextPos(const SearchText, Text: UnicodeString; BegPos: Integer): Integer; overload;
function PrevTextPos(const SearchText, Text: RawByteString; EndPos: Integer): Integer; overload;
function PrevTextPos(const SearchText, Text: UnicodeString; EndPos: Integer): Integer; overload;
function RangeTextPos(const SearchText, Text: RawByteString; BegPos, EndPos: Integer): Integer; overload;
function RangeTextPos(const SearchText, Text: UnicodeString; BegPos, EndPos: Integer): Integer; overload;
function StrAtPos(const ASubStr, AStr: RawByteString; Pos: Integer): Longbool; overload;
function StrAtPos(const ASubStr, AStr: UnicodeString; Pos: Integer): Longbool; overload;
function TextAtPos(const ASubText, AText: RawByteString; Pos: Integer): Longbool; overload;
function TextAtPos(const ASubText, AText: UnicodeString; Pos: Integer): Longbool; overload;
function IncludesStr(const S1: RawByteString; const AValues: Array of RawByteString): Integer; overload;
function IncludesStr(const S1: UnicodeString; const AValues: Array of UnicodeString): Integer; overload;
function IncludesText(const S1: RawByteString; const AValues: Array of RawByteString): Integer; overload;
function IncludesText(const S1: UnicodeString; const AValues: Array of UnicodeString): Integer; overload;
function IndexStr(const S1: RawByteString; const AValues: Array of AnsiString; MaxLen: Integer = -1): Integer; overload;
function IndexStr(const S1: UnicodeString; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Integer; overload;
function IndexPChar(PCh: PAnsiChar; const AValues: Array of AnsiString; MaxLen: Integer = -1): Integer; overload;
function IndexPChar(PCh: PWideChar; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Integer; overload;
function IndexText(const S1: RawByteString; const AValues: Array of AnsiString; MaxLen: Integer = -1): Integer; overload;
function IndexText(const S1: UnicodeString; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Integer; overload;
function IndexPCharI(PCh: PAnsiChar; const AValues: Array of AnsiString; MaxLen: Integer = -1): Integer; overload;
function IndexPCharI(PCh: PWideChar; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Integer; overload;
function MatchStr(const S1: RawByteString; const AValues: Array of AnsiString; MaxLen: Integer = -1): Boolean; overload;
function MatchStr(const S1: UnicodeString; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Boolean; overload;
function MatchPChar(PCh: PAnsiChar; const AValues: Array of AnsiString; MaxLen: Integer = -1): Boolean; overload;
function MatchPChar(PCh: PWideChar; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Boolean; overload;
function MatchText(const S1: RawByteString; const AValues: Array of AnsiString; MaxLen: Integer = -1): Boolean; overload;
function MatchText(const S1: UnicodeString; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Boolean; overload;
function MatchPCharI(PCh: PAnsiChar; const AValues: Array of AnsiString; MaxLen: Integer = -1): Boolean; overload;
function MatchPCharI(PCh: PWideChar; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Boolean; overload;
function SameStr(const Str1, Str2: RawByteString): Boolean; overload;
function SameStr(const Str1, Str2: UnicodeString): Boolean; overload;
function SameText(const Text1, Text2: RawByteString): Boolean; overload;
function SameText(const Text1, Text2: UnicodeString): Boolean; overload;
function SameTrimText(const Text1, Text2: RawByteString): Boolean; overload;
function SameTrimText(const Text1, Text2: UnicodeString): Boolean; overload;
function ContainsStr(const AStr, ASubStr: RawByteString): Boolean; overload;
function ContainsStr(const AStr, ASubStr: UnicodeString): Boolean; overload;
function ContainsText(const AText, ASubText: RawByteString): Boolean; overload;
function ContainsText(const AText, ASubText: UnicodeString): Boolean; overload;
function StartsStr(const ASubStr, AStr: RawByteString; MaxLen: Integer = -1): Boolean; overload;
function StartsStr(const ASubStr, AStr: UnicodeString; MaxLen: Integer = -1): Boolean; overload;
function StartsText(const ASubText, AText: RawByteString; MaxLen: Integer = -1): Boolean; overload;
function StartsText(const ASubText, AText: UnicodeString; MaxLen: Integer = -1): Boolean; overload;
function EndsStr(const ASubStr, AStr: RawByteString): Boolean; overload;
function EndsStr(const ASubStr, AStr: UnicodeString): Boolean; overload;
function EndsText(const ASubText, AText: RawByteString): Boolean; overload;
function EndsText(const ASubText, AText: UnicodeString): Boolean; overload;
function ExtractStr(var Source: RawByteString; Pos, Len: Integer; TrimSource: TTrimOptions = []): RawByteString; overload;
function ExtractStr(var Source: UnicodeString; Pos, Len: Integer; TrimSource: TTrimOptions = []): UnicodeString; overload;
function FromBase64(Str: PAnsiChar; Size: Integer): RawByteString; overload;
function FromBase64(Str: PWideChar; Size: Integer): UnicodeString; overload;
function FromBase64(const Str: RawByteString): RawByteString; overload;
function FromBase64(const Str: UnicodeString): UnicodeString; overload;
function ToBase64(const Str: PAnsiChar; Size: Integer; DropLineBreaks: Longbool = FALSE): RawByteString; overload;
function ToBase64(const Str: PWideChar; Size: Integer; DropLineBreaks: Longbool = FALSE): UnicodeString; overload;
function ToBase64(const Str: RawByteString; DropLineBreaks: Longbool = FALSE): RawByteString; overload;
function ToBase64(const Str: UnicodeString; DropLineBreaks: Longbool = FALSE): UnicodeString; overload;
function HashedString(const Str: RawByteString; Len: Integer): RawByteString; overload;
function HashedString(const Str: UnicodeString; Len: Integer): UnicodeString; overload;
function AdjustLineBreaks(const Str: RawByteString): RawByteString; overload;
function AdjustLineBreaks(const Str: UnicodeString): UnicodeString; overload;
function InsertLineBreaks(const Source: RawByteString; LineLen: Cardinal = 80): RawByteString; overload;
function InsertLineBreaks(const Source: UnicodeString; LineLen: Cardinal = 80): UnicodeString; overload;
function RemoveLineBreaks(const Str: RawByteString): RawByteString; overload;
function RemoveLineBreaks(const Str: UnicodeString): UnicodeString; overload;
function ReplaceCharAtPos(const Str: RawByteString; Old, New: AnsiChar; Pos: Integer = -1): RawByteString; overload;
function ReplaceCharAtPos(const Str: UnicodeString; Old, New: WideChar; Pos: Integer = -1): UnicodeString; overload;
function ReplaceChars(const Str: RawByteString; Old: TSysCharSet; New: AnsiChar): RawByteString; overload;
function ReplaceChars(const Str: UnicodeString; Old: TSysCharSet; New: WideChar): UnicodeString; overload;
function ReplaceCharsAsHex(const Str: RawByteString; Old: TSysCharSet): RawByteString; overload;
function ReplaceCharsAsHex(const Str: UnicodeString; Old: TSysCharSet): UnicodeString; overload;
function ReplaceMultiStrings(const Str: RawByteString; const ReplaceStrings: TReplaceStringsA; CaseInsens: Bool = TRUE): RawByteString; overload;
function ReplaceMultiStrings(const Str: UnicodeString; const ReplaceStrings: TReplaceStringsW; CaseInsens: Bool = TRUE): UnicodeString; overload;
function ReplaceStrings(const Str, Old, New: RawByteString; FullLineOnly: Longbool = FALSE): RawByteString; overload;
function ReplaceStrings(const Str, Old, New: UnicodeString; FullLineOnly: Longbool = FALSE): UnicodeString; overload;
function ReplaceStrAtPos(const Str, New: RawByteString; OldPos, OldLen: Integer): RawByteString; overload;
function ReplaceStrAtPos(const Str, New: UnicodeString; OldPos, OldLen: Integer): UnicodeString; overload;
function ReplaceText(const Str, Old, New: RawByteString; FullLineOnly: Longbool = FALSE): RawByteString; overload;
function ReplaceText(const Str, Old, New: UnicodeString; FullLineOnly: Longbool = FALSE): UnicodeString; overload;
function ReplaceTextAtPos(const Str, New: RawByteString; OldPos, OldLen: Integer): RawByteString; overload;
function ReplaceTextAtPos(const Str, New: UnicodeString; OldPos, OldLen: Integer): UnicodeString; overload;
function GetSystemFileVersion: Cardinal;
function GetSimpleFileVersion(const FileName: String; MainOnly: Longbool = FALSE): Cardinal;
function WinVersion(Simple: Longbool; AddNumber: Longbool = FALSE; AddSvcPack: Longbool = FALSE; AddCpuType: Longbool = FALSE): String;
function IsMinWinVersion(const MinWinVersion: String): Longbool;
function MainWinVersion: Cardinal;
function ComputerName: String;
function UserName: String;
function SafeCallErrorString(ErrorCode: HResult; ErrorAddr: Pointer): String;
function AppendHexAddress(const AText: String; Address: Pointer): String;
function AppendHexError(const AText: String; Error: Cardinal): String;
function AppendPeriod(const Str: String): String;
function SysErrorString(Code: Cardinal; Module: THandle = 0): WideString;
function SysFontInfo(const FontName: String): TExtLogFont;
function SysFontNames(const Fonts: TStrings; const FamilyNames: Array of String): Integer;
function SysFontNameAndSize(NCLFontType: TNCLFontType): TNCLFontNameSize;
function WinFixFont(const DefaultFont: String = 'Consolas'): String;
function WinVarFont: String;
function DateString(const ST: TSystemTime; const Format: String): String;
function TimeString(const ST: TSystemTime; const Format: String): String;
function UniqueString(const Str: RawByteString): RawByteString; overload;
function RegEntryExists(RegKey: THandle; const Path: String): Longbool;
function GetRegEntryType(RegKey: THandle; const Path, Entry: String): Cardinal;
function GetRegBinary(RegKey: THandle; const Path, Entry: String): RawByteString;
function SetRegBinary(RegKey: THandle; const Path, Entry: String; const Value: RawByteString): Longbool;
function GetRegDWord(RegKey: THandle; const Path, Entry: String): Cardinal;
function SetRegDWord(RegKey: THandle; const Path, Entry: String; Value: Cardinal): Longbool;
function GetRegString(RegKey: THandle; const Path, Entry: String): String;
function SetRegString(RegKey: HKEY; const Path, Entry, Value: String): Longbool;
function GetIniBool(const IniFile, Section, Item: String): Longbool;
function SetIniBool(const IniFile, Section, Item: String; Value: Longbool): Longbool;
function GetIniDWord(const IniFile, Section, Item: String): Cardinal;
function SetIniDWord(const IniFile, Section, Item: String; Value: Cardinal): Longbool;
function GetIniString(const IniFile, Section, Item: String; FullItem: Longbool = FALSE): String;
function SetIniString(const IniFile, Section, Item, Value: String): Longbool;
function ReadIniFileSection(const IniFile, Section: String; List: TStrings): Cardinal; overload;
function ReadIniFileSection(const IniFile, Section: String): String; overload;
function WriteIniFileSection(const FileName, Section, Data: String): Longbool;
function SystemBrowserPath: String;
function ResString(ID: Cardinal): String;
function ClassString(Window: THandle; ScrollbarParent: Longbool = FALSE): String;
function WindowStringA(Window: THandle): RawByteString;
function WindowStringW(Window: THandle): UnicodeString;
function DialogString(Dialog: THandle; ID: Integer): String;
function EnvironmentString(const Str: String): String;
function ExpandEnvironmentString(const Str: String): String;
function ModuleString(Module: THandle; ID: Cardinal): String;
function MenuString(Menu: THandle; ID, Flags: Cardinal): String;
function ColorToHtml(const Value: Variant): String;
function HtmlToColor(const Value: Variant): TColor;
function RGBCharsToWebHexColor(PHexColor, PRGBColor: PAnsiChar): PAnsiChar;
function RGBStringToWebHexColor(const RGBColor: RawByteString; PRGBLength: PCardinal = NIL): RawByteString; overload;
function RGBStringToWebHexColor(const RGBColor: UnicodeString; PRGBLength: PCardinal = NIL): UnicodeString; overload;
function GetDocType(const PeekData: RawByteString): String;
function HtmlDecode(const Html: RawByteString; NonAsciiOnly: Boolean = FALSE): RawByteString; overload;
function HtmlDecode(const Html: UnicodeString; NonAsciiOnly: Boolean = FALSE): UnicodeString; overload;
function HtmlEncode(const Html: RawByteString; NonAsciiOnly: Boolean = FALSE): RawByteString; overload;
function HtmlEncode(const Html: UnicodeString; NonAsciiOnly: Boolean = FALSE): UnicodeString; overload;
function RestoreNbsp(const Html: RawByteString): RawByteString; overload;
function RestoreNbsp(const Html: UnicodeString): UnicodeString; overload;
function RtfToString(const Rtf: RawByteString): RawByteString; overload;
function RtfToString(const Rtf: UnicodeString): UnicodeString; overload;
function StringToRtf(const Str: RawByteString): RawByteString; overload;
function StringToRtf(const Str: UnicodeString): UnicodeString; overload;
function PathCompact(const Path: RawByteString; MaxLen: Integer): RawByteString; overload;
function PathCompact(const Path: UnicodeString; MaxLen: Integer): UnicodeString; overload;
function FormatSettings(DecimalSeperator: Char = NUL): TFormatSettings;
function Format(const Format: RawByteString; const Args: Array of TVarRec): RawByteString; overload;
function Format(const Format: UnicodeString; const Args: Array of TVarRec): UnicodeString; overload;
function Format(const Format: RawByteString; const Args: Array of TVarRec;
		const FormatSettings: TFormatSettings): RawByteString; overload;
function Format(const Format: UnicodeString; const Args: Array of TVarRec;
		const FormatSettings: TFormatSettings): UnicodeString; overload;
function GetHtmlSysFontFormat(const Html: String): String;
function GetHtmlSizeInt(const Size: String; Default: Integer = 0): Integer;
function GetHtmlSizeFloat(const Size: String; Default: Integer = 0): Extended;
function RoundNearest(AValue: Extended): Integer;
function DivModPlus(Dividend: Cardinal; Divisor: Cardinal): Cardinal;
procedure DivMod(Dividend: Cardinal; Divisor: Cardinal; var Result, Remainder: Cardinal);


implementation

uses
  RichEdit, RichEd20, FileTools, RegistryTools, WindowTools, WOW64, Utf8;

type
  STIF_FLAGS = Integer;

const
  VER_EQUAL				= 1;
  VER_GREATER				= 2;
  VER_GREATER_EQUAL			= 3;
  VER_LESS				= 4;
  VER_LESS_EQUAL			= 5;
  VER_AND				= 6;
  VER_OR				= 7;

const
  VER_CONDITION_MASK			= 7;
  VER_NUM_BITS_PER_CONDITION_MASK	= 3;

const
  VER_NT_WORKSTATION			= $0000001;
  VER_NT_DOMAIN_CONTROLLER		= $0000002;
  VER_NT_SERVER				= $0000003;

function VerifyVersionInfo(var lpVersionInformation: TOSVersionInfoEx;
  dwTypeMask: DWORD; dwlConditionMask: DWORDLONG): BOOL; stdcall;
  external kernel32 name 'VerifyVersionInfoW';
function VerSetConditionMask(dwlConditionMask: DWORDLONG; dwTypeBitMask: DWORD;
  dwConditionMask: Byte): ULONGLONG; stdcall;
  external kernel32 name 'VerSetConditionMask';

const
  shlwapi32 = 'shlwapi.dll';

  function ChrCmpIA(w1, w2: Word): Bool; stdcall; external shlwapi32 name 'ChrCmpIA';
  function StrToInt64ExA(pszString: PAnsiChar; dwFlags: DWord; var pllRet: Int64): Bool; stdcall; external shlwapi32 name 'StrToInt64ExA';

  function ChrCmpIW(w1, w2: Word): Bool; stdcall; external shlwapi32 name 'ChrCmpIW';
  function StrToInt64ExW(pszString: PWideChar; dwFlags: DWord; var pllRet: Int64): Bool; stdcall; external shlwapi32 name 'StrToInt64ExW';


procedure Beep(BeepType: Cardinal = MB_ICONHAND); inline;
begin
  MessageBeep(BeepType);
end;

// URL protocol stuff ----------------------------------------------------------

function GetProtocolFromUrl(const Url: RawByteString; Start: Integer = 1): RawByteString;
begin
  Result := Copy(Url, Start, FirstCharPos(CL, Url) - pred(Start));
end;

function GetProtocolFromUrl(const Url: UnicodeString; Start: Integer = 1): UnicodeString;
begin
  Result := Copy(Url, Start, FirstCharPos(CL, Url) - pred(Start));
end;

function IsProtocolUrl(const Url: RawByteString): Boolean;
var
  sProtocol: RawByteString;
begin
  sProtocol := GetProtocolFromUrl(Url);
  Result := (Length((sProtocol)) > 2) and not StartsDSorBS(sProtocol);
end;

function IsProtocolUrl(const Url: UnicodeString): Boolean;
var
  sProtocol: UnicodeString;
begin
  sProtocol := GetProtocolFromUrl(Url);
  Result := (Length((sProtocol)) > 2) and not StartsDSorBS(sProtocol);
end;

function IsAboutUrl(const Url: RawByteString): Boolean;
begin
  Result := StrTools.StartsText(RAW(PROT_ABOUT), Url);
end;

function IsAboutUrl(const Url: UnicodeString): Boolean;
begin
  Result := StrTools.StartsText(PROT_ABOUT, Url);
end;

function IsBrowserUrl(const Url: RawByteString; CheckDelims: Longbool = FALSE): Boolean;
var
  Start: Integer;
begin
  Start := 1 + Ord(CheckDelims and (Url <> E) and SysUtils.CharInSet(Url[1], BLANKS + [QU, SQ, LT]));
  Result := StrTools.MatchText(GetProtocolFromUrl(Url, Start), [RAW(PROT_HTTPS), RAW(PROT_HTTP),
								RAW(PROT_FTPS), RAW(PROT_FTP),
								RAW(PROT_JSCRIPT), RAW(PROT_VBSCRIPT)]);
end;

function IsBrowserUrl(const Url: UnicodeString; CheckDelims: Longbool = FALSE): Boolean;
var
  Start: Integer;
begin
  Start := 1 + Ord(CheckDelims and (Url <> E) and SysUtils.CharInset(Url[1], BLANKS + [QU, SQ, LT]));
  Result := StrTools.MatchText(GetProtocolFromUrl(Url, Start), [UnicodeString(PROT_HTTPS), UnicodeString(PROT_HTTP),
								UnicodeString(PROT_FTPS), UnicodeString(PROT_FTP),
								UnicodeString(PROT_JSCRIPT), UnicodeString(PROT_VBSCRIPT)]);
end;

function IsCidUrl(const Url: RawByteString): Boolean;
begin
  Result := StrTools.StartsText(RAW(PROT_CID), Url);
end;

function IsCidUrl(const Url: UnicodeString): Boolean;
begin
  Result := StrTools.StartsText(PROT_CID, Url);
end;

function IsDataUrl(const Url: RawByteString): Boolean;
begin
  Result := StrTools.StartsText(RAW(PROT_DATA), Url);
end;

function IsDataUrl(const Url: UnicodeString): Boolean;
begin
  Result := StrTools.StartsText(PROT_DATA, Url);
end;

function IsHttpUrl(const Url: RawByteString): Boolean;
begin
  Result := StrTools.StartsText(RAW(PROT_HTTP), Url);
end;

function IsHttpUrl(const Url: UnicodeString): Boolean;
begin
  Result := StrTools.StartsText(RAW(PROT_HTTP), Url);
end;

function IsMailToUrl(const Url: RawByteString): Boolean;
begin
  Result := StrTools.StartsText(RAW(PROT_MAIL), Url);
end;

function IsMailToUrl(const Url: UnicodeString): Boolean;
begin
  Result := StrTools.StartsText(PROT_MAIL, Url);
end;

function IsResUrl(const Url: RawByteString): Boolean;
begin
  Result := StrTools.StartsText(RAW(PROT_RES), Url);
end;

function IsResUrl(const Url: UnicodeString): Boolean;
begin
  Result := StrTools.StartsText(PROT_RES, Url);
end;

function IsWWWUrl(const Url: RawByteString; CheckDelims: Longbool = FALSE): Boolean;
var
  Start: Integer;
begin
  Start := 1 + Ord(CheckDelims and (Url <> E) and SysUtils.CharInset(Url[1], BLANKS + [QU, SQ, LT]));
  Result := StrTools.TextAtPos(RAW(SDOM_WWW), Url, Start);
end;

function IsWWWUrl(const Url: UnicodeString; CheckDelims: Longbool = FALSE): Boolean;
var
  Start: Integer;
begin
  Start := 1 + Ord(CheckDelims and (Url <> E) and SysUtils.CharInset(Url[1], BLANKS + [QU, SQ, LT]));
  Result := StrTools.TextAtPos(SDOM_WWW, Url, Start);
end;


// All the rest of it ----------------------------------------------------------

procedure SilentSafeCallException(ErrorCode: Integer; ErrorAddr: Pointer);
begin
  Abort;
end;

function IsContextMenuShortCut: Longbool;
begin
  Result := Bool(Ord(GetKeyState(VK_APPS) < 0) * VK_APPS);
  if not Result then Result := Bool(Ord((GetKeyState(VK_F10) < 0) and (GetKeyState(VK_SHIFT) < 0)) * VK_APPS);
end;

function GetShiftState(KeyData: Longint): Word;
const
  AltMask = $20000000;
begin
  Result := (Ord(GetKeyState(VK_SHIFT) < 0) * scShift) or
	    (Ord(GetKeyState(VK_CONTROL) < 0) * scCtrl) or
	    (Ord(KeyData and AltMask <> 0) * scAlt);
end;

function GetContextMenuShortCut(KeyData: Longint): TShortCut;
begin
  Result := TShortCut(IsContextMenuShortCut) or GetShiftState(KeyData);
end;

function SendKeys(const Keys: Array of Cardinal): Longbool;
const
  MAXSIZE = MAXINT div SizeOf(TInput);
type
  PInputArray = ^TInputArray;
  TInputArray = Array[0..pred(MAXSIZE)] of TInput;
var
  iCount, iIndex, iItems, iKeyUp: Integer;
  pia: PInputArray;
begin
  iCount := succ(High(Keys));
  if iCount > 0 then begin
    iItems := iCount * 2;
    pia := AllocMem(iItems * SizeOf(TInput));
    try
      for iIndex := 0 to pred(iCount) do begin
	iKeyUp := pred(iItems - iIndex);
	with pia^[iIndex] do begin
	  Itype := INPUT_KEYBOARD;
	  with ki do begin
	    wVk := Keys[iIndex];
	    wScan := MapVirtualKey(wVk, MAPVK_VK_TO_VSC);
	    dwExtraInfo := MAXDWORD;
	  end;
	end;
	pia^[iKeyUp] := pia^[iIndex];
	pia^[iKeyUp].ki.dwFlags := KEYEVENTF_KEYUP;
      end;
      Result := (SendInput(iItems, pia[0], SizeOf(TInput)) <> 0);
      Exit;
    finally
      FreeMem(pia);
    end;
  end;
  Result := FALSE;
end;

function GetKeyDataToShiftState(KeyData: Longint): TShiftState;
const
  AltMask = $20000000;
begin
  Result := [];
  if GetKeyState(VK_SHIFT) < 0 then Include(Result, ssShift);
  if GetKeyState(VK_CONTROL) < 0 then Include(Result, ssCtrl);
  if KeyData and AltMask <> 0 then Include(Result, ssAlt);
end;

function GetShiftStateFromShiftStateEx(SSE: TShiftStateEx): TShiftState;
begin
  Result := [];
  if sseShift in SSE then Include(Result, ssShift);
  if sseCtrl in SSE then Include(Result, ssCtrl);
  if sseAlt in SSE then Include(Result, ssAlt);
  if sseLeft in SSE then Include(Result, ssLeft);
  if sseRight in SSE then Include(Result, ssRight);
  if sseMiddle in SSE then Include(Result, ssMiddle);
end;

function KeyboardStateToShiftStateEx: TShiftStateEx;
var
  KeyState: TKeyBoardState;
begin
  GetKeyboardState(KeyState);
  Result := KeyboardStateToShiftStateEx(KeyState);
end;

function KeyboardStateToShiftStateEx(const KeyboardState: TKeyboardState): TShiftStateEx;
begin
  Result := [];
  if KeyboardState[VK_SHIFT] and $80 <> 0 then Include(Result, sseShift);
  if KeyboardState[VK_CONTROL] and $80 <> 0 then Include(Result, sseCtrl);
  if KeyboardState[VK_MENU] and $80 <> 0 then Include(Result, sseAlt);
  if KeyboardState[VK_LBUTTON] and $80 <> 0 then Include(Result, sseLeft);
  if KeyboardState[VK_RBUTTON] and $80 <> 0 then Include(Result, sseRight);
  if KeyboardState[VK_MBUTTON] and $80 <> 0 then Include(Result, sseMiddle);
  if KeyboardState[VK_LSHIFT] and $80 <> 0 then Include(Result, sseLShift);
  if KeyboardState[VK_RSHIFT] and $80 <> 0 then Include(Result, sseRShift);
  if KeyboardState[VK_LCONTROL] and $80 <> 0 then Include(Result, sseLCtrl);
  if KeyboardState[VK_RCONTROL] and $80 <> 0 then Include(Result, sseRCtrl);
  if KeyboardState[VK_LMENU] and $80 <> 0 then Include(Result, sseLAlt);
  if KeyboardState[VK_RMENU] and $80 <> 0 then Include(Result, sseRAlt);
end;


// TStringArray ----------------------------------------------------------------

function TStringArray.Grow(Items: Integer): Integer;
begin
  Result := FCount;
  FCount := Result + Items;
  ReallocMem(FStrings, SizeOf(TAStrings) * FCount);
  Initialize(FStrings^[Result], Items);
end;

function TStringArray.AdjustString(const Str: String): String;
begin
  if Assigned(FStringAdjust) then
    Result := FStringAdjust(Str)
  else Result := Str;
end;

procedure TStringArray.New;
begin
  FCount := 0;
  FStrings := NIL;
end;

procedure TStringArray.Clear;
begin
  if FStrings <> NIL then begin
    if FCount > 0 then Finalize(FStrings^, FCount);
    FreeMem(FStrings);
    FStrings := NIL;
    FCount := 0;
  end;
end;

function TStringArray.Count: Integer;
begin
  Result := FCount;
end;

function TStringArray.GetString(Index: Integer): String;
begin
  if (Index >= 0) and (Index < FCount) then
    Result := FStrings^[Index]
  else Result := E;
end;

procedure TStringArray.SetString(Index: Integer; const Value: String);
begin
  if (Index >= 0) and (Index < FCount) then FStrings^[Index] := Value;
end;

function TStringArray.GetText: String;
var
  iIndex, iLen: Integer;
  pResult: PChar;
begin
  if FCount > 0 then begin
    iLen := 0;
    for iIndex := 0 to pred(FCount) do inc(iLen, Length(FStrings^[iIndex]) + Length(CRLF));
    System.SetString(Result, NIL, iLen);
    pResult := Pointer(Result);
    for iIndex := 0 to pred(FCount) do begin
      iLen := Length(FStrings^[iIndex]);
      if iLen > 0 then begin
	StrTools.StrPLCopy(pResult, FStrings^[iIndex], iLen);
	inc(pResult, iLen);
      end;
      PDWChar(pResult)^ := CRLF;
      inc(pResult, Length(CRLF));
    end;
    Exit;
  end;
  Result := E;
end;

procedure TStringArray.InsertLine(const Value: String; Index: Integer);
var
  iCount: Integer;
begin
  if (Value <> E) and (Index > 0) then begin
    iCount := Grow(1);
    if Index < iCount then begin
      Move(FStrings^[Index], FStrings^[succ(Index)], (iCount - Index) * SizeOf(String));
      FStrings^[Index] := Value;
    end
    else FStrings^[iCount] := Value;
  end;
end;

procedure TStringArray.AddLines(const Values: Array of String);
var
  iCurrent, iIndex: Integer;
begin
  iCurrent := Grow(succ(High(Values)));
  for iIndex := 0 to High(Values) do FStrings^[iCurrent + iIndex] := AdjustString(Values[iIndex]);
end;

procedure TStringArray.AddText(const Value: String);
var
  iOld, iNew, iBeg, iEnd, iIndex: Integer;
begin
  if Value <> E then begin
    iNew := GetCharCount(LF, Value);
    if not SysUtils.CharInSet(Value[Length(Value)], [LF, CR]) then inc(iNew);
    iOld := Grow(iNew);
    iBeg := 1;
    iEnd := 1;
    for iIndex := iOld to pred(FCount) do begin
      while not SysUtils.CharInSet(Value[iEnd], [NUL, LF, CR]) do inc(iEnd);
      if iEnd > iBeg then FStrings^[iIndex] := AdjustString(Copy(Value, iBeg, iEnd - iBeg));
      while SysUtils.CharInSet(Value[iEnd], [LF, CR]) do inc(iEnd);
      iBeg := iEnd;
    end;
  end;
end;


// TRawByteStringStream --------------------------------------------------------

constructor TRawByteStringStream.Create(const AString: RawByteString);
begin
  inherited Create;
  FDataString := AString;
end;

function TRawByteStringStream.Read(var Buffer; Count: Longint): Longint;
var
  pData: PAnsiChar;
begin
  Result := Length(FDataString) - FPosition;
  if Result > Count then Result := Count;
  pData := Pointer(FDataString);
  inc(pData, FPosition);
  Move(pData^, Buffer, Result);
  Inc(FPosition, Result);
end;

function TRawByteStringStream.Write(const Buffer; Count: Longint): Longint;
var
  pData: PAnsiChar;
begin
  Result := Count;
  if FPosition + Count > Size then SetSize(FPosition + Count);
  pData := Pointer(FDataString);
  inc(pData, FPosition);
  Move(Buffer, pData^, Result);
  Inc(FPosition, Result);
end;

function TRawByteStringStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
  case Origin of
    soFromBeginning: FPosition := Offset;
    soFromCurrent: Inc(FPosition, Offset);
    soFromEnd: FPosition := Length(FDataString) + Offset;
  end;
  if FPosition > Length(FDataString) then
    FPosition := Length(FDataString)
  else if FPosition <= 0 then FPosition := 0;
  Result := FPosition;
end;

function TRawByteStringStream.ReadString(Count: Longint): RawByteString;
var
  Len: Integer;
begin
  Len := Length(FDataString) - FPosition;
  if Len > Count then Len := Count;
  SetString(Result, PAnsiChar(@FDataString[FPosition + SizeOf(AnsiChar)]), Len);
  Inc(FPosition, Len);
end;

procedure TRawByteStringStream.WriteString(const AString: RawByteString);
begin
  Write(PAnsiChar(AString)^, Length(AString));
end;

procedure TRawByteStringStream.SetSize(NewSize: Longint);
begin
  SetLength(FDataString, NewSize);
  if FPosition > NewSize then FPosition := NewSize;
end;


// All the rest ----------------------------------------------------------------

function SCodeError(SCode: Longint): Cardinal; inline;
begin
  Result := SCODE_CODE(SCode);
end;

function HResultError(HRes: HResult): Cardinal; inline;
begin
  Result := HRESULT_CODE(HRes);
end;

function MakeSCode(Severity1Bit, Facility11Bits, Code8Bits: ULong): HResult; inline;
begin
  Result := MAKE_SCODE(Severity1Bit, Facility11Bits, Code8Bits);
end;

function MakeHResult(Severity1Bit, Facility11Bits, Code8Bits: ULong): HResult; inline;
begin
  Result := MAKE_HRESULT(Severity1Bit, Facility11Bits, Code8Bits);
end;

function GetRegKeyList(Key: HKEY; const Path: String; List: TStrings; const ReturnValue: String = E;
		       const RegNameFiltered: TRegNameFiltered = NIL): Cardinal;
var
  hRegKey: HKEY;
  dwSize: DWord;
  paSubKey: TPathArray;
  ftWritten: TFileTime;
  sSubKey, sValue: String;
begin
  Result := 0;
  if (List <> NIL) and (RegOpenKeyEx(Key, PChar(Path), 0, KEY_READ, hRegKey) = S_OK) then begin
    try
      List.BeginUpdate;
      repeat
	dwSize := SizeOf(TPathArray);
	if RegEnumKeyEx(hRegKey, Result, paSubKey, dwSize, NIL, NIL, NIL, @ftWritten) = S_OK then begin
	  sSubKey := StrTools.Trim(paSubKey);
	  if sSubKey <> E then begin
	    if ReturnValue <> E then begin
	      sValue := GetRegString(Key, Path + BS + sSubKey, ReturnValue);
	      if sValue <> E then begin
		if not Assigned(RegNameFiltered) or RegNameFiltered(sSubKey) then begin
		  List.Add(sSubKey + List.NameValueSeparator + sValue);
		end;
	      end;
	    end
	    else List.Add(sSubKey);
	    inc(Result);
	  end;
	end
	else Break;
      until Result = 0;
    finally
      RegCloseKey(hRegKey);
      List.EndUpdate;
    end;
  end;
end;

function WinNT: Longbool;
begin
  Result := (GetVersion < $80000000);
end;

function XPSP3: Longbool;
var
  viex: TOSVersionInfoEx;
  dwlConditionMask: DWordLong;
begin
  FillChar(viex, SizeOf(viex), 0);
  with viex do begin
    dwOSVersionInfoSize := SizeOf(viex);
    dwMajorVersion := 5;
    dwMinorVersion := 1;
    wServicePackMajor := 3;
    wServicePackMajor := 0;
  end;
  dwlConditionMask := 0;
  dwlConditionMask := VerSetConditionMask(dwlConditionMask, VER_MAJORVERSION, VER_GREATER_EQUAL);
  dwlConditionMask := VerSetConditionMask(dwlConditionMask, VER_MINORVERSION, VER_GREATER_EQUAL);
  dwlConditionMask := VerSetConditionMask(dwlConditionMask, VER_SERVICEPACKMAJOR, VER_GREATER_EQUAL);
  dwlConditionMask := VerSetConditionMask(dwlConditionMask, VER_SERVICEPACKMINOR, VER_GREATER_EQUAL);
  Result := VerifyVersionInfo(viex, VER_MAJORVERSION or VER_MINORVERSION or VER_SERVICEPACKMAJOR or VER_SERVICEPACKMINOR, dwlConditionMask);
end;

function IsLower(C: WideChar): Longbool;
begin
  {$IF CompilerVersion > 21.0}
  Result := C.IsLower;
  {$ELSE]}
  Result := TCharacter.IsLower(C);
  {$IFEND}
end;

function IsUpper(C: WideChar): Longbool;
begin
  {$IF CompilerVersion > 21.0}
  Result := C.IsUpper;
  {$ELSE]}
  Result := TCharacter.IsUpper(C);
  {$IFEND}
end;

function Asc(C: AnsiChar): Byte;
begin
  Result := Ord(C);
end;

function Asc(C: WideChar): Byte;
begin
  Result := Ord(C);
end;

function CharInSet(C: AnsiChar; const S: TSysCharSet): Longbool; inline;
begin
  Result := SysUtils.CharInSet(C, S);
end;

function CharInSet(C: WideChar; const S: TWideCharArray): Longbool;
var
  iIndex: Integer;
begin
  for iIndex := Low(S) to High(S) do begin
    Result := (C = S[iIndex]);
    if Result then Exit;
  end;
  Result := FALSE;
end;

function CharMove(Dest, Source: PAnsiChar): PAnsiChar;
begin
  Result := Dest;
  Result^ := Source^;
  inc(Result);
end;

function CharMove(Dest, Source: PWideChar): PWideChar;
begin
  Result := Dest;
  Result^ := Source^;
  inc(Result);
end;

function LoChar(C: AnsiChar): AnsiChar;
begin
  if Ord(C) in HI_CHARS then
    Result := AnsiChar(Ord(C) or $20)
  else Result := C;
end;

function LoChar(C: WideChar): WideChar;
begin
  if Ord(C) in HI_CHARS then
    Result := WideChar(Ord(C) or $20)
  else Result := C;
end;

function HiChar(C: AnsiChar): AnsiChar;
begin
  if Ord(C) in LO_CHARS then
    Result := AnsiChar(Ord(C) and $DF)
  else Result := C;
end;

function HiChar(C: WideChar): WideChar;
begin
  if Ord(C) in LO_CHARS then
    Result := WideChar(Ord(C) and $DF)
  else Result := C;
end;

function SamePDAChars(PDAC1, PDAC2: PDAChar): Longbool; register; assembler;
asm
  {$IFDEF CPUX64}
  // RCX = PDAC1, RDX = PDAC2
  MOV	CX,[RCX]
  MOV	DX,[RDX]
  XOR	RAX,RAX
  CMP	CL,DL
  JZ	@NEXT
  CMP	CL,'A'
  JB    @FALSE
  CMP	CL,'Z'
  JA	@LOWER1
  PUSH	CX
  OR	CL,020h
  CMP	CL,DL
  POP	CX
  JZ	@NEXT
  @LOWER1:
  CMP	CL,'a'
  JB	@FALSE
  CMP	CL,'z'
  JA	@FALSE
  AND	CL,0DFh
  CMP	CL,DL
  JNZ	@FALSE
  @NEXT:
  SHR	CX,08h
  SHR   DX,08h
  CMP	CL,DL
  JZ	@TRUE
  CMP	CL,'A'
  JB    @FALSE
  CMP	CL,'Z'
  JA	@LOWER2
  PUSH	CX
  OR	CL,020h
  CMP	CL,DL
  POP	CX
  JZ	@TRUE
  @LOWER2:
  CMP	CL,'a'
  JB	@FALSE
  CMP	CL,'z'
  JA	@FALSE
  AND	CL,0DFh
  CMP	CL,DL
  JNZ	@FALSE
  @TRUE:
  MOV	RAX,DWORD(-1)
  @FALSE:
  {$ELSE}
  // EAX = PDAC1, EDX = PDAC2
  MOV	CX,[EAX]
  MOV	DX,[EDX]
  XOR	EAX,EAX
  CMP	CL,DL
  JZ	@NEXT
  CMP	CL,'A'
  JB    @FALSE
  CMP	CL,'Z'
  JA	@LOWER1
  PUSH	CX
  OR	CL,020h
  CMP	CL,DL
  POP	CX
  JZ	@NEXT
  @LOWER1:
  CMP	CL,'a'
  JB	@FALSE
  CMP	CL,'z'
  JA	@FALSE
  AND	CL,0DFh
  CMP	CL,DL
  JNZ	@FALSE
  @NEXT:
  SHR	CX,08h
  SHR   DX,08h
  CMP	CL,DL
  JZ	@TRUE
  CMP	CL,'A'
  JB    @FALSE
  CMP	CL,'Z'
  JA	@LOWER2
  PUSH	CX
  OR	CL,020h
  CMP	CL,DL
  POP	CX
  JZ	@TRUE
  @LOWER2:
  CMP	CL,'a'
  JB	@FALSE
  CMP	CL,'z'
  JA	@FALSE
  AND	CL,0DFh
  CMP	CL,DL
  JNZ	@FALSE
  @TRUE:
  MOV	EAX,DWORD(-1)
  @FALSE:
  {$ENDIF}
end;

function SamePDWChars(PDWC1, PDWC2: PDWChar): Longbool; register; assembler;
asm
  {$IFDEF CPUX64}
  // RCX = PDAC1, RDX = PDAC2
  MOV	RCX,[RCX]
  MOV	RDX,[RDX]
  XOR	RAX,RAX
  CMP	CX,DX
  JZ	@NEXT
  CMP	CX,'A'
  JB    @FALSE
  CMP	CX,'Z'
  JA	@LOWER1
  PUSH	CX
  OR	CX,0020h
  CMP	CX,DX
  POP	CX
  JZ	@NEXT
  @LOWER1:
  CMP	CX,'a'
  JB	@FALSE
  CMP	CX,'z'
  JA	@FALSE
  AND	CX,00DFh
  CMP	CX,DX
  JNZ	@FALSE
  @NEXT:
  SHR	RCX,010h
  SHR   RDX,010h
  CMP	CX,DX
  JZ	@TRUE
  CMP	CX,'A'
  JB    @FALSE
  CMP	CX,'Z'
  JA	@LOWER2
  PUSH	CX
  OR	CX,0020h
  CMP	CX,DX
  POP	CX
  JZ	@TRUE
  @LOWER2:
  CMP	CX,'a'
  JB	@FALSE
  CMP	CX,'z'
  JA	@FALSE
  AND	CX,00DFh
  CMP	CX,DX
  JNZ	@FALSE
  @TRUE:
  MOV	RAX,DWORD(-1)
  @FALSE:
  {$ELSE}
  // EAX = PDAC1, EDX = PDAC2
  MOV	ECX,[EAX]
  MOV	EDX,[EDX]
  XOR	EAX,EAX
  CMP	CX,DX
  JZ	@NEXT
  CMP	CX,'A'
  JB    @FALSE
  CMP	CX,'Z'
  JA	@LOWER1
  PUSH	CX
  OR	CX,0020h
  CMP	CX,DX
  POP	CX
  JZ	@NEXT
  @LOWER1:
  CMP	CX,'a'
  JB	@FALSE
  CMP	CX,'z'
  JA	@FALSE
  AND	CX,00DFh
  CMP	CX,DX
  JNZ	@FALSE
  @NEXT:
  SHR	ECX,010h
  SHR   EDX,010h
  CMP	CX,DX
  JZ	@TRUE
  CMP	CX,'A'
  JB    @FALSE
  CMP	CX,'Z'
  JA	@LOWER2
  PUSH	CX
  OR	CX,0020h
  CMP	CX,DX
  POP	CX
  JZ	@TRUE
  @LOWER2:
  CMP	CX,'a'
  JB	@FALSE
  CMP	CX,'z'
  JA	@FALSE
  AND	CX,00DFh
  CMP	CX,DX
  JNZ	@FALSE
  @TRUE:
  MOV	EAX,DWORD(-1)
  @FALSE:
  {$ENDIF}
end;

function HexToDigitA(Hex: AnsiChar): Byte; register; assembler;
asm
  {$IFDEF CPUX64}
  // CL = Hex
  SUB	CL,'0'
  JB	@ERROR
  CMP	CL,0Ah
  JB	@EXIT
  AND	CL,0DFh
  SUB	CL,'A' - '0'
  JB	@ERROR
  CMP	CL,05h
  JA	@ERROR
  ADD	CL,0Ah
  @EXIT:
  MOVZX	RAX,CL
  RET
  @ERROR:
  XOR	RAX,RAX
  {$ELSE}
  // AL = Hex
  SUB	AL,'0'
  JB	@ERROR
  CMP	AL,0Ah
  JB	@EXIT
  AND	AL,0DFh
  SUB	AL,'A' - '0'
  JB	@ERROR
  CMP	AL,05h
  JA	@ERROR
  ADD	AL,0Ah
  @EXIT:
  MOVZX	EAX,AL
  RET
  @ERROR:
  XOR	EAX,EAX
  {$ENDIF}
end;

function HexToDigitW(Hex: WideChar): Byte; register; assembler;
asm
  {$IFDEF CPUX64}
  // CX = Hex
  SUB	CX,'0'
  JB	@ERROR
  CMP	CX,0Ah
  JB	@EXIT
  AND	CX,0FFDFh
  SUB	CX,'A' - '0'
  JB	@ERROR
  CMP	CX,05h
  JA	@ERROR
  ADD	CX,0Ah
  @EXIT:
  MOVZX	RAX,CX
  RET
  @ERROR:
  XOR	RAX,RAX
  {$ELSE}
  // AX = Hex
  SUB	AX,'0'
  JB	@ERROR
  CMP	AX,0Ah
  JB	@EXIT
  AND	AX,0FFDFh
  SUB	AX,'A' - '0'
  JB	@ERROR
  CMP	AX,05h
  JA	@ERROR
  ADD	AX,0Ah
  @EXIT:
  MOVZX	EAX,AX
  RET
  @ERROR:
  XOR	EAX,EAX
  {$ENDIF}
end;

function DigitToHexA(Digit: Byte): AnsiChar; register; assembler;
asm
  {$IFDEF CPUX64}
  // CL = Digit
  AND	CL,0Fh
  ADD	CL,'0'
  CMP	CL,'9'
  JBE	@EXIT
  ADD	CL,'A' - '0' - 0Ah
  @EXIT:
  MOVZX	RAX,CL
  {$ELSE}
  // AL = Digit
  AND	AL,0Fh
  ADD	AL,'0'
  CMP	AL,'9'
  JBE	@EXIT
  ADD	AL,'A' - '0' - 0Ah
  @EXIT:
  MOVZX	EAX,AL
  {$ENDIF}
end;

function DigitToHexW(Digit: Byte): WideChar; register; assembler;
asm
  {$IFDEF CPUX64}
  // CX = Digit
  AND	CX,0Fh
  ADD	CX,'0'
  CMP	CX,'9'
  JBE	@EXIT
  ADD	CX,'A' - '0' - 0Ah
  @EXIT:
  MOVZX	RAX,CX
  {$ELSE}
  // AX = Digit
  AND	AX,0Fh
  ADD	AX,'0'
  CMP	AX,'9'
  JBE	@EXIT
  ADD	AX,'A' - '0' - 0Ah
  @EXIT:
  MOVZX	EAX,AX
  {$ENDIF}
end;

function HexToCharA(Hex: PAnsiChar): AnsiChar;
begin
  Result := AnsiChar(HexToDigitA(Hex[0]) shl 4 + HexToDigitA(Hex[1]));
end;

function HexToCharW(Hex: PWideChar): WideChar;
begin
  Result := WideChar(HexToDigitW(Hex[0]) shl 4 + HexToDigitW(Hex[1]));
end;

function HexByteToDigit(HexByte: PAnsiChar): Byte;
asm
  {$IFDEF CPUX64}
  // RCX = @HexByte (two hex chars, returns NUL on error)
  MOV	AX,[RCX]
  SUB	AL,'0'
  JB	@ERROR
  CMP	AL,0Ah
  JB	@NEXT
  AND	AL,0DFh
  SUB	AL,'A' - '0'
  JB	@ERROR
  CMP	AL,05h
  JA	@ERROR
  ADD	AL,0Ah
  @NEXT:
  SUB	AH,'0'
  JB	@ERROR
  CMP	AH,0Ah
  JB	@EXIT
  AND	AH,0DFh
  SUB	AH,'A' - '0'
  JB	@ERROR
  CMP	AH,05h
  JA	@ERROR
  ADD	AH,0Ah
  @EXIT:
  SHL	AL,04h
  OR	AL,AH
  MOVZX	RAX,AL
  RET
  @ERROR:
  XOR	RAX,RAX
  {$ELSE}
  // EAX = @HexByte (two hex chars, returns NUL on error)
  MOV	AX,[EAX]
  @START:
  SUB	AL,'0'
  JB	@ERROR
  CMP	AL,0Ah
  JB	@NEXT
  AND	AL,0DFh
  SUB	AL,'A' - '0'
  JB	@ERROR
  CMP	AL,05h
  JA	@ERROR
  ADD	AL,0Ah
  @NEXT:
  SUB	AH,'0'
  JB	@ERROR
  CMP	AH,0Ah
  JB	@EXIT
  AND	AH,0DFh
  SUB	AH,'A' - '0'
  JB	@ERROR
  CMP	AH,05h
  JA	@ERROR
  ADD	AH,0Ah
  @EXIT:
  SHL	AL,04h
  OR	AL,AH
  MOVZX	EAX,AL
  RET
  @ERROR:
  XOR	EAX,EAX
  {$ENDIF}
end;

function HexByteToDigit(HexByte: PWideChar): Byte;
asm
  {$IFDEF CPUX64}
  // RAX = @HexByte (two hex chars, returns NUL on error)
  MOV	EAX,[RCX]
  SHL	AX,08h
  SHR	EAX,08h
  SUB	AL,'0'
  JB	@ERROR
  CMP	AL,0Ah
  JB	@NEXT
  AND	AL,0DFh
  SUB	AL,'A' - '0'
  JB	@ERROR
  CMP	AL,05h
  JA	@ERROR
  ADD	AL,0Ah
  @NEXT:
  SUB	AH,'0'
  JB	@ERROR
  CMP	AH,0Ah
  JB	@EXIT
  AND	AH,0DFh
  SUB	AH,'A' - '0'
  JB	@ERROR
  CMP	AH,05h
  JA	@ERROR
  ADD	AH,0Ah
  @EXIT:
  SHL	AL,04h
  OR	AL,AH
  MOVZX	RAX,AL
  RET
  @ERROR:
  XOR	RAX,RAX
  {$ELSE}
  // EAX = @HexByte (two hex chars, returns NUL on error)
  MOV	EAX,[EAX]
  SHL	AX,08h
  SHR	EAX,08h
  @START:
  SUB	AL,'0'
  JB	@ERROR
  CMP	AL,0Ah
  JB	@NEXT
  AND	AL,0DFh
  SUB	AL,'A' - '0'
  JB	@ERROR
  CMP	AL,05h
  JA	@ERROR
  ADD	AL,0Ah
  @NEXT:
  SUB	AH,'0'
  JB	@ERROR
  CMP	AH,0Ah
  JB	@EXIT
  AND	AH,0DFh
  SUB	AH,'A' - '0'
  JB	@ERROR
  CMP	AH,05h
  JA	@ERROR
  ADD	AH,0Ah
  @EXIT:
  SHL	AL,04h
  OR	AL,AH
  MOVZX	EAX,AL
  RET
  @ERROR:
  XOR	EAX,EAX
  {$ENDIF}
end;

function ByteToDigitChars(Value: Byte; PResult: PAnsiChar): Byte;
asm
  {$IFDEF CPUX64}
  // CL->RAX = Value, RDX = PResult, RCX = temp Count
  MOVZX	RAX,CL
  MOV	R8,RDX
  XOR	RCX,RCX
  PUSH	RDX
  ADD	R8,02h
  CMP	RAX,100
  JNB	@LOOP
  DEC	R8
  CMP	RAX,10
  JNB	@LOOP
  DEC	R8
  @LOOP:
  MOV	R9,10
  XOR	RDX,RDX
  DIV	R9
  ADD	RDX,'0'
  MOV	[R8],DL
  DEC	R8
  INC	RCX
  CMP	R8,[ESP]
  JNB	@LOOP
  POP	RDX
  MOV	RAX,RCX
  {$ELSE}
  // AL->EAX = Value, EDX = PResult, ECX = temp Count
  PUSH	EBX
  PUSH	EDX
  MOVZX	EAX,AL
  MOV	EBX,EDX
  XOR	ECX,ECX
  ADD	EBX,02h
  CMP	EAX,100
  JNB	@LOOP
  DEC	EBX
  CMP	EAX,10
  JNB	@LOOP
  DEC	EBX
  @LOOP:
  PUSH	EBX
  MOV	EBX,10
  XOR	EDX,EDX
  DIV	EBX
  ADD	EDX,'0'
  POP	EBX
  MOV	[EBX],DL
  DEC	EBX
  INC	ECX
  CMP	EBX,[ESP]
  JNB	@LOOP
  @EXIT:
  MOV	EAX,ECX
  POP	EDX
  POP	EBX
  {$ENDIF}
end;

function ByteToDigitChars(Value: Byte; PResult: PWideChar): Byte;
asm
  {$IFDEF CPUX64}
  // CL->RAX = Value, RDX = PResult, RCX = temp Count
  MOVZX	RAX,CL
  MOV	R8,RDX
  XOR	RCX,RCX
  PUSH	RDX
  ADD	R8,04h
  CMP	RAX,100
  JNB	@LOOP
  DEC	R8
  DEC	R8
  CMP	RAX,10
  JNB	@LOOP
  DEC	R8
  DEC	R8
  @LOOP:
  MOV	R9,10
  XOR	RDX,RDX
  DIV	R9
  ADD	RDX,'0'
  MOV	[R8],DX
  DEC	R8
  DEC	R8
  INC	RCX
  CMP	R8,[ESP]
  JNB	@LOOP
  POP	RDX
  MOV	RAX,RCX
  {$ELSE}
  // AL->EAX = Value, EDX = PResult, ECX = temp Count
  PUSH	EBX
  PUSH	EDX
  MOVZX	EAX,AX
  MOV	EBX,EDX
  XOR	ECX,ECX
  ADD	EBX,04h
  CMP	EAX,100
  JNB	@LOOP
  DEC	EBX
  DEC	EBX
  CMP	EAX,10
  JNB	@LOOP
  DEC	EBX
  DEC	EBX
  @LOOP:
  PUSH	EBX
  MOV	EBX,10
  XOR	EDX,EDX
  DIV	EBX
  ADD	EDX,'0'
  POP	EBX
  MOV	[EBX],DX
  DEC	EBX
  DEC	EBX
  INC	ECX
  CMP	EBX,[ESP]
  JNB	@LOOP
  @EXIT:
  MOV	EAX,ECX
  POP	EDX
  POP	EBX
  {$ENDIF}
end;

function DecCharsValue(DecChars: PAnsiChar; Count: Cardinal): Cardinal; register; assembler;
asm
  {$IFDEF CPUX64}
  // RCX = @DecChars, RDX = Count
  OR	RCX,RCX
  JNZ	@START
  OR	EDX,EDX
  JNZ	@START
  XOR	RAX,RAX
  JMP	@EXIT
  @START:
  MOV	R8,RCX
  XOR	RAX,RAX
  XOR	RCX,RCX
  @LOOP:
  MOVZX	RCX,BYTE PTR[R8]
  OR	RCX,RCX
  JE	@EXIT
  SUB	RCX,'0'
  JB	@EXIT
  CMP	RCX,09h
  JA	@EXIT
  @NEXT:
  PUSH	RAX
  PUSH	RDX
  MOV	RDX,0Ah
  MUL	RDX
  ADD	RAX,RCX
  POP	RDX
  POP	RCX
  INC	R8
  DEC	RDX
  JNE	@LOOP
  @EXIT:
  {$ELSE}
  // EAX = @DecChars, EDX = Count
  OR	EAX,EAX
  JZ	@END
  OR	EDX,EDX
  JNZ	@START
  XOR	EAX,EAX
  JMP	@END
  @START:
  PUSH	EBX
  MOV	EBX,EAX
  XOR	EAX,EAX
  XOR	ECX,ECX
  @LOOP:
  MOVZX	ECX,BYTE PTR[EBX]
  OR	ECX,ECX
  SUB	ECX,'0'
  JB	@EXIT
  CMP	ECX,09h
  JA	@EXIT
  @NEXT:
  PUSH	EAX
  PUSH	EDX
  MOV	EDX,0Ah
  MUL	EDX
  ADD	EAX,ECX
  POP	EDX
  POP	ECX
  INC	EBX
  DEC	EDX
  JNE	@LOOP
  @EXIT:
  POP	EBX
  @END:
  {$ENDIF}
end;

function DecCharsValue(DecChars: PWideChar; Count: Cardinal): Cardinal; register; assembler;
asm
  {$IFDEF CPUX64}
  // RCX = @DecChars, RDX = Count
  OR	RCX,RCX
  JNZ	@START
  OR	EDX,EDX
  JNZ	@START
  XOR	RAX,RAX
  JMP	@EXIT
  @START:
  MOV	R8,RCX
  XOR	RAX,RAX
  XOR	RCX,RCX
  @LOOP:
  MOVZX	RCX,WORD PTR[R8]
  OR	RCX,RCX
  JE	@EXIT
  SUB	RCX,'0'
  JB	@EXIT
  CMP	RCX,09h
  JA	@EXIT
  @NEXT:
  PUSH	RAX
  PUSH	RDX
  MOV	RDX,0Ah
  MUL	RDX
  ADD	RAX,RCX
  POP	RDX
  POP	RCX
  INC	R8
  INC	R8
  DEC	RDX
  JNE	@LOOP
  @EXIT:
  {$ELSE}
  // EAX = @DecChars, EDX = Count
  OR	EAX,EAX
  JZ	@END
  OR	EDX,EDX
  JNZ	@START
  XOR	EAX,EAX
  JMP	@END
  @START:
  PUSH	EBX
  MOV	EBX,EAX
  XOR	EAX,EAX
  XOR	ECX,ECX
  @LOOP:
  MOVZX	ECX,WORD PTR[EBX]
  OR	ECX,ECX
  JE	@EXIT
  SUB	ECX,'0'
  JB	@EXIT
  CMP	ECX,09h
  JA	@EXIT
  @NEXT:
  PUSH	EAX
  PUSH	EDX
  MOV	EDX,0Ah
  MUL	EDX
  ADD	EAX,ECX
  POP	EDX
  POP	ECX
  INC	EBX
  INC	EBX
  DEC	EDX
  JNE	@LOOP
  @EXIT:
  POP	EBX
  @END:
  {$ENDIF}
end;

function HexCharsValue(HexChars: PAnsiChar; Count: Cardinal): Cardinal; register; assembler;
asm
  {$IFDEF CPUX64}
  // RCX = @HexChars, RDX = Count
  OR	RCX,RCX
  JNZ	@START
  OR	EDX,EDX
  JNZ	@START
  XOR	RAX,RAX
  JMP	@EXIT
  @START:
  MOV	R8,RCX
  XOR	RAX,RAX
  XOR	RCX,RCX
  @LOOP:
  MOVZX	RCX,BYTE PTR[R8]
  OR	RCX,RCX
  JE	@EXIT
  SUB	RCX,'0'
  JB	@EXIT
  CMP	RCX,0Ah
  JB	@NEXT
  AND	RCX,0DFh
  SUB	RCX,'A' - '0'
  JB	@EXIT
  CMP	RCX,05h
  JA	@EXIT
  ADD	RCX,0Ah
  @NEXT:
  SHL	RAX,04h
  ADD	RAX,RCX
  INC	R8
  DEC	RDX
  JNE	@LOOP
  @EXIT:
  {$ELSE}
  // EAX = @HexChars, EDX = Count
  OR	EAX,EAX
  JZ	@END
  OR	EDX,EDX
  JNZ	@START
  XOR	EAX,EAX
  JMP	@END
  @START:
  PUSH	EBX
  MOV	EBX,EAX
  XOR	EAX,EAX
  XOR	ECX,ECX
  @LOOP:
  MOVZX	ECX,BYTE PTR[EBX]
  OR	ECX,ECX
  JE	@EXIT
  SUB	ECX,'0'
  JB	@EXIT
  CMP	ECX,0Ah
  JB	@NEXT
  AND	ECX,0DFh
  SUB	ECX,'A' - '0'
  JB	@EXIT
  CMP	ECX,05h
  JA	@EXIT
  ADD	ECX,0Ah
  @NEXT:
  SHL	EAX,04h
  ADD	EAX,ECX
  INC	EBX
  DEC	EDX
  JNE	@LOOP
  @EXIT:
  POP	EBX
  @END:
  {$ENDIF}
end;

function HexCharsValue(HexChars: PWideChar; Count: Cardinal): Cardinal; register; assembler;
asm
  {$IFDEF CPUX64}
  // RCX = @HexChars, RDX = Count
  OR	RCX,RCX
  JNZ	@START
  OR	EDX,EDX
  JNZ	@START
  XOR	RAX,RAX
  JMP	@EXIT
  @START:
  MOV	R8,RCX
  XOR	RAX,RAX
  XOR	RCX,RCX
  @LOOP:
  MOVZX	RCX,WORD PTR[R8]
  OR	RCX,RCX
  JE	@EXIT
  SUB	RCX,'0'
  JB	@EXIT
  CMP	RCX,0Ah
  JB	@NEXT
  AND	RCX,0FFDFh
  SUB	RCX,'A' - '0'
  JB	@EXIT
  CMP	RCX,05h
  JA	@EXIT
  ADD	RCX,0Ah
  @NEXT:
  SHL	RAX,04h
  ADD	RAX,RCX
  INC	R8
  INC	R8
  DEC	RDX
  JNE	@LOOP
  @EXIT:
  {$ELSE}
  // EAX = @HexChars, EDX = Count
  OR	EAX,EAX
  JZ	@END
  OR	EDX,EDX
  JNZ	@START
  XOR	EAX,EAX
  JMP	@END
  @START:
  PUSH	EBX
  MOV	EBX,EAX
  XOR	EAX,EAX
  XOR	ECX,ECX
  @LOOP:
  MOVZX	ECX,WORD PTR[EBX]
  OR	ECX,ECX
  JE	@EXIT
  SUB	ECX,'0'
  JB	@EXIT
  CMP	ECX,0Ah
  JB	@NEXT
  AND	ECX,0FFDFh
  SUB	ECX,'A' - '0'
  JB	@EXIT
  CMP	ECX,05h
  JA	@EXIT
  ADD	ECX,0Ah
  @NEXT:
  SHL	EAX,04h
  ADD	EAX,ECX
  INC	EBX
  INC	EBX
  DEC	EDX
  JNE	@LOOP
  @EXIT:
  POP	EBX
  @END:
  {$ENDIF}
end;

function HexToBytes(const Hex: RawByteString): RawByteString;
var
  iLen: Integer;
begin
  iLen := Length(Hex) shr 1;
  SetString(Result, NIL, iLen);
  HexToBin(PAnsiChar(StrTools.LowerCase(Hex)), PAnsiChar(Result), iLen);
end;

function HexToBytes(const Hex: UnicodeString): UnicodeString;
var
  iLen: Integer;
begin
  iLen := Length(Hex) shr 1;
  SetString(Result, NIL, iLen);
  HexToBin(PWideChar(StrTools.LowerCase(Hex)), PWideChar(Result), iLen);
end;

function BytesToHex(const Bytes: RawByteString): RawByteString;
var
  iLen: Integer;
begin
  iLen := Length(Bytes);
  SetString(Result, NIL, iLen shl 1);
  BinToHex(PAnsiChar(Bytes), PAnsiChar(Result), iLen);
end;

function BytesToHex(const Bytes: UnicodeString): UnicodeString;
var
  iLen: Integer;
begin
  iLen := Length(Bytes);
  SetString(Result, NIL, iLen shl 1);
  BinToHex(PWideChar(Bytes), PWideChar(Result), iLen);
end;

function BytesToHexA(const Bytes: PByte; Count: Cardinal; UpperCase: Longbool = TRUE): RawByteString;
const
  HEXCHARS: PAnsiChar = ('0123456789ABCDEF');
var
  iIndex: Integer;
  pResult: PAnsiChar;
begin
  SetString(Result, NIL, Count shl 01);
  pResult := Pointer(Result);
  for iIndex := 0 to pred(Count) do begin
    pResult^ := HEXCHARS[Bytes[iIndex] shr 04];
    inc(pResult);
    pResult^ := HEXCHARS[Bytes[iIndex] and $F];
    inc(pResult);
  end;
  if not UpperCase then CharLowerA(Pointer(Result));
end;

function BytesToHexW(const Bytes: PByte; Count: Cardinal; UpperCase: Longbool = TRUE): UnicodeString;
const
  HEXCHARS: PWideChar = ('0123456789ABCDEF');
var
  iIndex: Integer;
  pResult: PWideChar;
begin
  SetString(Result, NIL, Count shl 01);
  pResult := Pointer(Result);
  for iIndex := 0 to pred(Count) do begin
    pResult^ := HEXCHARS[Bytes[iIndex] shr 04];
    inc(pResult);
    pResult^ := HEXCHARS[Bytes[iIndex] and $F];
    inc(pResult);
  end;
  if not UpperCase then CharLowerW(Pointer(Result));
end;

function BytesToHexChars(const Bytes: PByte; Count: Cardinal; HexChars: PAnsiChar; UpperCase: Longbool = TRUE): Integer;
const
  HEX: PAnsiChar = ('0123456789ABCDEF');
var
  iIndex: Integer;
begin
  Result := Count shl 01;
  for iIndex := 0 to pred(Count) do begin
    HexChars^ := HEX[Bytes[iIndex] shr 04];
    inc(HexChars);
    HexChars^ := HEX[Bytes[iIndex] and $F];
    inc(HexChars);
  end;
  if not UpperCase then begin
    dec(HexChars, Result);
    CharLowerA(HexChars);
  end;
end;

function BytesToHexChars(const Bytes: PByte; Count: Cardinal; HexChars: PWideChar; UpperCase: Longbool = TRUE): Integer;
const
  HEX: PWideChar = ('0123456789ABCDEF');
var
  iIndex: Integer;
begin
  Result := Count shl 01;
  for iIndex := 0 to pred(Count) do begin
    HexChars^ := HEX[Bytes[iIndex] shr 04];
    inc(HexChars);
    HexChars^ := HEX[Bytes[iIndex] and $F];
    inc(HexChars);
  end;
  if not UpperCase then begin
    dec(HexChars, Result);
    CharLowerW(HexChars);
  end;
end;

function HexCharsToDigitCharsA(Digits, HexChars: PAnsiChar; HexLen: Cardinal): PAnsiChar;
var
  cIndex, cCount: Cardinal;
begin
  if (Digits <> NIL) and (HexChars <> NIL) and (HexLen > 0) then begin
    cIndex := 0;
    while cIndex < HexLen do begin
      cCount := ByteToDigitChars(HexByteToDigit(HexChars), Digits);
      inc(Digits, cCount);
      inc(HexChars, 2);
      inc(cIndex, 2);
    end;
  end;
  Result := Digits; // if (Digits <> NIL) and (HexLen > 0) then Result = StrEnd(Digits)
end;

function HexCharsToDigitCharsW(Digits, HexChars: PWideChar; HexLen: Cardinal): PWideChar;
var
  cIndex, cCount: Cardinal;
begin
  if (Digits <> NIL) and (HexChars <> NIL) and (HexLen > 0) then begin
    cIndex := 0;
    while cIndex < HexLen do begin
      cCount := ByteToDigitChars(HexByteToDigit(HexChars), Digits);
      inc(Digits, cCount);
      inc(HexChars, 2);
      inc(cIndex, 2);
    end;
  end;
  Result := Digits; // if (Digits <> NIL) and (HexLen > 0) then Result = StrEnd(Digits)
end;

function HexCharsToDigitsA(HexChars: PAnsiChar; HexLen: Cardinal): RawByteString;
var
  pResult: PAnsiChar;
  cIndex, cCount: Cardinal;
begin
  if (HexChars <> NIL) and (HexLen > 0) then begin
    SetString(Result, NIL, MulDiv(HexLen, 3, 2));
    pResult := Pointer(Result);
    cIndex := 0;
    while cIndex < HexLen do begin
      cCount := ByteToDigitChars(HexByteToDigit(HexChars), pResult);
      inc(pResult, cCount);
      inc(HexChars, 2);
      inc(cIndex, 2);
    end;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end;
end;

function HexCharsToDigitsW(HexChars: PWideChar; HexLen: Cardinal): UnicodeString;
var
  pResult: PWideChar;
  cIndex, cCount: Cardinal;
begin
  if (HexChars <> NIL) and (HexLen > 0) then begin
    SetString(Result, NIL, MulDiv(HexLen, 3, 2));
    pResult := Pointer(Result);
    cIndex := 0;
    while cIndex < HexLen do begin
      cCount := ByteToDigitChars(HexByteToDigit(HexChars), pResult);
      inc(pResult, cCount);
      inc(HexChars, 2);
      inc(cIndex, 2);
    end;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end;
end;

function HexCharsGroup(const Hex: RawByteString; GroupChars: Integer): RawByteString;
var
  iLen, iCount: Integer;
  pResult, pHex: PAnsiChar;
begin
  iLen := Length(Hex);
  if (GroupChars > 0) and (iLen > GroupChars) then begin
    SetString(Result, NIL, iLen + (iLen div GroupChars));
    pResult := Pointer(Result);
    pHex := Pointer(Hex);
    repeat
      iCount := 0;
      repeat
	pResult^ := pHex^;
	inc(pResult);
	inc(iCount);
	inc(pHex);
      until (iCount = GroupChars) or (pHex^ = NUL);
      if pHex^ <> NUL then begin
	pResult^ := SP;
	inc(pResult);
      end
      else begin
	Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
	Exit;
      end;
    until pHex^ = NUL;
  end;
  Result := Hex;
end;

function HexCharsGroup(const Hex: UnicodeString; GroupChars: Integer): UnicodeString;
var
  iLen, iCount: Integer;
  pResult, pHex: PWideChar;
begin
  iLen := Length(Hex);
  if (GroupChars > 0) and (iLen > GroupChars) then begin
    SetString(Result, NIL, iLen + (iLen div GroupChars));
    pResult := Pointer(Result);
    pHex := Pointer(Hex);
    repeat
      iCount := 0;
      repeat
	pResult^ := pHex^;
	inc(pResult);
	inc(iCount);
	inc(pHex);
      until (iCount = GroupChars) or (pHex^ = NUL);
      if pHex^ <> NUL then begin
	pResult^ := SP;
	inc(pResult);
      end
      else begin
	Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
	Exit;
      end;
    until pHex^ = NUL;
  end;
  Result := Hex;
end;

function CardToHexA(Value: UInt64; Digits: Cardinal = 8; const Prefix: AnsiString = '0x'): AnsiString; inline;
begin
  Result := Prefix + AnsiString(IntToHex(Value, Digits));
end;

function CardToHexW(Value: UInt64; Digits: Cardinal = 8; const Prefix: UnicodeString = '0x'): UnicodeString; inline;
begin
  Result := Prefix + SysUtils.IntToHex(Value, Digits);
end;

function CardToStrA(Value: UInt64): AnsiString; inline;
begin
  Result := AnsiString(UIntToStr(Value));
end;

function CardToStrW(Value: UInt64): UnicodeString; inline;
begin
  Result := UIntToStr(Value);
end;

function StrToCardEx(const Str: RawByteString): UInt64;
var
  cIndex: Cardinal;
begin
  if Str <> E then begin
    cIndex := 1;
    while SysUtils.CharInSet(Str[cIndex], DIGITS) do inc(cIndex);
    Result := StrToCard(Copy(Str, 1, pred(cIndex)));
    Exit;
  end;
  Result := 0;
end;

function StrToCardEx(const Str: UnicodeString): UInt64;
var
  cIndex: Cardinal;
begin
  if Str <> E then begin
    cIndex := 1;
    while SysUtils.CharInSet(Str[cIndex], DIGITS) do inc(cIndex);
    Result := StrToCard(Copy(Str, 1, pred(cIndex)));
    Exit;
  end;
  Result := 0;
end;

function StrToCardDef(const Str: RawByteString; Def: UInt64): UInt64; inline;
begin
  Result := StrLenToCard(Str, 1, Length(Str));
end;

function StrToCardDef(const Str: UnicodeString; Def: UInt64): UInt64; inline;
begin
  Result := StrLenToCard(Str, 1, Length(Str));
end;

function StrToCard(const Str: RawByteString; AddHexPrefix: Longbool = FALSE): UInt64; inline;
begin
  Result := StrLenToCard(Str, 1, Length(Str), AddHexPrefix);
end;

function StrToCard(const Str: UnicodeString; AddHexPrefix: Longbool = FALSE): UInt64; inline;
begin
  Result := StrLenToCard(Str, 1, Length(Str), AddHexPrefix);
end;

function StrToInt64Def(const Str: RawByteString; Default: Int64 = 0): UInt64;
var
  i64Result: Int64;
begin
  if StrToInt64ExA(PAnsiChar(Str), STIF_SUPPORT_HEX, i64Result) then
    Result := i64Result
  else Result := Default;
end;

function StrToInt64Def(const Str: UnicodeString; Default: Int64 = 0): UInt64;
var
  i64Result: Int64;
begin
  if StrToInt64ExW(PWideChar(Str), STIF_SUPPORT_HEX, i64Result) then
    Result := i64Result
  else Result := Default;
end;

function StrLenToCard(const Str: RawByteString; Pos, Len: Integer; AddHexPrefix: Longbool = FALSE): UInt64;
const
  HEX = '0x';
var
  rsHex: RawByteString;
begin
  if (Pos > 0) and (Len > 0) then begin
    if Str[Pos] = US then
      rsHex := HEX + Copy(Str, succ(Pos), pred(Len))
    else if AddHexPrefix then
      rsHex := HEX + Copy(Str, Pos, Len)
    else rsHex := Copy(Str, Pos, Len);
    Result := StrToInt64Def(rsHex);
    Exit;
  end;
  Result := 0;
end;

function StrLenToCard(const Str: UnicodeString; Pos, Len: Integer; AddHexPrefix: Longbool = FALSE): UInt64;
const
  HEX = '0x';
var
  usHex: UnicodeString;
begin
  if (Pos > 0) and (Len > 0) then begin
    if Str[Pos] = US then
      usHex := HEX + Copy(Str, succ(Pos), pred(Len))
    else if AddHexPrefix then
      usHex := HEX + Copy(Str, Pos, Len)
    else usHex := Copy(Str, Pos, Len);
    Result := StrToInt64Def(usHex);
    Exit;
  end;
  Result := 0;
end;

function StrToFloatEx(const Value: RawByteString; Default: Extended = -1): Extended;
var
  cIndex: Cardinal;
  fs: TFormatSettings;
begin
  if Value <> E then begin
    cIndex := 1;
    while SysUtils.CharInSet(Value[cIndex], FLOATCHARS) do inc(cIndex);
    {$IF CompilerVersion > 21.0}
    fs := TFormatSettings.Create('en-US');
    if AnsiStrings.TextToFloat(PAnsiChar(StrTools.Copy(Value, 1, pred(cIndex))), Result, fvExtended, fs) then Exit;
    {$ELSE}
    GetLocaleFormatSettings(1033, fs);
    if SysUtils.TextToFloat(PAnsiChar(StrTools.Copy(Value, 1, pred(cIndex))), Result, fvExtended, fs) then Exit;
    {$IFEND}
  end;
  Result := Default;
end;

function StrToFloatEx(const Value: UnicodeString; Default: Extended = -1): Extended;
var
  cIndex: Cardinal;
  fs: TFormatSettings;
begin
  if Value <> E then begin
    cIndex := 1;
    while SysUtils.CharInSet(Value[cIndex], FLOATCHARS) do inc(cIndex);
    {$IF CompilerVersion > 21.0}
    fs := TFormatSettings.Create('en-US');
    if SysUtils.TextToFloat(PWideChar(StrTools.Copy(Value, 1, pred(cIndex))), Result, fvExtended, fs) then Exit;
    {$ELSE}
    GetLocaleFormatSettings(1033, fs);
    if SysUtils.TextToFloat(PWideChar(StrTools.Copy(Value, 1, pred(cIndex))), Result, fvExtended, fs) then Exit;
    {$IFEND}
  end;
  Result := Default;
end;

function FixUnicodeLineBreaks(const Text: String; var Fixed: Boolean): String;
var
  pResult, pText, pLen: PWideChar;
begin
  Fixed := FALSE;
  if Text <> E then begin
    SetString(Result, NIL, Length(Text));
    pResult := Pointer(Result);
    pText := Pointer(Text);
    pLen := pText + Length(Text);
    // prevent access violation on first char
    pResult^ := pText^;
    inc(pResult);
    inc(pText);
    while pText < pLen do begin
      if (pText^ = LF) and (PDAChar(pText - 1)[1] = CR) then begin
	PDWChar(pResult - 1)^ := CRLF;
	Fixed := TRUE;
      end
      else pResult^ := pText^;
      inc(pResult);
      inc(pText);
    end;
  end;
end;

function DelAmp(const Str: RawByteString): RawByteString;
var
  pStr, pResult: PAnsiChar;
  iLen: Integer;
begin
  if Str <> E then begin
    pStr := Pointer(Str);
    while not SysUtils.CharInSet(pStr^, [NUL, AM]) do inc(pStr);
    if pStr^ = AM then begin
      SetString(Result, NIL, pred(Length(Str)));
      pResult := Pointer(Result);
      iLen := pStr - Pointer(Str);
      StrTools.StrLCopy(pResult, Pointer(Str), iLen);
      inc(pResult, iLen);
      inc(iLen);
      inc(pStr);
      iLen := Length(Str) - iLen;
      StrTools.StrLCopy(pResult, pStr, iLen);
      Exit;
    end;
  end;
  Result := Str;
end;

function DelAmp(const Str: UnicodeString): UnicodeString;
var
  pStr, pResult: PWideChar;
  iLen: Integer;
begin
  if Str <> E then begin
    pStr := Pointer(Str);
    while not SysUtils.CharInSet(pStr^, [NUL, AM]) do inc(pStr);
    if pStr^ = AM then begin
      SetString(Result, NIL, pred(Length(Str)));
      pResult := Pointer(Result);
      iLen := pStr - Pointer(Str);
      StrTools.StrLCopy(pResult, Pointer(Str), iLen);
      inc(pResult, iLen);
      inc(iLen);
      inc(pStr);
      iLen := Length(Str) - iLen;
      StrTools.StrLCopy(pResult, pStr, iLen);
      Exit;
    end;
  end;
  Result := Str;
end;

function SetPasString(Text: PAnsiChar; Size: Integer; Trim: Longbool = FALSE): RawByteString;
var
  pText: PAnsiChar;
begin
  if Trim then begin
    pText := Text + Size - 1;
    while (pText >= Text) and (pText^ <= SP) do dec(pText);
    if pText > Text then while (Text^ > NUL) and (Text^ <= SP) do inc(Text);
    Size := succ(pText - Text);
  end;
  SetString(Result, Text, Size);
end;

function SetPasString(Text: PWideChar; Size: Integer; Trim: Longbool = FALSE): UnicodeString;
var
  pText: PWideChar;
begin
  if Trim then begin
    pText := Text + Size - 1;
    while (pText >= Text) and (pText^ <= SP) do dec(pText);
    if pText > Text then while (Text^ > NUL) and (Text^ <= SP) do inc(Text);
    Size := succ(pText - Text);
  end;
  SetString(Result, Text, Size);
end;

function Crop(const Text: RawByteString; Count: Integer): RawByteString; inline;
begin
  if Count < 0 then Result := Crop(Text, Length(Text) + succ(Count), MAXINT);
end;

function Crop(const Text: RawByteString; Index, Count: Integer): RawByteString; inline;
begin
  Result := Text;
  if (Index > 0) and (Count > 0) then System.Delete(Result, Index, Count);
end;

function Crop(const Text: UnicodeString; Count: Integer): UnicodeString; inline;
begin
  if Count < 0 then Result := Crop(Text, Length(Text) + succ(Count), MAXINT);
end;

function Crop(const Text: UnicodeString; Index, Count: Integer): UnicodeString; inline;
begin
  Result := Text;
  if (Index > 0) and (Count > 0) then System.Delete(Result, Index, Count);
end;

function InsStr(const Insert, Text: RawByteString; Index: Integer): RawByteString; inline;
begin
  Result := Text;
  System.Insert(Insert, Result, Index);
end;

function InsStr(const Insert, Text: UnicodeString; Index: Integer): UnicodeString; inline;
begin
  Result := Text;
  System.Insert(Insert, Result, Index);
end;

function NoText(const Str: RawByteString): Longbool;
var
  pStr: PAnsiChar;
begin
  Result := FALSE;
  if Str = E then Exit;
  pStr := Pointer(Str);
  while (pStr^ > NUL) and (pStr^ <= SP) do inc(pStr);
  Result := (pStr^ <= SP);
end;

function NoText(const Str: UnicodeString): Longbool;
var
  pStr: PWideChar;
begin
  Result := FALSE;
  if Str = E then Exit;
  pStr := Pointer(Str);
  while (pStr^ > NUL) and (pStr^ <= SP) do inc(pStr);
  Result := (pStr^ <= SP);
end;

function SetStr(Text: PAnsiChar; Size: Integer): RawByteString; inline;
begin
  SetString(Result, Text, Size);
end;

function SetStr(Text: PWideChar; Size: Integer): UnicodeString; inline;
begin
  SetString(Result, Text, Size);
end;

procedure Capitalize(var Text: RawByteString; AllWords: Longbool = false);
var
  pText: PAnsiChar;
begin
  if Text <> E then begin
    pText := Pointer(Text);
    inc(pText, CharUpperBuffA(pText, 1));
    if AllWords then begin
      repeat
	while pText^ > SP do inc(pText);
	while (pText^ > NUL) and (pText^ <= SP) do inc(pText);
	if pText^ <> NUL then inc(pText, CharUpperBuffA(pText, 1));
      until pText^ = NUL;
    end;
  end;
end;

procedure Capitalize(var Text: UnicodeString; AllWords: Longbool = false);
var
  pText: PWideChar;
begin
  if Text <> E then begin
    pText := Pointer(Text);
    inc(pText, CharUpperBuffW(pText, 1));
    if AllWords then begin
      repeat
	while pText^ > SP do inc(pText);
	while (pText^ > NUL) and (pText^ <= SP) do inc(pText);
	if pText^ <> NUL then inc(pText, CharUpperBuffW(pText, 1));
      until pText^ = NUL;
    end;
  end;
end;

procedure Hyphenize(var Text: RawByteString; AllWords: Longbool = false);
var
  pPos: PAnsiChar;
begin
  if Text <> E then begin
    pPos := Pointer(Text);
    while pPos^ <> NUL do begin
      if pPos^ = SP then begin
	inc(pPos);
	if not AllWords then Exit;
      end;
    end;
  end;
end;

procedure Hyphenize(var Text: UnicodeString; AllWords: Longbool = false);
var
  pPos: PWideChar;
begin
  if Text <> E then begin
    pPos := Pointer(Text);
    while pPos^ <> NUL do begin
      if pPos^ = SP then begin
	pPos^ := DA;
	if not AllWords then Exit;
      end;
      inc(pPos);
    end;
  end;
end;

function SpaceCompress(const Text: RawByteString): RawByteString;
var
  pResult: PAnsiChar;
  iBeg, iEnd, iLen: Integer;
begin
  if Text <> E then begin
    SetString(Result, NIL, Length(Text));
    pResult := Pointer(Result);
    iBeg := 1;
    repeat
      iEnd := iBeg;
      while not SysUtils.CharInSet(Text[iEnd], [NUL, SP]) do inc(iEnd);
      iLen := iEnd - iBeg;
      if iLen > 0 then begin
	StrTools.StrLCopy(pResult, @Text[iBeg], iLen);
	inc(pResult, iLen);
      end;
      iBeg := iEnd;
      while Text[iBeg] = SP do inc(iBeg);
      if (Text[iBeg] > SP) and (Text[pred(iEnd)] > SP) then begin
	pResult^ := SP;
	inc(pResult);
      end;
    until Text[iBeg] = NUL;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := Trim(Text);
end;

function SpaceCompress(const Text: UnicodeString): UnicodeString;
var
  pResult: PWideChar;
  iBeg, iEnd, iLen: Integer;
begin
  if Text <> E then begin
    SetString(Result, NIL, Length(Text));
    pResult := Pointer(Result);
    iBeg := 1;
    repeat
      iEnd := iBeg;
      while not SysUtils.CharInSet(Text[iEnd], [NUL, SP]) do inc(iEnd);
      iLen := iEnd - iBeg;
      if iLen > 0 then begin
	StrTools.StrLCopy(pResult, @Text[iBeg], iLen);
	inc(pResult, iLen);
      end;
      iBeg := iEnd;
      while Text[iBeg] = SP do inc(iBeg);
      if (Text[iBeg] > SP) and (Text[pred(iEnd)] > SP) then begin
	pResult^ := SP;
	inc(pResult);
      end;
    until Text[iBeg] = NUL;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := StrTools.Trim(Text);
end;

function Trim(const Str: RawByteString): RawByteString; inline;
begin
  Result := AnsiStrings.Trim(Str);
end;

function Trim(const Str: UnicodeString): UnicodeString; inline;
begin
  Result := SysUtils.Trim(Str);
end;

function TrimLeft(const Str: RawByteString): RawByteString; inline;
begin
  Result := AnsiStrings.TrimLeft(Str);
end;

function TrimLeft(const Str: UnicodeString): UnicodeString; inline;
begin
  Result := SysUtils.TrimLeft(Str);
end;

function TrimRight(const Str: RawByteString): RawByteString; inline;
begin
  Result := AnsiStrings.TrimRight(Str);
end;

function TrimRight(const Str: UnicodeString): UnicodeString; inline;
begin
  Result := SysUtils.TrimRight(Str);
end;

procedure TrimSelf(var Str: RawByteString);
var
  pBeg, pEnd: PAnsiChar;
  iLen: Integer;
begin
  if Str <> E then begin
    pBeg := Pointer(Str);
    pEnd := pBeg + Length(Str);
    while (pEnd >= pBeg) and (pEnd^ <= SP) do dec(pEnd);
    while (pBeg <= pEnd) and (pBeg^ <= SP) do inc(pBeg);
    iLen := succ(pEnd - pBeg);
    if iLen > 0 then begin
      StrTools.StrLCopy(PAnsiChar(@Str[1]), pBeg, iLen);
      Delete(Str, succ(iLen), MAXINT);
      Exit;
    end;
  end;
  Str := E;
end;

procedure TrimSelf(var Str: UnicodeString);
var
  pBeg, pEnd: PWideChar;
  iLen: Integer;
begin
  if Str <> E then begin
    pBeg := Pointer(Str);
    pEnd := pBeg + Length(Str);
    while (pEnd >= pBeg) and (pEnd^ <= SP) do dec(pEnd);
    while (pBeg <= pEnd) and (pBeg^ <= SP) do inc(pBeg);
    iLen := succ(pEnd - pBeg);
    if iLen > 0 then begin
      StrTools.StrLCopy(PWideChar(@Str[1]), pBeg, iLen);
      Delete(Str, succ(iLen), MAXINT);
      Exit;
    end;
  end;
  Str := E;
end;

function TrimCopy(const Str: RawByteString; Index, Count: Integer): RawByteString;
var
  pBeg, pEnd: PAnsiChar;
  iLen: Integer;
begin
  if Str <> E then begin
    pBeg := Pointer(Str);
    iLen := Length(Str);
    if Index < 1 then
      Index := 1
    else if Index > iLen then Index := iLen;
    inc(pBeg, pred(Index));
    if Count - Index > iLen then
      pEnd := pBeg + iLen
    else pEnd := pBeg + pred(Count);
    while (pEnd >= pBeg) and (pEnd^ <= SP) do dec(pEnd);
    while (pBeg <= pEnd) and (pBeg^ <= SP) do inc(pBeg);
    iLen := succ(pEnd - pBeg);
    if iLen > 0 then begin
      SetString(Result, pBeg, iLen);
      Exit;
    end;
  end;
  Result := E;
end;

function TrimCopy(const Str: UnicodeString; Index, Count: Integer): UnicodeString;
var
  pBeg, pEnd: PWideChar;
  iLen: Integer;
begin
  if Str <> E then begin
    pBeg := Pointer(Str);
    iLen := Length(Str);
    if Index < 1 then
      Index := 1
    else if Index > iLen then Index := iLen;
    inc(pBeg, pred(Index));
    if Count - Index > iLen then
      pEnd := pBeg + iLen
    else pEnd := pBeg + pred(Count);
    while (pEnd >= pBeg) and (pEnd^ <= SP) do dec(pEnd);
    while (pBeg <= pEnd) and (pBeg^ <= SP) do inc(pBeg);
    iLen := succ(pEnd - pBeg);
    if iLen > 0 then begin
      SetString(Result, pBeg, iLen);
      Exit;
    end;
  end;
  Result := E;
end;

function TrimCustom(const Str: RawByteString; TrimCharSet: TSysCharSet; TrimOptions: TTrimOptions = TO_BOTH): RawByteString;
var
  iLen, iEnd, iBeg: Integer;
begin
  if Str <> E then begin
    TrimCharSet := TrimCharSet + TRIMCHARS;
    iLen := Length(Str);
    iEnd := iLen;
    iBeg := 1;
    if toRight in TrimOptions then while (iEnd > 0) and SysUtils.CharInSet(Str[iEnd], TrimCharSet + [NUL]) do dec(iEnd);
    if iEnd > 0 then begin
      if toLeft in TrimOptions then while SysUtils.CharInSet(Str[iBeg], TrimCharSet) do inc(iBeg);
      iLen := succ(iEnd - iBeg);
      if iLen < Length(Str) then
	Result := Copy(Str, iBeg, iLen)
      else Result := Str;
      Exit;
    end;
  end;
  Result := E
end;

function TrimCustom(const Str: UnicodeString; TrimCharSet: TSysCharSet; TrimOptions: TTrimOptions = TO_BOTH): UnicodeString;
var
  iLen, iEnd, iBeg: Integer;
begin
  if Str <> E then begin
    TrimCharSet := TrimCharSet + TRIMCHARS;
    iLen := Length(Str);
    iEnd := iLen;
    iBeg := 1;
    if toRight in TrimOptions then while (iEnd > 0) and SysUtils.CharInSet(Str[iEnd], TrimCharSet + [NUL]) do dec(iEnd);
    if iEnd > 0 then begin
      if toLeft in TrimOptions then while SysUtils.CharInSet(Str[iBeg], TrimCharSet) do inc(iBeg);
      iLen := succ(iEnd - iBeg);
      if iLen < Length(Str) then
	Result := Copy(Str, iBeg, iLen)
      else Result := Str;
      Exit;
    end;
  end;
  Result := E
end;

function DoWrap(const Source: RawByteString; LineLen: Cardinal = 80): RawByteString; inline;
begin
  Result := InsertLineBreaks(Source, LineLen);
end;

function DoWrap(const Source: UnicodeString; LineLen: Cardinal = 80): UnicodeString; inline;
begin
  Result := InsertLineBreaks(Source, LineLen);
end;

function Unwrap(Source: PAnsiChar; SrcLen: Integer; TrimCharSet: TSysCharSet = []; ReplaceChar: AnsiChar = NUL): RawByteString;
var
  pResult, pEnd, pPos: PAnsiChar;
begin
  if (Source <> NIL) and (SrcLen > 0) then begin
    TrimCharSet := TrimCharSet + TRIMCHARS;
    SetString(Result, NIL, SrcLen);
    pResult := Pointer(Result);
    pEnd := Source + SrcLen;
    while SysUtils.CharInSet(Source^, TrimCharSet) do inc(Source);
    while SysUtils.CharInSet(pEnd^, TrimCharSet) do dec(pEnd);
    pPos := Source;
    repeat
      while (pPos < pEnd) and not SysUtils.CharInSet(pPos^, [LF, CR]) do inc(pPos);
      pResult := StrTools.StrLECopy(pResult, Source, pPos - Source);
      if pPos = pEnd then Break;
      while SysUtils.CharInSet(pPos^, [LF, CR]) do inc(pPos);
      if pPos = pEnd then Break;
      if (ReplaceChar > NUL) and (pResult > Pointer(Result)) then begin
	if (pResult - 1)^ <> ReplaceChar then begin
	  pResult^ := ReplaceChar;
	  inc(pResult);
	end;
	while (pPos < pEnd) and SysUtils.CharInSet(pPos^, [TAB, SP]) do inc(pPos);
      end;
      Source := pPos;
    until Source^ = pEnd;
    while (pResult > Pointer(Result)) and SysUtils.CharInSet((pResult - 1)^, TrimCharSet) do dec(pResult);
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
    Exit;
  end;
  Result := E;
end;

function Unwrap(Source: PWideChar; SrcLen: Integer; TrimCharSet: TSysCharSet = []; ReplaceChar: WideChar = NUL): UnicodeString;
var
  pResult, pEnd, pPos: PWideChar;
begin
  if (Source <> NIL) and (SrcLen > 0) then begin
    TrimCharSet := TrimCharSet + TRIMCHARS;
    SetString(Result, NIL, SrcLen);
    pResult := Pointer(Result);
    pEnd := Source + SrcLen;
    while SysUtils.CharInSet(Source^, TrimCharSet) do inc(Source);
    while SysUtils.CharInSet(pEnd^, TrimCharSet) do dec(pEnd);
    pPos := Source;
    repeat
      while (pPos < pEnd) and not SysUtils.CharInSet(pPos^, [LF, CR]) do inc(pPos);
      pResult := StrTools.StrLECopy(pResult, Source, pPos - Source);
      if pPos = pEnd then Break;
      while SysUtils.CharInSet(pPos^, [LF, CR]) do inc(pPos);
      if pPos = pEnd then Break;
      if (ReplaceChar > NUL) and (pResult > Pointer(Result)) then begin
	if (pResult - 1)^ <> ReplaceChar then begin
	  pResult^ := ReplaceChar;
	  inc(pResult);
	end;
	while (pPos < pEnd) and SysUtils.CharInSet(pPos^, [TAB, SP]) do inc(pPos);
      end;
      Source := pPos;
    until Source^ = pEnd;
    while (pResult > Pointer(Result)) and SysUtils.CharInSet((pResult - 1)^, TrimCharSet) do dec(pResult);
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
    Exit;
  end;
  Result := E;
end;

function Unwrap(const Str: RawByteString; TrimCharSet: TSysCharSet = []; ReplaceChar: AnsiChar = NUL): RawByteString; inline;
begin
  Result := Unwrap(PAnsiChar(Str), Length(Str), TrimCharSet, ReplaceChar);
end;

function Unwrap(const Str: UnicodeString; TrimCharSet: TSysCharSet = []; ReplaceChar: WideChar = NUL): UnicodeString; inline;
begin
  Result := Unwrap(PWideChar(Str), Length(Str), TrimCharSet, ReplaceChar);
end;

function Unquote(const Str: RawByteString; CBeg: AnsiChar = QU; CEnd: AnsiChar = QU): RawByteString;
var
  pBeg, pEnd: PAnsiChar;
  iLen: Integer;
begin
  if Str <> E then begin
    pBeg := Pointer(Str);
    pEnd := pBeg;
    inc(pEnd, Length(Str));
    while (pEnd > pBeg) and (pEnd^ <= SP) do dec(pEnd);
    while (pBeg < pEnd) and (pBeg^ <= SP) do inc(pBeg);
    while pEnd > pBeg do begin
      if (pBeg^ = CBeg) and (pEnd^ = CEnd) then begin
	inc(pBeg);
	dec(pEnd);
      end
      else Break;
    end;
    iLen := succ(pEnd - pBeg);
    if (pBeg > Pointer(Str)) or (iLen < Length(Str)) then
      SetString(Result, pBeg, iLen)
    else Result := Str;
    Exit;
  end;
  Result := E;
end;

function Unquote(const Str: UnicodeString; CBeg: WideChar = QU; CEnd: WideChar = QU): UnicodeString;
var
  pBeg, pEnd: PWideChar;
  iLen: Integer;
begin
  if Str <> E then begin
    pBeg := Pointer(Str);
    pEnd := pBeg;
    inc(pEnd, Length(Str));
    while (pEnd > pBeg) and (pEnd^ <= SP) do dec(pEnd);
    while (pBeg < pEnd) and (pBeg^ <= SP) do inc(pBeg);
    while pEnd > pBeg do begin
      if (pBeg^ = CBeg) and (pEnd^ = CEnd) then begin
	inc(pBeg);
	dec(pEnd);
      end
      else Break;
    end;
    iLen := succ(pEnd - pBeg);
    if (pBeg > Pointer(Str)) or (iLen < Length(Str)) then
      SetString(Result, pBeg, iLen)
    else Result := Str;
    Exit;
  end;
  Result := E;
end;

function Unquote(const Str: RawByteString; QuoteChar: AnsiChar; var WasQuoted: Longbool): RawByteString;
var
  pBeg, pEnd: PAnsiChar;
  iLen: Integer;
begin
  WasQuoted := FALSE;
  if Str <> E then begin
    pBeg := Pointer(Str);
    pEnd := pBeg;
    inc(pEnd, Length(Str));
    while (pEnd > pBeg) and (pEnd^ <= SP) do dec(pEnd);
    while (pBeg < pEnd) and (pBeg^ <= SP) do inc(pBeg);
    while pEnd > pBeg do begin
      if (pBeg^ = QuoteChar) and (pEnd^ = QuoteChar) then begin
        WasQuoted := TRUE;
	inc(pBeg);
	dec(pEnd);
      end
      else Break;
    end;
    iLen := succ(pEnd - pBeg);
    if (pBeg > Pointer(Str)) or (iLen < Length(Str)) then
      SetString(Result, pBeg, iLen)
    else Result := Str;
    Exit;
  end;
  Result := E;
end;

function Unquote(const Str: UnicodeString; QuoteChar: WideChar; var WasQuoted: Longbool): UnicodeString;
var
  pBeg, pEnd: PWideChar;
  iLen: Integer;
begin
  WasQuoted := FALSE;
  if Str <> E then begin
    pBeg := Pointer(Str);
    pEnd := pBeg;
    inc(pEnd, Length(Str));
    while (pEnd > pBeg) and (pEnd^ <= SP) do dec(pEnd);
    while (pBeg < pEnd) and (pBeg^ <= SP) do inc(pBeg);
    while pEnd > pBeg do begin
      if (pBeg^ = QuoteChar) and (pEnd^ = QuoteChar) then begin
        WasQuoted := TRUE;
	inc(pBeg);
	dec(pEnd);
      end
      else Break;
    end;
    iLen := succ(pEnd - pBeg);
    if (pBeg > Pointer(Str)) or (iLen < Length(Str)) then
      SetString(Result, pBeg, iLen)
    else Result := Str;
    Exit;
  end;
  Result := E;
end;

function SetUnquote(const Str: RawByteString; SBeg: TQuoteChars = QUOTES; SEnd: TQuoteChars = QUOTES): RawByteString;
var
  pBeg, pEnd: PAnsiChar;
  iLen: Integer;
begin
  if Str <> E then begin
    pBeg := Pointer(Str);
    pEnd := pBeg;
    inc(pEnd, Length(Str));
    while (pEnd > pBeg) and (pEnd^ <= SP) do dec(pEnd);
    while (pBeg < pEnd) and (pBeg^ <= SP) do inc(pBeg);
    while pEnd > pBeg do begin
      if SysUtils.CharInSet(pBeg^, SBeg) and SysUtils.CharInSet(pEnd^, SEnd) then begin
	inc(pBeg);
	dec(pEnd);
      end
      else Break;
    end;
    iLen := succ(pEnd - pBeg);
    if (pBeg > Pointer(Str)) or (iLen < Length(Str)) then
      SetString(Result, pBeg, iLen)
    else Result := Str;
    Exit;
  end;
  Result := E;
end;

function SetUnquote(const Str: UnicodeString; SBeg: TQuoteChars = QUOTES; SEnd: TQuoteChars = QUOTES): UnicodeString;
var
  pBeg, pEnd: PWideChar;
  iLen: Integer;
begin
  if Str <> E then begin
    pBeg := Pointer(Str);
    pEnd := pBeg;
    inc(pEnd, Length(Str));
    while (pEnd > pBeg) and (pEnd^ <= SP) do dec(pEnd);
    while (pBeg < pEnd) and (pBeg^ <= SP) do inc(pBeg);
    while pEnd > pBeg do begin
      if SysUtils.CharInSet(pBeg^, SBeg) and SysUtils.CharInSet(pEnd^, SEnd) then begin
	inc(pBeg);
	dec(pEnd);
      end
      else Break;
    end;
    iLen := succ(pEnd - pBeg);
    if (pBeg > Pointer(Str)) or (iLen < Length(Str)) then
      SetString(Result, pBeg, iLen)
    else Result := Str;
    Exit;
  end;
  Result := E;
end;

function Unescape(const Str: RawByteString; SkipSpaces: Longbool = FALSE; EscapeChar: AnsiChar = PC;
		  MakeIXBreaks: Longbool = FALSE): Utf8String;
const
  BREAKS = [LF, CR];
var
  pSource: PAnsiChar;
  rsUtf8: RawByteString;
  pUtf8: PAnsiChar;
begin
  if Str <> E then begin
    SetString(rsUtf8, NIL, Length(Str));
    pUtf8 := Pointer(rsUtf8);
    pSource := Pointer(Str);
    repeat
      while not SysUtils.CharInSet(pSource^, [NUL, EscapeChar]) do begin
	pUtf8^ := AnsiChar(pSource^);
	inc(pSource);
	inc(pUtf8);
      end;
      if pSource^ = EscapeChar then begin
	// if not quoted-printable
	if (EscapeChar <> EQ) or not SysUtils.CharInSet((pSource + 1)^, BREAKS) then begin
	  if MakeIXBreaks and (HexToCharA(pSource + 1) = CR) then
	    pUtf8^ := LF
	  else pUtf8^ := AnsiChar(HexToCharA(pSource + 1));
	  if (pUtf8^ > NUL) and (not SkipSpaces or (pUtf8^ <> SP)) then begin
	    inc(pSource, 3);
	    inc(pUtf8);
	  end
	  else begin
	    pUtf8^ := AnsiChar(EscapeChar);
	    inc(pSource);
	  end;
	end
	else begin
	  repeat
	    inc(pSource);
	  until not SysUtils.CharInSet(pSource^, BREAKS);
	end;
      end
      else Break;
    until pSource^ = NUL;
    Delete(rsUtf8, succ(pUtf8 - Pointer(rsUtf8)), MAXINT);
    if Length(rsUtf8) < Length(Str) then
      Result := rsUtf8
    else Result := Str;
  end
  else Result := Str;
end;

function Unescape(const Str: UnicodeString; SkipSpaces: Longbool = FALSE; EscapeChar: WideChar = PC;
		  MakeIXBreaks: Longbool = FALSE): UnicodeString;
const
  BREAKS = [LF, CR];
var
  pSource: PWideChar;
  rsUtf8: RawByteString;
  pUtf8: PAnsiChar;
begin
  if Str <> E then begin
    SetString(rsUtf8, NIL, Length(Str));
    pUtf8 := Pointer(rsUtf8);
    pSource := Pointer(Str);
    repeat
      while not SysUtils.CharInSet(pSource^, [NUL, EscapeChar]) do begin
	pUtf8^ := AnsiChar(pSource^);
	inc(pSource);
	inc(pUtf8);
      end;
      if pSource^ = EscapeChar then begin
	// if not quoted-printable
	if (EscapeChar <> EQ) or not SysUtils.CharInSet((pSource + 1)^, BREAKS) then begin
	  if MakeIXBreaks and (HexToCharW(pSource + 1) = CR) then
	    pUtf8^ := LF
	  else pUtf8^ := AnsiChar(HexToCharW(pSource + 1));
	  if (pUtf8^ > NUL) and (not SkipSpaces or (pUtf8^ <> SP)) then begin
	    inc(pSource, 3);
	    inc(pUtf8);
	  end
	  else begin
	    pUtf8^ := AnsiChar(EscapeChar);
	    inc(pSource);
	  end;
	end
	else begin
	  repeat
	    inc(pSource);
	  until not SysUtils.CharInSet(pSource^, BREAKS);
	end;
      end
      else Break;
    until pSource^ = NUL;
    Delete(rsUtf8, succ(pUtf8 - Pointer(rsUtf8)), MAXINT);
    if Length(rsUtf8) < Length(Str) then begin
      Result := Utf8CharsToWideString(PAnsiChar(rsUtf8), Length(rsUtf8));
      if Result = E then SetString(Result, PAnsiChar(rsUtf8), Length(rsUtf8));
    end
    else Result := Str;
  end
  else Result := Str;
end;

function UrlEscapeText(const Text: RawByteString; MakeIXBreaks: Longbool = TRUE): RawByteString;
const
  LF_Char: AnsiChar = LF;
var
  pResult, pText: PAnsiChar;
begin
  if Text <> E then begin
    SetString(Result, NIL, Length(Text) * 3);
    pResult := Pointer(Result);
    pText := Pointer(Text);
    while pText^ > NUL do begin
      if (pText^ <= SP) or (pText^ > TLD) then begin
	pResult^ := PC;
	if not (MakeIXBreaks and (pText^ = CR)) then
	  inc(pResult, succ(BytesToHexChars(PByte(pText), 1, pResult + 1)))
	else inc(pResult, succ(BytesToHexChars(@LF_Char, 1, pResult + 1)));
      end
      else begin
	pResult^ := pText^;
	inc(pResult);
      end;
      inc(pText);
    end;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end;
end;

function UrlEscapeText(const Text: UnicodeString; MakeIXBreaks: Longbool = TRUE): UnicodeString;
const
  LF_Char: WideChar = LF;
var
  pResult, pText: PWideChar;
begin
  if Text <> E then begin
    SetString(Result, NIL, Length(Text) * 3);
    pResult := Pointer(Result);
    pText := Pointer(Text);
    while pText^ > NUL do begin
      if (pText^ <= SP) or (pText^ > TLD) then begin
	pResult^ := PC;
	if not (MakeIXBreaks and (pText^ = CR)) then
	  inc(pResult, succ(BytesToHexChars(PByte(pText), 1, pResult + 1)))
	else inc(pResult, succ(BytesToHexChars(@LF_Char, 1, pResult + 1)));
      end
      else begin
	pResult^ := pText^;
	inc(pResult);
      end;
      inc(pText);
    end;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end;
end;

function CheckAST_QMRegEx(const Target, RegEx: String): Longbool;
var
  cTLen, cRLen, cIndex: Cardinal;
  pRegEx, pTarget: PChar;
begin
  Result := FALSE;
  if (Target = E) or (RegEx = E) then Exit;
  cTLen := Length(Target);
  cRLen := Length(RegEx);
  pRegEx := Pointer(RegEx);
  pTarget := Pointer(Target);
  if (pRegEx[0] = AST) or (pRegEx[cRLen - 1] = AST) then begin
    // leading asterisk
    if (pRegEx[0] = AST) and (pRegEx[cRLen - 1] <> AST) then begin
      Result := StrTools.EndsStr(Copy(RegEx, 2, cRLen - 1), Target);
      Exit;
    end
    // enclosing asterisks
    else if (pRegEx[0] = AST) and (pRegEx[cRLen - 1] = AST) then begin
      Result := StrTools.ContainsStr(Target, Copy(RegEx, 2, cRLen - 2));
      Exit;
    end
    // trailing asterisk
    else if (pRegEx[0] <> AST) and (pRegEx[cRLen - 1] = AST) then begin
      Result := StrTools.StartsStr(Copy(RegEx, 1, cRLen - 1), Target);
      Exit;
    end;
  end;
  if cTLen <> cRLen then Exit;
  // accept any char for question marks
  for cIndex := 0 to pred(cTLen) do if (pRegEx[cIndex] <> QM) and (pTarget[cIndex] <> pRegEx[cIndex]) then Exit;
  Result := TRUE;
end;

function GetAnyString(StringPointer: Pointer): String;
var
  cLen: Cardinal;
begin
  cLen := Cardinal(StringPointer);
  if cLen > 0 then begin
    cLen := PCardinal(cLen - SizeOf(Cardinal))^;
    if StrTools.StrLen(PAnsiChar(StringPointer)) <> cLen then
      Result := String(PWideChar(StringPointer))
    else Result := String(PAnsiChar(StringPointer));
    Exit;
  end;
  Result := E;
end;

function GetStringAddress(AnyObject: TObject; const FieldName: String; ClassInfo: Pointer = NIL): Pointer;
var
  rc: TRttiContext;
  rt: TRttiType;
  rf: TRttiField;
begin
  Result := NIL;
  rc := TRttiContext.Create;
  try
    if ClassInfo <> NIL then
      rt := rc.GetType(ClassInfo)
    else rt := rc.GetType(AnyObject.ClassInfo);
    if rt = NIL then Exit;
    rf := rt.GetField(FieldName);
    if rf = NIL then Exit;
    Result := PPointer(Pointer((UInt64(AnyObject) + UInt64(rf.Offset))))^;
  finally
    rc.Free;
  end;
end;

function StringFromHandleA(Hdl: THandle; Size: Integer = -1; Free: Longbool = FALSE; Local: Longbool = FALSE): RawByteString;
var
  pResult: PAnsiChar;
begin
  Result := E;
  if Hdl <> 0 then begin
    try
      if Size = 0 then begin
	if Local then begin
	  Size := LocalSize(Hdl);
	  SetString(Result, PAnsiChar(LocalLock(Hdl)), Size div SizeOf(AnsiChar));
	end
	else begin
	  Size := GlobalSize(Hdl);
	  SetString(Result, PAnsiChar(GlobalLock(Hdl)), Size div SizeOf(AnsiChar));
	end;
	Exit;
      end
      else begin
	if Local then
	  pResult := Pointer(LocalLock(Hdl))
	else pResult := Pointer(GlobalLock(Hdl));
	if pResult <> NIL then begin
	  if Size = -1 then
	    SetString(Result, pResult, StrTools.StrLen(pResult))
	  else if Size > 0 then begin
	    while pResult[pred(Size)] = NUL do dec(Size);
	    SetString(Result, pResult, Size);
	  end;
	end;
      end;
    finally
      if Local then begin
	LocalUnlock(Hdl);
	if Free then LocalFree(Hdl);
      end
      else begin
	GlobalUnlock(Hdl);
	if Free then GlobalFree(Hdl);
      end;
    end;
  end;
end;

function StringFromHandleW(Hdl: THandle; Size: Integer = -1; Free: Longbool = FALSE; Local: Longbool = FALSE): UnicodeString;
var
  pResult: PWideChar;
  fHasBOM: Boolean;
begin
  Result := E;
  if Hdl <> 0 then begin
    try
      if Size = 0 then begin
	if Local then begin
	  Size := LocalSize(Hdl);
	  SetString(Result, PWideChar(LocalLock(Hdl)), Size div SizeOf(WideChar));
	end
	else begin
	  Size := GlobalSize(Hdl);
	  SetString(Result, PWideChar(GlobalLock(Hdl)), Size div SizeOf(WideChar));
	end;
	Exit;
      end
      else begin
	if Local then
	  pResult := Pointer(LocalLock(Hdl))
	else pResult := Pointer(GlobalLock(Hdl));
	if pResult <> NIL then begin
	  fHasBOM := (pResult^ = UNICODE_BOM);
	  inc(pResult, Ord(fHasBOM));
	  if Size = -1 then
	    SetString(Result, pResult, StrTools.StrLen(pResult))
	  else if Size > 0 then begin
	    while pResult[pred(Size)] = NUL do dec(Size);
	    SetString(Result, pResult, Size - Ord(fHasBOM));
	  end;
	end;
      end;
    finally
      if Local then begin
	LocalUnlock(Hdl);
	if Free then LocalFree(Hdl);
      end
      else begin
	GlobalUnlock(Hdl);
	if Free then GlobalFree(Hdl);
      end;
    end;
  end;
end;

function GetPCharHandle(Text: PAnsiChar; Size: Integer = -1; ReallocHandle: THandle = 0; Local: Longbool = FALSE): THandle;
var
  pResult: PAnsiChar;
  cMemSize: Cardinal;
begin
  Result := 0;
  if Text <> NIL then begin
    if Size = -1 then Size := StrTools.StrLen(Text);
    inc(Size);
    cMemSize := Size * (SizeOf(Text[0]));
    if Local then begin
      if ReallocHandle <> 0 then
	Result := LocalReAlloc(ReallocHandle, GHND, cMemSize)
      else Result := LocalAlloc(GHND, cMemSize);
    end
    else begin
      if ReallocHandle <> 0 then
	Result := GlobalReAlloc(ReallocHandle, GHND, cMemSize)
      else Result := GlobalAlloc(GHND, cMemSize);
    end;
    if Result <> 0 then begin
      if Local then
	pResult := LocalLock(Result)
      else pResult := GlobalLock(Result);
      try
	dec(Size);
	Move(Pointer(Text)^, pResult^, cMemSize);
	inc(pResult, Size);
	pResult^ := NUL;
      finally
	if Local then
	  LocalUnlock(Result)
	else GlobalUnlock(Result);
      end;
    end;
  end;
end;

function GetPCharHandle(Text: PWideChar; Size: Integer = -1; ReallocHandle: THandle = 0; Local: Longbool = FALSE): THandle;
var
  pResult: PWideChar;
  cMemSize: Cardinal;
begin
  Result := 0;
  if Text <> NIL then begin
    if Size = -1 then Size := StrTools.StrLen(Text);
    inc(Size);
    cMemSize := Size * (SizeOf(Text[0]));
    if Local then begin
      if ReallocHandle <> 0 then
	Result := LocalReAlloc(ReallocHandle, GHND, cMemSize)
      else Result := LocalAlloc(GHND, cMemSize);
    end
    else begin
      if ReallocHandle <> 0 then
	Result := GlobalReAlloc(ReallocHandle, GHND, cMemSize)
      else Result := GlobalAlloc(GHND, cMemSize);
    end;
    if Result <> 0 then begin
      if Local then
        pResult := LocalLock(Result)
      else pResult := GlobalLock(Result);
      try
        dec(Size);
	Move(Pointer(Text)^, pResult^, cMemSize);
	inc(pResult, Size);
	pResult^ := NUL;
      finally
	if Local then
	  LocalUnlock(Result)
	else GlobalUnlock(Result);
      end;
    end;
  end;
end;

function GetStringHandle(const Str: RawByteString; ReallocHandle: THandle = 0; Local: Longbool = FALSE): THandle; inline;
begin
  Result := GetPCharHandle(PAnsiChar(Str), Length(Str), ReallocHandle, Local);
end;

function GetStringHandle(const Str: UnicodeString; ReallocHandle: THandle = 0; Local: Longbool = FALSE): THandle; inline;
begin
  Result := GetPCharHandle(PWideChar(Str), Length(Str), ReallocHandle, Local);
end;

function GetTextAsHandle(const Strings: Array of RawByteString): THandle;
var
  cSize: Cardinal;
  iIndex, iLength: Integer;
  pText: PAnsiChar;
begin
  Result := 0;
  cSize := 0;
  for iIndex := Low(Strings) to High(Strings) do inc(cSize, Length(Strings[iIndex]));
  if cSize = 0 then Exit;
  cSize := succ(cSize) * SizeOf(pText[0]);
  Result := GlobalAlloc(GHND, cSize);
  if Result = 0 then Exit;
  pText := GlobalLock(Result);
  try
    for iIndex := Low(Strings) to High(Strings) do begin
      iLength := Length(Strings[iIndex]);
      StrTools.StrPLCopy(pText, Strings[iIndex], iLength);
      inc(pText, iLength);
    end;
    pText^ := NUL;
  finally
    GlobalUnlock(Result);
  end;
end;

function GetTextAsHandle(const Strings: Array of UnicodeString): THandle;
var
  cSize: Cardinal;
  iIndex, iLength: Integer;
  pText: PWideChar;
begin
  Result := 0;
  cSize := 0;
  for iIndex := Low(Strings) to High(Strings) do inc(cSize, Length(Strings[iIndex]));
  if cSize = 0 then Exit;
  cSize := succ(cSize) * SizeOf(pText[0]);
  Result := GlobalAlloc(GHND, cSize);
  if Result = 0 then Exit;
  pText := GlobalLock(Result);
  try
    for iIndex := Low(Strings) to High(Strings) do begin
      iLength := Length(Strings[iIndex]);
      StrTools.StrPLCopy(pText, Strings[iIndex], iLength);
      inc(pText, iLength);
    end;
    pText^ := NUL;
  finally
    GlobalUnlock(Result);
  end;
end;

function Copy(const Str: RawByteString; Pos, Len: Integer): RawByteString;
begin
  if (Pos > 0) and (Len > 0) then
    Result := System.Copy(Str, Pos, Len)
  else Result := E;
end;

function Copy(const Str: UnicodeString; Pos, Len: Integer): UnicodeString;
begin
  if (Pos > 0) and (Len > 0) then
    Result := System.Copy(Str, Pos, Len)
  else Result := E;
end;

function CopyLine(const Str: RawByteString; Num: Integer): RawByteString;
var
  pStr, pPos: PAnsiChar;
  iPos: Integer;
begin
  if (Str <> E) and (Num > 0) then begin
    pStr := Pointer(Str);
    pPos := pStr;
    iPos := 1;
    while Num > 0 do begin
      while not SysUtils.CharInSet(pStr^, [NUL, CR]) do inc(pStr);
      dec(Num);
      if Num = 0 then begin
	Result := Copy(Str, iPos, pStr - pPos);
	Exit;
      end
      else if pStr^ = NUL then begin
	Result := Copy(Str, iPos, pStr - pPos);
	Exit;
      end
      else begin
	while SysUtils.CharInSet(pStr^, [LF, CR]) do inc(pStr);
	if (pStr^ = NUL) and (Num > 0) then begin
	  while (pStr > pPos) and (pStr^ < SP) do dec(pStr);
	  Result := Copy(Str, iPos, succ(pStr - pPos));
	  Exit;
	end
	else begin
	  iPos := succ(pStr - Pointer(Str));
	  pPos := pStr;
	end;
      end;
    end;
  end;
end;

function CopyLine(const Str: UnicodeString; Num: Integer): UnicodeString;
var
  pStr, pPos: PWideChar;
  iPos: Integer;
begin
  if (Str <> E) and (Num > 0) then begin
    pStr := Pointer(Str);
    pPos := pStr;
    iPos := 1;
    while Num > 0 do begin
      while not CharInSet(pStr^, [NUL, CR]) do inc(pStr);
      dec(Num);
      if Num = 0 then begin
	Result := Copy(Str, iPos, pStr - pPos);
	Exit;
      end
      else if pStr^ = NUL then begin
	Result := Copy(Str, iPos, pStr - pPos);
	Exit;
      end
      else begin
	while CharInSet(pStr^, [LF, CR]) do inc(pStr);
	if (pStr^ = NUL) and (Num > 0) then begin
	  while (pStr > pPos) and (pStr^ < SP) do dec(pStr);
	  Result := Copy(Str, iPos, succ(pStr - pPos));
	  Exit;
	end
	else begin
	  iPos := succ(pStr - Pointer(Str));
	  pPos := pStr;
	end;
      end;
    end;
  end;
end;

function GetFirstLine(const Str: RawByteString): RawByteString;
var
  pStr: PAnsiChar;
begin
  if Str <> E then begin
    pStr := Pointer(Str);
    while not SysUtils.CharInSet(pStr^, [NUL, LF, CR]) do inc(pStr);
    while SysUtils.CharInSet(pStr^, [LF, CR]) do inc(pStr);
    Result := Copy(Str, 1, pStr - Pointer(Str));
    Exit;
  end;
  Result := E;
end;

function GetFirstLine(const Str: UnicodeString): UnicodeString;
var
  pStr: PWideChar;
begin
  if Str <> E then begin
    pStr := Pointer(Str);
    while not CharInSet(pStr^, [NUL, LF, CR]) do inc(pStr);
    while CharInSet(pStr^, [LF, CR]) do inc(pStr);
    Result := Copy(Str, 1, pStr - Pointer(Str));
    Exit;
  end;
  Result := E;
end;

function GetLastLine(const Str: RawByteString): RawByteString;
var
  pStr, pEnd: PAnsiChar;
begin
  Result := E;
  if Str <> E then begin
    pStr := Pointer(Str);
    pEnd := StrTools.StrEnd(pStr);
    while (pEnd >= pStr) and SysUtils.CharInSet(pEnd^, [NUL, LF, CR]) do dec(pEnd);
    if pEnd < pStr then
      Exit
    else if pEnd = pStr then begin
      Result := Copy(Str, 1, MAXINT);
      Exit;
    end
    else while (pEnd > pStr) and not SysUtils.CharInSet(pEnd^, [LF, CR]) do dec(pEnd);
    if (pEnd > pStr) or not SysUtils.CharInSet(pEnd^, [LF, CR]) then begin
      if SysUtils.CharInSet(pEnd^, [LF, CR]) then
	inc(pEnd, 2)
      else inc(pEnd);
      Result := Copy(Str, pEnd - pStr, MAXINT);
      Exit;
    end;
  end;
end;

function GetLastLine(const Str: UnicodeString): UnicodeString;
var
  pStr, pEnd: PWideChar;
begin
  Result := E;
  if Str <> E then begin
    pStr := Pointer(Str);
    pEnd := StrTools.StrEnd(pStr);
    while (pEnd >= pStr) and CharInSet(pEnd^, [NUL, LF, CR]) do dec(pEnd);
    if pEnd < pStr then
      Exit
    else if pEnd = pStr then begin
      Result := Copy(Str, 1, MAXINT);
      Exit;
    end
    else while (pEnd > pStr) and not CharInSet(pEnd^, [LF, CR]) do dec(pEnd);
    if (pEnd > pStr) or not CharInSet(pEnd^, [LF, CR]) then begin
      if CharInSet(pEnd^, [LF, CR]) then
	inc(pEnd, 2)
      else inc(pEnd);
      Result := Copy(Str, pEnd - pStr, MAXINT);
      Exit;
    end;
  end;
end;

function LowerCase(const Str: RawByteString): RawByteString; inline;
begin
  Result := AnsiStrings.LowerCase(Str);
end;

function LowerCase(const Str: UnicodeString): UnicodeString; inline;
begin
  Result := SysUtils.LowerCase(Str);
end;

function UpperCase(const Str: RawByteString): RawByteString; inline;
begin
  Result := AnsiStrings.UpperCase(Str);
end;

function UpperCase(const Str: UnicodeString): UnicodeString; inline;
begin
  Result := SysUtils.UpperCase(Str);
end;

procedure StrLower(var Str: RawByteString); inline;
begin
  if Str <> E then CharLowerA(PAnsiChar(Str));
end;

procedure StrLower(var Str: UnicodeString); inline;
begin
  if Str <> E then CharLowerW(PWideChar(Str));
end;

procedure StrUpper(var Str: RawByteString); inline;
begin
  if Str <> E then CharUpperA(PAnsiChar(Str));
end;

procedure StrUpper(var Str: UnicodeString); inline;
begin
  if Str <> E then CharUpperW(PWideChar(Str));
end;

function ToLower(Chr: AnsiChar): AnsiChar;
var
  aChar: AnsiChar;
begin
  aChar := Chr;
  CharLowerBuffA(@aChar, 1);
  Result := aChar;
end;

function ToLower(Chr: WideChar): WideChar;
var
  wChar: WideChar;
begin
  wChar := Chr;
  CharLowerBuffW(@wChar, 1);
  Result := wChar;
end;

function ToUpper(Chr: AnsiChar): AnsiChar;
var
  aChar: AnsiChar;
begin
  aChar := Chr;
  CharUpperBuffA(@aChar, 1);
  Result := aChar;
end;

function ToUpper(Chr: WideChar): WideChar;
var
  wChar: WideChar;
begin
  wChar := Chr;
  CharUpperBuffW(@wChar, 1);
  Result := wChar;
end;

function ToLower(var Str: RawByteString; Pos: Integer = 1; Len: Integer = 1): Integer;
begin
  if Pos <= Length(Str) then begin
    if pred(Pos + Len) > Length(Str) then Len := succ(Length(Str) - Pos);
    Result := CharLowerBuffA(@Str[Pos], Len);
    Exit;
  end;
  Result := 0;
end;

function ToLower(var Str: UnicodeString; Pos: Integer = 1; Len: Integer = 1): Integer;
begin
  if Pos <= Length(Str) then begin
    if pred(Pos + Len) > Length(Str) then Len := succ(Length(Str) - Pos);
    Result := CharLowerBuffW(@Str[Pos], Len);
    Exit;
  end;
  Result := 0;
end;

function ToUpper(var Str: RawByteString; Pos: Integer = 1; Len: Integer = 1): Integer;
begin
  if Pos <= Length(Str) then begin
    if pred(Pos + Len) > Length(Str) then Len := succ(Length(Str) - Pos);
    Result := CharUpperBuffA(@Str[Pos], Len);
    Exit;
  end;
  Result := 0;
end;

function ToUpper(var Str: UnicodeString; Pos: Integer = 1; Len: Integer = 1): Integer;
begin
  if Pos <= Length(Str) then begin
    if pred(Pos + Len) > Length(Str) then Len := succ(Length(Str) - Pos);
    Result := CharUpperBuffW(@Str[Pos], Len);
    Exit;
  end;
  Result := 0;
end;

function StrScan(const Str: PAnsiChar; Char: AnsiChar): PAnsiChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrScan(Str, Char);
  {$ELSE}
  Result := SysUtils.StrScan(Str, Char);
  {$IFEND}
end;

function StrScan(const Str: PWideChar; Char: WideChar): PWideChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrRScan(Str, Char);
  {$ELSE}
  Result := SysUtils.StrRScan(Str, Char);
  {$IFEND}
end;

function StrRScan(const Str: PAnsiChar; Char: AnsiChar): PAnsiChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrScan(Str, Char);
  {$ELSE}
  Result := SysUtils.StrScan(Str, Char);
  {$IFEND}
end;

function StrRScan(const Str: PWideChar; Char: WideChar): PWideChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrRScan(Str, Char);
  {$ELSE}
  Result := SysUtils.StrRScan(Str, Char);
  {$IFEND}
end;

function ScanBack(const Str: PAnsiChar; Char: AnsiChar): PAnsiChar;  inline; deprecated 'No bounds check, ensure Char exists!';
begin
  Result := Str;
  if Str <> NIL then while Result^ <> Char do dec(Result);
end;

function ScanBack(const Str: PWideChar; Char: WideChar): PWideChar;  inline; deprecated 'No bounds check, ensure Char exists!';
begin
  Result := Str;
  if Str <> NIL then while Result^ <> Char do dec(Result);
end;

function ScanFore(const Str: PAnsiChar; Char: AnsiChar): PAnsiChar; inline;
begin
  Result := Str;
  if Str <> NIL then while (Result^ <> NUL) and (Result^ = Char) do inc(Result);
end;

function ScanFore(const Str: PWideChar; Char: WideChar): PWideChar; inline;
begin
  Result := Str;
  if Str <> NIL then while (Result^ <> NUL) and (Result^ = Char) do inc(Result);
end;

function StrLen(const Str: PAnsiChar): Cardinal; inline;
begin
  Result := lstrlenA(Str);
end;

function StrLen(const Str: PWideChar): Cardinal; inline;
begin
  Result := lstrlenW(Str);
end;

function StrEnd(const Str: PAnsiChar): PAnsiChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrEnd(Str);
  {$ELSE}
  Result := SysUtils.StrEnd(Str);
  {$IFEND}
end;

function StrEnd(const Str: PWideChar): PWideChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrEnd(Str);
  {$ELSE}
  Result := SysUtils.StrEnd(Str);
  {$IFEND}
end;

function StrEnd(const Str: RawByteString): PAnsiChar; inline;
begin
  if Str <> E then begin
    Result := Pointer(Str);
    inc(Result, Length(Str));
    Exit;
  end;
  Result := NIL;
end;

function StrEnd(const Str: UnicodeString): PWideChar; inline;
begin
  if Str <> E then begin
    Result := Pointer(Str);
    inc(Result, Length(Str));
    Exit;
  end;
  Result := NIL;
end;

function StrPos(Str, SubStr: PAnsiChar): PAnsiChar; inline;
begin
  if (Str = NIL) or (SubStr = NIL) then begin
    Result := NIL;
    Exit;
  end;
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrPos(Str, SubStr);
  {$ELSE}
  Result := SysUtils.StrPos(Str, SubStr);
  {$IFEND}
end;

function StrPos(Str, SubStr: PWideChar): PWideChar; inline;
begin
  if (Str = NIL) or (SubStr = NIL) then begin
    Result := NIL;
    Exit;
  end;
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrPos(Str, SubStr);
  {$ELSE}
  Result := SysUtils.StrPos(Str, SubStr);
  {$IFEND}
end;

function StrPos(const Str, SubStr: RawByteString): Integer;
var
  pResult: PAnsiChar;
begin
  pResult := StrTools.StrPos(PAnsiChar(Str), PAnsiChar(SubStr));
  if pResult <> NIL then begin
    Result := succ(pResult - Pointer(Str));
    Exit;
  end;
  Result := 0;
end;

function StrPos(const Str, SubStr: UnicodeString): Integer;
var
  pResult: PWideChar;
begin
  pResult := StrTools.StrPos(PWideChar(Str), PWideChar(SubStr));
  if pResult <> NIL then begin
    Result := succ(pResult - Pointer(Str));
    Exit;
  end;
  Result := 0;
end;

function TextPos(Str, SubStr: PAnsiChar): PAnsiChar; inline;
begin
  if (Str = NIL) or (SubStr = NIL) then begin
    Result := NIL;
    Exit;
  end;
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.TextPos(Str, SubStr);
  {$ELSE}
  Result := SysUtils.TextPos(Str, SubStr);
  {$IFEND}
end;

function TextPos(Str, SubStr: PWideChar): PWideChar; inline;
begin
  if (Str = NIL) or (SubStr = NIL) then begin
    Result := NIL;
    Exit;
  end;
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.TextPos(Str, SubStr);
  {$ELSE}
  Result := SysUtils.TextPos(Str, SubStr);
  {$IFEND}
end;

function TextPos(const Str, SubStr: RawByteString): Integer;
var
  pResult: PAnsiChar;
begin
  pResult := StrTools.TextPos(PAnsiChar(Str), PAnsiChar(SubStr));
  if pResult <> NIL then begin
    Result := succ(pResult - Pointer(Str));
    Exit;
  end;
  Result := 0;
end;

function TextPos(const Str, SubStr: UnicodeString): Integer;
var
  pResult: PWideChar;
begin
  pResult := StrTools.TextPos(PWideChar(Str), PWideChar(SubStr));
  if pResult <> NIL then begin
    Result := succ(pResult - Pointer(Str));
    Exit;
  end;
  Result := 0;
end;

function StrCharLength(const Str: PAnsiChar): Integer; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrCharLength(Str);
  {$ELSE}
  Result := SysUtils.StrCharLength(Str);
  {$IFEND}
end;

function StrCharLength(const Str: PWideChar): Integer; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrCharLength(Str);
  {$ELSE}
  Result := SysUtils.StrCharLength(Str);
  {$IFEND}
end;

function StrComp(S1, S2: PAnsiChar): Integer; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrComp(S1, S2);
  {$ELSE}
  Result := SysUtils.StrComp(S1, S2);
  {$IFEND}
end;

function StrComp(S1, S2: PWideChar): Integer; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrComp(S1, S2);
  {$ELSE}
  Result := SysUtils.StrComp(S1, S2);
  {$IFEND}
end;

function StrIComp(S1, S2: PAnsiChar): Integer; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrIComp(S1, S2);
  {$ELSE}
  Result := SysUtils.StrIComp(S1, S2);
  {$IFEND}
end;

function StrIComp(S1, S2: PWideChar): Integer; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrIComp(S1, S2);
  {$ELSE}
  Result := SysUtils.StrIComp(S1, S2);
  {$IFEND}
end;

function StrLComp(S1, S2: PAnsiChar; MaxLen: Cardinal): Integer; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrLComp(S1, S2, MaxLen);
  {$ELSE}
  Result := SysUtils.StrLComp(S1, S2, MaxLen);
  {$IFEND}
end;

function StrLComp(S1, S2: PWideChar; MaxLen: Cardinal): Integer; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrLComp(S1, S2, MaxLen);
  {$ELSE}
  Result := SysUtils.StrLComp(S1, S2, MaxLen);
  {$IFEND}
end;

function StrAtPos(S1, S2: PAnsiChar; MaxLen: Cardinal): Longbool; inline;
begin
  if not ((S1 = NIL) xor (S2 = NIL)) then
    Result := (StrTools.StrLComp(S1, S2, MaxLen) = 0)
  else Result := FALSE;
end;

function StrAtPos(S1, S2: PWideChar; MaxLen: Cardinal): Longbool; inline;
begin
  if not ((S1 = NIL) xor (S2 = NIL)) then
    Result := (StrTools.StrLIComp(S1, S2, MaxLen) = 0)
  else Result := FALSE;
end;


function StrLIComp(S1, S2: PAnsiChar; MaxLen: Cardinal): Integer; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrLIComp(S1, S2, MaxLen);
  {$ELSE}
  Result := SysUtils.StrLIComp(S1, S2, MaxLen);
  {$IFEND}
end;

function StrLIComp(S1, S2: PWideChar; MaxLen: Cardinal): Integer; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrLIComp(S1, S2, MaxLen);
  {$ELSE}
  Result := SysUtils.StrLIComp(S1, S2, MaxLen);
  {$IFEND}
end;

function StrIAtPos(S1, S2: PAnsiChar; MaxLen: Cardinal): Longbool; inline;
begin
  if not ((S1 = NIL) xor (S2 = NIL)) then
    Result := (StrTools.StrLIComp(S1, S2, MaxLen) = 0)
  else Result := FALSE;
end;

function StrIAtPos(S1, S2: PWideChar; MaxLen: Cardinal): Longbool; inline;
begin
  if not ((S1 = NIL) xor (S2 = NIL)) then
    Result := (StrTools.StrLIComp(S1, S2, MaxLen) = 0)
  else Result := FALSE;
end;

function StrCat(Dest: PAnsiChar; const Source: PAnsiChar): PAnsiChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrCat(Dest, Source);
  {$ELSE}
  Result := SysUtils.StrCat(Dest, Source);
  {$IFEND}
end;

function StrCat(Dest: PWideChar; const Source: PWideChar): PWideChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrCat(Dest, Source);
  {$ELSE}
  Result := SysUtils.StrCat(Dest, Source);
  {$IFEND}
end;

function StrLCat(Dest: PAnsiChar; const Source: PAnsiChar; MaxLen: Cardinal): PAnsiChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrLCat(Dest, Source, MaxLen);
  {$ELSE}
  Result := SysUtils.StrLCat(Dest, Source, MaxLen);
  {$IFEND}
end;

function StrLCat(Dest: PWideChar; const Source: PWideChar; MaxLen: Cardinal): PWideChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrLCat(Dest, Source, MaxLen);
  {$ELSE}
  Result := SysUtils.StrLCat(Dest, Source, MaxLen);
  {$IFEND}
end;

function StrCopy(Dest: PAnsiChar; const Source: PAnsiChar): PAnsiChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrCopy(Dest, Source);
  {$ELSE}
  Result := SysUtils.StrCopy(Dest, Source);
  {$IFEND}
end;

function StrCopy(Dest: PWideChar; const Source: PWideChar): PWideChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrCopy(Dest, Source);
  {$ELSE}
  Result := SysUtils.StrCopy(Dest, Source);
  {$IFEND}
end;

function StrECopy(Dest: PAnsiChar; const Source: PAnsiChar): PAnsiChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrECopy(Dest, Source);
  {$ELSE}
  Result := SysUtils.StrECopy(Dest, Source);
  {$IFEND}
end;

function StrECopy(Dest: PWideChar; const Source: PWideChar): PWideChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrECopy(Dest, Source);
  {$ELSE}
  Result := SysUtils.StrECopy(Dest, Source);
  {$IFEND}
end;

function StrLCopy(Dest: PAnsiChar; const Source: PAnsiChar; MaxLen: Cardinal): PAnsiChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrLCopy(Dest, Source, MaxLen);
  {$ELSE}
  Result := SysUtils.StrLCopy(Dest, Source, MaxLen);
  {$IFEND}
end;

function StrLCopy(Dest: PWideChar; const Source: PWideChar; MaxLen: Cardinal): PWideChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrLCopy(Dest, Source, MaxLen);
  {$ELSE}
  Result := SysUtils.StrLCopy(Dest, Source, MaxLen);
  {$IFEND}
end;

function StrLECopy(Dest: PAnsiChar; const Source: PAnsiChar; Length: Cardinal): PAnsiChar; inline;
begin
  Result := StrTools.StrMove(Dest, Source, Length) + Length;
end;

function StrLECopy(Dest: PWideChar; const Source: PWideChar; Length: Cardinal): PWideChar; inline;
begin
  Result := StrTools.StrMove(Dest, Source, Length) + Length;
end;

function StrPLCopy(Dest: PAnsiChar; const Source: RawByteString; MaxLen: Cardinal): PAnsiChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrPLCopy(Dest, Source, MaxLen);
  {$ELSE}
  Result := SysUtils.StrPLCopy(Dest, Source, MaxLen);
  {$IFEND}
end;

function StrPLCopy(Dest: PWideChar; const Source: UnicodeString; MaxLen: Cardinal): PWideChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrPLCopy(Dest, Source, MaxLen);
  {$ELSE}
  Result := SysUtils.StrPLCopy(Dest, Source, MaxLen);
  {$IFEND}
end;

function StrLECatW(PResult: PWideChar; const Strings: Array of TStringRecW): PWideChar;
var
  iIndex: Integer;
begin
  Result := PResult;
  for iIndex := Low(Strings) to High(Strings) do begin
    with Strings[iIndex] do Result := StrTools.StrMove(Result, PWChar, PWLen) + PWLen;
  end;
end;

function StrLECatA(PResult: PAnsiChar; const Strings: Array of TStringRecA): PAnsiChar;
var
  iIndex: Integer;
begin
  Result := PResult;
  for iIndex := Low(Strings) to High(Strings) do begin
    with Strings[iIndex] do Result := StrTools.StrMove(Result, PAChar, PALen) + PALen;
  end;
end;

function StrMove(Dest: PAnsiChar; const Source: PAnsiChar; Count: Cardinal): PAnsiChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrMove(Dest, Source, Count);
  {$ELSE}
  Result := SysUtils.StrMove(Dest, Source, Count);
  {$IFEND}
end;

function StrMove(Dest: PWideChar; const Source: PWideChar; Count: Cardinal): PWideChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrMove(Dest, Source, Count);
  {$ELSE}
  Result := SysUtils.StrMove(Dest, Source, Count);
  {$IFEND}
end;

function StrLPasCopy(Source: PAnsiChar; Length: Integer): RawByteString;
begin
  if (Source <> NIL) and (Length > 0) then begin
    SetString(Result, Source, Length);
    Exit;
  end;
  Result := E;
end;

function StrLPasCopy(Source: PWideChar; Length: Integer): UnicodeString;
begin
  if (Source <> NIL) and (Length > 0) then begin
    SetString(Result, Source, Length);
    Exit;
  end;
  Result := E;
end;

function AnsiStrAlloc(Size: Cardinal): PAnsiChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.AnsiStrAlloc(Size);
  {$ELSE}
  Result := SysUtils.AnsiStrAlloc(Size);
  {$IFEND}
end;

function StrAlloc(Size: Cardinal): PWideChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrAlloc(Size);
  {$ELSE}
  Result := SysUtils.StrAlloc(Size);
  {$IFEND}
end;

function StrBufSize(const Str: PAnsiChar): Cardinal; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrBufSize(Str);
  {$ELSE}
  Result := SysUtils.StrBufSize(Str);
  {$IFEND}
end;

function StrBufSize(const Str: PWideChar): Cardinal; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrBufSize(Str);
  {$ELSE}
  Result := SysUtils.StrBufSize(Str);
  {$IFEND}
end;

function StrNew(const Str: PAnsiChar): PAnsiChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.StrNew(Str);
  {$ELSE}
  Result := SysUtils.StrNew(Str);
  {$IFEND}
end;

function StrNew(const Str: PWideChar): PWideChar; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.StrNew(Str);
  {$ELSE}
  Result := SysUtils.StrNew(Str);
  {$IFEND}
end;

procedure StrDispose(Str: PAnsiChar); inline;
begin
  {$IF CompilerVersion > 21.0}
  AnsiStrings.StrDispose(Str);
  {$ELSE}
  SysUtils.StrDispose(Str);
  {$IFEND}
end;

procedure StrDispose(Str: PWideChar); inline;
begin
  {$IF CompilerVersion > 21.0}
  SysUtils.StrDispose(Str);
  {$ELSE}
  SysUtils.StrDispose(Str);
  {$IFEND}
end;

function AnsiStrEnd(Str: PAnsiChar): PAnsiChar; inline;
begin
  Result := Str;
  inc(Result, StrTools.StrLen(Str));
end;

function AnsiStrEnd(Str: PWideChar): PWideChar; inline;
begin
  Result := Str;
  inc(Result, StrTools.StrLen(Str));
end;

function AnsiStrPos(Str, SubStr: PAnsiChar): PAnsiChar; inline;
begin
  Result := StrStrA(Str, SubStr);
end;

function AnsiStrPos(Str, SubStr: PWideChar): PWideChar; inline;
begin
  Result := StrStrW(Str, SubStr);
end;

function AnsiStrRPos(Str, SubStr: PAnsiChar): PAnsiChar;
var
  iLen: Integer;
begin
  Result := NIL;
  iLen := StrTools.StrLen(SubStr);
  repeat
    Result := StrRStrIA(Str, Result, SubStr);
  until (Result = NIL) or StrTools.AnsiStrAtPos(Result, SubStr, iLen);
end;

function AnsiStrRPos(Str, SubStr: PWideChar): PWideChar;
var
  iLen: Integer;
begin
  Result := NIL;
  iLen := StrTools.StrLen(SubStr);
  repeat
    Result := StrRStrIW(Str, Result, SubStr);
  until (Result = NIL) or StrTools.AnsiStrAtPos(Result, SubStr, iLen);
end;

function AnsiTextPos(Str, SubStr: PAnsiChar): PAnsiChar; inline;
begin
  Result := StrStrIA(Str, SubStr);
end;

function AnsiTextPos(Str, SubStr: PWideChar): PWideChar; inline;
begin
  Result := StrStrIW(Str, SubStr);
end;

function AnsiTextRPos(Str, SubStr: PAnsiChar): PAnsiChar; inline;
begin
  Result := StrRStrIA(Str, NIL, SubStr);
end;

function AnsiTextRPos(Str, SubStr: PWideChar): PWideChar; inline;
begin
  Result := StrRStrIW(Str, NIL, SubStr);
end;

function AnsiStrComp(S1, S2: PAnsiChar): Integer; inline;
begin
  Result := StrCmpA(S1, S2);
end;

function AnsiStrComp(S1, S2: PWideChar): Integer; inline;
begin
  Result := StrCmpW(S1, S2);
end;

function AnsiStrIComp(S1, S2: PAnsiChar): Integer; inline;
begin
  Result := StrCmpIA(S1, S2);
end;

function AnsiStrIComp(S1, S2: PWideChar): Integer; inline;
begin
  Result := StrCmpIW(S1, S2);
end;

function AnsiStrLComp(S1, S2: PAnsiChar; MaxLen: Cardinal): Integer; inline;
begin
  Result := StrCmpNA(S1, S2, MaxLen);
end;

function AnsiStrLComp(S1, S2: PWideChar; MaxLen: Cardinal): Integer; inline;
begin
  Result := StrCmpNW(S1, S2, MaxLen);
end;

function AnsiStrAtPos(S1, S2: PAnsiChar; MaxLen: Cardinal): Longbool; inline;
begin
  Result := (StrTools.AnsiStrLComp(S1, S2, MaxLen) = 0);
end;

function AnsiStrAtPos(S1, S2: PWideChar; MaxLen: Cardinal): Longbool; inline;
begin
  Result := (StrTools.AnsiStrLComp(S1, S2, MaxLen) = 0);
end;

function AnsiStrLIComp(S1, S2: PAnsiChar; MaxLen: Cardinal): Integer; inline;
begin
  Result := StrCmpNIA(S1, S2, MaxLen);
end;

function AnsiStrLIComp(S1, S2: PWideChar; MaxLen: Cardinal): Integer; inline;
begin
  Result := StrCmpNIW(S1, S2, MaxLen);
end;

function AnsiStrIAtPos(S1, S2: PAnsiChar; MaxLen: Cardinal): Longbool; inline;
begin
  Result := (StrTools.AnsiStrLIComp(S1, S2, MaxLen) = 0);
end;

function AnsiStrIAtPos(S1, S2: PWideChar; MaxLen: Cardinal): Longbool; inline;
begin
  Result := (StrTools.AnsiStrLIComp(S1, S2, MaxLen) = 0);
end;

function AnsiExtractQuotedStr(Src: PAnsiChar; Quote: AnsiChar = QU): AnsiString; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := AnsiStrings.AnsiExtractQuotedStr(Src, Quote);
  {$ELSE}
  Result := SysUtils.AnsiExtractQuotedStr(Src, Quote);
  {$IFEND}
end;

function AnsiExtractQuotedStr(Src: PWideChar; Quote: WideChar = QU): UnicodeString; inline;
begin
  {$IF CompilerVersion > 21.0}
  Result := SysUtils.AnsiExtractQuotedStr(Src, Quote);
  {$ELSE}
  Result := SysUtils.AnsiExtractQuotedStr(Src, Quote);
  {$IFEND}
end;

function IsStringUpperCase(const Str: RawByteString): Longbool; inline;
begin
  Result := StrTools.AnsiSameStr(Str, StrTools.AnsiUpperCase(Str));
end;

function IsStringUpperCase(const Str: UnicodeString): Longbool; inline;
begin
  Result := StrTools.AnsiSameStr(Str, StrTools.AnsiUpperCase(Str));
end;

function IsUTF8String(const Str: RawByteString): Longbool;
begin
  Result := WideStrUtils.IsUTF8String(Str);
end;

function SearchBuf(Buf: PAnsiChar; BufLen: Integer; SelStart, SelLength: Integer;
		   SearchString: RawByteString; Options: TStringSearchOptions;
		   const WordDelimiters: TSysCharSet = WORDDELIMITERS): PAnsiChar;
var
  SearchCount, I: Integer;
  C: AnsiChar;
  Direction: Shortint;
  ShadowMap: Array[0..256] of AnsiChar;
  CharMap: Array [AnsiChar] of AnsiChar absolute ShadowMap;

  function FindNextWordStart(var BufPtr: PAnsiChar): Boolean;
  begin
    if Direction = 1 then begin
      while (SearchCount > 0) and (BufPtr^ in WordDelimiters) do begin
	inc(BufPtr, Direction);
	dec(SearchCount);
      end;
    end;
    if Direction = -1 then begin
      while (SearchCount > 0) and (BufPtr^ in WordDelimiters) do begin
	inc(BufPtr, Direction);
	dec(SearchCount);
      end;
    end;
    Result := SearchCount > 0;
    if Direction = -1 then begin
      dec(BufPtr, Direction);
      inc(SearchCount);
    end;
  end;

begin
  Result := NIL;
  if BufLen <= 0 then Exit;
  if soDown in Options then begin
    Direction := 1;
    inc(SelStart, SelLength);
    SearchCount := BufLen - SelStart - Length(SearchString) + 1;
    if SearchCount < 0 then Exit;
    if SelStart + SearchCount > BufLen then Exit;
  end
  else begin
    Direction := -1;
    dec(SelStart, Length(SearchString));
    SearchCount := SelStart + 1;
  end;
  if (SelStart < 0) or (SelStart > BufLen) then Exit;
  Result := @Buf[SelStart];
  for C := Low(CharMap) to High(CharMap) do CharMap[(C)] := (C);
  ShadowMap[256] := #0;
  if not (soMatchCase in Options) then begin
    CharUpperBuffA(PAnsiChar(@CharMap), Length(CharMap));
    CharUpperBuffA(@SearchString[1], Length(SearchString));
  end;
  while SearchCount > 0 do begin
    if (soWholeWord in Options) and (Result <> @Buf[SelStart]) then if not FindNextWordStart(Result) then Break;
    I := 0;
    while (CharMap[(Result[I])] = (SearchString[I+1])) do begin
      inc(I);
      if I >= Length(SearchString) then begin
	if (not (soWholeWord in Options)) or (SearchCount = 0) or (Result[I] in WordDelimiters) then Exit;
	Break;
      end;
    end;
    inc(Result, Direction);
    dec(SearchCount);
  end;
  Result := NIL;
end;

function SearchBuf(Buf: PWideChar; BufLen: Integer; SelStart, SelLength: Integer;
		   SearchString: UnicodeString; Options: TStringSearchOptions): PWideChar;
var
  SearchCount, I: Integer;
  Direction: Shortint;

  function FindNextWordStart(var BufPtr: PWideChar): Boolean;
  begin
    if Direction = 1 then begin
      while (SearchCount > 0) and
      {$IF CompilerVersion > 21.0}
      (not BufPtr^.IsLetterOrDigit) do begin
      {$ELSE}
      (not TCharacter.IsLetterOrDigit(BufPtr^)) do begin
      {$IFEND}
	inc(BufPtr, Direction);
	dec(SearchCount);
      end;
    end;
    if Direction = -1 then begin
      while (SearchCount > 0) and
      {$IF CompilerVersion > 21.0}
      (not BufPtr^.IsLetterOrDigit) do begin
      {$ELSE}
      (not TCharacter.IsLetterOrDigit(BufPtr^)) do begin
      {$IFEND}
	inc(BufPtr, Direction);
	dec(SearchCount);
      end;
    end;
    Result := (SearchCount > 0);
    if Direction = -1 then begin
      dec(BufPtr, Direction);
      inc(SearchCount);
    end;
  end;

  function NextChar(S: PWideChar): WideChar;
  begin
    if not (soMatchCase in Options) then
      Result := WideChar(CharUpperW(PWideChar(S[I])))
    else Result := S[I];
  end;

begin
  Result := NIL;
  if BufLen <= 0 then Exit;
  if soDown in Options then begin
    Direction := 1;
    inc(SelStart, SelLength);
    SearchCount := BufLen - SelStart - Length(SearchString) + 1;
    if SearchCount < 0 then Exit;
    if SelStart + SearchCount > BufLen then Exit;
  end
  else begin
    Direction := -1;
    dec(SelStart, Length(SearchString));
    SearchCount := SelStart + 1;
  end;
  if (SelStart < 0) or (SelStart > BufLen) then Exit;
  Result := @Buf[SelStart];
  if not (soMatchCase in Options) then CharUpperBuffW(@SearchString[1], Length(SearchString));
  while SearchCount > 0 do begin
    if (soWholeWord in Options) and (Result <> @Buf[SelStart]) then if not FindNextWordStart(Result) then Break;
    I := 0;
    while NextChar(Result) = SearchString[I+1] do begin
      inc(I);
      if I >= Length(SearchString) then begin
	if (not (soWholeWord in Options)) or (SearchCount = 0) or
	{$IF CompilerVersion > 21.0}
	(not WideChar(Result[I]).IsLetterOrDigit) then begin
	{$ELSE}
	(not TCharacter.IsLetterOrDigit(Result[I])) then begin
	{$IFEND}
	  Exit;
	end;
	Break;
      end;
    end;
    inc(Result, Direction);
    dec(SearchCount);
  end;
  Result := NIL;
end;

function StringPos(const SearchStr, Str: RawByteString; BegPos: Integer; Options: TStringSearchOptions = [soDown]): Integer;
var
  iLen: Integer;
  ui: UInt64;
begin
  Result := 0;
  if SearchStr = E then Exit;
  iLen := Length(Str);
  if (iLen = 0) or (BegPos = 0) or (BegPos > iLen) then Exit;
  ui := UInt64(StrTools.SearchBuf(PAnsiChar(Str), iLen, 0, pred(BegPos), SearchStr, Options + [soDown, soMatchCase]));
  if ui <> 0 then Result := succ((ui - UInt64(Pointer(Str))) div SizeOf(SearchStr[1]));
end;

function StringPos(const SearchStr, Str: UnicodeString; BegPos: Integer; Options: TStringSearchOptions = [soDown]): Integer;
var
  iLen: Integer;
  ui: UInt64;
begin
  Result := 0;
  if SearchStr = E then Exit;
  iLen := Length(Str);
  if (iLen = 0) or (BegPos = 0) or (BegPos > iLen) then Exit;
  ui := UInt64(StrTools.SearchBuf(PWideChar(Str), iLen, 0, pred(BegPos), SearchStr, Options + [soDown, soMatchCase]));
  if ui <> 0 then Result := succ((ui - UInt64(Pointer(Str))) div SizeOf(SearchStr[1]));
end;

function TextPos(const SearchText, Text: RawByteString; BegPos: Integer; Options: TStringSearchOptions = [soDown]): Integer;
var
  iLen: Integer;
  ui: UInt64;
begin
  Result := 0;
  if SearchText = E then Exit;
  iLen := Length(Text);
  if (iLen = 0) or (BegPos = 0) or (BegPos > iLen) then Exit;
  ui := UInt64(StrTools.SearchBuf(PAnsiChar(Text), iLen, 0, pred(BegPos), SearchText, Options + [soDown]));
  if ui <> 0 then Result := succ((ui - UInt64(Pointer(Text))) div SizeOf(SearchText[1]));
end;

function TextPos(const SearchText, Text: UnicodeString; BegPos: Integer; Options: TStringSearchOptions = [soDown]): Integer;
var
  iLen: Integer;
  ui: UInt64;
begin
  Result := 0;
  if SearchText = E then Exit;
  iLen := Length(Text);
  if (iLen = 0) or (BegPos = 0) or (BegPos > iLen) then Exit;
  ui := UInt64(StrTools.SearchBuf(PWideChar(Text), iLen, 0, pred(BegPos), SearchText, Options + [soDown]));
  if ui <> 0 then Result := succ((ui - UInt64(Pointer(Text))) div SizeOf(SearchText[1]));
end;

function CollapseWhiteSpace(const Str, MaxLineBreak: RawByteString): RawByteString;
var
  pResult, pMax, pBeg, pEnd: PAnsiChar;
  iLen: Integer;
begin
  if Str <> E then begin
    SetString(Result, NIL, Length(Str));
    pMax := PAnsiChar(MaxLineBreak);
    pResult := Pointer(Result);
    pBeg := Pointer(Str);
    pEnd := pBeg;
    repeat
      while pEnd^ > SP do inc(pEnd);
      if StrTools.StrLComp(pEnd, pMax, Length(MaxLineBreak)) = 0 then begin
	iLen := pEnd - pBeg;
	inc(pEnd, Length(MaxLineBreak));
	StrTools.StrLCopy(pResult, pBeg, iLen);
	inc(pResult, iLen);
	StrTools.StrLCopy(pResult, pMax, Length(MaxLineBreak));
	inc(pResult, Length(MaxLineBreak));
	pBeg := pEnd;
	while AnsiChar(pBeg^) in TRIMCHARS do inc(pBeg);
	if pBeg^ = NUL then Break;
	pEnd := pBeg;
      end
      else if pEnd^ = NUL then Break;
      inc(pEnd);
    until pEnd^ = NUL;
    if pEnd > pBeg then begin
      iLen := pEnd - pBeg;
      StrTools.StrLCopy(pResult, pBeg, iLen);
      inc(pResult, iLen);
    end;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end;
end;

function CollapseWhiteSpace(const Str, MaxLineBreak: UnicodeString): UnicodeString;
var
  pResult, pMax, pBeg, pEnd: PWideChar;
  iLen: Integer;
begin
  if Str <> E then begin
    SetString(Result, NIL, Length(Str));
    pMax := PWideChar(MaxLineBreak);
    pResult := Pointer(Result);
    pBeg := Pointer(Str);
    pEnd := pBeg;
    repeat
      while pEnd^ > SP do inc(pEnd);
      if StrTools.StrLComp(pEnd, pMax, Length(MaxLineBreak)) = 0 then begin
	iLen := pEnd - pBeg;
	inc(pEnd, Length(MaxLineBreak));
	StrTools.StrLCopy(pResult, pBeg, iLen);
	inc(pResult, iLen);
	StrTools.StrLCopy(pResult, pMax, Length(MaxLineBreak));
	inc(pResult, Length(MaxLineBreak));
	pBeg := pEnd;
	while AnsiChar(pBeg^) in TRIMCHARS do inc(pBeg);
	if pBeg^ = NUL then Break;
	pEnd := pBeg;
      end
      else if pEnd^ = NUL then Break;
      inc(pEnd);
    until pEnd^ = NUL;
    if pEnd > pBeg then begin
      iLen := pEnd - pBeg;
      StrTools.StrLCopy(pResult, pBeg, iLen);
      inc(pResult, iLen);
    end;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end;
end;

// returns position in Longbool!
function Contains(Str, SubStr: AnsiString; BegPos: Integer = 1; EndPos: Integer = MAXINT): Longbool;
var
  iStrLen, iSubStrLen, iIndex: Integer;
  pStr, pBeg, pEnd, pSubStr: PAnsiChar;
begin
  if (Str <> E) and (SubStr <> E) then begin
    iStrLen := Length(Str);
    iSubStrLen := Length(SubStr);
    if EndPos > iStrLen then EndPos := iStrLen;
    dec(EndPos, pred(iSubStrLen));
    if (BegPos > 0) and (EndPos > BegPos) then begin
      pStr := Pointer(Str);
      pBeg := pStr + pred(BegPos);
      pEnd := pStr + pred(EndPos);
      pSubStr := Pointer(SubStr);
      dec(iSubStrLen);
      while pBeg <= pEnd do begin
	for iIndex := 0 to iSubStrLen do begin
	  if not SameChar(pBeg[iIndex], pSubStr[iIndex]) then begin
	    inc(pBeg);
	    Break;
	  end
	  else if iIndex = iSubStrLen then begin
	    Result := Longbool(succ(pBeg - pStr));
	    Exit;
	  end;
	end;
      end;
    end;
  end;
  Result := FALSE;
end;

// returns position in Longbool!
function Contains(Str, SubStr: UnicodeString; BegPos: Integer = 1; EndPos: Integer = MAXINT): Longbool;
var
  iStrLen, iSubStrLen, iIndex: Integer;
  pStr, pBeg, pEnd, pSubStr: PAnsiChar;
begin
  if (Str <> E) and (SubStr <> E) then begin
    iStrLen := Length(Str);
    iSubStrLen := Length(SubStr);
    if EndPos > iStrLen then EndPos := iStrLen;
    dec(EndPos, pred(iSubStrLen));
    if (BegPos > 0) and (EndPos > BegPos) then begin
      pStr := Pointer(Str);
      pBeg := pStr + pred(BegPos);
      pEnd := pStr + pred(EndPos);
      pSubStr := Pointer(SubStr);
      dec(iSubStrLen);
      while pBeg <= pEnd do begin
	for iIndex := 0 to iSubStrLen do begin
	  if not SameChar(pBeg[iIndex], pSubStr[iIndex]) then begin
	    inc(pBeg);
	    Break;
	  end
	  else if iIndex = iSubStrLen then begin
	    Result := Longbool(succ(pBeg - pStr));
	    Exit;
	  end;
	end;
      end;
    end;
  end;
  Result := FALSE;
end;

function StrContains(PBegPos, PEndPos, PSubStr: PAnsiChar; SubStrLen: Cardinal): PAnsiChar;
var
  cIndex: Cardinal;
begin
  if (PBegPos <> NIL) and (PEndPos > PBegPos) and (PSubStr <> E) and (SubStrLen > 0) then begin
    dec(SubStrLen);
    dec(PEndPos, SubStrLen);
    if PEndPos >= PBegPos then begin
      while PBegPos <= PEndPos do begin
	for cIndex := 0 to SubStrLen do begin
	  if not SameChar((PBegPos + cIndex)^, (PSubStr + cIndex)^) then begin
	    inc(PBegPos);
	    Break;
	  end
	  else if cIndex = SubStrLen then begin
	    Result := PBegPos;
	    Exit;
	  end;
	end;
      end;
    end;
  end;
  Result := NIL;
end;

function StrContains(PBegPos, PEndPos, PSubStr: PWideChar; SubStrLen: Cardinal): PWideChar;
var
  cIndex: Cardinal;
begin
  if (PBegPos <> NIL) and (PEndPos > PBegPos) and (PSubStr <> E) and (SubStrLen > 0) then begin
    dec(SubStrLen);
    dec(PEndPos, SubStrLen);
    if PEndPos >= PBegPos then begin
      while PBegPos <= PEndPos do begin
	for cIndex := 0 to SubStrLen do begin
	  if not SameChar((PBegPos + cIndex)^, (PSubStr + cIndex)^) then begin
	    inc(PBegPos);
	    Break;
	  end
	  else if cIndex = SubStrLen then begin
	    Result := PBegPos;
	    Exit;
	  end;
	end;
      end;
    end;
  end;
  Result := NIL;
end;

function GetPadding(const Text: RawByteString; var Left, Right: RawByteString): TTrimOptions;
var
  pText, pPos, pEnd, pResult: PAnsiChar;
begin
  Result := [];
  if Text <> E then begin
    pText := Pointer(Text);
    pPos := pText;
    while (pPos^ > NUL) and (pPos^ <= SP) do inc(pPos);
    if pPos > pText then begin
      Include(Result, toLeft);
      SetString(Left, NIL, pPos - pText);
      pResult := Pointer(Left);
      StrTools.StrLCopy(pResult, pText, Length(Left));
    end;
    pEnd := pText + pred(Length(Text));
    pPos := pEnd;
    while (pPos >= pText) and (pPos^ <= SP) do dec(pPos);
    if pPos < pEnd then begin
      Include(Result, toRight);
      SetString(Right, NIL, pEnd - pPos);
      inc(pPos);
      pResult := Pointer(Right);
      StrTools.StrLCopy(pResult, pPos, Length(Right));
    end;
  end;
end;

function GetPadding(const Text: UnicodeString; var Left, Right: UnicodeString): TTrimOptions;
var
  pText, pPos, pEnd, pResult: PWideChar;
begin
  Result := [];
  if Text <> E then begin
    pText := Pointer(Text);
    pPos := pText;
    while (pPos^ > NUL) and (pPos^ <= SP) do inc(pPos);
    if pPos > pText then begin
      Include(Result, toLeft);
      SetString(Left, NIL, pPos - pText);
      pResult := Pointer(Left);
      StrTools.StrLCopy(pResult, pText, Length(Left));
    end;
    pEnd := pText + pred(Length(Text));
    pPos := pEnd;
    while (pPos >= pText) and (pPos^ <= SP) do dec(pPos);
    if pPos < pEnd then begin
      Include(Result, toRight);
      SetString(Right, NIL, pEnd - pPos);
      inc(pPos);
      pResult := Pointer(Right);
      StrTools.StrLCopy(pResult, pPos, Length(Right));
    end;
  end;
end;

function GetLineCount(const Text: RawByteString; TestTrimmed: Longbool = FALSE): Integer;
var
  pText: PAnsiChar;
begin
  Result := 0;
  if Text = E then Exit;
  pText := Pointer(Text);
  if TestTrimmed then while (pText^ > NUL) and (pText^ <= SP) do inc(pText);
  if pText^ = NUL then Exit;
  inc(Result);
  repeat
    while not SysUtils.CharInSet(pText^, [NUL, CR]) do inc(pText);
    if pText > NUL then begin
      if (pText + 1)^ = LF then begin
	inc(pText, Length(CRLF));
	inc(Result);
      end
      else inc(pText);
    end
    else Break;
  until pText^ = NUL;
  if TestTrimmed then begin
    while (pText >= Pointer(Text)) and (pText^ <= SP) do begin
      dec(pText);
      if (pText^ = LF) and ((pText - 1)^ = CR) then begin
	dec(pText, Length(CRLF));
	dec(Result);
      end;
    end;
  end;
  if Result = 0 then inc(Result);
end;

function GetLineCount(const Text: UnicodeString; TestTrimmed: Longbool = FALSE): Integer;
var
  pText: PWideChar;
begin
  Result := 0;
  if Text = E then Exit;
  pText := Pointer(Text);
  if TestTrimmed then while (pText^ > NUL) and (pText^ <= SP) do inc(pText);
  if pText^ = NUL then Exit;
  inc(Result);
  repeat
    while not CharInSet(pText^, [NUL, CR]) do inc(pText);
    if pText > NUL then begin
      if (pText + 1)^ = LF then begin
	inc(pText, Length(CRLF));
	inc(Result);
      end
      else inc(pText);
    end
    else Break;
  until pText^ = NUL;
  if TestTrimmed then begin
    while (pText >= Pointer(Text)) and (pText^ <= SP) do begin
      dec(pText);
      if (pText^ = LF) and ((pText - 1)^ = CR) then begin
	dec(pText, Length(CRLF));
	dec(Result);
      end;
    end;
  end;
  if Result = 0 then inc(Result);
end;

function SameChar(Chr1, Chr2: AnsiChar): Longbool;
begin
  Result := not ChrCmpIA(Byte(Chr1), Byte(Chr2));
end;

function SameChar(Chr1, Chr2: WideChar): Longbool;
begin
  Result := not ChrCmpIW(Word(Chr1), Word(Chr2));
end;

function CharDiff(OpenChar, CloseChar: AnsiChar; BegChars, EndChars: PAnsiChar): Integer;
begin
  Result := 0;
  if (BegChars <> NIL) and (EndChars > BegChars) then begin
    while BegChars <= EndChars do begin
      if BegChars^ = OpenChar then
	inc(Result)
      else if BegChars^ = CloseChar then
	dec(Result)
      else if BegChars^ = NUL then Break;
      inc(BegChars);
    end;
  end;
end;

function CharDiff(OpenChar, CloseChar: WideChar; BegChars, EndChars: PWideChar): Integer;
begin
  Result := 0;
  if (BegChars <> NIL) and (EndChars > BegChars) then begin
    while BegChars <= EndChars do begin
      if BegChars^ = OpenChar then
	inc(Result)
      else if BegChars^ = CloseChar then
	dec(Result)
      else if BegChars^ = NUL then Break;
      inc(BegChars);
    end;
  end;
end;

function GetCharsCount(const Chars: TSysCharSet; const Text: RawByteString): Integer;
var
  pText: PAnsiChar;
begin
  Result := 0;
  if Text = E then Exit;
  pText := Pointer(Text);
  while pText^ > NUL do begin
    if SysUtils.CharInSet(pText^, Chars) then inc(Result);
    inc(pText);
  end;
end;

function GetCharsCount(const Chars: TSysCharSet; const Text: UnicodeString): Integer;
var
  pText: PWideChar;
begin
  Result := 0;
  if Text = E then Exit;
  pText := Pointer(Text);
  while pText^ > NUL do begin
    if CharInSet(pText^, Chars) then inc(Result);
    inc(pText);
  end;
end;

function GetCharCount(Chr: AnsiChar; const Text: RawByteString): Integer;
var
  pText: PAnsiChar;
begin
  Result := 0;
  if Text = E then Exit;
  pText := Pointer(Text);
  while pText^ > NUL do begin
    if pText^ = Chr then inc(Result);
    inc(pText);
  end;
end;

function GetCharCount(Chr: WideChar; const Text: UnicodeString): Integer;
var
  pText: PWideChar;
begin
  Result := 0;
  if Text = E then Exit;
  pText := Pointer(Text);
  while pText^ > NUL do begin
    if pText^ = Chr then inc(Result);
    inc(pText);
  end;
end;

function FirstCharInSet(const Text: RawByteString; Chars: TSysCharSet): Integer;
var
  pText: PAnsiChar;
begin
  if (Text <> E) and (Chars <> []) then begin
    Chars := Chars + [NUL];
    pText := Pointer(Text);
    while not SysUtils.CharInSet(pText^, Chars) do inc(pText);
    if pText^ <> NUL then begin
      Result := succ(pText - Pointer(Text));
      Exit;
    end;
  end;
  Result := 0;
end;

function FirstCharInSet(const Text: UnicodeString; Chars: TSysCharSet): Integer;
var
  pText: PWideChar;
begin
  if (Text <> E) and (Chars <> []) then begin
    Chars := Chars + [NUL];
    pText := Pointer(Text);
    while not CharInSet(pText^, Chars) do inc(pText);
    if pText^ <> NUL then begin
      Result := succ(pText - Pointer(Text));
      Exit;
    end;
  end;
  Result := 0;
end;

function FirstCharPos(Chr: AnsiChar; const Text: RawByteString): Integer; inline;
begin
  Result := NextCharPos(Chr, Text, 1);
end;

function FirstCharPos(Chr: WideChar; const Text: UnicodeString): Integer; inline;
begin
  Result := NextCharPos(Chr, Text, 1);
end;

function LastCharPos(Chr: AnsiChar; const Text: RawByteString): Integer; inline;
begin
  Result := PrevCharPos(Chr, Text, Length(Text));
end;

function LastCharPos(Chr: WideChar; const Text: UnicodeString): Integer; inline;
begin
  Result := PrevCharPos(Chr, Text, Length(Text));
end;

function NextCharPos(Chr: AnsiChar; const Text: RawByteString; BegPos: Integer): Integer; inline;
begin
  Result := RangeCharPos(Chr, Text, BegPos, Length(Text));
end;

function NextCharPos(Chr: WideChar; const Text: UnicodeString; BegPos: Integer): Integer; inline;
begin
  Result := RangeCharPos(Chr, Text, BegPos, Length(Text));
end;

function PrevCharPos(Chr: AnsiChar; const Text: RawByteString; EndPos: Integer): Integer; inline;
begin
  Result := RangeCharPos(Chr, Text, EndPos, 1);
end;

function PrevCharPos(Chr: WideChar; const Text: UnicodeString; EndPos: Integer): Integer; inline;
begin
  Result := RangeCharPos(Chr, Text, EndPos, 1);
end;

function RangeCharPos(Chr: AnsiChar; const Text: RawByteString; BegPos, EndPos: Integer): Integer;
var
  iLen: Integer;
  paBeg, paEnd: PAnsiChar;
begin
  Result := 0;
  iLen := Length(Text);
  if iLen = 0 then Exit;
  if (BegPos <= EndPos) and (EndPos <= iLen) then begin
    paBeg := PAnsiChar(Text) + pred(BegPos);
    paEnd := PAnsiChar(Text) + pred(EndPos);
    while (paBeg <= paEnd) and (paBeg^ <> Chr) do inc(paBeg);
    if paBeg^ = Chr then Result := succ(paBeg - PAnsiChar(Text));
    Exit;
  end;
  if (BegPos >= EndPos) and (EndPos >= 1) then begin
    paBeg := PAnsiChar(Text) + pred(BegPos);
    paEnd := PAnsiChar(Text) + pred(EndPos);
    while (paBeg >= paEnd) and (paBeg^ <> Chr) do dec(paBeg);
    if paBeg^ = Chr then Result := succ(paBeg - PAnsiChar(Text));
    Exit;
  end;
end;

function RangeCharPos(Chr: WideChar; const Text: UnicodeString; BegPos, EndPos: Integer): Integer;
var
  iLen: Integer;
  pwBeg, pwEnd: PWideChar;
begin
  Result := 0;
  iLen := Length(Text);
  if iLen = 0 then Exit;
  if (BegPos <= EndPos) and (EndPos <= iLen) then begin
    pwBeg := PWideChar(Text) + pred(BegPos);
    pwEnd := PWideChar(Text) + pred(EndPos);
    while (pwBeg <= pwEnd) and (pwBeg^ <> Chr) do inc(pwBeg);
    if pwBeg^ = Chr then Result := succ(pwBeg - PWideChar(Text));
    Exit;
  end;
  if (BegPos >= EndPos) and (EndPos >= 1) then begin
    pwBeg := PWideChar(Text) + pred(BegPos);
    pwEnd := PWideChar(Text) + pred(EndPos);
    while (pwBeg >= pwEnd) and (pwBeg^ <> Chr) do dec(pwBeg);
    if pwBeg^ = Chr then Result := succ(pwBeg - PWideChar(Text));
    Exit;
  end;
end;

function AnsiPos(const SubStr, Str: RawByteString): Integer; inline;
begin
  Result := AnsiStrings.AnsiPos(SubStr, Str);
end;

function AnsiPos(const SubStr, Str: UnicodeString): Integer; inline;
begin
  Result := SysUtils.AnsiPos(SubStr, Str);
end;

function AnsiQuotedStr(const Str: RawByteString; Quote: AnsiChar = QU): RawByteString; inline;
begin
  Result := AnsiStrings.AnsiQuotedStr(Str, Quote);
end;

function AnsiQuotedStr(const Str: UnicodeString; Quote: WideChar = QU): UnicodeString; inline;
begin
  Result := SysUtils.AnsiQuotedStr(Str, Quote);
end;

function AnsiSameStr(const S1, S2: RawByteString): Longbool; inline;
begin
  Result := (AnsiStrings.AnsiCompareStr(S1, S2) = 0);
end;

function AnsiSameStr(const S1, S2: UnicodeString): Longbool; inline;
begin
  Result := (SysUtils.AnsiCompareStr(S1, S2) = 0);
end;

function AnsiSameText(const S1, S2: RawByteString): Longbool; inline;
begin
  Result := (AnsiStrings.AnsiCompareStr(S1, S2) = 0);
end;

function AnsiSameText(const S1, S2: UnicodeString): Longbool; inline;
begin
  Result := (SysUtils.AnsiCompareStr(S1, S2) = 0);
end;

function AnsiCompareStr(const S1, S2: RawByteString): Integer; inline;
begin
  Result := AnsiStrings.AnsiCompareStr(S1, S2);
end;

function AnsiCompareStr(const S1, S2: UnicodeString): Integer; inline;
begin
  Result := SysUtils.AnsiCompareStr(S1, S2);
end;

function AnsiCompareText(const S1, S2: RawByteString): Integer; inline;
begin
  Result := AnsiStrings.AnsiCompareText(S1, S2);
end;

function AnsiCompareText(const S1, S2: UnicodeString): Integer; inline;
begin
  Result := SysUtils.AnsiCompareText(S1, S2);
end;

function AnsiLowerCase(const Str: RawByteString): RawByteString; inline;
begin
  Result := AnsiStrings.AnsiLowerCase(Str);
end;

function AnsiLowerCase(const Str: UnicodeString): UnicodeString; inline;
begin
  Result := SysUtils.AnsiLowerCase(Str);
end;

function AnsiUpperCase(const Str: RawByteString): RawByteString; inline;
begin
  Result := AnsiStrings.AnsiUpperCase(Str);
end;

function AnsiUpperCase(const Str: UnicodeString): UnicodeString; inline;
begin
  Result := SysUtils.AnsiUpperCase(Str);
end;

function GetStrCount(const SearchStr, Str: RawByteString): Integer;
var
  iPos, iLen: Integer;
begin
  Result := 0;
  iPos := 1;
  iLen := Length(SearchStr);
  if (SearchStr = E) or (Str = E) then Exit;
  repeat
    iPos := NextStrPos(SearchStr, Str, iPos);
    if iPos > 0 then begin
      inc(iPos, iLen);
      inc(Result);
    end
    else Break;
  until iPos = 0;
end;

function GetStrCount(const SearchStr, Str: UnicodeString): Integer;
var
  iPos, iLen: Integer;
begin
  Result := 0;
  iPos := 1;
  iLen := Length(SearchStr);
  if (SearchStr = E) or (Str = E) then Exit;
  repeat
    iPos := NextStrPos(SearchStr, Str, iPos);
    if iPos > 0 then begin
      inc(iPos, iLen);
      inc(Result);
    end
    else Break;
  until iPos = 0;
end;

function FirstStrPos(const SearchStr, Str: RawByteString): Integer; inline;
begin
  Result := NextStrPos(SearchStr, Str, 1);
end;

function FirstStrPos(const SearchStr, Str: UnicodeString): Integer; inline;
begin
  Result := NextStrPos(SearchStr, Str, 1);
end;

function LastStrPos(const SearchStr, Str: RawByteString): Integer; inline;
begin
  Result := PrevStrPos(SearchStr, Str, Length(Str));
end;

function LastStrPos(const SearchStr, Str: UnicodeString): Integer; inline;
begin
  Result := PrevStrPos(SearchStr, Str, Length(Str));
end;

function NextStrPos(const SearchStr, Str: RawByteString; BegPos: Integer): Integer;
var
  iLen: Integer;
  pStr: PAnsiChar;
begin
  Result := 0;
  if SearchStr = E then Exit;
  iLen := Length(Str);
  if (iLen = 0) or (BegPos = 0) or (BegPos > iLen) then Exit;
  pStr := Pointer(Str);
  inc(pStr, pred(BegPos));
  pStr := StrStrA(pStr, Pointer(SearchStr));
  if pStr <> NIL then Result := succ(UInt64(pStr - Pointer(Str)));
end;

function NextStrPos(const SearchStr, Str: UnicodeString; BegPos: Integer): Integer;
var
  iLen: Integer;
  pStr: PWideChar;
begin
  Result := 0;
  if SearchStr = E then Exit;
  iLen := Length(Str);
  if (iLen = 0) or (BegPos = 0) or (BegPos > iLen) then Exit;
  pStr := Pointer(Str);
  inc(pStr, pred(BegPos));
  pStr := StrStrW(pStr, Pointer(SearchStr));
  if pStr <> NIL then Result := succ(UInt64(pStr - Pointer(Str)));
end;

function PrevStrPos(const SearchStr, Str: RawByteString; EndPos: Integer): Integer;
var
  iLen: Integer;
  pLast: PAnsiChar;
begin
  Result := 0;
  if SearchStr = E then Exit;
  iLen := Length(Str);
  if (iLen = 0) or (EndPos = 0) or (EndPos > iLen) then Exit;
  pLast := Pointer(Str);
  inc(pLast, pred(EndPos));
  iLen := Length(SearchStr);
  repeat
    pLast := StrRStrIA(Pointer(Str), pLast, Pointer(SearchStr));
  until (pLast = NIL) or AnsiStrAtPos(pLast, Pointer(SearchStr), iLen);
  if pLast <> NIL then Result := succ(UInt64(pLast - Pointer(Str)));
end;

function PrevStrPos(const SearchStr, Str: UnicodeString; EndPos: Integer): Integer;
var
  iLen: Integer;
  pLast: PWideChar;
begin
  Result := 0;
  if SearchStr = E then Exit;
  iLen := Length(Str);
  if (iLen = 0) or (EndPos = 0) or (EndPos > iLen) then Exit;
  pLast := Pointer(Str);
  inc(pLast, pred(EndPos));
  iLen := Length(SearchStr);
  repeat
    pLast := StrRStrIW(Pointer(Str), pLast, Pointer(SearchStr));
  until (pLast = NIL) or AnsiStrAtPos(pLast, Pointer(SearchStr), iLen);
  if pLast <> NIL then Result := succ(UInt64(pLast - Pointer(Str)));
end;

function RangeStrPos(const SearchStr, Str: RawByteString; BegPos, EndPos: Integer): Integer;
var
  iSearchStrLen, iStrLen: Integer;
  pStr, pMax: PAnsiChar;
  acFirst: AnsiChar;
begin
  Result := 0;
  if (SearchStr = E) or (Str = E) then Exit;
  iSearchStrLen := Length(SearchStr);
  iStrLen := Length(Str);
  if BegPos = 0 then inc(BegPos);
  if EndPos > iStrLen then EndPos := iStrLen;
  if EndPos < BegPos then Exit;
  pStr := Pointer(Str);
  pMax := pStr;
  acFirst := SearchStr[1];
  inc(pStr, pred(BegPos));
  inc(pMax, EndPos - iSearchStrLen);
  repeat
    while (pStr < pMax) and (pStr^ <> acFirst) do inc(pStr);
    if pStr^ = acFirst then begin
      if StrTools.AnsiStrLComp(Pointer(SearchStr), pStr, iSearchStrLen) = 0 then begin
	Result := succ(pStr - Pointer(Str));
	Exit;
      end;
    end;
    inc(pStr);
  until pStr >= pMax;
end;

function RangeStrPos(const SearchStr, Str: UnicodeString; BegPos, EndPos: Integer): Integer;
var
  iSearchStrLen, iStrLen: Integer;
  pStr, pMax: PWideChar;
  acFirst: WideChar;
begin
  Result := 0;
  if (SearchStr = E) or (Str = E) then Exit;
  iSearchStrLen := Length(SearchStr);
  iStrLen := Length(Str);
  if BegPos = 0 then inc(BegPos);
  if EndPos > iStrLen then EndPos := iStrLen;
  if EndPos < BegPos then Exit;
  pStr := Pointer(Str);
  pMax := pStr;
  acFirst := SearchStr[1];
  inc(pStr, pred(BegPos));
  inc(pMax, EndPos - iSearchStrLen);
  repeat
    while (pStr < pMax) and (pStr^ <> acFirst) do inc(pStr);
    if pStr^ = acFirst then begin
      if StrTools.AnsiStrLComp(Pointer(SearchStr), pStr, iSearchStrLen) = 0 then begin
	Result := succ(pStr - Pointer(Str));
	Exit;
      end;
    end;
    inc(pStr);
  until pStr >= pMax;
end;

function GetTextCount(const SearchText, Text: RawByteString): Integer;
var
  iPos, iLen: Integer;
begin
  Result := 0;
  iPos := 1;
  iLen := Length(SearchText);
  if (SearchText = E) or (Text = E) then Exit;
  repeat
    iPos := NextTextPos(SearchText, Text, iPos);
    if iPos > 0 then begin
      inc(iPos, iLen);
      inc(Result);
    end
    else Break;
  until iPos = 0;
end;

function GetTextCount(const SearchText, Text: UnicodeString): Integer;
var
  iPos, iLen: Integer;
begin
  Result := 0;
  iPos := 1;
  iLen := Length(SearchText);
  if (SearchText = E) or (Text = E) then Exit;
  repeat
    iPos := NextTextPos(SearchText, Text, iPos);
    if iPos > 0 then begin
      inc(iPos, iLen);
      inc(Result);
    end
    else Break;
  until iPos = 0;
end;

function FirstTextPos(const SearchText, Text: RawByteString): Integer; inline;
begin
  Result := NextTextPos(SearchText, Text, 1);
end;

function FirstTextPos(const SearchText, Text: UnicodeString): Integer; inline;
begin
  Result := NextTextPos(SearchText, Text, 1);
end;

function LastTextPos(const SearchText, Text: RawByteString): Integer; inline;
begin
  Result := PrevTextPos(SearchText, Text, Length(Text));
end;

function LastTextPos(const SearchText, Text: UnicodeString): Integer; inline;
begin
  Result := PrevTextPos(SearchText, Text, Length(Text));
end;

function NextTextPos(const SearchText, Text: RawByteString; BegPos: Integer): Integer;
var
  iLen: Integer;
  pText: PAnsiChar;
begin
  Result := 0;
  if SearchText = E then Exit;
  iLen := Length(Text);
  if (iLen = 0) or (BegPos = 0) or (BegPos > iLen) then Exit;
  pText := Pointer(Text);
  inc(pText, pred(BegPos));
  pText := StrStrIA(pText, Pointer(SearchText));
  if pText <> NIL then Result := succ(UInt64(pText - Pointer(Text)));
end;

function NextTextPos(const SearchText, Text: UnicodeString; BegPos: Integer): Integer;
var
  iLen: Integer;
  pText: PWideChar;
begin
  Result := 0;
  if SearchText = E then Exit;
  iLen := Length(Text);
  if (iLen = 0) or (BegPos = 0) or (BegPos > iLen) then Exit;
  pText := Pointer(Text);
  inc(pText, pred(BegPos));
  pText := StrStrIW(pText, Pointer(SearchText));
  if pText <> NIL then Result := succ(UInt64(pText - Pointer(Text)));
end;

function PrevTextPos(const SearchText, Text: RawByteString; EndPos: Integer): Integer;
var
  iLen: Integer;
  pLast: PAnsiChar;
begin
  Result := 0;
  if SearchText = E then Exit;
  iLen := Length(Text);
  if (iLen = 0) or (EndPos = 0) or (EndPos > iLen) then Exit;
  pLast := Pointer(Text);
  inc(pLast, pred(EndPos));
  pLast := StrRStrIA(Pointer(Text), pLast, Pointer(SearchText));
  if pLast <> NIL then Result := succ(UInt64(pLast - Pointer(Text)));
end;

function PrevTextPos(const SearchText, Text: UnicodeString; EndPos: Integer): Integer;
var
  iLen: Integer;
  pLast: PWideChar;
begin
  Result := 0;
  if SearchText = E then Exit;
  iLen := Length(Text);
  if (iLen = 0) or (EndPos = 0) or (EndPos > iLen) then Exit;
  pLast := Pointer(Text);
  inc(pLast, pred(EndPos));
  pLast := StrRStrIW(Pointer(Text), pLast, Pointer(SearchText));
  if pLast <> NIL then Result := succ(UInt64(pLast - Pointer(Text)));
end;

function RangeTextPos(const SearchText, Text: RawByteString; BegPos, EndPos: Integer): Integer;
var
  cSearchTextLen, cTextLen: Integer;
  pText, pMax: PAnsiChar;
  acFirst: AnsiChar;
begin
  Result := 0;
  if (SearchText = E) or (Text = E) then Exit;
  cSearchTextLen := Length(SearchText);
  cTextLen := Length(Text);
  if BegPos = 0 then inc(BegPos);
  if EndPos > cTextLen then EndPos := cTextLen;
  if EndPos < BegPos then Exit;
  pText := Pointer(Text);
  pMax := pText;
  acFirst := SearchText[1];
  inc(pText, pred(BegPos));
  inc(pMax, EndPos - cSearchTextLen);
  repeat
    while (pText < pMax) and (pText^ <> acFirst) do inc(pText);
    if pText^ = acFirst then begin
      if StrTools.AnsiStrLIComp(Pointer(SearchText), pText, cSearchTextLen) = 0 then begin
	Result := succ(pText - Pointer(Text));
	Exit;
      end;
    end;
    inc(pText);
  until pText >= pMax;
end;

function RangeTextPos(const SearchText, Text: UnicodeString; BegPos, EndPos: Integer): Integer;
var
  cSearchTextLen, cTextLen: Integer;
  pText, pMax: PWideChar;
  acFirst: WideChar;
begin
  Result := 0;
  if (SearchText = E) or (Text = E) then Exit;
  cSearchTextLen := Length(SearchText);
  cTextLen := Length(Text);
  if BegPos = 0 then inc(BegPos);
  if EndPos > cTextLen then EndPos := cTextLen;
  if EndPos < BegPos then Exit;
  pText := Pointer(Text);
  pMax := pText;
  acFirst := SearchText[1];
  inc(pText, pred(BegPos));
  inc(pMax, EndPos - cSearchTextLen);
  repeat
    while (pText < pMax) and (pText^ <> acFirst) do inc(pText);
    if pText^ = acFirst then begin
      if StrTools.AnsiStrLIComp(Pointer(SearchText), pText, cSearchTextLen) = 0 then begin
	Result := succ(pText - Pointer(Text));
	Exit;
      end;
    end;
    inc(pText);
  until pText >= pMax;
end;

function StrAtPos(const ASubStr, AStr: RawByteString; Pos: Integer): Longbool;
begin
  Result := (ASubStr <> E) and (Pos > 0) and (Pos <= Length(AStr)) and
	    (StrTools.AnsiStrLComp(PAnsiChar(ASubStr), @AStr[Pos], Length(ASubStr)) = 0);
end;

function StrAtPos(const ASubStr, AStr: UnicodeString; Pos: Integer): Longbool;
begin
  Result := (ASubStr <> E) and (Pos > 0) and (Pos <= Length(AStr)) and
	    (StrTools.AnsiStrLComp(PWideChar(ASubStr), @AStr[Pos], Length(ASubStr)) = 0);
end;

function TextAtPos(const ASubText, AText: RawByteString; Pos: Integer): Longbool;
begin
  Result := (ASubText <> E) and (Pos > 0) and (Pos <= Length(AText)) and
	    (StrTools.AnsiStrLIComp(PAnsiChar(ASubText), @AText[Pos], Length(ASubText)) = 0);
end;

function TextAtPos(const ASubText, AText: UnicodeString; Pos: Integer): Longbool;
begin
  Result := (ASubText <> E) and (Pos > 0) and (Pos <= Length(AText)) and
	    (StrTools.AnsiStrLIComp(PWideChar(ASubText), @AText[Pos], Length(ASubText)) = 0);
end;

function IncludesStr(const S1: RawByteString; const AValues: Array of RawByteString): Integer;
begin
  for Result := Low(AValues) to High(AValues) do begin
    if StrTools.ContainsStr(S1, AValues[Result]) then Exit;
  end;
  Result := -1;
end;

function IncludesStr(const S1: UnicodeString; const AValues: Array of UnicodeString): Integer;
begin
  for Result := Low(AValues) to High(AValues) do begin
    if StrTools.ContainsStr(S1, AValues[Result]) then Exit;
  end;
  Result := -1;
end;

function IncludesText(const S1: RawByteString; const AValues: Array of RawByteString): Integer;
begin
  for Result := Low(AValues) to High(AValues) do begin
    if StrTools.ContainsText(S1, AValues[Result]) then Exit;
  end;
  Result := -1;
end;

function IncludesText(const S1: UnicodeString; const AValues: Array of UnicodeString): Integer;
begin
  for Result := Low(AValues) to High(AValues) do begin
    if StrTools.ContainsText(S1, AValues[Result]) then Exit;
  end;
  Result := -1;
end;

function IndexStr(const S1: RawByteString; const AValues: Array of AnsiString; MaxLen: Integer = -1): Integer;
begin
  if MaxLen > 0 then begin
    for Result := Low(AValues) to High(AValues) do begin
      if StrAtPos(PAnsiChar(S1), Pointer(AValues[Result]), MaxLen) then Exit;
    end;
    Result := -1;
  end
  else begin
    for Result := Low(AValues) to High(AValues) do begin
      if StrAtPos(PAnsiChar(S1), Pointer(AValues[Result]), Length(AValues[Result])) then Exit;
    end;
    Result := -1;
  end;
end;

function IndexStr(const S1: UnicodeString; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Integer;
begin
  if MaxLen > 0 then begin
    for Result := Low(AValues) to High(AValues) do begin
      if StrAtPos(PWideChar(S1), Pointer(AValues[Result]), MaxLen) then Exit;
    end;
    Result := -1;
  end
  else begin
    for Result := Low(AValues) to High(AValues) do begin
      if StrAtPos(PWideChar(S1), Pointer(AValues[Result]), Length(AValues[Result])) then Exit;
    end;
    Result := -1;
  end;
end;

function IndexPChar(PCh: PAnsiChar; const AValues: Array of AnsiString; MaxLen: Integer = -1): Integer;
begin
  Result := StrTools.IndexStr(PCh, AValues, MaxLen);
end;

function IndexPChar(PCh: PWideChar; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Integer;
begin
  Result := StrTools.IndexStr(PCh, AValues, MaxLen);
end;

function IndexText(const S1: RawByteString; const AValues: Array of AnsiString; MaxLen: Integer = -1): Integer;
begin
  if MaxLen > 0 then begin
    for Result := Low(AValues) to High(AValues) do begin
      if StrIAtPos(PAnsiChar(S1), Pointer(AValues[Result]), MaxLen) then Exit;
    end;
    Result := -1;
  end
  else begin
    for Result := Low(AValues) to High(AValues) do begin
      if StrIAtPos(PAnsiChar(S1), Pointer(AValues[Result]), Length(AValues[Result])) or ((S1 = E) and ((AValues[Result]) = E)) then Exit;
    end;
    Result := -1;
  end;
end;

function IndexText(const S1: UnicodeString; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Integer;
begin
  if MaxLen > 0 then begin
    for Result := Low(AValues) to High(AValues) do begin
      if StrIAtPos(PWideChar(S1), Pointer(AValues[Result]), MaxLen) then Exit;
    end;
    Result := -1;
  end
  else begin
    for Result := Low(AValues) to High(AValues) do begin
      if StrIAtPos(PWideChar(S1), Pointer(AValues[Result]), Length(AValues[Result])) or ((S1 = E) and ((AValues[Result]) = E)) then Exit;
    end;
    Result := -1;
  end;
end;

function IndexPCharI(PCh: PAnsiChar; const AValues: Array of AnsiString; MaxLen: Integer = -1): Integer;
begin
  Result := StrTools.IndexText(PCh, AValues, MaxLen);
end;

function IndexPCharI(PCh: PWideChar; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Integer;
begin
  Result := StrTools.IndexText(PCh, AValues, MaxLen);
end;

function MatchStr(const S1: RawByteString; const AValues: Array of AnsiString; MaxLen: Integer = -1): Boolean;
begin
  Result := (StrTools.IndexStr(S1, AValues, MaxLen) > -1);
end;

function MatchStr(const S1: UnicodeString; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Boolean;
begin
  Result := (StrTools.IndexStr(S1, AValues, MaxLen) > -1);
end;

function MatchPChar(PCh: PAnsiChar; const AValues: Array of AnsiString; MaxLen: Integer = -1): Boolean;
begin
  Result := StrTools.MatchStr(PCh, AValues, MaxLen);
end;

function MatchPChar(PCh: PWideChar; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Boolean;
begin
  Result := StrTools.MatchStr(PCh, AValues, MaxLen);
end;

function MatchText(const S1: RawByteString; const AValues: Array of AnsiString; MaxLen: Integer = -1): Boolean;
begin
  Result := (StrTools.IndexText(S1, AValues, MaxLen) > -1);
end;

function MatchText(const S1: UnicodeString; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Boolean;
begin
  Result := (StrTools.IndexText(S1, AValues, MaxLen) > -1);
end;

function MatchPCharI(PCh: PAnsiChar; const AValues: Array of AnsiString; MaxLen: Integer = -1): Boolean;
begin
  Result := StrTools.MatchText(PCh, AValues, MaxLen);
end;

function MatchPCharI(PCh: PWideChar; const AValues: Array of UnicodeString; MaxLen: Integer = -1): Boolean;
begin
  Result := StrTools.MatchText(PCh, AValues, MaxLen);
end;

function SameStr(const Str1, Str2: RawByteString): Boolean; inline;
begin
  Result := AnsiStrings.SameStr(Str1, Str2);
end;

function SameStr(const Str1, Str2: UnicodeString): Boolean; inline;
begin
  Result := SysUtils.SameStr(Str1, Str2);
end;

function SameText(const Text1, Text2: RawByteString): Boolean; inline;
begin
  Result := AnsiStrings.SameText(Text1, Text2);
end;

function SameText(const Text1, Text2: UnicodeString): Boolean; inline;
begin
  Result := SysUtils.SameText(Text1, Text2);
end;

function SameTrimText(const Text1, Text2: RawByteString): Boolean;
begin
  Result := AnsiStrings.SameText(StrTools.Trim(Text1), StrTools.Trim(Text2));
end;

function SameTrimText(const Text1, Text2: UnicodeString): Boolean;
begin
  Result := SysUtils.SameText(StrTools.Trim(Text1), StrTools.Trim(Text2));
end;

function ContainsStr(const AStr, ASubStr: RawByteString): Boolean; inline;
begin
  Result := (FirstStrPos(ASubStr, AStr) > 0);
end;

function ContainsStr(const AStr, ASubStr: UnicodeString): Boolean; inline;
begin
  Result := (FirstStrPos(ASubStr, AStr) > 0);
end;

function ContainsText(const AText, ASubText: RawByteString): Boolean; inline;
begin
  Result := (FirstTextPos(ASubText, AText) > 0);
end;

function ContainsText(const AText, ASubText: UnicodeString): Boolean; inline;
begin
  Result := (FirstTextPos(ASubText, AText) > 0);
end;

function StartsStr(const ASubStr, AStr: RawByteString; MaxLen: Integer = -1): Boolean;
begin
  if MaxLen = -1 then
    Result := StrAtPos(ASubStr, AStr, 1)
  else Result := (StrTools.StrLComp(PAnsiChar(ASubStr), PAnsiChar(AStr), MaxLen) = 0);
end;

function StartsStr(const ASubStr, AStr: UnicodeString; MaxLen: Integer = -1): Boolean;
begin
  if MaxLen = -1 then
    Result := StrAtPos(ASubStr, AStr, 1)
  else Result := (StrTools.StrLComp(PWideChar(ASubStr), PWideChar(AStr), MaxLen) = 0);
end;

function StartsText(const ASubText, AText: RawByteString; MaxLen: Integer = -1): Boolean;
begin
  if MaxLen = -1 then
    Result := TextAtPos(ASubText, AText, 1)
  else Result := (StrTools.StrLIComp(PAnsiChar(ASubText), PAnsiChar(AText), MaxLen) = 0);
end;

function StartsText(const ASubText, AText: UnicodeString; MaxLen: Integer = -1): Boolean;
begin
  if MaxLen = -1 then
    Result := TextAtPos(ASubText, AText, 1)
  else Result := (StrTools.StrLIComp(PWideChar(ASubText), PWideChar(AText), MaxLen) = 0);
end;

function EndsStr(const ASubStr, AStr: RawByteString): Boolean; inline;
begin
  Result := StrAtPos(aSubStr, AStr, succ(Length(AStr) - Length(ASubStr)));
end;

function EndsStr(const ASubStr, AStr: UnicodeString): Boolean; inline;
begin
  Result := StrAtPos(aSubStr, AStr, succ(Length(AStr) - Length(ASubStr)));
end;

function EndsText(const ASubText, AText: RawByteString): Boolean; inline;
begin
  Result := TextAtPos(aSubText, AText, succ(Length(AText) - Length(ASubText)));
end;

function EndsText(const ASubText, AText: UnicodeString): Boolean; inline;
begin
  Result := TextAtPos(aSubText, AText, succ(Length(AText) - Length(ASubText)));
end;

function ExtractStr(var Source: RawByteString; Pos, Len: Integer; TrimSource: TTrimOptions = []): RawByteString;
var
  iIndex: Integer;
begin
  Result := Copy(Source, Pos, Len);
  if toLeft in TrimSource then begin
    iIndex := pred(Pos);
    while (iIndex > 0) and (Source[iIndex] <= SP) do dec(iIndex);
    if Source[iIndex] > SP then inc(iIndex);
    inc(Len, Pos - iIndex);
    Pos := iIndex;
  end;
  if toRight in TrimSource then begin
    iIndex := Pos + Len;
    while (iIndex <= Length(Source)) and (Source[iIndex] <= SP) do inc(iIndex);
    Len := iIndex - Pos;
  end;
  Delete(Source, Pos, Len);
end;

function ExtractStr(var Source: UnicodeString; Pos, Len: Integer; TrimSource: TTrimOptions = []): UnicodeString;
var
  iIndex: Integer;
begin
  Result := Copy(Source, Pos, Len);
  if toLeft in TrimSource then begin
    iIndex := pred(Pos);
    while (iIndex > 0) and (Source[iIndex] <= SP) do dec(iIndex);
    if Source[iIndex] > SP then inc(iIndex);
    inc(Len, Pos - iIndex);
    Pos := iIndex;
  end;
  if toRight in TrimSource then begin
    iIndex := Pos + Len;
    while (iIndex <= Length(Source)) and (Source[iIndex] <= SP) do inc(iIndex);
    Len := iIndex - Pos;
  end;
  Delete(Source, Pos, Len);
end;

function FromBase64(Str: PAnsiChar; Size: Integer): RawByteString;
var
  dwLen: DWord;
begin
  if Str <> E then begin
    dwLen := 0;
    if CryptStringToBinaryA(Str, Size, CRYPT_STRING_BASE64, NIL, dwLen, NIL, NIL) and (dwLen > 0) then begin
      SetString(Result, NIL, dwLen);
      if CryptStringToBinaryA(Str, Size, CRYPT_STRING_BASE64, Pointer(Result), dwLen, NIL, NIL) then Exit;
    end;
  end;
  Result := E;
end;

function FromBase64(Str: PWideChar; Size: Integer): UnicodeString;
var
  dwLen: DWord;
begin
  if Str <> E then begin
    dwLen := 0;
    if CryptStringToBinaryW(Str, Size, CRYPT_STRING_BASE64, NIL, dwLen, NIL, NIL) and (dwLen > 0) then begin
      SetString(Result, NIL, dwLen);
      if CryptStringToBinaryW(Str, Size, CRYPT_STRING_BASE64, Pointer(Result), dwLen, NIL, NIL) then Exit;
    end;
  end;
  Result := E;
end;

function FromBase64(const Str: RawByteString): RawByteString; inline;
begin
  Result := FromBase64(PAnsiChar(Str), Length(Str));
end;

function FromBase64(const Str: UnicodeString): UnicodeString; inline;
begin
  Result := FromBase64(PWideChar(Str), Length(Str));
end;

function ToBase64(const Str: PAnsiChar; Size: Integer; DropLineBreaks: Longbool = FALSE): RawByteString;
var
  dwLen: DWord;
begin
  if Str <> E then begin
    dwLen := 0;
    if CryptBinaryToStringA(PByte(Str), Size * SizeOf(AnsiChar), CRYPT_STRING_BASE64, NIL, dwLen) and (dwLen > 0) then begin
      SetString(Result, NIL, pred(dwLen));
      if CryptBinaryToStringA(PByte(Str), Size * SizeOf(AnsiChar), CRYPT_STRING_BASE64, Pointer(Result), dwLen) then begin
	if DropLineBreaks then Result := ReplaceStrings(Result, RAW(CRLF), RAW(E));
	Exit;
      end;
    end;
  end;
  Result := E;
end;

function ToBase64(const Str: PWideChar; Size: Integer; DropLineBreaks: Longbool = FALSE): UnicodeString;
var
  dwLen: DWord;
begin
  if Str <> E then begin
    dwLen := 0;
    if CryptBinaryToStringW(PByte(Str), Size * SizeOf(WideChar), CRYPT_STRING_BASE64, NIL, dwLen) and (dwLen > 0) then begin
      SetString(Result, NIL, pred(dwLen));
      if CryptBinaryToStringW(PByte(Str), Size * SizeOf(WideChar), CRYPT_STRING_BASE64, Pointer(Result), dwLen) then begin
	if DropLineBreaks then Result := ReplaceStrings(Result, CRLF, E);
	Exit;
      end;
    end;
  end;
  Result := E;
end;

function ToBase64(const Str: RawByteString; DropLineBreaks: Longbool = FALSE): RawByteString; inline;
begin
  Result := ToBase64(PAnsiChar(Str), Length(Str));
end;

function ToBase64(const Str: UnicodeString; DropLineBreaks: Longbool = FALSE): UnicodeString; inline;
begin
  Result := ToBase64(PWideChar(Str), Length(Str));
end;

function HashedString(const Str: RawByteString; Len: Integer): RawByteString;
var
  rsHash: RawByteString;
begin
  if Str <> E then begin
    Len := Len div 2;
    SetLength(rsHash, Len);
    if HashData(Pointer(Str), Length(Str) * SizeOf(AnsiChar), Pointer(rsHash), Len) = S_OK then begin
      SetString(Result, NIL, Len * 2);
      BinToHex(PAnsiChar(rsHash), PAnsiChar(Result), Len);
      Exit;
    end;
  end;
  Result := E;
end;

function HashedString(const Str: UnicodeString; Len: Integer): UnicodeString;
var
  rsHash: RawByteString;
begin
  if Str <> E then begin
    Len := Len div 2;
    SetLength(rsHash, Len);
    if HashData(Pointer(Str), Length(Str) * SizeOf(WideChar), Pointer(rsHash), Len) = S_OK then begin
      SetString(Result, NIL, Len * 2);
      BinToHex(PAnsiChar(rsHash), PWideChar(Result), Len);
      Exit;
    end;
  end;
  Result := E;
end;

// Windows only SysUtils function
function AdjustLineBreaks(const Str: RawByteString): RawByteString;
var
  pSource, pLast, pDest: PAnsiChar;
  iDestLen, iCopyLen: Integer;
begin
  pSource := Pointer(Str);
  iDestLen := Length(Str);
  pLast := pSource + iDestLen;
  while pSource < pLast do begin
    case pSource^ of
      LF: inc(iDestLen);
      CR: if pSource[1] = LF then
	inc(pSource)
      else inc(iDestLen);
    else
      if SysUtils.CharInSet(pSource^, LeadBytes) then begin
	inc(pSource);
	Continue;
      end;
    end;
    inc(pSource);
  end;
  if iDestLen = Length(pSource) then
    Result := Str
  else begin
    pSource := Pointer(Str);
    SetString(Result, NIL, iDestLen);
    pDest := Pointer(Result);
    while pSource < pLast do begin
      case pSource^ of
	LF: begin
	  pDest^ := CR;
	  inc(pDest);
	  pDest^ := LF;
	  inc(pDest);
	  inc(pSource);
	end;
	CR: begin
	  pDest^ := CR;
	  inc(pDest);
	  pDest^ := LF;
	  inc(pDest);
	  inc(pSource);
	  if pSource^ = LF then inc(pSource);
	end;
      else
	if SysUtils.CharInSet(pSource^, LeadBytes) then begin
	  iCopyLen := StrTools.StrCharLength(pSource);
	  StrTools.StrLCopy(pDest, pSource, iCopyLen);
	  inc(pSource, iCopyLen);
	  inc(pDest, iCopyLen);
	  Continue;
	end;
	pDest^ := pSource^;
	inc(pSource);
	inc(pDest);
      end;
    end;
  end;
end;

// Windows only SysUtils function
function AdjustLineBreaks(const Str: UnicodeString): UnicodeString;
var
  pSource, pLast, pDest: PWideChar;
  iDestLen, iCopyLen: Integer;
begin
  pSource := Pointer(Str);
  iDestLen := Length(Str);
  pLast := pSource + iDestLen;
  while pSource < pLast do begin
    case pSource^ of
      LF: inc(iDestLen);
      CR: if pSource[1] = LF then
	inc(pSource)
      else inc(iDestLen);
    else
      if CharInSet(pSource^, LeadBytes) then begin
	inc(pSource);
	Continue;
      end;
    end;
    inc(pSource);
  end;
  if iDestLen = Length(pSource) then
    Result := Str
  else begin
    pSource := Pointer(Str);
    SetString(Result, NIL, iDestLen);
    pDest := Pointer(Result);
    while pSource < pLast do begin
      case pSource^ of
	LF: begin
	  pDest^ := CR;
	  inc(pDest);
	  pDest^ := LF;
	  inc(pDest);
	  inc(pSource);
	end;
	CR: begin
	  pDest^ := CR;
	  inc(pDest);
	  pDest^ := LF;
	  inc(pDest);
	  inc(pSource);
	  if pSource^ = LF then inc(pSource);
	end;
      else
	if CharInSet(pSource^, LeadBytes) then begin
	  iCopyLen := StrTools.StrCharLength(pSource);
	  StrTools.StrLCopy(pDest, pSource, iCopyLen);
	  inc(pSource, iCopyLen);
	  inc(pDest, iCopyLen);
	  Continue;
	end;
	pDest^ := pSource^;
	inc(pSource);
	inc(pDest);
      end;
    end;
  end;
end;

function InsertLineBreaks(const Source: RawByteString; LineLen: Cardinal = 80): RawByteString;
var
  cLen, cPos: Cardinal;
  pSource, pResult: PAnsiChar;
begin
  if (Source <> E) and (LineLen <> 0) then begin
    cLen := Length(Source);
    pSource := PAnsiChar(Source);
    SetString(Result, NIL, cLen + DivModPlus(cLen, LineLen) * Cardinal(Length(CRLF)));
    pResult := PAnsiChar(Result);
    cPos := 0;
    while cPos + LineLen <= cLen do begin
      StrTools.StrLCopy(pResult, pSource, LineLen);
      inc(pSource, LineLen);
      inc(pResult, LineLen);
      inc(cPos, LineLen);
      PDAChar(pResult)^ := CRLF;
      inc(pResult, Length(CRLF));
    end;
    dec(cLen, cPos);
    if cLen <> 0 then begin
      StrTools.StrLCopy(pResult, pSource, cLen);
      inc(pResult, cLen);
      PDAChar(pResult)^ := CRLF;
    end;
  end;
end;

function InsertLineBreaks(const Source: UnicodeString; LineLen: Cardinal = 80): UnicodeString;
var
  cLen, cPos: Cardinal;
  pSource, pResult: PWideChar;
begin
  if (Source <> E) and (LineLen <> 0) then begin
    cLen := Length(Source);
    pSource := PWideChar(Source);
    SetString(Result, NIL, cLen + DivModPlus(cLen, LineLen) * Cardinal(Length(CRLF)));
    pResult := PWideChar(Result);
    cPos := 0;
    while cPos + LineLen <= cLen do begin
      StrTools.StrLCopy(pResult, pSource, LineLen);
      inc(pSource, LineLen);
      inc(pResult, LineLen);
      inc(cPos, LineLen);
      PDWChar(pResult)^ := CRLF;
      inc(pResult, Length(CRLF));
    end;
    dec(cLen, cPos);
    if cLen <> 0 then begin
      StrTools.StrLCopy(pResult, pSource, cLen);
      inc(pResult, cLen);
      PDWChar(pResult)^ := CRLF;
    end;
  end;
end;

function RemoveLineBreaks(const Str: RawByteString): RawByteString;
var
  pStr, pResult: PAnsiChar;
begin
  if Str <> E then begin
    SetString(Result, NIL, Length(Str));
    pResult := Pointer(Result);
    pStr := Pointer(Str);
    repeat
      case pStr^ of
	NUL: Break;
	LF, CR: inc(pStr);
      else
	pResult^ := pStr^;
	inc(pResult);
	inc(pStr);
      end;
    until pStr^ = NUL;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
    Exit;
  end;
  Result := E;
end;

function RemoveLineBreaks(const Str: UnicodeString): UnicodeString;
var
  pStr, pResult: PWideChar;
begin
  if Str <> E then begin
    SetString(Result, NIL, Length(Str));
    pResult := Pointer(Result);
    pStr := Pointer(Str);
    repeat
      case pStr^ of
	NUL: Break;
	LF, CR: inc(pStr);
      else
	pResult^ := pStr^;
	inc(pResult);
	inc(pStr);
      end;
    until pStr^ = NUL;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
    Exit;
  end;
  Result := E;
end;

function ReplaceCharAtPos(const Str: RawByteString; Old, New: AnsiChar; Pos: Integer = -1): RawByteString;
var
  pResult: PAnsiChar;
begin
  Result := Str;
  UniqueString(Result);
  if Result <> E then begin
    pResult := Pointer(Result);
    if Pos = -1 then Pos := FirstCharPos(Old, Str);
    if Pos >=0 then pResult[pred(Pos)] := New;
  end;
end;

function ReplaceCharAtPos(const Str: UnicodeString; Old, New: WideChar; Pos: Integer = -1): UnicodeString;
var
  pResult: PWideChar;
begin
  Result := Str;
  UniqueString(Result);
  if Result <> E then begin
    pResult := Pointer(Result);
    if Pos = -1 then Pos := FirstCharPos(Old, Str);
    if Pos >=0 then pResult[pred(Pos)] := New;
  end;
end;

function ReplaceChars(const Str: RawByteString; Old: TSysCharSet; New: AnsiChar): RawByteString;
var
  pResult: PAnsiChar;
begin
  Result := Str;
  UniqueString(Result);
  if Result <> E then begin
    pResult := Pointer(Result);
    repeat
      while not SysUtils.CharInSet(pResult^, Old + [NUL]) do inc(pResult);
      while SysUtils.CharInSet(pResult^, Old) do begin
	pResult^ := New;
	inc(pResult);
      end;
    until pResult^ = NUL;
  end;
end;

function ReplaceChars(const Str: UnicodeString; Old: TSysCharSet; New: WideChar): UnicodeString;
var
  pResult: PWideChar;
begin
  Result := Str;
  UniqueString(Result);
  if Result <> E then begin
    pResult := Pointer(Result);
    repeat
      while not CharInSet(pResult^, Old + [NUL]) do inc(pResult);
      while CharInSet(pResult^, Old) do begin
	pResult^ := New;
	inc(pResult);
      end;
    until pResult^ = NUL;
  end;
end;

function ReplaceCharsAsHex(const Str: RawByteString; Old: TSysCharSet): RawByteString;
var
  pStr, pResult: PAnsiChar;
begin
  if Str <> E then begin
    SetString(Result, NIL, Length(Str) * 3);
    pResult := Pointer(Result);
    pStr := Pointer(Str);
    repeat
      while not SysUtils.CharInSet(pStr^, Old + [NUL]) do begin
        pResult^ := pStr^;
        inc(pResult);
	inc(pStr);
      end;
      while SysUtils.CharInSet(pStr^, Old) do begin
	pResult := StrLECopy(pResult, PAnsiChar(CardToHexA(Ord(SL), 2, PC)), 3);
	inc(pStr);
      end;
    until pStr^ = NUL;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end;
end;

function ReplaceCharsAsHex(const Str: UnicodeString; Old: TSysCharSet): UnicodeString;
var
  pStr, pResult: PWideChar;
begin
  if Str <> E then begin
    SetString(Result, NIL, Length(Str) * 3);
    pResult := Pointer(Result);
    pStr := Pointer(Str);
    repeat
      while not CharInSet(pStr^, Old + [NUL]) do begin
        pResult^ := pStr^;
        inc(pResult);
	inc(pStr);
      end;
      while CharInSet(pStr^, Old) do begin
	pResult := StrLECopy(pResult, PWideChar(CardToHexW(Ord(SL), 2, PC)), 3);
	inc(pStr);
      end;
    until pStr^ = NUL;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end;
end;

function ReplaceMultiStrings(const Str: RawByteString; const ReplaceStrings: TReplaceStringsA; CaseInsens: Bool = TRUE): RawByteString;
var
  iLen, iPos, iIndex, iCount, iStrLen, iNewLen, iOldLen: Integer;
  pBeg, pEnd, pStrEnd, pResult: PAnsiChar;
  setSearchChars: Set of AnsiChar;
begin
  with ReplaceStrings do if (Str <> E) and (Length(FromStr) = Length(ToStr)) and (Length(FromStr) > 0) then begin
    iLen := 0;
    iPos := 0;
    iStrLen := Length(Str);
    pStrEnd := PAnsiChar(Str) + iStrLen;
    setSearchChars := [NUL];
    pBeg := Pointer(Str);
    pEnd := pBeg;
    for iIndex := Low(FromStr) to High(FromStr) do begin
      iCount := Length(ToStr[iIndex]) - Length(FromStr[iIndex]);
      Include(setSearchChars, StrTools.ToLower(FromStr[iIndex][1]));
      if iCount > iLen then begin
	iLen := iCount;
	iPos := iIndex;
      end;
    end;
    if iLen > 0 then
      SetString(Result, NIL, iStrLen + MulDiv(iStrLen, iLen, Length(FromStr[iPos])))
    else SetString(Result, NIL, iStrLen);
    pResult := Pointer(Result);
    if iStrLen > 0 then begin
      repeat
	while not StrTools.CharInSet(ToLower(pEnd^), setSearchChars) do inc(pEnd);
	if pEnd^ > NUL then begin
	  for iIndex := Low(FromStr) to High(FromStr) do begin
	    if CaseInsens then
	      iPos := IndexPCharI(pEnd, FromStr)
	    else iPos := IndexPChar(pEnd, FromStr);
	    if iPos > -1 then Break;
	  end;
	  if iPos > -1 then begin
	    iLen := pEnd - pBeg;
	    if iLen >= 0 then begin
	      iNewLen := Length(ToStr[iPos]);
	      iOldLen := Length(FromStr[iPos]);
	      if iLen > 0 then pResult := StrLECopy(pResult, pBeg, iLen);
	      if iNewLen > 0 then pResult := StrLECopy(pResult, Pointer(ToStr[iPos]), iNewLen);
	      inc(pEnd, iOldLen);
	      pBeg := pEnd;
	    end;
	  end
	  else inc(pEnd);
	end
	else begin
	  if pBeg = Pointer(Str) then begin
	    Result := Str;
	    Exit;
	  end
	  else if pBeg > Pointer(Str) then begin
	    iLen := pStrEnd - pBeg;
	    pResult := StrLECopy(pResult, pBeg, iLen);
	  end;
	  Break;
	end;
      until pEnd > pStrEnd;
    end;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
    Exit;
  end;
  Result := Str;
end;

function ReplaceMultiStrings(const Str: UnicodeString; const ReplaceStrings: TReplaceStringsW; CaseInsens: Bool = TRUE): UnicodeString;
var
  iLen, iPos, iIndex, iCount, iStrLen, iNewLen, iOldLen: Integer;
  pBeg, pEnd, pStrEnd, pResult: PWideChar;
  setSearchChars: TWideCharArray;
begin
  with ReplaceStrings do if (Str <> E) and (Length(FromStr) = Length(ToStr)) and (Length(FromStr) > 0) then begin
    iLen := 0;
    iPos := 0;
    iStrLen := Length(Str);
    pStrEnd := PWideChar(Str) + iStrLen;
    SetLength(setSearchChars, succ(Length(FromStr)));
    setSearchChars[0] := NUL;
    pBeg := Pointer(Str);
    pEnd := pBeg;
    for iIndex := Low(FromStr) to High(FromStr) do begin
      iCount := Length(ToStr[iIndex]) - Length(FromStr[iIndex]);
      setSearchChars[succ(iIndex)] := StrTools.ToLower(FromStr[iIndex][1]);
      if iCount > iLen then begin
	iLen := iCount;
	iPos := iIndex;
      end;
    end;
    if iLen > 0 then
      SetString(Result, NIL, iStrLen + MulDiv(iStrLen, iLen, Length(FromStr[iPos])))
    else SetString(Result, NIL, iStrLen);
    pResult := Pointer(Result);
    if iStrLen > 0 then begin
      repeat
	while not StrTools.CharInSet(StrTools.ToLower(pEnd^), setSearchChars) do inc(pEnd);
	if pEnd^ > NUL then begin
	  for iIndex := Low(FromStr) to High(FromStr) do begin
	    if CaseInsens then
	      iPos := IndexPCharI(pEnd, FromStr)
	    else iPos := IndexPChar(pEnd, FromStr);
	    if iPos > -1 then Break;
	  end;
	  if iPos > -1 then begin
	    iLen := pEnd - pBeg;
	    if iLen >= 0 then begin
	      iNewLen := Length(ToStr[iPos]);
	      iOldLen := Length(FromStr[iPos]);
	      if iLen > 0 then pResult := StrLECopy(pResult, pBeg, iLen);
	      if iNewLen > 0 then pResult := StrLECopy(pResult, Pointer(ToStr[iPos]), iNewLen);
	      inc(pEnd, iOldLen);
	      pBeg := pEnd;
	    end;
	  end
	  else inc(pEnd);
	end
	else begin
	  if pBeg = Pointer(Str) then begin
	    Result := Str;
	    Exit;
	  end
	  else if pBeg > Pointer(Str) then begin
	    iLen := pStrEnd - pBeg;
	    pResult := StrLECopy(pResult, pBeg, iLen);
	  end;
	  Break;
	end;
      until pEnd > pStrEnd;
    end;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
    Exit;
  end;
  Result := Str;
end;

function ReplaceStrings(const Str, Old, New: RawByteString; FullLineOnly: Longbool = FALSE): RawByteString;
var
  iStrLen, iOldLen, iNewLen, iLen: Integer;
  pStrEnd, pResult, pOld, pNew, pBeg, pEnd: PAnsiChar;
begin
  iStrLen := Length(Str);
  iOldLen := Length(Old);
  iNewLen := Length(New);
  if (iStrLen > 0) and (iOldLen > 0) then begin
    if FullLineOnly then begin
      if StrTools.StartsStr(Old, Str) then
	pBeg := PAnsiChar(Str) + iOldLen
      else pBeg := Pointer(Str);
    end
    else pBeg := Pointer(Str);
    if iNewLen > iOldLen then
      SetString(Result, NIL, iStrLen + MulDiv(iStrLen, iNewLen - iOldLen, iOldLen))
    else SetString(Result, NIL, iStrLen);
    pStrEnd := PAnsiChar(Str) + iStrLen;
    pResult := Pointer(Result);
    pOld := Pointer(Old);
    pNew := Pointer(New);
    repeat
      pEnd := StrTools.AnsiStrPos(pBeg, pOld);
      iLen := pEnd - pBeg;
      if iLen >= 0 then begin
	if iLen > 0 then begin
	  if FullLineOnly and not (AnsiChar((pEnd - 1)^) in WSP) then inc(iLen, iOldLen);
	  StrTools.StrLCopy(pResult, pBeg, iLen);
	  inc(pResult, iLen);
	end;
	if (iNewLen > 0) and (not FullLineOnly or (AnsiChar((pEnd - 1)^) in WSP)) then begin
	  StrTools.StrLCopy(pResult, pNew, iNewLen);
	  inc(pResult, iNewLen);
	end;
	pBeg := pEnd + iOldLen;
      end
      else begin
	if pBeg = Pointer(Str) then begin
	  Result := Str;
	  Exit;
	end
	else if pBeg > Pointer(Str) then begin
	  iLen := pStrEnd - pBeg;
	  StrTools.StrLCopy(pResult, pBeg, iLen);
	  inc(pResult, iLen);
	end;
	Break;
      end;
    until pBeg > pStrEnd;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
    Exit;
  end;
  Result := Str;
end;

function ReplaceStrings(const Str, Old, New: UnicodeString; FullLineOnly: Longbool = FALSE): UnicodeString;
var
  iStrLen, iOldLen, iNewLen, iLen: Integer;
  pStrEnd, pResult, pOld, pNew, pBeg, pEnd: PWideChar;
begin
  iStrLen := Length(Str);
  iOldLen := Length(Old);
  iNewLen := Length(New);
  if (iStrLen > 0) and (iOldLen > 0) then begin
    if FullLineOnly then begin
      if StrTools.StartsStr(Old, Str) then
	pBeg := PWideChar(Str) + iOldLen
      else pBeg := Pointer(Str);
    end
    else pBeg := Pointer(Str);
    if iNewLen > iOldLen then
      SetString(Result, NIL, iStrLen + MulDiv(iStrLen, iNewLen - iOldLen, iOldLen))
    else SetString(Result, NIL, iStrLen);
    pStrEnd := PWideChar(Str) + iStrLen;
    pResult := Pointer(Result);
    pOld := Pointer(Old);
    pNew := Pointer(New);
    repeat
      pEnd := StrTools.AnsiStrPos(pBeg, pOld);
      iLen := pEnd - pBeg;
      if iLen >= 0 then begin
	if iLen > 0 then begin
	  if FullLineOnly and not (AnsiChar((pEnd - 1)^) in WSP) then inc(iLen, iOldLen);
	  StrTools.StrLCopy(pResult, pBeg, iLen);
	  inc(pResult, iLen);
	end;
	if (iNewLen > 0) and (not FullLineOnly or (AnsiChar((pEnd - 1)^) in WSP)) then begin
	  StrTools.StrLCopy(pResult, pNew, iNewLen);
	  inc(pResult, iNewLen);
	end;
	pBeg := pEnd + iOldLen;
      end
      else begin
	if pBeg = Pointer(Str) then begin
	  Result := Str;
	  Exit;
	end
	else if pBeg > Pointer(Str) then begin
	  iLen := pStrEnd - pBeg;
	  StrTools.StrLCopy(pResult, pBeg, iLen);
	  inc(pResult, iLen);
	end;
	Break;
      end;
    until pBeg > pStrEnd;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
    Exit;
  end;
  Result := Str;
end;

function ReplaceStrAtPos(const Str, New: RawByteString; OldPos, OldLen: Integer): RawByteString;
var
  iStrLen, iNewLen, iCopy: Integer;
  pStr, pNew, pResult: PAnsiChar;
begin
  iStrLen := Length(Str);
  iNewLen := Length(New);
  if (iStrLen > 0) and (iNewLen > 0) and (OldPos > 0) then begin
    pStr := Pointer(Str);
    pNew := Pointer(New);
    if OldPos > iStrLen then begin
      SetString(Result, NIL, iStrLen + iNewLen);
      pResult := Pointer(Result);
      StrTools.StrLCopy(pResult, pStr, iStrLen);
      inc(pResult, iStrLen);
      StrTools.StrLCopy(pResult, pNew, iNewLen);
    end
    else begin
      if OldPos + OldLen > succ(iStrLen) then OldLen := succ(iStrLen - OldPos);
      SetString(Result, NIL, iStrLen + iNewLen - OldLen);
      pResult := Pointer(Result);
      iCopy := pred(OldPos);
      if iCopy > 0 then begin
	StrTools.StrLCopy(pResult, pStr, iCopy);
	inc(pResult, iCopy);
      end;
      inc(iCopy, OldLen);
      inc(pStr, iCopy);
      StrTools.StrLCopy(pResult, pNew, iNewLen);
      inc(pResult, iNewLen);
      iCopy := iStrLen - iCopy;
      if iCopy > 0 then StrTools.StrLCopy(pResult, pStr, iCopy);
    end;
    Exit;
  end;
  Result := Str;
end;

function ReplaceStrAtPos(const Str, New: UnicodeString; OldPos, OldLen: Integer): UnicodeString;
var
  iStrLen, iNewLen, iCopy: Integer;
  pStr, pNew, pResult: PWideChar;
begin
  iStrLen := Length(Str);
  iNewLen := Length(New);
  if (iStrLen > 0) and (iNewLen > 0) and (OldPos > 0) then begin
    pStr := Pointer(Str);
    pNew := Pointer(New);
    if OldPos > iStrLen then begin
      SetString(Result, NIL, iStrLen + iNewLen);
      pResult := Pointer(Result);
      StrTools.StrLCopy(pResult, pStr, iStrLen);
      inc(pResult, iStrLen);
      StrTools.StrLCopy(pResult, pNew, iNewLen);
    end
    else begin
      if OldPos + OldLen > succ(iStrLen) then OldLen := succ(iStrLen - OldPos);
      SetString(Result, NIL, iStrLen + iNewLen - OldLen);
      pResult := Pointer(Result);
      iCopy := pred(OldPos);
      if iCopy > 0 then begin
	StrTools.StrLCopy(pResult, pStr, iCopy);
	inc(pResult, iCopy);
      end;
      inc(iCopy, OldLen);
      inc(pStr, iCopy);
      StrTools.StrLCopy(pResult, pNew, iNewLen);
      inc(pResult, iNewLen);
      iCopy := iStrLen - iCopy;
      if iCopy > 0 then StrTools.StrLCopy(pResult, pStr, iCopy);
    end;
    Exit;
  end;
  Result := Str;
end;

function ReplaceText(const Str, Old, New: RawByteString; FullLineOnly: Longbool = FALSE): RawByteString;
var
  iStrLen, iOldLen, iNewLen, iLen: Integer;
  pStrEnd, pResult, pOld, pNew, pBeg, pEnd: PAnsiChar;
begin
  iStrLen := Length(Str);
  iOldLen := Length(Old);
  iNewLen := Length(New);
  if (iStrLen > 0) and (iOldLen > 0) then begin
    if FullLineOnly then begin
      if StrTools.StartsText(Old, Str) then
	pBeg := PAnsiChar(Str) + iOldLen
      else pBeg := Pointer(Str);
    end
    else pBeg := Pointer(Str);
    if iNewLen > iOldLen then
      SetString(Result, NIL, iStrLen + MulDiv(iStrLen, iNewLen - iOldLen, iOldLen))
    else SetString(Result, NIL, iStrLen);
    pStrEnd := PAnsiChar(Str) + iStrLen;
    pResult := Pointer(Result);
    pOld := Pointer(Old);
    pNew := Pointer(New);
    repeat
      pEnd := StrTools.AnsiTextPos(pBeg, pOld);
      iLen := pEnd - pBeg;
      if iLen >= 0 then begin
	if iLen > 0 then begin
	  if FullLineOnly and not (AnsiChar((pEnd - 1)^) in WSP) then inc(iLen, iOldLen);
	  StrTools.StrLCopy(pResult, pBeg, iLen);
	  inc(pResult, iLen);
	end;
	if (iNewLen > 0) and (not FullLineOnly or (AnsiChar((pEnd - 1)^) in WSP)) then begin
	  StrTools.StrLCopy(pResult, pNew, iNewLen);
	  inc(pResult, iNewLen);
	end;
	pBeg := pEnd + iOldLen;
      end
      else begin
	if pBeg = Pointer(Str) then begin
	  Result := Str;
	  Exit;
	end
	else if pBeg > Pointer(Str) then begin
	  iLen := pStrEnd - pBeg;
	  StrTools.StrLCopy(pResult, pBeg, iLen);
	  inc(pResult, iLen);
	end;
	Break;
      end;
    until pBeg > pStrEnd;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
    Exit;
  end;
  Result := Str;
end;

function ReplaceText(const Str, Old, New: UnicodeString; FullLineOnly: Longbool = FALSE): UnicodeString;
var
  iStrLen, iOldLen, iNewLen, iLen: Integer;
  pStrEnd, pResult, pOld, pNew, pBeg, pEnd: PWideChar;
begin
  iStrLen := Length(Str);
  iOldLen := Length(Old);
  iNewLen := Length(New);
  if (iStrLen > 0) and (iOldLen > 0) then begin
    if FullLineOnly then begin
      if StrTools.StartsText(Old, Str) then
	pBeg := PWideChar(Str) + iOldLen
      else pBeg := Pointer(Str);
    end
    else pBeg := Pointer(Str);
    if iNewLen > iOldLen then
      SetString(Result, NIL, iStrLen + MulDiv(iStrLen, iNewLen - iOldLen, iOldLen))
    else SetString(Result, NIL, iStrLen);
    pStrEnd := PWideChar(Str) + iStrLen;
    pResult := Pointer(Result);
    pOld := Pointer(Old);
    pNew := Pointer(New);
    repeat
      pEnd := StrTools.AnsiTextPos(pBeg, pOld);
      iLen := pEnd - pBeg;
      if iLen >= 0 then begin
	if iLen > 0 then begin
	  if FullLineOnly and not (AnsiChar((pEnd - 1)^) in WSP) then inc(iLen, iOldLen);
	  StrTools.StrLCopy(pResult, pBeg, iLen);
	  inc(pResult, iLen);
	end;
	if (iNewLen > 0) and (not FullLineOnly or (AnsiChar((pEnd - 1)^) in WSP)) then begin
	  StrTools.StrLCopy(pResult, pNew, iNewLen);
	  inc(pResult, iNewLen);
	end;
	pBeg := pEnd + iOldLen;
      end
      else begin
	if pBeg = Pointer(Str) then begin
	  Result := Str;
	  Exit;
	end
	else if pBeg > Pointer(Str) then begin
	  iLen := pStrEnd - pBeg;
	  StrTools.StrLCopy(pResult, pBeg, iLen);
	  inc(pResult, iLen);
	end;
	Break;
      end;
    until pBeg > pStrEnd;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
    Exit;
  end;
  Result := Str;
end;

function ReplaceTextAtPos(const Str, New: RawByteString; OldPos, OldLen: Integer): RawByteString; inline;
begin
  Result := ReplaceStrAtPos(Str, New, OldPos, OldLen);
end;

function ReplaceTextAtPos(const Str, New: UnicodeString; OldPos, OldLen: Integer): UnicodeString; inline;
begin
  Result := ReplaceStrAtPos(Str, New, OldPos, OldLen);
end;

function GetSystemFileVersion: Cardinal;
begin
  Result := GetSimpleFileVersion(GetSysDir + 'Kernel32.dll');
end;

function GetSimpleFileVersion(const FileName: String; MainOnly: Longbool = FALSE): Cardinal;
var
  fvi: TFileVersionInfo;
  iFirstSepPos, iNextSepPos: Integer;
begin
  FillChar(fvi, SizeOf(TFileVersionInfo), 0);
  with fvi do begin
    QueryFile := FileName;
    Include(QueryParameters, fviFileVersion);
    if FileVersionInfo(fvi) then begin
      iFirstSepPos := 1;
      while CharInSet(FileVersion[iFirstSepPos], DIGITS) do inc(iFirstSepPos);
      if iFirstSepPos > 1 then begin
	iNextSepPos := succ(iFirstSepPos);
	while CharInSet(FileVersion[iNextSepPos], DIGITS) do inc(iNextSepPos);
	Result := StrToIntDef(Copy(FileVersion, 1, pred(iFirstSepPos)), 0);
	if MainOnly then Exit;
	Result := (Result shl 16) or Cardinal(StrToIntDef(Copy(FileVersion, succ(iFirstSepPos), pred(iNextSepPos - iFirstSepPos)), 0));
	Exit;
      end;
    end;
  end;
  Result := 0;
end;

function WinVersion(Simple: Longbool; AddNumber: Longbool = FALSE; AddSvcPack: Longbool = FALSE; AddCpuType: Longbool = FALSE): String;
const
  REGKEY_WINVER = REGKEY_SW + REGKEY_MS + REGKEY_WIN_NT + REGKEY_CV;
var
  vi: TOSVersionInfoExW;
  si: TSystemInfo;
  fv: Cardinal;

  function GetSimpleVersion(var VersionInfo: TOSVersionInfoExW): Cardinal;
  begin
    with VersionInfo do begin
      if dwMajorVersion >= 6 then begin
	Result := GetSystemFileVersion;
	dwMajorVersion := Result shr 16;
	dwMinorVersion := Result and $FFFF;
	if dwMajorVersion = 6 then begin
	  case dwMinorVersion of
	    1: begin
	      dwMajorVersion := 7;
	      dwMinorVersion := 0;
	    end;
	    2: begin
	      dwMajorVersion := 8;
	      dwMinorVersion := 0;
	    end;
	    3: begin
	      dwMajorVersion := 8;
	      dwMinorVersion := 1;
	    end;
	  end;
	end;
      end;
      Result := (dwMajorVersion shl 16) or dwMinorVersion;
    end;
  end;

begin
  with vi do begin
    FillChar(vi, SizeOf(vi), 0);
    dwOSVersionInfoSize := SizeOf(vi);
    if not GetVersionEx(vi) then begin
      dec(dwOSVersionInfoSize, 8);
      GetVersionEx(vi);
    end;
    if Simple then begin
      if dwPlatformId = VER_PLATFORM_WIN32_NT then GetSimpleVersion(vi);
      Result := IntToStr(dwMajorVersion) + DOT + IntToStr(dwMinorVersion);
    end
    else begin
      case dwPlatformId of
	VER_PLATFORM_WIN32s: Result := '32s';
	VER_PLATFORM_WIN32_WINDOWS: begin
	  case dwMinorVersion of
	     0: Result := '95';
	    10: Result := '98';
	    90: Result := 'ME';
	  end;
	end;
	VER_PLATFORM_WIN32_NT: begin
	  if dwMajorVersion >= 6 then begin
	    fv := GetSimpleVersion(vi);
	    dwMajorVersion := fv shr 16;
	    dwMinorVersion := fv and $FFFF;
	    if dwMajorVersion < 10 then
	      Result := GetRegString(HKEY_LOCAL_MACHINE, REGKEY_WINVER, 'ProductName') + SP
	    else Result := 'Windows' + SP;
	    Result := Result + CardToStrW(dwMajorVersion) + DOT + CardToStrW(dwMinorVersion) + SP +
		      GetRegString(HKEY_LOCAL_MACHINE, REGKEY_WINVER, 'DisplayVersion') + ' Build ' +
		      GetRegString(HKEY_LOCAL_MACHINE, REGKEY_WINVER, 'CurrentBuild');
	    AddNumber := FALSE;
	    AddCpuType := FALSE;
	  end;
	  case dwMajorVersion of
	    5: case dwMinorVersion of
	      0: Result := '2K';
	      1: Result := 'XP';
	      2: Result := 'S3';
	    end;
	  else
	    if dwMajorVersion < 5 then Result := 'NT';
	  end;
	end;
      end;
      if dwMajorVersion < 6 then Result := 'Win' + SP + Result;
    end;
    if AddNumber then Result := Result + SP + LB + IntToStr(dwMajorVersion) + DOT + IntToStr(dwMinorVersion) + RB;
    if AddSvcPack then Result := Result + ' SP ' + IntToStr(wServicePackMajor);
    if AddCpuType then begin
      if dwMajorVersion >= 5 then
	GetNativeSystemInfo(si)
      else GetSystemInfo(si);
      case si.wProcessorArchitecture of
	PROCESSOR_ARCHITECTURE_INTEL..PROCESSOR_ARCHITECTURE_ARM: Result := Result + ' - 32bit';
	PROCESSOR_ARCHITECTURE_IA64..PROCESSOR_ARCHITECTURE_ALPHA64,
	PROCESSOR_ARCHITECTURE_AMD64..PROCESSOR_ARCHITECTURE_IA32_ON_WIN64: Result := Result + ' - 64bit';
	PROCESSOR_ARCHITECTURE_MSIL: Result := Result + ' - MSIL';
      else
	Result := Result + ' - ????';
      end;
    end;
  end;
end;

function IsMinWinVersion(const MinWinVersion: String): Longbool;

  function GetMajorVersion(const Version: String; SepPos: Integer): Cardinal;
  begin
    Result := StrToIntDef(Copy(Version, 1, pred(SepPos)), 0);
  end;

  function GetMinorVersion(const Version: String; SepPos: Integer): Cardinal;
  begin
    Result := StrToIntDef(Copy(Version, succ(SepPos), MAXINT), 0);
  end;

var
  sCurrentVersion: String;
  iMinPos, iCurPos: Integer;
  cMinVersion, cCurVersion: Cardinal;
begin
  sCurrentVersion := WinVersion(TRUE);
  if (MinWinVersion <> E) and (sCurrentVersion <> E) then begin
    iMinPos := 1;
    while CharInSet(MinWinVersion[iMinPos], DIGITS) do inc(iMinPos);
    iCurPos := 1;
    while CharInSet(sCurrentVersion[iCurPos], DIGITS) do inc(iCurPos);
    cMinVersion := GetMajorVersion(MinWinVersion, iMinPos) shl 16 + GetMinorVersion(MinWinVersion, iMinPos);
    cCurVersion := GetMajorVersion(sCurrentVersion, iCurPos) shl 16 + GetMinorVersion(sCurrentVersion, iCurPos);
    Result := (cCurVersion >= cMinVersion);
    Exit;
  end;
  Result := FALSE;
end;

function MainWinVersion: Cardinal;
begin
  Result := StrToCardEx(WinVersion(TRUE));
end;

function ComputerName: String;
var
  dwLen: DWord;
begin
  dwLen := MAX_COMPUTERNAME_LENGTH;
  SetString(Result, NIL, dwLen);
  if GetComputerName(Pointer(Result), dwLen) then
    Delete(Result, succ(dwLen), MAXINT)
  else Result := E;
end;

function UserName: String;
var
  dwLen: DWord;
begin
  dwLen := 0;
  WNetGetUser(NIL, NIL, dwLen);
  SetString(Result, NIL, pred(dwLen));
  WNetGetUser(NIL, Pointer(Result), dwLen);
end;

function SafeCallErrorString(ErrorCode: HResult; ErrorAddr: Pointer): String;
begin
  Result := AppendHexAddress(AppendHexError(SysErrorString(ErrorCode), ErrorCode), ErrorAddr);
end;

function AppendHexAddress(const AText: String; Address: Pointer): String;
begin
  if AText <> E then begin
    Result := AText + SP + AT + CardToHexW(UInt64(Address), 8);
    Exit;
  end;
  Result := E;
end;

function AppendHexError(const AText: String; Error: Cardinal): String;
begin
  if AText <> E then begin
    Result := AText + SP + LB + HEX_MARKER + CardToHexW(Error, 8) + RB;
    Exit;
  end;
  Result := E;
end;

function AppendPeriod(const Str: String): String;
begin
  if (Str <> E) and (Str[Length(Str)] <> DOT) then
    Result := Str + DOT
  else Result := Str;
end;

function SysErrorString(Code: Cardinal; Module: THandle = 0): WideString;
const
  DEFAULT_FLAGS = FORMAT_MESSAGE_ALLOCATE_BUFFER or FORMAT_MESSAGE_FROM_SYSTEM;
  MODULE_FLAG = FORMAT_MESSAGE_FROM_HMODULE;
  E_UNSPECIFIED = $80000008;
var
  pResult: PChar;
  ei: IErrorInfo;
begin
  if (Code and not MINLONG) <> ERROR_SUCCESS then begin
    if FormatMessage(DEFAULT_FLAGS or Ord(Module <> 0) * MODULE_FLAG, Pointer(Module), Code, 0, @pResult, 0, NIL) <> 0 then begin
      try
	Result := pResult;
      finally
	LocalFree(HLOCAL(pResult));
      end;
    end
    else if (GetErrorInfo(0, ei) = S_OK) then begin
      ei.GetDescription(Result);
      SetErrorInfo(0, ei);
    end;
    Result := TrimCustom(Result, TRIMCHARS + [DOT]);
    if (Result = E) and (Code <> E_UNSPECIFIED) then begin
      if Code < MINLONG then
	Result := SysErrorString(E_UNSPECIFIED) + SP + LB + HT + IntToStr(Code) + RB
      else Result := SysErrorString(E_UNSPECIFIED) + SP + LB + HEX_MARKER + IntToHex(Code, 8) + RB;
    end;
    Exit;
  end;
  Result := E;
end;

function EnumFontsProc(const ELFE: TEnumLogFontEx; const NTM: TNewTextMetric;
		       FontType: Cardinal; PLF: PLogFont): Integer; stdcall;
begin
  Result := 1;
  if (FontType and TRUETYPE_FONTTYPE) = TRUETYPE_FONTTYPE then with ELFE.elfLogFont do begin
    if not (lfCharset in [SYMBOL_CHARSET, MAC_CHARSET]) and (lfWeight = 400)
    and (lfItalic + lfUnderline + lfStrikeOut = 0) then with NTM do begin
      if (tmInternalLeading >= 0) and ((ntmFlags and NTM_REGULAR) <> 0)
      and ((tmDescent > 0) and ((tmAscent div tmDescent) >= 3)) then begin
	PLF^ := ELFE.elfLogFont;
	Result := 0;
      end;
    end;
  end;
end;

function SysFontInfo(const FontName: String): TExtLogFont;
var
  ftSys: TFont;
  lf: TLogFont;
  cotmSize: Cardinal;
  potm: POutlineTextmetric;
  hScreenDC, hFont: THandle;
begin
  FillChar(Result, SizeOf(Result), 0);
  if FontName <> E then begin
    hFont := 0;
    hScreenDC := 0;
    ftSys := TFont.Create;
    if ftSys <> NIL then begin
      try
	hScreenDC := GetDC(0);
	ftSys.Name := FontName;
	hFont := SelectObject(hScreenDC, ftSys.Handle);
	lf := Result.elfLogFont;
	lf.lfCharset := DEFAULT_CHARSET;
	StrTools.StrPLCopy(lf.lfFaceName, FontName, LF_FACESIZE);
	EnumFontFamiliesEx(hScreenDC, lf, @EnumFontsProc, LPARAM(@Result.elfLogFont), 0);
	cotmSize := GetOutlineTextMetrics(hScreenDC, 0, NIL);
	if cotmSize > 0 then begin
	  potm := AllocMem(cotmSize);
	  if potm <> NIL then begin
	    try
	      if GetOutlineTextMetrics(hScreenDC, cotmSize, potm) <> 0 then Result.elfPanose := potm^.otmPanoseNumber;
	    finally
	      FreeMem(potm);
	    end;
	  end;
	end;
      finally
	if hFont <> 0 then SelectObject(hScreenDC, hFont);
	ReleaseDC(WindowFromDC(hScreenDC), hScreenDC);
	ftSys.Free;
      end;
    end;
  end;
end;

function EnumFontFamilyProc(const ELFE: TEnumLogFontEx; const NTM: TNewTextMetric;
			   FontType: Cardinal; const Fonts: TStrings): Integer; stdcall;
var
  sFont: String;
begin
  Result := 1;
  if (FontType and TRUETYPE_FONTTYPE) = TRUETYPE_FONTTYPE then begin
    sFont := ELFE.elfFullName;
    if Fonts.IndexOf(sFont) = -1 then Fonts.Add(sFont);
  end;
end;

function EnumFontFamExProc(const ELFE: TEnumLogFontEx; const NTM: TNewTextMetric;
			   FontType: Cardinal; const Fonts: TStrings): Integer; stdcall;
const
  NAMED_STYLES: Array[0..6] of String = ('Bold', 'Cond', 'Heavy', 'Italic', 'Light', 'Script', 'Symbol');
var
  sFont: String;
begin
  Result := 1;
  if (FontType and TRUETYPE_FONTTYPE) = TRUETYPE_FONTTYPE then with ELFE.elfLogFont do begin
    if (lfWeight = 400) and (lfItalic + lfUnderline + lfStrikeOut = 0) then with NTM do begin
      if (tmInternalLeading >= 0) and ((ntmFlags and NTM_REGULAR) <> 0)
      and ((tmDescent > 0) and ((tmAscent div tmDescent) >= 3)) then begin
	sFont := lfFaceName;
	if not ((sFont = E) or (PChar(sFont)^ = AT) or (StrTools.IncludesText(sFont, NAMED_STYLES) > -1))
	and (Fonts.IndexOf(sFont) = -1) then begin
	  Fonts.Add(sFont);
	end;
      end;
    end;
  end;
 end;

function SysFontNames(const Fonts: TStrings; const FamilyNames: Array of String): Integer;
var
  lf: TLogFont;
  hScreenDC: THandle;
  iIndex: Integer;
begin
  if Fonts <> NIL then begin
    hScreenDC := GetDC(0);
    Fonts.BeginUpdate;
    try
      FillChar(lf, SizeOf(lf), 0);
      lf.lfCharSet := ANSI_CHARSET;
      if Length(FamilyNames) > 0 then begin
	for iIndex := Low(FamilyNames) to High(FamilyNames) do begin
	  StrTools.StrPLCopy(lf.lfFaceName, FamilyNames[iIndex], LF_FACESIZE);
	  EnumFontFamiliesEx(hScreenDC, lf, @EnumFontFamilyProc, LPARAM(Fonts), 0);
	end;
      end
      else EnumFontFamiliesEx(hScreenDC, lf, @EnumFontFamExProc, LPARAM(Fonts), 0);
    finally
      ReleaseDC(WindowFromDC(hScreenDC), hScreenDC);
      Fonts.EndUpdate;
    end;
    Result := Fonts.Count;
  end
  else Result := -1;
end;

function SysFontNameAndSize(NCLFontType: TNCLFontType): TNCLFontNameSize;
var
  ncm: TNonClientMetrics;
begin
  with Result do begin
    fnsName := E;
    fnsSize := 0;
  end;
  ncm.cbSize := SizeOf(ncm);
  if SystemParametersInfo(SPI_GETNONCLIENTMETRICS, SizeOf(ncm), @ncm, 0) then with ncm, Result do begin
    case NCLFontType of
      ftCaption: begin
	fnsName := lfCaptionFont.lfFaceName;
	fnsSize := lfCaptionFont.lfheight;
      end;
      ftSmCaption: begin
	fnsName := lfSmCaptionFont.lfFaceName;
	fnsSize := lfSmCaptionFont.lfheight;
      end;
      ftMenu: begin
	fnsName := lfMenuFont.lfFaceName;
	fnsSize := lfMenuFont.lfheight;
      end;
      ftStatus: begin
	fnsName := lfStatusFont.lfFaceName;
	fnsSize := lfStatusFont.lfheight;
      end;
      ftMessage: begin
	fnsName := lfMessageFont.lfFaceName;
	fnsSize := lfMessageFont.lfheight;
      end;
    end;
  end;
  with Result do if fnsSize <> 0 then fnsSize := PointsFromPixels(fnsSize);
end;

function WinFixFont(const DefaultFont: String = FONT_CONSOLE): String;
const
  Outdated = 'Courier';
var
  lf: TLogFont;
begin
  if GetObject(GetStockObject(ANSI_FIXED_FONT), SizeOf(lf), @lf) <> 0 then Result := lf.lfFaceName;
  if (DefaultFont <> E) and StrTools.SameText(lf.lfFaceName, Outdated) then Result := DefaultFont;
  if StrTools.SameText(Result, Outdated) then Result := FONT_COURIER;
end;

function WinVarFont: String;
var
  ncm: TNonClientMetrics;
begin
  FillChar(ncm, SizeOf(ncm), 0);
  ncm.cbSize := SizeOf(ncm);
  if SystemParametersInfo(SPI_GETNONCLIENTMETRICS, SizeOf(ncm), @ncm, 0) then Result := ncm.lfMessageFont.lfFaceName;
  if not StrTools.SameText(Result, FONT_SEGOE) then Result := FONT_ARIAL;
end;

function DateString(const ST: TSystemTime; const Format: String): String;
var
  iLen: Integer;
begin
  iLen := GetDateFormat(LOCALE_USER_DEFAULT, 0, @ST, PChar(Format), NIL, 0);
  SetString(Result, NIL, iLen);
  Delete(Result, GetDateFormat(LOCALE_USER_DEFAULT, 0, @ST, PChar(Format), Pointer(Result), iLen), MAXINT);
end;

function TimeString(const ST: TSystemTime; const Format: String): String;
var
  iLen: Integer;
begin
  iLen := GetTimeFormat(LOCALE_USER_DEFAULT, 0, @ST, PChar(Format), NIL, 0);
  SetString(Result, NIL, iLen);
  Delete(Result, GetTimeFormat(LOCALE_USER_DEFAULT, 0, @ST, PChar(Format), Pointer(Result), iLen), MAXINT);
end;

function RegOpenKeyRead(RegKey: THandle; const Path: String): HKEY;
begin
  if RegOpenKeyEx(RegKey, PChar(Path), 0, KEY_READ, Result) <> S_OK then Result := 0;
end;

function UniqueString(const Str: RawByteString): RawByteString;
begin
  Result := Str;
  System.UniqueString(AnsiString(Result));
end;

function RegEntryExists(RegKey: THandle; const Path: String): Longbool;
var
  hRegKey: HKEY;
begin
  hRegKey := RegOpenKeyRead(RegKey, Path);
  if hRegKey <> 0 then begin
    RegCloseKey(hRegKey);
    Result := TRUE;
    Exit;
  end;
  Result := FALSE;
end;

function GetRegEntryType(RegKey: THandle; const Path, Entry: String): Cardinal;
var
  hRegKey: HKEY;
begin
  hRegKey := RegOpenKeyRead(RegKey, Path);
  if hRegKey <> 0 then begin
    try
      RegQueryValueEx(hRegKey, PChar(Entry), NIL, @Result, NIL, NIL);
      Exit;
    finally
      RegCloseKey(hRegKey);
    end;
  end;
  Result := REG_NONE;
end;

function GetRegBinary(RegKey: THandle; const Path, Entry: String): RawByteString;
var
  hRegKey: HKEY;
  dwType, dwSize: DWord;
begin
  hRegKey := RegOpenKeyRead(RegKey, Path);
  if hRegKey <> 0 then begin
    try
      dwType := 0;
      dwSize := 0;
      if (RegQueryValueEx(hRegKey, PChar(Entry), NIL, @dwType, NIL, @dwSize) = S_OK) and (dwType = REG_BINARY) then begin
	SetString(Result, NIL, dwSize);
	if RegQueryValueEx(hRegKey, PChar(Entry), NIL, NIL, Pointer(Result), @dwSize) = S_OK then Exit;
      end;
    finally
      RegCloseKey(hRegKey);
    end;
  end;
  Result := E;
end;

function SetRegBinary(RegKey: THandle; const Path, Entry: String; const Value: RawByteString): Longbool;
var
  hRegKey: HKEY;
begin
  if RegOpenKeyEx(RegKey, PChar(Path), 0, KEY_SET_VALUE, hRegKey) = S_OK then begin
    try
      Result := (RegSetValueEx(hRegKey, PChar(Entry), 0, REG_BINARY, Pointer(Value), Length(Value)) = S_OK);
      Exit;
    finally
      RegCloseKey(hRegKey);
    end;
  end;
  Result := FALSE;
end;

function GetRegDWord(RegKey: THandle; const Path, Entry: String): Cardinal;
var
  hRegKey: HKEY;
  dwType, dwSize: DWord;
begin
  hRegKey := RegOpenKeyRead(RegKey, Path);
  if hRegKey <> 0 then begin
    try
      dwType := 0;
      dwSize := 0;
      if (RegQueryValueEx(hRegKey, PChar(Entry), NIL, @dwType, NIL, @dwSize) = S_OK) and (dwType = REG_DWORD)
      and (RegQueryValueEx(hRegKey, PChar(Entry), NIL, NIL, @Result, @dwSize) = S_OK) then begin
	Exit;
      end;
    finally
      RegCloseKey(hRegKey);
    end;
  end;
  Result := 0;
end;

function SetRegDWord(RegKey: THandle; const Path, Entry: String; Value: Cardinal): Longbool;
var
  hRegKey: HKEY;
begin
  if RegOpenKeyEx(RegKey, PChar(Path), 0, KEY_SET_VALUE, hRegKey) = S_OK then begin
    try
      Result := (RegSetValueEx(hRegKey, PChar(Entry), 0, REG_DWORD, @Value, SizeOf(Value)) = S_OK);
      Exit;
    finally
      RegCloseKey(hRegKey);
    end;
  end;
  Result := FALSE;
end;

function GetRegString(RegKey: THandle; const Path, Entry: String): String;
var
  hRegKey: HKEY;
  dwSize: DWord;
begin
  hRegKey := RegOpenKeyRead(RegKey, Path);
  if hRegKey <> 0 then begin
    try
      dwSize := 0;
      if RegQueryValueEx(hRegKey, PChar(Entry), NIL, NIL, NIL, @dwSize) = S_OK then begin
	SetString(Result, NIL, pred(dwSize div SizeOf(Char)));
	if RegQueryValueEx(hRegKey, PChar(Entry), NIL, NIL, Pointer(Result), @dwSize) = S_OK then Exit;
      end;
    finally
      RegCloseKey(hRegKey);
    end;
  end;
  Result := E;
end;

function SetRegString(RegKey: HKEY; const Path, Entry, Value: String): Longbool;
var
  hRegKey: HKEY;
  dwSize: DWord;
begin
  if RegOpenKeyEx(RegKey, PChar(Path), 0, KEY_SET_VALUE, hRegKey) = S_OK then begin
    try
      dwSize := succ(Length(Value)) * SizeOf(Char);
      if RegSetValueEx(hRegKey, PChar(Entry), 0, REG_SZ, PChar(Value), dwSize) = S_OK then begin
	Result := TRUE;
	Exit;
      end;
    finally
      RegCloseKey(hRegKey);
    end;
  end;
  Result := FALSE;
end;

function GetIniBool(const IniFile, Section, Item: String): Longbool;
var
  sl: TStringList;
begin
  if Section = E then begin
    sl := TStringList.Create;
    if sl <> NIL then with sl do begin
      try
	LoadFromFile(IniFile);
	if Count > 0 then begin
	  Result := (StrToIntDef(Values[Item], 0) <> 0);
	  Exit;
	end;
      finally
	sl.Free;
      end;
    end;
    Result := FALSE;
  end
  else Result := (GetPrivateProfileInt(PChar(Section), PChar(Item), 0, PChar(IniFile)) <> 0);
end;

function SetIniBool(const IniFile, Section, Item: String; Value: Longbool): Longbool;
begin
  Result := WritePrivateProfileString(PChar(Section), PChar(Item), PChar(IntToStr(ord(Value = true))), PChar(IniFile));
end;

function GetIniDWord(const IniFile, Section, Item: String): Cardinal;
var
  sl: TStringList;
begin
  if Section = E then begin
    sl := TStringList.Create;
    if sl <> NIL then with sl do begin
      try
	LoadFromFile(IniFile);
	if Count > 0 then begin
	  Result := StrToIntDef(Values[Item], 0);
	  Exit;
	end;
      finally
	sl.Free;
      end;
    end;
    Result := 0;
  end
  else Result := GetPrivateProfileInt(PChar(Section), PChar(Item), 0, PChar(IniFile))
end;

function SetIniDWord(const IniFile, Section, Item: String; Value: Cardinal): Longbool;
begin
  Result := WritePrivateProfileString(PChar(Section), PChar(Item), PChar(IntToStr(Value)), PChar(IniFile));
end;

function GetIniString(const IniFile, Section, Item: String; FullItem: Longbool = FALSE): String;
var
  iIndex, iLen: Integer;
  sl: TStringList;
begin
  if Section = E then begin
    sl := TStringList.Create;
    if sl <> NIL then with sl do begin
      try
	LoadFromFile(IniFile);
	if Count > 0 then begin
	  iIndex := IndexOfName(Item);
	  if FullItem then
	    Result := Strings[iIndex]
	  else Result := ValueFromIndex[iIndex];
	  Exit;
	end;
      finally
	sl.Free;
      end;
    end;
    Result := E;
  end
  else begin
    iLen := $FFFF;
    SetString(Result, NIL, iLen);
    iLen := GetPrivateProfileString(PChar(Section), PChar(Item), E, Pointer(Result), iLen, PChar(IniFile));
    Delete(Result, succ(iLen), MAXINT);
  end;
end;

function SetIniString(const IniFile, Section, Item, Value: String): Longbool;
begin
  Result := WritePrivateProfileString(PChar(Section), PChar(Item), PChar(Value), PChar(IniFile));
end;

function ReadIniFileSection(const IniFile, Section: String; List: TStrings): Cardinal;
var
  iLen, iIndex: Integer;
  sSection: String;
  pSection: PWideChar;
begin
  Result := 0;
  if List <> NIL then begin
    iLen := FileSize(IniFile);
    SetString(sSection, NIL, iLen);
    iLen := GetPrivateProfileSection(PChar(Section), Pointer(sSection), iLen, PChar(IniFile));
    Delete(sSection, succ(iLen), MAXINT);
    if iLen > 0 then with List do begin
      pSection := Pointer(sSection);
      BeginUpdate;
      try
	while pSection^ > NUL do begin
	  iIndex := Add(pSection);
	  if iIndex > -1 then begin
	    iLen := Length(Strings[iIndex]);
	    if iLen > 0 then inc(pSection, succ(iLen));
	  end;
	end;
      finally
	EndUpdate;
      end;
    end;
    Result := List.Count;
  end;
end;

function ReadIniFileSection(const IniFile, Section: String): String;
var
  iLen: Integer;
begin
  if FileExists(IniFile) then begin
    iLen := FileSize(IniFile);
    SetString(Result, NIL, iLen);
    iLen := GetPrivateProfileSection(PChar(Section), Pointer(Result), iLen, PChar(IniFile));
    Delete(Result, succ(iLen), MAXINT);
    Exit;
  end;
  Result := E;
end;

function WriteIniFileSection(const FileName, Section, Data: String): Longbool;
begin
  Result := WritePrivateProfileSection(PChar(Section), PChar(Data), PChar(FileName));
end;

function SystemBrowserPath: String;
const
  REGKEY_CLIENTS_INTERNET_COMMAND = REGKEY_CLIENTS_INTERNET + '\%s\' + REGKEY_SHELL_OPEN_CMD;

  function ReadBrowserEntry(RootKey: THandle; var Browser: String): String;
  begin
    Result := E;
    if Browser = E then Browser := GetRegString(RootKey, REGKEY_CLIENTS_INTERNET, E);
    if Browser <> E then Result := Unquote(GetRegString(RootKey, StrTools.Format(REGKEY_CLIENTS_INTERNET_COMMAND, [Browser]), E));
  end;

  function ReadAssocStrExecutable(const AssocStr: String): String;
  var
    dwSize: DWord;
  begin
    dwSize := 0;
    if AssocQueryString(0, ASSOCSTR_EXECUTABLE, PChar(AssocStr), NIL, NIL, @dwSize) = S_FALSE then begin
      SetString(Result, NIL, pred(dwSize));
      if AssocQueryString(0, ASSOCSTR_EXECUTABLE, PChar(AssocStr), NIL, Pointer(Result), @dwSize) = S_OK then Exit;
    end;
    Result := E;
  end;

  function ReadHttpHandler: String;
  begin
    Result := ReadAssocStrExecutable('https');
    if Result = E then Result := ReadAssocStrExecutable('http');
  end;

  function ReadHtmlHandler: String;
  begin
    Result := ReadAssocStrExecutable('html');
    if Result = E then Result := ReadAssocStrExecutable('htm');
  end;

var
  sBrowser: String;
begin
  Result := ReadHttpHandler;
  if Result = E then Result := ReadHtmlHandler;
  if Result = E then Result := ReadBrowserEntry(HKEY_CURRENT_USER, sBrowser);
  if Result = E then Result := ReadBrowserEntry(HKEY_LOCAL_MACHINE, sBrowser);
end;

function ResString(ID: Cardinal): String;
begin
  Result := ModuleString(HInstance, ID);
end;

function ClassString(Window: THandle; ScrollbarParent: Longbool = FALSE): String;
begin
  SetString(Result, NIL, MAX_PATH);
  Delete(Result, succ(GetClassName(Window, Pointer(Result), MAX_PATH)), MAXINT);
  StrUpper(Result);
  if ScrollbarParent and StrTools.SameText(Result, 'SCROLLBAR') then Result := ClassString(GetParent(Window));
end;

function WindowStringA(Window: THandle): RawByteString;
var
  iLen: Integer;
begin
  iLen := succ(GetWindowTextLengthA(Window));
  SetString(Result, NIL, iLen);
  Delete(Result, succ(GetWindowTextA(Window, Pointer(Result), iLen)), MAXINT);
end;

function WindowStringW(Window: THandle): UnicodeString;
var
  iLen: Integer;
begin
  iLen := succ(GetWindowTextLengthW(Window));
  SetString(Result, NIL, iLen);
  Delete(Result, succ(GetWindowTextW(Window, Pointer(Result), iLen)), MAXINT);
end;

function DialogString(Dialog: THandle; ID: Integer): String;
begin
  Result := WindowStringW(GetDlgItem(Dialog, ID));
end;

function EnvironmentString(const Str: String): String;
var
  sVariable: String;
  iLen: Integer;
begin
  if Str <> E then begin
    iLen := Length(Str);
    if (Str[1] <> PC) and (Str[iLen] <> PC) then
      sVariable := PC + Str + PC
    else sVariable := Str;
    Result := ExpandEnvironmentString(sVariable);
    Exit;
  end;
  Result := E;
end;

function ExpandEnvironmentString(const Str: String): String;
var
  iLen: Integer;
begin
  if Str <> E then begin
    iLen := ExpandEnvironmentStrings(PChar(Str), NIL, 0);
    SetString(Result, NIL, pred(iLen));
    iLen := ExpandEnvironmentStrings(PChar(Str), Pointer(Result), iLen);
    Delete(Result, iLen, 1);
    Exit;
  end;
   Result := E;
end;

function ModuleString(Module: THandle; ID: Cardinal): String;
var
  iLen: Integer;
begin
  iLen := $FFFF;
  SetString(Result, NIL, iLen);
  Delete(Result, succ(LoadString(Module, ID, Pointer(Result), iLen)), MAXINT);
end;

function MenuString(Menu: THandle; ID, Flags: Cardinal): String;
var
  iLen: Integer;
begin
  iLen := GetMenuString(Menu, ID, NIL, 0, Flags);
  if iLen > 0 then begin
    SetString(Result, NIL, iLen);
    GetMenuString(Menu, ID, Pointer(Result), succ(iLen), Flags);
    Exit;
  end;
  Result := E;
end;

function ColorToHtml(const Value: Variant): String;
begin
  try
    case VarType(Value) of
      varOleStr, varStrArg, varString, varUString: Result := Value;
    else
      Result := ColorToWebColorStr(Value);
    end;
  except
    Result := E;
  end;
end;

function HtmlToColor(const Value: Variant): TColor;
begin
  try
    case VarType(Value) of
      varOleStr, varStrArg, varString, varUString: begin
	if StrTools.StartsStr(HT, Value) then
	  Result := WebColorStrToColor(Value)
	else Result := WebColorNameToColor(Value);
      end;
    else
      Result := Value;
    end;
  except
    Result := clNone;
  end;
end;

function RGBCharsToWebHexColor(PHexColor, PRGBColor: PAnsiChar): PAnsiChar;
// rgb(127, 128, 129) => #7F8081, percents allowed, prefix not required, returns StrEnd(RGBColor)
const
  INHERIT: PAnsiChar = 'inherit';
var
  iLength, iValue: Integer;
  bValue: Byte;
begin
  Result := PRGBColor;
  if (PHexColor <> NIL) and (PRGBColor <> NIL) then begin
    if StrTools.StrLIComp(PRGBColor, 'rgb', 3) = 0 then begin
      inc(PRGBColor, 3);
      if ToLower(PRGBColor^) = 'a' then inc(PRGBColor);
    end;
    if PRGBColor^ = LB then begin
      PHexColor^ := HT;
      inc(PHexColor);
      iLength := 0;
      repeat
	iValue := 0;
	inc(iLength);
	repeat
	  while (PRGBColor^ <> NUL) and not (PRGBColor^ in DIGITS) do inc(PRGBColor);
	  if PRGBColor = NUL then begin
	    Move(INHERIT^, PHexColor^, Length(INHERIT));
	    inc(PRGBColor, Length(INHERIT));
	    Result := PRGBColor;
	    Exit;
	  end;
	  while (PRGBColor^ <> NUL) and (PRGBColor^ in DIGITS) do begin
	    iValue := iValue * 10 + (Ord(PRGBColor^) - Ord('0'));
	    inc(PRGBColor);
	  end;
	  if PRGBColor = NUL then begin
	    Move(INHERIT^, PHexColor^, Length(INHERIT));
	    inc(PRGBColor, Length(INHERIT));
	    Result := PRGBColor;
	    Exit;
	  end;
	until PRGBColor^ in [NUL, PC, RB, CM, SL, SC];
	if PRGBColor^ = PC then begin
	  iValue := MulDiv(iValue, 255, 100);
	  if iValue > 255 then iValue := 255;
	end;
	bValue := iValue;
	inc(PHexColor, BytesToHexChars(@bValue, 1, PHexColor, TRUE));
      until iLength = 3;
      while not (PRGBColor^ in [NUL, RB, SC]) do inc(PRGBColor);
      if PRGBColor^ > NUL then inc(PRGBColor);
      Result := PRGBColor;
    end;
  end;
end;

function RGBStringToWebHexColor(const RGBColor: RawByteString; PRGBLength: PCardinal = NIL): RawByteString;
// rgb(127, 128, 129) => #7F8081, percents allowed, prefix not required, optionally returns RGB length
const
  INHERIT: PAnsiChar = 'inherit';
var
  pRGB, pStart, pResult: PAnsiChar;
  iLength, iValue: Integer;
  bValue: Byte;
begin
  if RGBColor <> E then begin
    pRGB := Pointer(RGBColor);
    iLength := 7;
    pStart := pRGB;
    if StrTools.StrLIComp(pRGB, 'rgb', 3) = 0 then begin
      inc(pRGB, 3);
      if ToLower(pRGB^) = 'a' then begin
	inc(pRGB);
	inc(iLength);
      end;
    end;
    if pRGB^ = LB then begin
      SetString(Result, NIL, iLength);
      pResult := Pointer(Result);
      pResult^ := HT;
      inc(pResult);
      iLength := 0;
      repeat
	iValue := 0;
	inc(iLength);
	repeat
	  while (pRGB^ <> NUL) and not (pRGB^ in DIGITS) do inc(pRGB);
	  if pRGB = NUL then begin
	    Result := INHERIT;
	    Exit;
	  end;
	  while (pRGB^ <> NUL) and (pRGB^ in DIGITS) do begin
	    iValue := iValue * 10 + (Ord(pRGB^) - Ord('0'));
	    inc(pRGB);
	  end;
	  if pRGB = NUL then begin
	    Result := INHERIT;
	    Exit;
	  end;
	until pRGB^ in [NUL, PC, RB, CM, SL, SC];
	if pRGB^ = PC then begin
	  iValue := MulDiv(iValue, 255, 100);
	  if iValue > 255 then iValue := 255;
	end;
	bValue := iValue;
	inc(pResult, BytesToHexChars(@bValue, 1, pResult, TRUE));
      until iLength = 3;
      if PRGBLength <> NIL then begin
	while not (pRGB^ in [NUL, RB, SC]) do inc(pRGB);
	PRGBLength^ := succ(pRGB - pStart);
      end;
    end;
    Exit;
  end;
  Result := INHERIT;
end;

function RGBStringToWebHexColor(const RGBColor: UnicodeString; PRGBLength: PCardinal = NIL): UnicodeString;
begin
  Result := String(RGBStringToWebHexColor(AnsiString(RGBColor), PRGBLength));
end;

function GetDocType(const PeekData: RawByteString): String;
const
  DOCTYPE = '!DOCTYPE';
var
  pStr, pTagBeg, pTagEnd: PAnsiChar;
begin
  pStr := PAnsiChar(RemoveLineBreaks(PeekData));
  if pStr <> NIL then begin
    pTagBeg := pStr;
    repeat
      while not (pTagBeg^ in [NUL, LT]) do inc(pTagBeg);
      if pTagBeg^ = LT then begin
	inc(pTagBeg);
	pTagEnd := pTagBeg;
	while not (pTagEnd^ in [NUL, LT, GT]) do inc(pTagEnd);
	case pTagEnd^ of
	  NUL: Break;
	  GT: begin
	    if StrIAtPos(DOCTYPE, pTagBeg, Length(DOCTYPE)) then begin
	      inc(pTagBeg, succ(Length(DOCTYPE)));
	      if StrIAtPos(DOC_HTML, pTagBeg, Length(DOC_HTML)) then
		Result := DOC_HTML
	      else if StrIAtPos(DOC_SVG, pTagBeg, Length(DOC_SVG)) then
		Result := DOC_SVG;
	      Exit;
	    end
	    else if StrIAtPos(DOC_SVG, pTagBeg, Length(DOC_SVG)) then begin
	      Result := DOC_SVG;
	      Exit;
	    end;
	  end;
	end;
	pTagBeg := pTagEnd;
	inc(pTagBeg);
      end
      else Break;
    until pTagBeg^ = NUL;
  end;
  Result := E;
end;

function HtmlDecode(const Html: RawByteString; NonAsciiOnly: Boolean = FALSE): RawByteString;
var
  pHtml, pResult: PAnsiChar;
  iLen, iVal: Integer;
begin
  if Html <> E then begin
    SetString(Result, NIL, Length(Html) * (Length(IntToStr(MAXBYTE)) + 3));
    pResult := Pointer(Result);
    pHtml := Pointer(Html);
    pResult^ := pHtml^;
    repeat
      while not SysUtils.CharInSet(pHtml^, [NUL, AM]) do begin
	pResult^ := pHtml^;
	inc(pResult);
	inc(pHtml);
      end;
      case pHtml^ of
	NUL: begin
	  pResult^ := NUL;
	  Break;
	end;
	AM: begin
	  if (pHtml + 1)^ = HT then begin
	    inc(pHtml, 2);
	    iLen := 0;
	    iVal := 0;
	    if LoChar(pHtml^) = 'x' then begin
	      inc(pHtml);
	      while SysUtils.CharInSet(AnsiChar(HiChar(pHtml[iLen])), HEXCHARS) do inc(iLen);
	      if pHtml[iLen] = SC then
		iVal := HexCharsValue(pHtml, iLen)
	      else dec(pHtml, 2);
	    end
	    else begin
	      while SysUtils.CharInSet(AnsiChar(pHtml[iLen]), DIGITS) do inc(iLen);
	      if pHtml[iLen] = SC then
		iVal := DecCharsValue(pHtml, iLen)
	      else dec(pHtml, 2);
	    end;
	    if (pHtml[iLen] = SC) and (iVal > MAXCHAR * Byte(NonAsciiOnly)) and (iVal <= MAXBYTE) then begin
	      pResult^ := AnsiChar(iVal);
	      inc(pHtml, succ(iLen));
	      inc(pResult);
	      Continue;
	    end
	    else begin
              while pHtml^ <> AM do dec(pHtml);
	      pResult^ := pHtml^;
	      inc(pResult);
	      inc(pHtml);
	    end;
	  end
	  else if not NonAsciiOnly then begin
	    iLen := 0;
	    inc(pHtml);
	    if StrIAtPos(pHtml, 'amp;', 4) then begin
	      pResult^ := AM;
	      iLen := 4;
	    end
	    else if StrIAtPos(pHtml, 'lt;', 3) then begin
	      pResult^ := LT;
	      iLen := 3;
	    end
	    else if StrIAtPos(pHtml, 'gt;' , 3) then begin
	      pResult^ := GT;
	      iLen := 3;
	    end
	    else if StrIAtPos(pHtml, 'tab;' , 3) then begin
	      pResult^ := TAB;
	      iLen := 3;
	    end
	    else if StrIAtPos(pHtml, 'apos;', 5) then begin
	      pResult^ := SQ;
	      iLen := 5;
	    end
	    else if StrIAtPos(pHtml, 'nbsp;', 5) then begin
	      pResult^ := NBSP;
	      iLen := 5;
	    end
	    else if StrIAtPos(pHtml, 'quot;', 5) then begin
	      pResult^ := QU;
	      iLen := 5;
	    end;
	    if iLen > 0 then begin
	      inc(pHtml, iLen);
	      inc(pResult);
	    end
	    else begin
	      pResult^ := (pHtml - 1)^;
	      inc(pResult);
	    end;
	  end
	  else begin
	    pResult^ := pHtml^;
	    inc(pResult);
	    inc(pHtml);
	  end;
	end;
      end;
    until pHtml^ = NUL;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := Html;
end;

function HtmlDecode(const Html: UnicodeString; NonAsciiOnly: Boolean = FALSE): UnicodeString;
var
  pHtml, pResult: PWideChar;
  iLen, iVal: Integer;
begin
  if Html <> E then begin
    SetString(Result, NIL, Length(Html) * (Length(IntToStr(MAXWORD)) + 3));
    pResult := Pointer(Result);
    pHtml := Pointer(Html);
    pResult^ := pHtml^;
    repeat
      while not CharInSet(pHtml^, [NUL, AM]) do begin
	pResult^ := pHtml^;
	inc(pResult);
	inc(pHtml);
      end;
      case pHtml^ of
	NUL: begin
	  pResult^ := NUL;
	  Break;
	end;
	AM: begin
	  if (pHtml + 1)^ = HT then begin
	    inc(pHtml, 2);
	    iLen := 0;
	    iVal := 0;
	    if LoChar(pHtml^) = 'x' then begin
	      inc(pHtml);
	      while CharInSet(WideChar(HiChar(pHtml[iLen])), HEXCHARS) do inc(iLen);
	      if pHtml[iLen] = SC then
		iVal := HexCharsValue(pHtml, iLen)
	      else dec(pHtml, 2);
	    end
	    else begin
	      while CharInSet(WideChar(pHtml[iLen]), DIGITS) do inc(iLen);
	      if pHtml[iLen] = SC then
		iVal := DecCharsValue(pHtml, iLen)
	      else dec(pHtml, 2);
	    end;
	    if (pHtml[iLen] = SC) and (iVal > MAXCHAR * Byte(NonAsciiOnly)) and (iVal <= MAXWORD) then begin
	      pResult^ := WideChar(iVal);
	      inc(pHtml, succ(iLen));
	      inc(pResult);
	      Continue;
	    end
	    else begin
	      while pHtml^ <> AM do dec(pHtml);
	      pResult^ := pHtml^;
	      inc(pResult);
	      inc(pHtml);
	    end;
	  end
	  else if not NonAsciiOnly then begin
	    iLen := 0;
	    inc(pHtml);
	    if StrIAtPos(pHtml, 'amp;', 4) then begin
	      pResult^ := AM;
	      iLen := 4;
	    end
	    else if StrIAtPos(pHtml, 'lt;', 3) then begin
	      pResult^ := LT;
	      iLen := 3;
	    end
	    else if StrIAtPos(pHtml, 'gt;' , 3) then begin
	      pResult^ := GT;
	      iLen := 3;
	    end
	    else if StrIAtPos(pHtml, 'tab;' , 3) then begin
	      pResult^ := TAB;
	      iLen := 3;
	    end
	    else if StrIAtPos(pHtml, 'apos;', 5) then begin
	      pResult^ := SQ;
	      iLen := 5;
	    end
	    else if StrIAtPos(pHtml, 'nbsp;', 5) then begin
	      pResult^ := NBSP;
	      iLen := 5;
	    end
	    else if StrIAtPos(pHtml, 'quot;', 5) then begin
	      pResult^ := QU;
	      iLen := 5;
	    end;
	    if iLen > 0 then begin
	      inc(pHtml, iLen);
	      inc(pREsult);
	    end
	    else begin
	      pResult^ := (pHtml - 1)^;
	      inc(pResult);
	    end;
	  end
	  else begin
	    pResult^ := pHtml^;
	    inc(pResult);
	    inc(pHtml);
	  end;
	end;
      end;
    until pHtml^ = NUL;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := Html;
end;

function HtmlEncode(const Html: RawByteString; NonAsciiOnly: Boolean = FALSE): RawByteString;
var
  pResult, pHtml, pCopy: PAnsiChar;
  iLen: Integer;
begin
  if Html <> E then begin
    SetString(Result, NIL, Length(Html) * (Length(IntToStr(MAXBYTE)) + 3));
    pResult := Pointer(Result);
    pHtml := Pointer(Html);
    repeat
      if Ord(pHtml^) > MAXCHAR then begin
	pCopy := PAnsiChar(AM + HT + CardToStrA(Ord(pHtml^)) + SC);
	pResult := StrLECopy(pResult, Pointer(pCopy), StrTools.StrLen(pCopy));
      end
      else begin
	if not NonAsciiOnly then begin
	  case pHtml^ of
	    TAB: begin
	      pCopy := '&tab;';
	      iLen := 5;
	    end;
	    QU: begin
	      pCopy := '&quot;';
	      iLen := 6;
	    end;
	    AM: begin
	      pCopy := '&amp;';
	      iLen := 5;
	    end;
	    SQ: begin
	      pCopy := '&apos;';
	      iLen := 6;
	    end;
	    LT: begin
	      pCopy := '&lt;';
	      iLen := 4;
	    end;
	    GT: begin
	      pCopy := '&gt;';
	      iLen := 4;
	    end;
	    NBSP: begin
	      pCopy := '&#xA0;';
	      iLen := 6;
	    end;
	  else
	    pResult^ := pHtml^;
	    inc(pResult);
	    pCopy := NIL;
	    iLen := 0;
	  end;
	  if iLen > 0 then pResult := StrLECopy(pResult, Pointer(pCopy), iLen);
	end
	else begin
	  pResult^ := pHtml^;
	  inc(pResult);
	end;
      end;
      if pHtml^ <> NUL then inc(pHtml);
    until pHtml^ = NUL;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := Html;
end;

function HtmlEncode(const Html: UnicodeString; NonAsciiOnly: Boolean = FALSE): UnicodeString;
var
  pResult, pHtml, pCopy: PWideChar;
  iLen: Integer;
begin
  if Html <> E then begin
    SetString(Result, NIL, Length(Html) * (Length(IntToStr(MAXWORD)) + 3));
    pResult := Pointer(Result);
    pHtml := Pointer(Html);
    repeat
      if Ord(pHtml^) > MAXCHAR then begin
	pCopy := PWideChar(AM + HT + CardToStrW(Ord(pHtml^)) + SC);
	pResult := StrLECopy(pResult, Pointer(pCopy), StrTools.StrLen(pCopy));
      end
      else begin
	if not NonAsciiOnly then begin
	  case pHtml^ of
	    TAB: begin
	      pCopy := '&tab;';
	      iLen := 5;
	    end;
	    QU: begin
	      pCopy := '&quot;';
	      iLen := 6;
	    end;
	    AM: begin
	      pCopy := '&amp;';
	      iLen := 5;
	    end;
	    SQ: begin
	      pCopy := '&#xA0;';
	      iLen := 6;
	    end;
	    LT: begin
	      pCopy := '&lt;';
	      iLen := 4;
	    end;
	    GT: begin
	      pCopy := '&gt;';
	      iLen := 4;
	    end;
	    NBSP: begin
	      pCopy := '&nbsp;';
	      iLen := 6;
	    end;
	  else
	    pResult^ := pHtml^;
	    inc(pResult);
	    pCopy := NIL;
	    iLen := 0;
	  end;
	  if iLen > 0 then pResult := StrLECopy(pResult, Pointer(pCopy), iLen);
	end
	else begin
	  pResult^ := pHtml^;
	  inc(pResult);
	end;
      end;
      if pHtml^ <> NUL then inc(pHtml);
    until pHtml^ = NUL;
    Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
  end
  else Result := Html;
end;

function RestoreNbsp(const Html: RawByteString): RawByteString;
const
  X160: PAnsiChar = '&#160;';
  X0A0: PAnsiChar = '&#xA0;';
  NBSP: PAnsiChar = '&nbsp;';
var
  pResult, pBeg, pEnd: PAnsiChar;
begin
  if Html <> E then begin
    SetString(Result, NIL, Length(Html));
    pResult := Pointer(Result);
    pBeg := Pointer(Html);
    pEnd := pBeg;
    repeat
      while not SysUtils.CharInSet(pEnd^, [NUL, LF, CR, AM]) do inc(pEnd);
      case pEnd^ of
	NUL: begin
	  pResult := StrLECopy(pResult, pBeg, pEnd - pBeg);
	  Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
	  Break;
	end;
	LF, CR: begin
	  pResult := StrLECopy(pResult, pBeg, pEnd - pBeg);
	  while SysUtils.CharInSet(pEnd^, [LF, CR]) do inc(pEnd);
	  pResult^ := SP;
	  inc(pResult);
	  pBeg := pEnd;
	end;
	AM: begin
	  if StrIAtPos(pEnd, X160, Length(X160)) or StrIAtPos(pEnd, NBSP, Length(NBSP)) then begin
	    pResult := StrLECopy(pResult, pBeg, pEnd - pBeg);
	    pResult := StrLECopy(pResult, X0A0, Length(X0A0));
	    inc(pEnd, Length(X0A0));
	    pBeg := pEnd;
	  end
	  else inc(pEnd);
	end;
      else
	inc(pEnd);
      end;
    until pEnd^ = NUL;
  end;
end;

function RestoreNbsp(const Html: UnicodeString): UnicodeString;
const
  X160: PWideChar = '&#160;';
  X0A0: PWideChar = '&#xA0;';
  NBSP: PWideChar = '&nbsp;';
var
  pResult, pBeg, pEnd: PWideChar;
begin
  if Html <> E then begin
    SetString(Result, NIL, Length(Html));
    pResult := Pointer(Result);
    pBeg := Pointer(Html);
    pEnd := pBeg;
    repeat
      while not CharInSet(pEnd^, [NUL, LF, CR, AM]) do inc(pEnd);
      case pEnd^ of
	NUL: begin
	  pResult := StrLECopy(pResult, pBeg, pEnd - pBeg);
	  Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
	  Break;
	end;
	LF, CR: begin
	  pResult := StrLECopy(pResult, pBeg, pEnd - pBeg);
	  while CharInSet(pEnd^, [LF, CR]) do inc(pEnd);
	  pResult^ := SP;
	  inc(pResult);
	  pBeg := pEnd;
	end;
	AM: begin
	  if StrIAtPos(pEnd, X160, Length(X160)) or StrIAtPos(pEnd, NBSP, Length(NBSP)) then begin
	    pResult := StrLECopy(pResult, pBeg, pEnd - pBeg);
	    pResult := StrLECopy(pResult, X0A0, Length(X0A0));
	    inc(pEnd, Length(X0A0));
	    pBeg := pEnd;
	  end
	  else inc(pEnd);
	end;
      else
	inc(pEnd);
      end;
    until pEnd^ = NUL;
  end;
end;

function RtfToString(const Rtf: RawByteString): RawByteString;
var
  rtc: TRtfTxtConverter;
begin
  if StrTools.StartsText(RAW(RTF_MARKER), Rtf) then begin
    rtc := TRtfTxtConverter.Create;
    if rtc <> NIL then with rtc do begin
      try
	Result := RAW(rtc.RtfToText(Rtf, FALSE));
      finally
	rtc.Free;
      end;
    end;
  end
  else Result := Rtf;
end;

function RtfToString(const Rtf: UnicodeString): UnicodeString;
var
  rtc: TRtfTxtConverter;
begin
  if StrTools.StartsText(RTF_MARKER, Rtf) then begin
    rtc := TRtfTxtConverter.Create;
    if rtc <> NIL then with rtc do begin
      try
	Result := rtc.RtfToText(RAW(Rtf), FALSE);
      finally
	rtc.Free;
      end;
    end;
  end
  else Result := Rtf;
end;

function StringToRtf(const Str: RawByteString): RawByteString;
var
  lf: TLogFont;
  rtfFont: TRtfFont;
  rtc: TRtfTxtConverter;
begin
  if Str <> E then begin
    if GetObject(GetStockObject(ANSI_VAR_FONT), SizeOf(TLogFont), @lf) <> 0 then begin
      with rtfFont do begin
	Name := lf.lfFaceName;
	Size := lf.lfHeight;
	Styles := 0;
      end;
    end;
    rtc := TRtfTxtConverter.Create;
    if rtc <> NIL then with rtc do begin
      try
	Result := rtc.TextToRtf(UnicodeString(Str), rtfFont);
      finally
	rtc.Free;
      end;
    end;
  end;
end;

function StringToRtf(const Str: UnicodeString): UnicodeString;
var
  lf: TLogFont;
  rtfFont: TRtfFont;
  rtc: TRtfTxtConverter;
begin
  if Str <> E then begin
    if GetObject(GetStockObject(ANSI_VAR_FONT), SizeOf(TLogFont), @lf) <> 0 then begin
      with rtfFont do begin
	Name := lf.lfFaceName;
	Size := lf.lfHeight;
	Styles := 0;
      end;
    end;
    rtc := TRtfTxtConverter.Create;
    if rtc <> NIL then with rtc do begin
      try
	Result := UnicodeString(rtc.TextToRtf(Str, rtfFont));
      finally
	rtc.Free;
      end;
    end;
  end;
end;

function PathCompact(const Path: RawByteString; MaxLen: Integer): RawByteString;
begin
  if Length(Path) <= MAX_PATH then begin
    SetLength(Result, MaxLen);
    PathCompactPathExA(Pointer(Result), Pointer(Path), succ(MaxLen), 0);
  end
  else Result := Copy(Path, 1, MaxLen - 4) + ' ...';
end;

function PathCompact(const Path: UnicodeString; MaxLen: Integer): UnicodeString;
begin
  if Length(Path) <= MAX_PATH then begin
    SetLength(Result, MaxLen);
    PathCompactPathExW(Pointer(Result), Pointer(Path), succ(MaxLen), 0);
  end
  else Result := Copy(Path, 1, MaxLen - 4) + ' ...';
end;

function FormatSettings(DecimalSeperator: Char = NUL): TFormatSettings;
begin
  {$IF CompilerVersion > 21.0}
  Result := TFormatSettings.Create;
  {$ELSE}
  GetLocaleFormatSettings(GetUserDefaultLCID, Result);
  {$IFEND}
  if DecimalSeperator <> NUL then Result.DecimalSeparator := DecimalSeperator;
end;

function Format(const Format: RawByteString; const Args: Array of TVarRec): RawByteString;
begin
  Result := AnsiStrings.Format(Format, Args);
end;

function Format(const Format: UnicodeString; const Args: Array of TVarRec): UnicodeString;
begin
  Result := SysUtils.Format(Format, Args);
end;

function Format(const Format: RawByteString; const Args: Array of TVarRec;
		const FormatSettings: TFormatSettings): RawByteString;
begin
  Result := AnsiStrings.Format(Format, Args, FormatSettings);
end;

function Format(const Format: UnicodeString; const Args: Array of TVarRec;
		const FormatSettings: TFormatSettings): UnicodeString;
begin
  Result := SysUtils.Format(Format, Args, FormatSettings);
end;

function GetHtmlSysFontFormat(const Html: String): String;
const
  DB = '<div style="font-family: %s; font-size: %dpt;">';
  DE = '</div>';
  FB = '<font face="%s"; size="2">';
  FE = '</font>';
begin
  if MainWinVersion < 6 then
    Result := StrTools.Format(FB, [SysFontNameAndSize(ftStatus).fnsName]) + Html + FE
  else with SysFontNameAndSize(ftStatus) do Result := StrTools.Format(DB, [fnsName, succ(fnsSize)]) + Html + DE;
end;

function GetHtmlSizeInt(const Size: String; Default: Integer = 0): Integer;
var
  pSize: PWideChar;
  rm: TFPURoundingMode;
  iLen: Integer;
begin
  Result := 0;
  if Size <> E then begin
    pSize := Pointer(Size);
    if (pSize^ = PLS) or (pSize^ = MNS) then inc(pSize);
    while CharInSet(pSize^, DIGITS + [DOT]) do inc(pSize);
    if pSize > Pointer(Size) then begin
      rm := SetRoundMode(rmNearest);
      try
	iLen := pSize - Pointer(Size);
	Result := Round(StrToFloatDef(Copy(Size, 1, iLen), Default, FormatSettings(DOT)));
	if iLen < Length(Size) then begin
	  if (pSize^ = PC) and ((pSize + 1)^ = NUL) then begin
	    Result := MulDiv(Result, Default, 100);
	    Exit;
	  end;
	  if StrTools.TextPos(pSize, PWideChar('pt')) <> NIL then begin
	    Result := PixelsFromPoints(Result);
	    Exit;
	  end;
	end;
      finally
	SetRoundMode(rm);
      end;
    end;
  end;
end;

function GetHtmlSizeFloat(const Size: String; Default: Integer = 0): Extended;
var
  pSize: PWideChar;
  IsFloat: Longbool;
  rm: TFPURoundingMode;
  iLen: Integer;
begin
  Result := 0;
  if Size <> E then begin
    pSize := Pointer(Size);
    if (pSize^ = PLS) or (pSize^ = MNS) then inc(pSize);
    IsFloat := FALSE;
    while CharInSet(pSize^, DIGITS + [DOT]) do begin
      if pSize^ = DOT then IsFloat := TRUE;
      inc(pSize);
    end;
    if pSize > Pointer(Size) then begin
      rm := SetRoundMode(rmNearest);
      try
	iLen := pSize - Pointer(Size);
	Result := StrToFloatDef(Copy(Size, 1, iLen), Default, FormatSettings(DOT));
	if IsFloat and (Default <> 0) then begin
	  Result := Result * Default;
	  Exit;
	end;
	if iLen < Length(Size) then begin
	  if (pSize^ = PC) and ((pSize + 1)^ = NUL) then begin
	    Result := (Result * Default)/100;
	    Exit;
	  end;
	  if StrTools.TextPos(pSize, PWideChar('pt')) <> NIL then begin
	    Result := PixelsFromPoints(Round(Result));
	    Exit;
	  end;
	end;
      finally
	SetRoundMode(rm);
      end;
    end;
  end;
end;

function RoundNearest(AValue: Extended): Integer;
var
  rm: TFPURoundingMode;
begin
  rm := SetRoundMode(rmNearest);
  try
    Result := Round(AValue);
  finally
    SetRoundMode(rm);
  end;
end;

function DivModPlus(Dividend: Cardinal; Divisor: Cardinal): Cardinal;
asm
  {$IFDEF CPUX64}
  // ECX = Dividend; EDX = Divisor; => adds 1 for remainder
  MOV	EAX,ECX
  MOV	ECX,EDX
  XOR	EDX,EDX
  DIV	ECX
  OR	EDX,EDX
  JE	@EXIT
  INC	EAX
  @EXIT:
  {$ELSE}
  // EAX = Dividend; EDX = Divisor; => adds 1 for remainder
  MOV	ECX,EDX
  XOR	EDX,EDX
  DIV	ECX
  OR	EDX,EDX
  JE	@EXIT
  INC	EAX
  @EXIT:
  {$ENDIF}
end;

procedure DivMod(Dividend: Cardinal; Divisor: Cardinal; var Result, Remainder: Cardinal); assembler;
asm
  {$IFDEF CPUX64}
  // RCX = Dividend; RDX = Divisor; R8 = @Result; R9 = @Remainder
  MOV	EAX,ECX
  MOV	ECX,EDX
  XOR	EDX,EDX
  DIV	ECX
  MOV	[R8],EAX
  MOV	[R9],EDX
  {$ELSE}
  // RAX = Dividend; RDX = Divisor; RCX = @Result; Stack = @Remainder
  PUSH	EBX
  MOV	EBX,EDX
  XOR	EDX,EDX
  DIV	EBX
  MOV	EBX,Remainder
  MOV	[ECX],EAX
  MOV	[EBX],EDX
  POP	EBX
  {$ENDIF}
end;

procedure Init;
begin
  LoadLibrary(LIB_RICHED10);
  LoadLibrary(LIB_RICHED20);
  LoadLibrary(LIB_MSFTEDIT);
end;

begin
  Init;
end.

